/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "../basetypes.h"

class HandCard;
class Game;

/** Class for a card
 **/
class Card {
  public:
    static Card const empty;
    static Card const unknown;

    static Card const fox;
    static Card const dolle;
    static Card const charlie;

    static Card const club_ace;
    static Card const spade_ace;
    static Card const heart_ace;
    static Card const diamond_ace;
    static Card const club_ten;
    static Card const spade_ten;
    static Card const heart_ten;
    static Card const diamond_ten;
    static Card const club_king;
    static Card const spade_king;
    static Card const heart_king;
    static Card const diamond_king;
    static Card const club_queen;
    static Card const spade_queen;
    static Card const heart_queen;
    static Card const diamond_queen;
    static Card const club_jack;
    static Card const spade_jack;
    static Card const heart_jack;
    static Card const diamond_jack;
    static Card const club_nine;
    static Card const spade_nine;
    static Card const heart_nine;
    static Card const diamond_nine;

    static unsigned constexpr number_of_colors = 4u;
    static unsigned constexpr number_of_tcolors = number_of_colors + 1;
    static unsigned constexpr number_of_values = 6u;
    static unsigned constexpr number_of_cards = number_of_colors * number_of_values;

    enum Value : int8_t {
      nine = 0,
      jack = 2,
      queen = 3,
      king = 4,
      ten = 10,
      ace = 11,
      nocardvalue = 100, // for simpler error-tracking
      unknowncardvalue = 101, // for simpler error-tracking
      maxcardvalue = unknowncardvalue
    }; // enum Value

    enum TColor : int8_t {
      diamond = 0,
      heart,
      spade,
      club,
      trump,
      nocardcolor = 100, // for simpler error-tracking
      unknowncardcolor = 101, // for simpler error-tracking
      maxcardcolor = unknowncardcolor
    }; // enum TColor
    using Color = TColor;

    // converts the string in a value
    static Value value(string const& value);
    // converts the string in a color
    static Color color(string const& color);

  public:

    constexpr Card() noexcept = default;
    constexpr Card(Card const& c) noexcept = default;
    constexpr Card& operator=(Card const& card) noexcept = default;
    constexpr Card(Color c, Value v) noexcept :
      color_(c), value_(v)
      { }
    Card(string const& name);

    //virtual ~Card() = default;

    istream& read(istream& istr);

    constexpr Color color() const noexcept
    { return color_; }
    constexpr Value value() const noexcept
    { return value_; }

    constexpr bool is_empty() const noexcept
    { return color_ == nocardcolor  || value_ == nocardvalue; }
    constexpr explicit operator bool() const noexcept
    { return !this->is_empty(); }
    constexpr bool is_unknown() const noexcept
    { return color_ == unknowncardcolor  || value_ == unknowncardvalue; }

    // convert to an integer (between 1 and 24)
    int to_int() const noexcept;

    constexpr unsigned points() const noexcept
    { return static_cast<unsigned>(value_); }

    Card::TColor tcolor(Game const& game) const;
    bool istrump(Game const& game) const;
    bool istrump(GameType gametype, bool dollen) const;
    bool isdolle(Game const& game) const;
    bool isdolle(GameType gametype, bool dollen) const;

    bool less(HandCard const& card) const;

    constexpr bool operator==(Card const& rhs) const noexcept
    { return color_ == rhs.color_ && value_ == rhs.value_; }
    constexpr bool operator!=(Card const& rhs) const noexcept
    { return color_ != rhs.color_ || value_ != rhs.value_; }
    constexpr bool operator<(Card const& rhs) const noexcept
    { return (color_ < rhs.color_ || (color_ == rhs.color_ && value_ < rhs.value_)); }
  private:
    Color color_ = nocardcolor;
    Value value_ = nocardvalue;
}; // class Card

// whether the marriage is determined by a 'tcolor'-trick
bool is_selector(Card::TColor tcolor, MarriageSelector marriage_selector);

string to_string(Card::Value value) noexcept;
string to_string(Card::TColor tcolor) noexcept;
string to_string(Card const& card) noexcept;
string gettext(Card::Value value);
string gettext(Card::Color color);
string gettext(Card const& card);

ostream& operator<<(ostream& ostr, Card::Value value);
istream& operator>>(istream& istr, Card::Value& value);
ostream& operator<<(ostream& ostr, Card::TColor tcolor);
istream& operator>>(istream& istr, Card::Color& color);
ostream& operator<<(ostream& ostr, Card const& card);
istream& operator>>(istream& istr, Card& card);
