/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "../basetypes.h"
#include "hand_cards.h"
class Trick;
class Player;

/**
 ** Hand (of a player) with all cards (played and not played)
 **/
class Hand {
  public:
  using const_iterator = HandCards::const_iterator;

    Hand() = default;
    explicit Hand(Player const& player);
    Hand(Player const& player, Hand const& h);
    explicit Hand(vector<Card> const& cards);
    Hand(Player const& player, vector<Card> const& cards);
    Hand(Player const& player, vector<Card> const& unplayed_cards,
         vector<Card> const& played_cards);
    explicit Hand(string const& text);
    explicit Hand(istream& istr);
    Hand(Hand const& h);
    Hand& operator=(Hand const& h);

    ~Hand() = default;

    operator HandCards() const;
    operator vector<Card>() const;
    const_iterator begin() const;
    const_iterator end() const;

    HandCards const& cards() const;

    bool self_check() const;

    Player const& player() const;
    void set_player(Player const& player);

    Game const& game() const;

    bool empty() const;
    // the number of cards (which are not played)
    size_t cardsnumber() const;
    // the number of cards (all)
    size_t cardsnumber_all() const;

    // the points on the hand
    unsigned points() const;

    // set an unknown card to a known one
    void set_known(Card const& card, unsigned n = 1);

    // the i'th card (of the unplayed)
    HandCard const& card(unsigned i) const;
    // the i'th card (of all)
    HandCard const& card_all(unsigned i) const;
    // whether the i'th card is played
    bool played(unsigned i) const;
    // in wich trick the i'th card was played
    unsigned played_trick(unsigned i) const;

    // all cards without double ones
    HandCards cards_single() const;
    // the cards and its number
    map<Card, unsigned> counted_cards() const;

    // the highest value
    Card::Value highest_value() const;
    // the highest value of the given tcolor
    Card::Value highest_value(Card::TColor const tcolor) const;
    // the lowest value
    Card::Value lowest_value() const;
    // the lowest value of the given tcolor
    Card::Value lowest_value(Card::TColor const tcolor) const;

    // the highest card 
    HandCard const& highest_card() const;
    // the highest trump 
    HandCard const& highest_trump() const;
    // the highest trump till the given card
    HandCard const& highest_trump_till(Card const& card_limit) const;
    // the highest card of the given tcolor
    HandCard const& highest_card(Card::TColor const tcolor) const;
    // the lowest trump 
    HandCard const& lowest_trump() const;
    // the lowest card of the given tcolor
    HandCard const& lowest_card(Card::TColor const tcolor) const;
    // whether there exists a higher card
    bool higher_card_exists(Card const& card) const;
    // whether there exists a lower card
    bool lower_card_exists(Card const& card) const;
    // the number of higher cards
    unsigned higher_cards_no(Card const& card) const;
    // the next jabbing card of 'card'
    HandCard const& next_jabbing_card(Card const& card) const;
    // the next higher card then 'card'
    HandCard const& next_higher_card(Card const& card) const;
    // the card 'card' or next higher card then 'card'
    HandCard const& same_or_higher_card(Card const& card) const;
    // the next lower card then 'card' (in the same color)
    HandCard const& next_lower_card(Card const& card) const;
    // the card 'card' or next lower card then 'card'
    HandCard const& same_or_lower_card(Card const& card) const;
    // whether the trick can be jabbed
    bool can_jab(Trick const& trick) const;

    // get the position of the card
    unsigned getpos(Card const& card) const;
    // get the position of the card (in all cards)
    unsigned getpos_all(Card const& card) const;

    // transform the position in the unplayed cards to the position in all cards
    unsigned pos_to_pos_all(unsigned pos) const;
    // transform the position in all cards to the position in the unplayed cards
    unsigned pos_all_to_pos(unsigned pos) const;



    // changing

    // play the card 'card'
    void playcard(HandCard const& card);
    // play the card at the position 'pos'
    void playcard(unsigned pos);
    // unplaye the card 'card'
    void unplaycard(HandCard const& card);

    // add the card to the hand
    void add(Card const& card, unsigned n = 1);
    // add the cards to the hand
    void add(vector<Card> const& cards);
    // add the hand to the hand
    void add(Hand const& hand);

    // add the played card to the hand
    void add_played(Card const& card, unsigned n = 1);
    // add the played cards to the hand
    void add_played(vector<Card> const& cards);

    // remove the card from the hand
    unsigned remove(HandCard const& card);
    unsigned remove(Card card);
    // remove the cards from the hand
    void remove(HandCards const& cards);



    // card request

    unsigned requested_position() const;
    // the requested card
    HandCard requested_card() const;
    // the requested position
    unsigned requested_position_all() const;
    // request the card at the position (in the unplayed cards)
    void request_position(unsigned position);
    // request the card at the position (in all cards)
    void request_position_all(unsigned position_all);
    // request the card
    void request_card(Card const& card);
    // forget the request
    void forget_request();



    // counting

    // the valid cards for the trick (do not use in for-loop!)
    HandCards validcards(Trick const& trick) const;
    // the cards of a given tcolor
    HandCards cards(Card::TColor tcolor) const;
    // the trumps
    HandCards trumps() const;

    // whether there is a trump card in the hand
    bool hastrump() const;
    // whether there is a color card in the hand
    bool hascolor() const;

    // whether the hand does contain possible genscher cards
    bool has_possible_genscher() const;
    // whether the hand does contain swines (or can announce them)
    bool has_swines() const;
    // whether the hand does contain hyperswines (or can announce them)
    bool has_hyperswines() const;

    // the number of cards with the value
    unsigned count(Card::Value value) const;
    // the number of cards with 10 or 11 points
    unsigned count_rich_cards() const;
    // the number of cards with the trump color
    unsigned count(Card::TColor tcolor) const;
    // the number of cards 'card'
    unsigned count(Card const& card) const;
    // the number of cards 'Card(tcolor, value)'
    unsigned count(Card::TColor tcolor,
                   Card::Value value) const;

    // the number of cards 'Card(color)' without counting trump
    unsigned count(Card::Color color, Game const& game) const;
    unsigned count(Card::Color color,
                   GameType gametype) const;
    unsigned count(Card::Color color,
                   GameType gametype,
                   bool dollen) const;
    // the number of fehl
    unsigned count_fehl() const;
    unsigned count_fehl_colors(GameType game_type) const;

    // the number of played cards 'card'
    unsigned played(Card const& card) const;

    // the number of trumps
    unsigned count_trumps() const;
    // the number of trumps
    unsigned count_trumps(GameType game_type) const;
    // the number of single tens
    unsigned count_single_tens(GameType game_type) const;
    // the number of trumps aces
    unsigned count_trump_aces() const;
    // the number of trumps aces
    unsigned count_trump_aces(GameType game_type) const;
    // the number of trumps nines
    unsigned count_trump_nines() const;
    // the number of trump kings
    unsigned count_trump_kings() const;
    // the number of nines
    unsigned count_nines() const;
    // the number of kings
    unsigned count_kings() const;
    // the number of club queens
    unsigned count_club_queens() const;
    // the number of dolle
    unsigned count_dolle() const;
    // the number of swines
    unsigned count_swines() const;
    // the number of hyperswines
    unsigned count_hyperswines() const;

    // the number of cards to shift in the poverty
    unsigned count_poverty_cards() const;
    // whether the hand has a poverty
    bool has_poverty() const;

    // the number of cards 'Card(tcolor, value)' in all cards
    unsigned count_all(Card::TColor tcolor, Card::Value value) const;
    // the number of cards 'card' in all cards
    unsigned count_all(Card const& card) const;
    // the number trump aces in all cards
    unsigned count_all_trump_aces() const;


    // the number of cards which are greater or equal in the game
    unsigned count_ge(Card const& card) const;
    // the number of cards which are less or equal in the game
    unsigned count_le(Card const& card) const;
    // the number of trump cards which are greater or equal in the game
    unsigned count_ge_trump(Card const& card) const;
    // the number of trump cards which are less or equal in the game
    unsigned count_le_trump(Card const& card) const;
    // the relative position of the card in the trumps
    float rel_pos_trump(Card const& card) const;

    // checking

    // whether the 'hand' has the same cards as 'this'
    bool has_same_cards(Hand const& hand) const;

    bool contains(Card::TColor tcolor) const;
    bool contains(Card const& card) const;
    bool contains(Card::Color color, Card::Value value) const;
    bool contains(Hand const& hand) const;
    HandCards missing_cards(Hand const& hand) const;

    bool contains_ten_points_in_trump() const;

  private:
    // the cards on the hand
    HandCards cards_;
    // all cards on the hand
    HandCards cards_all_;
    // whether a card is played and in which trick
    vector<unsigned> played_;

    // which card is requested
    unsigned requested_position_ = UINT_MAX;

    // the player who owns the hand
    Player const* player_ = nullptr;
}; // class Hand

inline auto begin(Hand const& hand)
{ return hand.begin(); }
inline auto end(Hand const& hand)
{ return hand.end(); }

ostream& operator<<(ostream& ostr, Hand const& hand);

bool operator==(Hand const& a, Hand const& b);
bool operator!=(Hand const& a, Hand const& b);
