/**********************************************************************
 *
 *   Some routines for effective Debugging
 *
 *   Copyright (C) 2001 – 2018 by Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *
 *********************************************************************/

#include <iostream>
#include <ctime>
#include "debug.h"
#include "utils.h"

Debug debug;
// Filter
// table: Zeichnen des Tisches
// gameplay: Spielverlauf mit ausgeben
// heuristics

// initialize the pointer to the debug function
DebugFunction debug_function = nullptr;

namespace ExceptionTerminateHandler {

unique_ptr<UncaughtException> uncaughtException_;

// neue terminate Funktion, die what, CPOS und Bugreport ausgibt
// fuer eine nicht gefangene Exception
void
terminate()
{   	
  ostringstream ostr;	
  if (uncaughtException_) {
    ostr << "EXCEPTION(file = " << uncaughtException_->file_
      << ", line = " << uncaughtException_->line_ 
      << "): " << uncaughtException_->what_ << endl;
  } else {
    ostr << "EXCEPTION(unknown)" << endl;
  }

  if (::debug_function)
    (*debug_function)(ostr.str());		
  else					
    cerr << ostr.str();		
  abort();
} // void terminate()

void set(string const& file, int const line, string const& what)
{
  uncaughtException_ = make_unique<UncaughtException>(file, line, what);
  set_terminate(&terminate);
} // static void set(const char* file, const int line, const char* what)

void unset()
{
  uncaughtException_.reset();
} // static void unset

}; // namespace ExceptionTerminateHandler

/** constructor
 **/
Debug::Debug()
{
  this->add("");
  this->add("codeline");
}

/** destructor
 **/
Debug::~Debug()
= default;

/** @return   the output stream
 **/
ostream&
Debug::ostr()
{ return *this->ostr_; }

/** write the debug information into the output stream
 **
 ** @param    ostr   output stream
 **/
void
Debug::set_ostr(ostream& ostr)
{
  this->ostr_ = &ostr;
} // void Debug::set_ostr(ostream& ostr)

/** write the debug information into the file
 **
 ** @param    file   file name
 **/
void
Debug::set_file(string const file)
{
  if (   file.empty()
      || file == "-"
      || file == "cerr")
    this->ostr_ = &cerr;
  else
    this->ostr_ = new ofstream(file);
} // void Debug::set_file(string file)

/** set the debug output for the filter
 **
 ** @param    filter   filter name
 **/
void
Debug::add(string const filter)
{
  this->filter_.insert(filter);
}

/** 
 ** remove the debug output for the filter
 **
 ** @param    filter   filter name
 **/
void
Debug::remove(string const filter)
{
  this->filter_.erase(filter);
}

/** -> result
 **
 ** @param    filter   filter name
 **
 ** @return   whether the filter is set
 **/
bool
Debug::filter(string const filter)
{
  return this->filter_.count(filter);
}

/** 
 ** -> result
 **
 ** @param    filter   filter name
 **
 ** @return   whether the filter is set
 **/
bool
Debug::operator()(string const filter) const
{
  return (filter.empty() || this->filter_.count(filter));
} // bool Debug::operator()(string filter) const

/** 
 ** -> result
 ** If the filter is set, returns the output stream
 ** If the filter is not set, returns the nullstream
 **
 ** @param    filter   filter name
 **
 ** @return   output stream for the filter
 **/
ostream&
Debug::ostr(string const filter)
{
  if (filter.empty() || this->filter_.count(filter)) {
    if (this->ostr_)
      return *this->ostr_;
    else
      return cerr;
  } else {
    return ::null_ostr;
  }
} // ostream& Debug::ostr(string filter)

/** 
 ** If the filter is set, returns the output stream
 ** If the filter is not set, returns the nullstream
 ** write the file, line and time into the stream
 **
 ** @param    filter          filter name
 ** @param    file            source file
 ** @param    line            source line
 ** @param    function_name   name of the function
 **
 ** @return   output stream for the filter
 **/
ostream&
Debug::ostr(string const filter,
                  string const file, long const line, string const function_name)
{
  auto& ostr = this->ostr(filter);
  if (&ostr == &::null_ostr)
    return ostr;
  if (this->filter("time"))
    ostr << Time::current() << ' ';
  if (this->filter("clock")) {
    auto const c = std::clock();
    ostr << Time(c / CLOCKS_PER_SEC, (c / (CLOCKS_PER_SEC / 1000) % 1000)) << ' ';
  }
  if (this->filter("codeline"))
    ostr << file << '#' << line << ' ';
  if (this->filter("function"))
    ostr << function_name << "() ";
  return ostr;
} // ostream& Debug::ostr(string filter, string file, long line)
