/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *  
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "game.h"

#include "../player/player.h"
#include "../party/rule.h"
#include "../card/trick.h"

/** -> result
 **
 ** @param    player  player
 **
 ** @return   points, the player has got
 **/
unsigned
Game::points_of_player(Player const& player) const
{
  unsigned points = 0;

  for (auto const& trick : this->tricks()) {
    if (   (trick.winnerplayer() == player)
        && trick.intrickpile()) {
      points += trick.points();
    }
  }

  return points;
} // unsigned Game::points_of_player(Player player) const

/** -> result
 **
 ** @param    playerno   number of the player
 **
 ** @return   points, the player has got
 **/
unsigned
Game::points_of_player(unsigned const playerno) const
{
  return this->points_of_player(this->player(playerno));
} // unsigned Game::points_of_player(unsigned playerno) const

/** -> result
 **
 ** @param    player  player
 ** @param    trickno last trick to view
 **
 ** @return   points, the player has got til trickno
 **/
unsigned
Game::points_of_player(Player const& player, unsigned trickno) const
{
  unsigned points = 0;

  if (trickno != UINT_MAX)
    trickno += 1;
  for (auto trick = this->tricks().begin();
       ((trickno > 0)
        && (trick != this->tricks().end()));
       ++trick, --trickno) {
    if (   (trick->winnerplayer() == player)
        && trick->intrickpile()) {
      points += trick->points();
    }
  }

  return points;
} // unsigned Game::points_of_player(Player const& player, unsigned trickno) const

/** -> result
 **
 ** @param    playerno   number of the player
 ** @param    trickno    last trick to view
 **
 ** @return   points, the player has got till trickno
 **/
unsigned
Game::points_of_player(unsigned const playerno, unsigned const trickno) const
{
  return this->points_of_player(this->player(playerno), trickno);
} // unsigned Game::points_of_player(unsigned playerno, unsigned trickno) const

/** -> result
 ** uses hidden information
 **
 ** @param    team    team
 **
 ** @return   points, the team has got
 **/
unsigned
Game::points_of_team(Team const& team) const
{
  unsigned points = 0;

  for (auto const& p : this->players()) {
    if (p.team() == team)
      points += this->points_of_player(p);
  }

  return points;
} // unsigned Game::points_of_team(Team const& team) const

/** @return   the points played so far in the game
 **/
unsigned
Game::played_points() const
{
  unsigned points = 0;

  for (auto const& p : this->players())
    points += this->points_of_player(p);
  if (::game_status < GameStatus::game_finished)
    points += this->tricks().current().points();

  return points;
} // unsigned Game::played_points() const

/** @return   the remaining points to be played
 **/
unsigned
Game::remaining_points() const
{
  return (240 - this->played_points());
} // unsigned Game::remaining_points() const

/** -> result
 **
 ** @param    team   team to check
 **
 ** @return   the points needed for 'team' to win the game
 **/
unsigned
Game::needed_points_to_win(Team const& team) const
{
  Announcement const highest_bid_re  
    = this->announcements().last(Team::re);
  Announcement const highest_bid_contra
    = this->announcements().last(Team::contra);

  switch (team) {
  case Team::re:
    if (   (highest_bid_re == Announcement::noannouncement)
        && (highest_bid_contra == Announcement::noannouncement) )
      return 121;
    else if (highest_bid_re == Announcement::noannouncement)
      return point_limit_for_opposite_team(highest_bid_contra);
    else 
      return max(needed_points(highest_bid_re),
                 point_limit_for_opposite_team(highest_bid_contra));
  case Team::contra:
    if (   (highest_bid_contra == Announcement::noannouncement)
        && (highest_bid_re == Announcement::noannouncement) )
      return 120;
    else if (   (highest_bid_contra == Announcement::no120)
             && (highest_bid_re == Announcement::no120) )
      return 120;
    else if (highest_bid_contra == Announcement::noannouncement)
      return point_limit_for_opposite_team(highest_bid_re);
    else 
      return max(needed_points(highest_bid_contra),
                 point_limit_for_opposite_team(highest_bid_re));
  default:
    return 0;
  } // switch(team)
} // unsigned Game::needed_points_to_win(Team const& team) const

/** @return   the winner team
 **/
Team
Game::winnerteam() const
{
  DEBUG_ASSERTION((this->tricks().remaining_no() == 0),
                  "Game::winnerteam():\n"
                  "  game is not finished, yet\n"
                  << "current no: " << this->tricks().current_no() << '\n'
                  << "remaining no: " << this->tricks().remaining_no() << '\n');

  // the points the team has made
  unsigned const points_re     = this->points_of_team(Team::re);
  unsigned const points_contra = this->points_of_team(Team::contra);

  // The official rules (http://www.doko-verband.de/download/turnierspielregeln.pdf)
  // distinguish between two kinds of announcements:
  // "ANsagen" (re and contra (both as opening and reply), let's call them "calls") and
  // "ABsagen" (no90, no60 and so on, let's call them "bids").
  // This distinction is important when deciding on the winner team, because bids determine how
  // many points a party needs to win, while all calls are ignored with one exception:
  // if a game ends 120-120 and the only announcement was a contra call, the contra team
  // will lose (it wins all other 120-120 situations unless it has made a bid).
  //
  // FreeDoko follows this distinction only if strict_limit is false, it treats all announcements
  // as bids if strict_limit is true. This has a number of consequences:
  // * instead of a single reply announcement (the reply call), there are several reply bids:
  //   no120_reply (treated as a NO121 bid)
  //    no90_reply (treated as a NO151 bid)
  //    no60_reply (treated as a NO181 bid)
  //    no30_reply (treated as a NO211 bid)
  //     no0_reply (treated as a bid for at least one trick)
  // * if no120_reply was the highest bid of the re team (which implies a contra announcement),
  //   the re team will win a 120-120 game because it fulfilled its bid
  // * if both teams announced no120 and the game ends 120-120, there will be no winner
  // * there'll be many more games without a winner because an opening re (or contra) will
  //   stand as a no120 bid until the end. This way, even 120 points against a no0 bid may
  //   not be sufficient to win the game.

  Announcement highest_bid_re     = this->announcements().last(Team::re);
  Announcement highest_bid_contra = this->announcements().last(Team::contra);

  // a re/contra (without strict limit) is seen as no announcement
  // a reply can only be made, when the other team has made an announcement
  if (!this->rule()(Rule::Type::announcement_strict_limit)) {
    // special case: only contra has announced 'contra'
    if (   (highest_bid_contra == Announcement::no120)
        && (highest_bid_re == Announcement::noannouncement) )
      if (this->points_of_team(Team::re) >= 120)
        return Team::re;
      else
        return Team::contra;

    if (   (highest_bid_re == Announcement::no120)
        || (highest_bid_re == Announcement::reply) )
      highest_bid_re = Announcement::noannouncement;
    if (   (highest_bid_contra == Announcement::no120)
        || (highest_bid_contra == Announcement::reply) )
      highest_bid_contra = Announcement::noannouncement;
  } // if (!this->rule()(Rule::Type::announcement_strict_limit))

  // Now determine the winning team

  // Case #1: no party made a bid
  // * there's always a winner
  // * if no call was made or both teams made calls or only the re team made a call:
  // * Re-Team wins with 121 or more points [according to official rules article 7.1.2,1-3]
  // * Contra-Team wins with 120 or more points [7.1.3,1-3]
  // * if only the contra team made a call:
  // * Re-Team wins with 120 ore more points [7.1.2,4] (see above)
  // * Contra-Team wins with 121 or more points [7.1.3,4] (see above)

  if (   (highest_bid_re == Announcement::noannouncement)
      && (highest_bid_contra == Announcement::noannouncement) ) {
    if (points_re > 120)
      return Team::re;
    else
      return Team::contra;
  }

  // Case #2: at least one of the teams made a bid
  // * both teams are allowed to make bids but it's only possible for one team to fulfill its bid =>
  // * Re-Team wins if it fulfills its bid [7.1.2,5-6]
  // * Contra-Team wins if it fulfills its bid [7.1.3,5-6]

  // check if re fulfilled its bid
  if (   (   (highest_bid_re == Announcement::no120)  // call gets treated as bid if strict_limit is true
          && (points_contra < 120) )
      || (   (highest_bid_re == Announcement::no90)
          && (points_contra < 90 ) )
      || (   (highest_bid_re == Announcement::no60)
          && (points_contra < 60 ) )
      || (   (highest_bid_re == Announcement::no30)
          && (points_contra < 30 ) )
      || (   (highest_bid_re == Announcement::no0)
          && !this->tricks().has_trick(Team::contra) ) ) {
    return Team::re;
  }

  // these bids are only possible if strict_limit is true
  if (   (   (highest_bid_re == Announcement::no120_reply)
          && (points_re >= 120 ))
      || (   (highest_bid_re == Announcement::no90_reply)
          && (points_re >= 90 ))
      || (   (highest_bid_re == Announcement::no60_reply)
          && (points_re >= 60 ))
      || (   (highest_bid_re == Announcement::no30_reply)
          && (points_re >= 30 ))
      || (   (highest_bid_re == Announcement::no0_reply)
          && this->tricks().has_trick(Team::re) ) ) {
    return Team::re;
  }

  // check if contra fulfilled its bid
  if (   (  (highest_bid_contra == Announcement::no120)  // call gets treated as bid if strict_limit is true
          && (points_re < 120 ))
      || (   (highest_bid_contra == Announcement::no90)
          && (points_re < 90 ))
      || (   (highest_bid_contra == Announcement::no60)
          && (points_re < 60 ))
      || (   (highest_bid_contra == Announcement::no30)
          && (points_re < 30 ))
      || (   (highest_bid_contra == Announcement::no0)
          && !this->tricks().has_trick(Team::re) ) ) {
    return Team::contra;
  }

  // these bids are only possible if strict_limit is true
  if (   (   (highest_bid_contra == Announcement::no120_reply)
          && (points_contra >= 120 ))
      || (   (highest_bid_contra == Announcement::no90_reply)
          && (points_contra >= 90 ))
      || (   (highest_bid_contra == Announcement::no60_reply)
          && (points_contra >= 60 ))
      || (   (highest_bid_contra == Announcement::no30_reply)
          && (points_contra >= 30 ))
      || (   (highest_bid_contra == Announcement::no0_reply)
          && this->tricks().has_trick(Team::contra) ) ) {
    return Team::contra;
  }

  // Case #3: at least one bid has been made, but no bid has been fulfilled
  // the case that no team has made an annoucement / bid is Case #1,
  // so here there exists at least one team with a bid
  // * Re-Team wins if it didn't make a bid [7.1.2,7-8]
  // * Contra-Team wins if it didn't make a bid [7.1.3,7-8]
  // * no winner if both teams made bids [7.1.4]

  if (!highest_bid_re)
    // contra has not fulfilled its announcement
    return Team::re;
  else if (!highest_bid_contra)
    // re has not fulfilled its announcement
    return Team::contra;
  else
    // no player has made enough points for his announcement
    return Team::noteam;
} // Team Game::winnerteam() const
