/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "game.h"

#include "../party/rule.h"
#include "../player/player.h"
#include "../player/human/human.h"
#include "../card/trick.h"
#include "../ui/ui.h"

/** reset the teaminfo
 **/
void
Game::teaminfo_reset()
{                                                                  
  for (unsigned p = 0; p < this->playerno(); ++p) {
    this->player(p).set_team(Team::unknown);
  }
  this->teaminfo_ = vector<Team>(this->playerno(), Team::unknown);
  this->human_teaminfo_ = vector<Team>(this->playerno(), Team::unknown);
} // void Game::teaminfo_reset()

/** sets the teaminfo at the start of the game
 **
 ** @todo      marriage: team of the players (maybe re, maybe contra, ?)
 **/
void
Game::teaminfo_set_at_gamestart()
{                                                                  
  if (this->type() == GameType::normal) {
    this->teaminfo_ = vector<Team>(this->playerno(), Team::unknown);
    for (unsigned p = 0; p < this->playerno(); ++p) {
      this->player(p).set_team(this->player(p).hand().contains(Card::club_queen)
                               ? Team::re
                               : this->player(p).hand().contains(Card::unknown)
                               ? Team::unknown
                               : Team::contra);
    }
  } else if (type() == GameType::poverty) {
    // is done in 'this->poverty_shift()'

  } else if (this->type() == GameType::marriage) {
    // set all teams to unknown but of the soloplayer
    this->teaminfo_ = vector<Team>(this->playerno(), Team::unknown);
    this->teaminfo_[this->soloplayer().no()] = Team::re;
    for (unsigned p = 0; p < this->playerno(); ++p)
      this->player(p).set_team((this->player(p) == this->soloplayer())
                               ? Team::re
                               : Team::maybe_contra);

  } else if (::is_solo(this->type())) {
    // set all teams to contra but of the soloplayer
    this->teaminfo_ = vector<Team>(this->playerno(), Team::contra);
    this->teaminfo_[this->soloplayer().no()] = Team::re;
    for (unsigned p = 0; p < this->playerno(); ++p)
      this->player(p).set_team((this->player(p) == this->soloplayer())
                               ? Team::re
                               : Team::contra);
  } // if (this->type())

  this->human_teaminfo_ = vector<Team>(this->playerno(), Team::unknown);
  this->human_teaminfo_update();
} // void Game::teaminfo_set_at_gamestart()

/** updates the global teaminfo
 **/
void
Game::teaminfo_update()
{
  if (this->type() != GameType::normal)
    return ;

  if (!this->tricks().current().isstartcard()) {
    Player const& player
      = this->tricks().current().player_of_card(this->tricks().current().actcardno() - 1);

    if (!is_real(this->teaminfo(player))) {
      // the player has played a club queen
      if (this->tricks().current().card_of_player(player) == Card::club_queen) {
        this->set_teaminfo(player, Team::re);
      } else 

      // the startcard is trump, but the player has not played a trump
      if (   this->tricks().current().startcard().istrump()
          && (!this->tricks().current().card_of_player(player).istrump())
          && !player.hand().contains(Card::club_queen)) {
        this->set_teaminfo(player, Team::contra);
      } else

      // it is the last played card of the player
      if (   (player.cards_to_play() == 0)
          && !player.hand().contains(Card::club_queen)) {
        this->set_teaminfo(player, Team::contra);
      }

    } // if (team of 'player' unknown)
  } // if (!this->tricks().current().isstartcard())

  { // count the played club queens for a silent marriage
    for (auto const& player : this->players()) {
      if (this->teaminfo(player) == Team::re) {
        // We only want to know, whether the player has played his second
        // club queen, so he must already be known as re.
        unsigned no = 0;
        for (auto const& t : this->tricks()) {
          if (t.cardno_of_player(player) < t.actcardno())
            if (t.card_of_player(player) == Card::club_queen)
              no += 1;
        } // for (t : this->tricks())

        if (no == this->rule()(Rule::Type::number_of_same_cards)) {
          // all other players must be contra
          for (unsigned p = 0; p < this->playerno(); ++p)
            if (this->teaminfo(p) != Team::re)
              this->set_teaminfo(this->player(p), Team::contra);
        } // if (both club queens played)
      } // if (this->teaminfo_[player] == Team::re)
    } // for (player)
  } // count the played club queens for a silent marriage

  if (this->rule()(Rule::Type::swine_only_second)
      && this->swines_announced())
    // the swines owner has to say his team
    this->set_teaminfo(*this->swines_owner(),
                       this->swines_owner().team());

  { // count the 're's and 'contra's
    unsigned re_no = 0;
    unsigned contra_no = 0;
    for (auto const& t : this->teaminfo_) {
      if (t == Team::re)
        re_no += 1;
      else if (t == Team::contra)
        contra_no += 1;
    } // for (t)

    if (re_no == this->rule()(Rule::Type::number_of_players_per_team)) {
      for (unsigned p = 0; p < this->playerno(); ++p)
        if (this->teaminfo(p) != Team::re)
          this->set_teaminfo(this->player(p), Team::contra);
    } // if (re_no == 2)
    if (contra_no == this->playerno() - 1) {
      for (unsigned p = 0; p < this->playerno(); ++p)
        if (this->teaminfo(p) != Team::contra)
          this->set_teaminfo(this->player(p), Team::re);

      this->determine_silent_marriage();
    } // if (contra_no == this->playerno() - 1)

    if (contra_no == this->playerno() - 2) {
      if (!this->tricks().current().isstartcard()) {
        Player const& player
          = this->tricks().current().player_of_card(this->tricks().current().actcardno() - 1);
        if ((this->teaminfo(player) == Team::re)
            && (this->tricks().current().startcard().istrump()
                && !this->tricks().current().card_of_player(player).istrump())) {
          // the player is re but can play no more trump

          // all players but two are contra, one is re and does not have
          // any more club queen
          // so the remaining player must be re
          for (unsigned p = 0; p < this->playerno(); ++p)
            if (this->teaminfo(p) != Team::contra)
              this->set_teaminfo(this->player(p), Team::re);
        } // // if (the player is re but can play no more trump)
      } // if (!this->tricks().current().isstartcard())
    } // if (contra_no == this->playerno() - 2)
  } // count the 're's and 'contra's

  this->human_teaminfo_update();
} // void Game::teaminfo_update()


/** update the teaminfo of all humans
 **/
void
Game::human_teaminfo_update()
{
  DEBUG_ASSERTION(::game_status & GameStatus::game,
                  "Game::update_human_teaminfo()\n"
                  "  not in a game: " << ::game_status);

  // backup of the teaminfo (so changes can be seen)
  vector<Team> const teaminfo_bak = this->human_teaminfo_;

  { // update the teaminfo from the one of 'game'

    // whether some team has changed
    bool changed = false;
    for (unsigned p = 0; p < this->playerno(); ++p) {
      Player const& player = this->player(p);
      if (   ::is_real(this->teaminfo(player))
          && (this->teaminfo_for_humans(player) != this->teaminfo(player)) ) {
        this->human_teaminfo_[p] = this->teaminfo(player);
        changed = true;
      }
      if (   (player.type() == Player::Type::human)
          && (this->teaminfo_for_humans(player) != player.team())
          && ::is_real(player.team()) ) {
        this->human_teaminfo_[p] = player.team();
        changed = true;
      }
    } // for (p < this->playerno())

    if (!changed)
      return ;
  } // update the teaminfo from the one of 'game'

  if (this->type() != GameType::normal)
    return ;

  // the number of re players
  unsigned re_no = 0;
  // the number of human re players
  unsigned human_re_no = 0;
  // the number of club qeens still on the human hands
  unsigned human_club_queens_no = 0;
  // the number of contra players
  unsigned contra_no = 0;

  for (unsigned p = 0; p < this->playerno(); ++p) {
    Player const& player = this->player(p);
    if (this->teaminfo_for_humans(player) == Team::re)
      re_no += 1;
    else if (this->teaminfo_for_humans(player) == Team::contra)
      contra_no += 1;

    if (this->player(p).type() == Player::Type::human) {
      if (this->teaminfo_for_humans(player) == Team::re) {
        human_re_no += 1;
        human_club_queens_no += player.hand().count(Card::club_queen);
      }
    }
  } // for (p < this->playerno())

  // check whether all re players are known
  if (   (re_no == this->rule()(Rule::Type::number_of_players_per_team))
      || (this->cards().count_played(Card::club_queen) + human_club_queens_no
          == this->rule()(Rule::Type::number_of_same_cards))
     ) {
    for (unsigned p = 0; p < this->playerno(); ++p) {
      if (this->human_teaminfo_[p] != Team::re)
        this->human_teaminfo_[p] = Team::contra;
    } // for (p < this->playerno())
  } // if (all re players known)

  // check whether all contra players are known
  if (   (contra_no == this->playerno() - 1)
      || (contra_no == (this->playerno() - human_re_no - 1))
     ) {
    for (unsigned p = 0; p < this->playerno(); ++p) {
      if (this->human_teaminfo_[p] != Team::contra)
        this->human_teaminfo_[p] = Team::re;
    } // for (p < this->playerno())
  } // if (all contra players known


  if (!this->isvirtual()) {
    for (unsigned p = 0; p < this->playerno(); ++p) {
      if (this->teaminfo_for_humans(this->player(p)) != teaminfo_bak[p]) {
        ::ui->teaminfo_changed(this->player(p));
        this->signal_teaminfo_changed_(this->player(p));
      }
    }
  }
} // void Game::human_teaminfo_update()

/** -> result
 **
 ** @param    player - the player
 **
 ** @return   the team of player 'player'
 **/
Team
Game::team(Player const& player) const
{
  return player.team();
} // Team Game::team(Player const& player) const
