/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "trick.h"

#include "../../utils/string.h"

namespace GameplayActions {

/** constructor
 **
 ** @param    trick   full trick
 **/
TrickFull::TrickFull(Trick const& trick) :
  Base(),
  trick(trick)
{ }

/** constructor
 **
 ** @param    line   line with the data
 **/
TrickFull::TrickFull(string const& line) :
  Base()
{
  // ex: trick full = 0, diamond queen, diamond jack, spade queen, diamond nine, 2

  auto const pre = to_string(this->type()) + " =";
  if (string(line, 0, pre.length()) != pre)
    return ;

  // first split the line in the single parts
  auto const parts(String::split(string(line, pre.length()),
						    ','));
  auto p = parts.begin();
  if (p == parts.end())
    return ;

  unsigned startplayer = UINT_MAX;
  unsigned winnerplayer = UINT_MAX;
  vector<Card> cards;
  { // read the player
    if (*p == "-")
      return ;

    istringstream istr(*p);
    istr >> startplayer;
    ++p;
  } // read the player

  { // read the cards
    for (; p != parts.end(); ++p) {
      if (isdigit((*p)[0]))
	  break;
      cards.push_back(Card(*p));
    }
  } // read the cards
  { // read the winnerplayer
    if (p != parts.end()) {
      istringstream istr(*p);
      istr >> winnerplayer;
      ++p;
    }
  } // read the winnerplayer
  if (p != parts.end())
    return ;

  this->trick = Trick(startplayer, cards, winnerplayer);
} // TrickFull(string line)

/** constructor
 **
 ** @param    line   line with the data
 ** @param    istr   input stream (behind the line) (to be removed)
 **/
TrickFull::TrickFull(string const& line, istream& istr) :
  Base()
{
#ifndef OUTDATED
  // old format in the bug report:
  //   Played: Player 2: club ace
  if (string(line, 0, strlen("Trick full: ")) == "Trick full: ") {
    this->trick = Trick(istr);

    return ;
  }
#endif

#ifdef WORKAROUND
  // copied from 'TrickFull::TrickFull(string line)'

  // ex: trick full = 0, diamond queen, diamond jack, spade queen, diamond nine, 2

  { // new format: one line
    auto const pre = to_string(this->type()) + " =";
    if (string(line, 0, pre.length()) != pre)
      return ;

    // first split the line in the single parts
    auto const parts(String::split(string(line, pre.length()),
						      ','));
    auto p = parts.begin();
    if (p == parts.end())
      return ;

    unsigned startplayer = UINT_MAX;
    unsigned winnerplayer = UINT_MAX;
    vector<Card> cards;
    { // read the player
      if (*p == "-")
	return ;

      istringstream istr(*p);
      istr >> startplayer;
      ++p;
    } // read the player

    { // read the cards
      for (; p != parts.end(); ++p) {
	if (isdigit((*p)[0]))
	  break;
	cards.push_back(Card(*p));
      }
    } // read the cards
    { // read the winnerplayer
      if (p != parts.end()) {
	istringstream istr(*p);
	istr >> winnerplayer;
	++p;
      }
    } // read the winnerplayer
    if (p != parts.end())
      return ;

    this->trick = Trick(startplayer, cards, winnerplayer);
  } // new format: one line
#endif // #ifdef WORKAROUND
} // TrickFull(string line, istream& istr)

/** -> result
 **
 ** @param    action   action to compare with
 **
 ** @return   whether the action is equal to this
 **/
bool
TrickFull::equal(GameplayAction const& action) const
{
  return (   this->GameplayAction::equal(action)
	  && (this->trick == dynamic_cast<GameplayActions::TrickFull const&>(action).trick));
} // bool TrickFull::equal(GameplayAction action) const

/** writes the action in the output stream
 **
 ** @param    ostr   output stream
 **/
void
TrickFull::write(ostream& ostr) const
{
  ostr << this->type() << " = ";
  if (this->trick.startplayerno() == UINT_MAX) {
    ostr << "-";
    return ;
  }

  ostr << this->trick.startplayerno();
  for (unsigned c = 0; c < this->trick.actcardno(); ++c)
    ostr << ", " << this->trick.card(c);

  // cannot use 'Trick::isfull()' because that needs a corresponding game/party
  if (this->trick.winnerplayerno() != UINT_MAX)
    ostr << ", " << this->trick.winnerplayerno();
} // void write(ostream& ostr) const

/** @return   a translation representing the data
 **/
string
TrickFull::data_translation() const
{
  if (this->trick.startplayerno() == UINT_MAX)
    return "-";

  string translation;

  translation += std::to_string(this->trick.startplayerno());

  for (unsigned c = 0; c < this->trick.actcardno(); ++c)
    translation += ", " + _(this->trick.card(c));

  if (this->trick.winnerplayerno() != UINT_MAX)
    translation += ", " + std::to_string(this->trick.winnerplayerno());

  return translation;
} // virtual string TrickFull::data_translation() const

} // namespace GameplayActions
