/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "game.h"
#include "gameplay.h"
#include "gameplay_actions/swines.h"

#include "../party/rule.h"
#include "../ui/ui.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
GameSwines::GameSwines(Game& game) noexcept :
  game_(&game)
{ }

/** constructor
 **
 ** @param    swines   swines to copy
 ** @param    game     corresponding game
 **/
GameSwines::GameSwines(GameSwines const& swines, Game& game) :
  game_(&game),
  swines_owner_(swines.swines_owner_
                ? &game.players().player(swines.swines_owner_->no())
                : nullptr),
  first_fox_catcher_(swines.first_fox_catcher_
                     ? &game.players().player(swines.first_fox_catcher_->no())
                     : nullptr),
  swines_announced_(swines.swines_announced_),
  hyperswines_owner_(swines.hyperswines_owner_
                     ? &game.players().player(swines.hyperswines_owner_->no())
                     : nullptr),
  hyperswines_announced_(swines.hyperswines_announced_)
{ }

/** self check
 **/
void
GameSwines::self_check() const
{
  if (this->swines_owner_)
    DEBUG_ASSERTION((this->game_ == &this->swines_owner_->game()),
		    "GameSwines::self_check()\n"
		    "  the swines owner is not of the same game");
  if (this->first_fox_catcher_)
    DEBUG_ASSERTION((this->game_ == &this->first_fox_catcher_->game()),
		    "GameSwines::self_check()\n"
		    "  the first fox catcher is not of the same game");
  if (this->hyperswines_owner_)
    DEBUG_ASSERTION((this->game_ == &this->hyperswines_owner_->game()),
		    "GameSwines::self_check()\n"
		    "  the hyperswines owner is not of the same game");
}

/** reset the values
 **/
void
GameSwines::reset() noexcept
{
  this->swines_owner_ = {};
  this->first_fox_catcher_ = {};
  this->swines_announced_ = false;
  this->hyperswines_owner_ = {};
  this->hyperswines_announced_ = false;
}

/** @return   whether the first fox has been caught
 **/
bool
GameSwines::first_fox_caught() const noexcept
{
  return this->first_fox_catcher_;
}

/** @return   catcher of the first fox
 **/
Player const&
GameSwines::first_fox_catcher() const noexcept
{
  return *this->first_fox_catcher_;
}

/** @return   whether swines were announced
 **/
bool
GameSwines::swines_announced() const noexcept
{
  return this->swines_owner_;
}

/** @return   the player who has swines
 **   NULL, if no player has swines   
 **/
Player const&
GameSwines::swines_owner() const noexcept
{
  return *this->swines_owner_;
}

/** -> result
 ** for the rules look into the code
 ** In a reservation 'true' can be returned although the announcement cannot
 ** be made (see GameSwines::swines_announce).
 ** This is needed for the sorting in solo games.
 **
 ** @param    player   the player who wants to announce swines
 **
 ** @return   whether 'player' can announce swines
 **/
bool
GameSwines::swines_announcement_valid(Player const& player) const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();
  auto const& hand = player.hand();

  if (!rule(Rule::Type::swines))
    return false;

  // the swines must not be announced already
  if (   this->swines_announced()
      && (player == this->swines_owner()))
    return false;

  // check game
  switch (game.type()) {
  case GameType::normal:
    break;
  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::thrown_richness:
  case GameType::fox_highest_trump:
  case GameType::redistribute:
    return false;
  case GameType::poverty:
    if (!rule(Rule::Type::swines_in_poverty))
      return false;
    break;
  case GameType::genscher:
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
    break;
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
  case GameType::solo_koehler:
    return false;
  case GameType::solo_club:
  case GameType::solo_spade:
  case GameType::solo_heart:
  case GameType::solo_diamond:
    if (!rule(Rule::Type::swines_in_solo))
      return false;
    break;
  case GameType::solo_meatless:
    return false;
  } // switch (this->type())

  // check 'fox do not count' in poverty
  if (   (game.type() == GameType::poverty)
      && (player == (::game_status == GameStatus::game_reservation
                     ? game.players().current_player()
                     : game.players().soloplayer()) )
      && (hand.count_poverty_cards()
          > rule(Rule::Type::max_number_of_poverty_trumps)) )
    return false;

  // check cards number
  if (   (rule(Rule::Type::swine_only_second)
          ? (hand.count_trump_aces() != 1)
          : (hand.count_trump_aces()
             != rule(Rule::Type::number_of_same_cards))) )
    return false;

  // special case: shifted swines
  if (   (::game_status == GameStatus::game_poverty_shift)
      && this->swines_announced())
    return true;

  // 'swines announcement begin': must be in the reservation
  // (in a poverty before the shifting/accepting)
  if (   rule(Rule::Type::swines_announcement_begin)
      && (   (::game_status == GameStatus::game_start)
          || (::game_status == GameStatus::game_init)
          || (::game_status == GameStatus::game_reservation)
          || (   (::game_status == GameStatus::game_poverty_shift)
              && (   (game.poverty().shifted_cardno() == 0)
                  || (player.team() == Team::unknown))
              && !game.poverty().shifted())
          || (   (game.type() != GameType::poverty)
              && (game.cards().count_played() == 0) )
         )
      && !game.poverty().shifted()) {
    return true;
  }

  // normal case, announcement in the game: can announce
  if (   !rule(Rule::Type::swines_announcement_begin)
      && !rule(Rule::Type::swine_only_second))
    return true;

  // 'swine only second': first fox must be caught by the same team
  if (   rule(Rule::Type::swine_only_second)
      && (hand.count_all_trump_aces()
          == rule(Rule::Type::number_of_same_cards))
      && (this->first_fox_catcher_)
      && (   (game.teaminfo().get(this->first_fox_catcher())
              == game.teaminfo().get(player))
          && (   ::is_real(game.teaminfo().get(player))
              || (this->first_fox_catcher() == player) )
         )
     )
    return true;

  return false;
} // bool GameSwines::swines_announcement_valid(Player player) const

/** announce swines if they can be announced and they have not yet been
 ** announced
 **
 ** @param    player   player who wants to announce swines
 **
 ** @return   whether the swines are announced
 **/
bool
GameSwines::swines_announce(Player& player)
{
  auto& game = *this->game_;
  auto const& rule = game.rule();

  if (!this->swines_announcement_valid(player))
    return false;

  // in a reservation no announcement is allowed,
  // but 'valid' returns 'true', because of the cards ordering.
  // This is needed when solo games are selected
  if (::game_status == GameStatus::game_reservation)
    return false;

  this->swines_owner_ = &player;
  this->swines_announced_ = true;
  this->swines_owner_->hand_sort();
  // sort the hand of all players because of possible hyperswines

  if (rule(Rule::Type::swine_only_second)) {
    // the swines owner has to say his team
    game.teaminfo().update();
  }

  game.gameplay_->gameplay_action(GameplayActions::Swines(this->swines_owner().no()));
  if (!game.isvirtual()) {
    ::ui->swines_announced(this->swines_owner());
    this->signal_swines_announced_(this->swines_owner());
    ::ui->gameplay_action(GameplayActions::Swines(this->swines_owner().no()));
    game.signal_gameplay_action_(GameplayActions::Swines(this->swines_owner().no()));
  } // if (!this->isvirtual())

  for (auto& p : game.players()) {
    p.swines_announced(player);
    if (p.hand().has_hyperswines())
      p.hand_sort();
  }

  game.cards().update_swines();

  return true;
} // bool GameSwines::swines_announce(Player& player)

/** @return   whether swines were announced
 **/
bool
GameSwines::hyperswines_announced() const noexcept
{
  return this->hyperswines_owner_;
}

/** @return   the player who has hyperswines
 **   NULL, if no player has hyperswines   
 **/
Player const&
GameSwines::hyperswines_owner() const noexcept
{
  return *this->hyperswines_owner_;
} // Player const* GameSwines::hyperswines_owner() const

/** -> result
 ** for the rules look into the code
 **
 ** @param    player   the player who wants to announce hyperswines
 **
 ** @return   whether 'player' can announce hyperswines
 **/
bool
GameSwines::hyperswines_announcement_valid(Player const& player) const
{
  auto const& game = *this->game_;
  auto const& rule = game.rule();
  auto const& hand = player.hand();

  // check rule 'hyperswines'
  if (!rule(Rule::Type::hyperswines))
    return false;

  // check rule 'hyperswines in solo'
  if (   ::is_solo(game.type())
      && !rule(Rule::Type::hyperswines_in_solo))
    return false;

  // check rule 'hyperswines in poverty'
  if (   ::is_poverty(game.type())
      && !rule(Rule::Type::hyperswines_in_poverty))
    return false;

#ifdef WORKAROUND
  // in the following case we have a problem
  // * virtual games
  // * cards informations
  //   - the player can have diamond ace and diamond nine 
  //   - the player cannot have all four cards (2 diamond ace, 2 diamond nines)
  // @todo   use 'remaining_cards_no(player) < 4'
  if (   game.isvirtual()
      && !this->hyperswines_announced())
    return false;
#endif

  // swines must be announced or must be valid
  if (!(   this->swines_announced()
        || this->swines_announcement_valid(player)))
    return false;

  // check for joint swines and hyperswines
  if (   !rule(Rule::Type::swines_and_hyperswines_joint)
      && (   (this->swines_owner_ == &player)
          || (   !this->swines_announced()
              && this->swines_announcement_valid(player))))
    return false;

  // check for enough cards on the hand
  if (hand.count(game.cards().hyperswine())
      != rule(Rule::Type::number_of_same_cards))
    return false;

  // special case: shifted hyperswines
  if (   (::game_status == GameStatus::game_poverty_shift)
      && (   (game.players().current_player().team() == Team::unknown)
          || this->hyperswines_owner_) )
    return true;

  // the hyperswines must not be announced
  if (this->hyperswines_announced())
    return false;

  // check for right announcement time
  if (   rule(Rule::Type::hyperswines_announcement_begin)
      && (   (::game_status == GameStatus::game_init)
          || (::game_status == GameStatus::game_reservation)
          || (   (::game_status == GameStatus::game_poverty_shift)
              && (   (game.poverty().shifted_cardno() == 0)
                  || (player.team() == Team::unknown))
              && !game.poverty().shifted())
          || (::game_status == GameStatus::game_start) )
      && !game.poverty().shifted())
    return true;

  // normal case, announcement in the game: can announce
  if (!rule(Rule::Type::hyperswines_announcement_begin))
    return true;

  return false;
} // bool GameSwines::hyperswines_announcement_valid(Player const& player) const

/** announce hyperswines if they can be announced
 ** If the player has not annouced 'swines', they are announced automatically
 **
 ** @param    player   player who wants to announce hyperswines
 **
 ** @return   whether the hyperswines are announced
 **/
bool
GameSwines::hyperswines_announce(Player& player)
{
  auto& game = *this->game_;

  if (!this->hyperswines_announcement_valid(player))
    return false;

  if (!this->swines_announced())
    this->swines_announce(player);

  // this check is needed again, since after the 'swines announce' a player
  // can alread have announced hyperswines
  if (!this->hyperswines_announcement_valid(player))
    return false;

  this->hyperswines_owner_ = &player;
  this->hyperswines_announced_ = true;
  this->hyperswines_owner_->hand_sort();

  game.gameplay_->gameplay_action(GameplayActions::Hyperswines(this->hyperswines_owner().no()));
  if (!game.isvirtual()) {
    ::ui->hyperswines_announced(this->hyperswines_owner());
    this->signal_hyperswines_announced_(this->hyperswines_owner());
    ::ui->gameplay_action(GameplayActions::Hyperswines(this->hyperswines_owner().no()));
    game.signal_gameplay_action_(GameplayActions::Hyperswines(this->hyperswines_owner().no()));
  } // if (!this->isvirtual())

  for (auto& p : game.players())
    p.hyperswines_announced(player);

  game.cards().update_hyperswines();

  return true;
} // bool GameSwines::hyperswines_announce(Player& player)

/** test the reservations for swines
 **/
void
GameSwines::test_swines_from_reservations()
{
  auto& game = *this->game_;
  auto& players = game.players();
  auto const& rule = game.rule();

  if (!rule(Rule::Type::swines))
    return ;

  for (unsigned i = 0; i < game.playerno();
       i++, game.players().next_current_player()) {
    if (   game.reservation(game.players().current_player()).swines
        && this->swines_announcement_valid(game.players().current_player()) )
      this->swines_announce(game.players().current_player());

  } // for (p < number_of_players())

  if (rule(Rule::Type::swines_announcement_begin)) {
    for (auto& p : players)
      p.check_swines_announcement_at_game_start();
  }
} // void GameSwines::test_swines_from_reservations()

/** evaluate the trick for swines
 **/
void
GameSwines::evaluate_trick(Trick const& trick)
{
  auto& game = *this->game_;
  auto const& rule = game.rule();

  // test for catcher of the first fox of swines
  if (   rule(Rule::Type::swine_only_second)
      && !this->first_fox_catcher_) {
    for (auto& p : game.players()) {
      if (trick.card_of_player(p).istrumpace()) {
        this->first_fox_catcher_ = &trick.winnerplayer();
        p.hand_sort();
        break;
      } // if (is trump ace)
    } // for (p : this->player)
  } // if (fox in trick)
}

/** @return   signal, when swines are announced
 **/
Signal<void(Player const&)>&
GameSwines::signal_swines_announced() noexcept
{ return this->signal_swines_announced_; }

/** @return   signal, when hyperswines are announced
 **/
Signal<void(Player const&)>&
GameSwines::signal_hyperswines_announced() noexcept
{ return this->signal_hyperswines_announced_; }
