/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"
#include "teaminfo.h"
#include "game.h"
#include "../party/rule.h"
#include "../ui/ui.h"
#include "../player/human/human.h"

/** constructor
 **
 ** @param    game   corresponding game
 **/
GameTeamInfo::GameTeamInfo(Game& game) noexcept :
  game_(&game)
{ }

/** constructor
 **
 ** @param    teaminfo   teaminfo to copy
 ** @param    game   corresponding game
 **/
GameTeamInfo::GameTeamInfo(GameTeamInfo const& teaminfo,
                           Game& game) noexcept :
game_(&game),
  teaminfo_(teaminfo.teaminfo_)
{ }

/** reset the teaminformation
 **/
void
GameTeamInfo::reset()
{
  auto& game = *this->game_;
  auto const& rule = game.rule();
  auto const n = rule(Rule::Type::number_of_players_in_game);
  for (auto& player : game.players()) {
    player.set_team(Team::unknown);
  }
  this->teaminfo_ = vector<Team>(n, Team::unknown);
}

/** set the teaminformation at gamestart
 **/
void
GameTeamInfo::set_at_gamestart()
{
  auto& game = *this->game_;
  auto const n = game.rule()(Rule::Type::number_of_players_in_game);

  if (game.type() == GameType::normal) {
    this->teaminfo_ = vector<Team>(n, Team::unknown);
    for (auto& player : game.players()) {
      player.set_team(player.hand().contains(Card::club_queen)
                      ? Team::re
                      : player.hand().contains(Card::unknown)
                      ? Team::unknown
                      : Team::contra);
    }
    this->all_known_ = false;
  } else if (game.type() == GameType::poverty) {
    // is done in 'this->poverty_shift()'

    this->all_known_ = false;
  } else if (game.type() == GameType::marriage) {
    this->teaminfo_ = vector<Team>(n, Team::maybe_contra);
    this->teaminfo_[game.players().soloplayer().no()] = Team::re;
  } else if (::is_solo(game.type())) {
    this->all_known_ = false;
    // set all teams to contra but of the soloplayer
    this->teaminfo_ = vector<Team>(n, Team::contra);
    this->teaminfo_[game.players().soloplayer().no()] = Team::re;
    this->all_known_ = true;
  } // if (this->type())
}

/** update the team information
 **/
void
GameTeamInfo::update()
{
  if (this->all_known_)
    return ;

  if (::game_status < GameStatus::game_play)
    return ;

  auto& game = *this->game_;
  auto const& rule = game.rule();
  auto const& tricks = game.tricks();
  auto const& current_trick = tricks.current();

  if (!current_trick.isstartcard()) {
    Player const& player
      = current_trick.player_of_card(current_trick.actcardno() - 1);
    auto const card = current_trick.card_of_player(player);
    auto const startcard = current_trick.startcard();

    if (!is_real(this->get(player))) {
      if (card == Card::club_queen) {
        this->set(player, Team::re);
      } else if (   startcard.istrump()
                 && (!card.istrump())
                 && !player.hand().contains(Card::club_queen)) {
        this->set(player, Team::contra);
      } else if (   (player.cards_to_play() == 0)
                 && !player.hand().contains(Card::club_queen)) {
        this->set(player, Team::contra);
      }

    } // if (team of 'player' unknown)
  } // if (!current_trick.isstartcard())

  { // count the played club queens for a silent marriage
    for (auto const& player : game.players()) {
      if (this->get(player) == Team::re) {
        // We only want to know, whether the player has played his second
        // club queen, so he must already be known as re.
        unsigned n = 0;
        for (auto const& trick : tricks) {
          if (trick.cardno_of_player(player) < trick.actcardno()) {
            if (trick.card_of_player(player) == Card::club_queen) {
              n += 1;
            }
          }
        } // for (t : this->tricks())

        if (n == rule(Rule::Type::number_of_same_cards)) {
          this->set_unknown(Team::contra);
          return ;
        } // if (both club queens played)
      } // if (this->teaminfo_[player] == Team::re)
    } // for (player)
  } // count the played club queens for a silent marriage

  if (   rule(Rule::Type::swine_only_second)
      && game.swines().swines_announced()) {
    // the swines owner has to say his team
    this->set(game.swines().swines_owner(),
              game.swines().swines_owner().team());
  }

  { // count the 're's and 'contra's
    auto const re_no = static_cast<unsigned>(count(this->teaminfo_, Team::re));
    auto const contra_no = static_cast<unsigned>(count(this->teaminfo_, Team::contra));
    auto const players_per_team = rule(Rule::Type::number_of_players_per_team);
    auto const playerno = rule(Rule::Type::number_of_players_in_game);
    if (re_no == players_per_team) {
      this->set_unknown(Team::contra);
      return ;
    }
    if (contra_no == playerno - 1) {
      this->set_unknown(Team::re);
      game.marriage().determine_silent_marriage();
      return ;
    }

    if (contra_no == playerno - 2) {
      if (!current_trick.isstartcard()) {
        Player const& player
          = current_trick.player_of_card(current_trick.actcardno() - 1);
        if (   (this->get(player) == Team::re)
            && current_trick.startcard().istrump()
            && !current_trick.card_of_player(player).istrump()) {

          // all players but two are contra, one is re and does not have
          // any more club queen
          // so the remaining player must be re
          this->set_unknown(Team::re);
        } // // if (the player is re but can play no more trump)
      } // if (!current_trick.isstartcard())
    } // if (contra_no == this->playerno() - 2)

    if (re_no + contra_no == playerno) {
      this->all_known_ = true;
    }
  } // count the 're's and 'contra's
} // void GameTeamInfo::update()

/** @return  the team information of player
 **/
Team
GameTeamInfo::get(Player const& player) const
{
  return this->teaminfo_[player.no()];
}

/** @return  the team information of player, known by the human
 **/
Team
GameTeamInfo::get_human_teaminfo(Player const& player) const
{
  if (this->game_->players().count_humans() == 1) {
    auto const team = this->game_->players().human().team_information().team(player);
    if (is_real(team))
      return team;
  }
  return this->get(player);
}

/** @return   all teaminfo
 **/
vector<Team> const&
GameTeamInfo::get_all() const noexcept
{
  return this->teaminfo_;
}

/** set the team information of player
 **/
void
GameTeamInfo::set(Player const& player, Team const team)
{
  DEBUG_ASSERTION(is_real(team),
                  "Game::set_teaminfo(player, team):\n"
                  "  'team' not valid: " << team);
  DEBUG_ASSERTION(::game_status >= GameStatus::game_play
                  && !this->game_->isvirtual()
                  && !this->game_->rule()(Rule::Type::genscher)
                  && is_real(team)
                  && is_real(player.team())
                  ? player.team() == team
                  : true,
                  "Game::set_teaminfo(player, team):\n"
                  "  'team' is false: " << team << " != " << player.team() << " = team of player");

  auto const n = player.no();
  if (teaminfo_[n] == team)
    return ;
  this->teaminfo_[n] = team;

  Game& game = *this->game_;
  if (player.team() == Team::unknown)
    game.players().player(player.no()).set_team(team);

  this->signal_changed_(player);
  if (!game.isvirtual()) {
    ::ui->teaminfo_changed(player);
  }

  this->update();
  for (auto& player : game.players())
    player.teaminfo_update();
}

/** set the unknown teams to 'team'
 **
 ** @param   team   team to set to
 **/
void
GameTeamInfo::set_unknown(Team const team) noexcept
{
  for (auto& t : this->teaminfo_) {
    if (!is_real(t)) {
      t = team;
    }
  }
  this->all_known_ = true;
}

/** @return   signal, when the teaminfo is changed
 **/
Signal<void(Player const&)>&
GameTeamInfo::signal_changed() noexcept
{
  return this->signal_changed_;
}

/** write the teaminfo into the output stream
 **/
void
GameTeamInfo::write(ostream& ostr) const
{
  for (auto const& player : this->game_->players())
    ostr << player.no() << ": " << this->get(player) << '\n';
}
