/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 * 
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "references_check.h"

#include "../game/game.h"
#include "../player/player.h"
#include "../card/trick.h"
#include "../os/bug_report_replay.h"

#include "../utils/file.h"

#include <fstream>
#include <cstring>

#ifdef BOOST
#define BOOST_FILESYSTEM_NO_DEPRECATED
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#endif

// the references check
unique_ptr<ReferencesCheck> references_check = {};

/** constructor
 **/
ReferencesCheck::ReferencesCheck(string const& references_dir) :
  report_file_("References.report.csv")
{
  this->set_main_directory(references_dir);
  this->next_reference();
  //this->statistics_.reset();
} // ReferencesCheck::ReferencesCheck(string references_dir)

/** destructor
 **/
ReferencesCheck::~ReferencesCheck()
{ 
  this->write_statistics();
}

/** set the main directory
 **
 ** @param    references_main_dir   the main directory with the references
 **/
void
ReferencesCheck::set_main_directory(string const& references_main_dir)
{
  this->references_main_dir_ = references_main_dir;
  this->set_directory(references_main_dir);
#ifdef BOOST
  { // create the statistics file
    ofstream ostr((this->references_main_dir_ + "/" + this->report_file_).c_str());
    if (ostr.fail()) {
      this->report_file_.clear();
      return ;
    } // if (ostr.fail())

    ostr << "file;"
      << "trick;"
      << "player;"
      << "discrepancy;"
      << "real action;"
      << "ref action;"
      << "comment\n";

    cout << "Created file '"
      << this->references_main_dir_ + "/" + this->report_file_
      << "' for the references report.\n";

    this->statistics_.reset();
  } // load the next reference and create the statistics file
#endif
} // void ReferencesCheck::set_main_directory(string references_main_dir)

/** set the directory
 ** adds the subdirectories, but skip the ones whith a file “_skip” in it
 **
 ** @param    references_dir   the directory with the references
 **/
void
ReferencesCheck::set_directory(string const& references_dir)
{
  if (!File::isdirectory(references_dir)) {
    cerr << "ReferencesCheck:set_directory()\n"
      << "'" << references_dir << "' is no directory. Aborting."
      << endl;
    exit(EXIT_FAILURE);
  }
  this->references_dir_ = references_dir;
  if (this->references_dir_.empty())
    return ;

#ifdef BOOST
  { // add the subdirectories
    for (this->dir_itr
         = boost::filesystem::directory_iterator(this->references_dir_);
         this->dir_itr != boost::filesystem::directory_iterator();
         ++this->dir_itr) {
      try {
        if (this->dir_itr->path().filename() == this->report_file_)
          continue;
        if (   boost::filesystem::is_directory(*this->dir_itr)
            && (this->dir_itr->path().filename().string()[0] != '.')) {
          if (!boost::filesystem::is_regular_file(this->dir_itr->path().string() + "/_skip"))
            this->references_subdir_.push_back(this->dir_itr->path().string());
        }
      } catch (std::exception const& ex) {
        cerr << dir_itr->path().filename() << " " << ex.what() << endl;
        this->references_dir_ = "";
      }
    } // for (this->dir_itr)
  } // add the subdirectories
#endif
} // void ReferencesCheck::set_directory(string references_dir)

/** open the next file
 **
 ** @return   whether another file was loaded
 **/
bool
ReferencesCheck::next_reference()
{
#ifdef BOOST
  if (this->dir_itr == boost::filesystem::directory_iterator())
    this->dir_itr = boost::filesystem::directory_iterator(this->references_dir_);
  else
    ++this->dir_itr;
  for (;
       this->dir_itr != boost::filesystem::directory_iterator();
       ++this->dir_itr) {
    try {
      if (this->dir_itr->path().filename() == this->report_file_) {
        continue;
      }
      if (!boost::filesystem::is_directory(*this->dir_itr)
          && (this->dir_itr->path().filename().string().size()
              >= strlen(".Reference.FreeDoko"))
          && (this->dir_itr->path().filename().string().substr(this->dir_itr->path().filename().string().size()
                                                               - strlen(".Reference.FreeDoko"))
              == ".Reference.FreeDoko")
         ) {
        if (this->load_reference()) {
          if  (   (::game_status == GameStatus::programstart)
               || (::game_status == GameStatus::party_initial_loaded)
               || (::game_status == GameStatus::party_loaded) ) {
            return true;
          } else {
            if (::party->gameno() > 0)
              throw GameStatus::party_loaded;
            else
              throw GameStatus::party_new;
          }
        }
        return false;
      }
    } catch (std::exception const& ex) {
      cerr << this->references_dir_ << "/" << dir_itr->path().filename() << " " << ex.what() << endl;
      this->references_dir_ = "";
    }
  } // for (this->dir_itr)

  { // go to the next subdir

    if (this->references_subdir_.empty()) {
      throw GameStatus::quit;
    }
    string const subdir = this->references_subdir_.front();
    this->references_subdir_.erase(this->references_subdir_.begin());
    this->set_directory(subdir);
    this->next_reference();
    return true;
  } // go to the next subdir
#endif

  return false;
} // bool ReferencesCheck::next_reference()

/** open the reference
 **
 ** @return   whether the loading was successful
 **/
bool
ReferencesCheck::load_reference()
{
#ifdef BOOST
  if (this->dir_itr == boost::filesystem::directory_iterator())
    return false;

  ::bug_report_replay
    = make_unique<OS_NS::BugReportReplay>(this->dir_itr->path().string(),
                                          OS_NS::BugReportReplay::verbose_none);
  // ToDo: check for auto action end
  if (!bug_report_replay->loaded()) {
    cerr << "Error loading the reference '" << this->dir_itr->path().string()
      << "'" << endl;
    if (!this->report_file_.empty()) {
      ofstream ostr((this->references_main_dir_ + "/"
                     + this->report_file_).c_str(),
                    std::ios::app);
      if (!ostr.fail())
        ostr << this->dir_itr->path().string() << ";loading error\n";
    } // if (output set)
    ::bug_report_replay = {};
    return false;
  }
#endif

  this->statistics_.no_files += 1;

  return true;
} // bool ReferencesCheck::load_reference()

/** check the action of the reference against the action of the game
 **
 ** @param    game           the game
 ** @param    game_action    the action from the game
 ** @param    check_action   the action to check
 ** @param    comment        the comment of the check
 ** @param    discrepancy    the discrepancy between the two actions
 ** @param    last_check     whether this is the last check
 **/
void
ReferencesCheck::check(Game const& game,
                       GameplayAction const& game_action,
                       GameplayAction const& check_action,
                       string const& comment,
                       GameplayActions::Discrepancy const& discrepancy,
                       bool const last_check)
{
  this->statistics_.no_checks += 1;
  if (!discrepancy) {
    this->statistics_.no_success += 1;
  }

  string output;
  { // create the result line
    // format:
    //   seed (trick, player); OK; ; real action; text
    //   seed (trick, player); discrepancy; real action; check action; text
    // semicolon separated list:
    // 1) seed
    // 2) trick
    // 3) player
    // 4) discrepancy
    // 5) real gameplay action / empty
    // 6) gameplay action from the reference
    // 7) comment (optional)
    ostringstream ostr;
#ifdef BOOST
    ostr
      << string(this->references_dir_ + "/",
                this->references_main_dir_.size() + 1)
      << std::setfill('0') << setw(6) << this->dir_itr->path().filename().string()
      << ';' << setw(2) << game.tricks().current_no() << ';'
      << player(check_action) << ';';
#endif
    if (!discrepancy)
      ostr << "OK";
    else
      ostr << discrepancy;
    ostr << ';' << to_string(game_action)
      << ';' << to_string(check_action);
    if (!comment.empty())
      ostr << ';' << comment;
    output = ostr.str();
  } // create the result line

  // print the result line
  cout << "check: " << output << endl;

  { // save the result line
    if (this->references_dir_.empty()
        && !this->report_file_.empty())
      //this->set_directory(".");
      this->next_reference();
    if (!this->report_file_.empty()) {
      ofstream ostr((this->references_main_dir_ + "/"
                     + this->report_file_).c_str(),
                    std::ios::app);
      if (ostr.fail()) {
        this->report_file_.clear();
        return ;
      } // if (ostr.fail())
      ostr << output << '\n';
    } // if (output set)
  } // save the result line

  if (discrepancy || last_check) {
    throw GameStatus::party_new;
    //exit(static_cast<int>(discrepancy));
  }
} // void ReferencesCheck::check(Game game, GameplayAction game_action, GameplayAction check_action, string comment, GameplayAction::Discrepancy discrepancy, bool last_check)

/** write the statistics
 **/
void
ReferencesCheck::write_statistics() const
{
  // print the statistics
  cout << '\n'
    << "reference checks\n"
    << "  files:   "
    << setw(static_cast<int>(ceil(log10(this->statistics_.no_checks))))
    << this->statistics_.no_files << '\n'
    << "  checks:  "
    << this->statistics_.no_checks << '\n'
    << "  success: "
    << setw(static_cast<int>(ceil(log10(this->statistics_.no_checks))))
    << this->statistics_.no_success << '\n'
    << "  failed:  "
    << setw(static_cast<int>(ceil(log10(this->statistics_.no_checks))))
    << (this->statistics_.no_checks
        - this->statistics_.no_success) << '\n'
    << '\n';

  { // save the result line
    if (this->references_dir_.empty()
        && !this->report_file_.empty()) {
      return ;
    }
    if (!this->report_file_.empty()) {
      ofstream ostr((this->references_main_dir_ + "/"
                     + this->report_file_).c_str(),
                    std::ios::app);
      if (ostr.fail()) {
        return ;
      } // if (ostr.fail())
      ostr << '\n'
        << "files;"
        << this->statistics_.no_files << '\n'
        << "checks;"
        << this->statistics_.no_checks << '\n'
        << "success;"
        << this->statistics_.no_success << '\n'
        << "failed;"
        << (this->statistics_.no_checks
            - this->statistics_.no_success) << '\n';
    } // if (output set)
  } // save the result line
} // ReferencesCheck::write_statistics() const
