/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 * 
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "setting.cardsorder.h"

#include "../utils/string.h"

#include "../party/party.h"
#include "../game/game.h"

#include "../ui/ui.h"

/** default constructor
 **
 ** @param    setting   corresponding setting
 **/
Setting::CardsOrder::CardsOrder(Setting& setting) :
  setting_(&setting),
  sorted_(false),
  pos_(Card::number_of_tcolors, Card::nocardcolor),
  direction_(Card::number_of_tcolors, Direction::down)
{
  this->pos_[Card::trump]   = 0;
  this->pos_[Card::club]    = 1;
  this->pos_[Card::heart]   = 2;
  this->pos_[Card::spade]   = 3;
  this->pos_[Card::diamond] = 4;
} // Setting::CardsOrder::CardsOrder(Setting& setting)

/** constructor with the order
 **
 ** @param    setting   corresponding setting
 ** @param    c0   first color
 ** @param    d0   direction of the first
 ** @param    c1   second color
 ** @param    d1   direction of the second
 ** @param    c2   third color
 ** @param    d2   direction of the third
 ** @param    c3   fourth color
 ** @param    d3   direction of the fourth
 ** @param    c4   fifth color
 ** @param    d4   direction of the fifth
 **/
Setting::CardsOrder::CardsOrder(Setting& setting,
                                Card::TColor c0, Direction d0,
                                Card::TColor c1, Direction d1,
                                Card::TColor c2, Direction d2,
                                Card::TColor c3, Direction d3,
                                Card::TColor c4, Direction d4) :
  setting_(&setting),
  sorted_(true),
  pos_(Card::number_of_tcolors, Card::nocardcolor),
  direction_(Card::number_of_tcolors, Direction::down)
{
  // the colors should be different
  this->pos_[c0] = 0;
  this->pos_[c1] = 1;
  this->pos_[c2] = 2;
  this->pos_[c3] = 3;
  this->pos_[c4] = 4;

  this->direction_[c0] = d0;
  this->direction_[c1] = d1;
  this->direction_[c2] = d2;
  this->direction_[c3] = d3;
  this->direction_[c4] = d4;

  this->ensure_validity();

  return ;
} // Setting::CardsOrder::CardsOrder(TColor, Direction, ...)

/** constructor from a string
 **
 ** @param    setting   corresponding setting
 ** @param    value   string representation of the order
 **/
Setting::CardsOrder::CardsOrder(Setting& setting, string const& value) :
  setting_(&setting),
  sorted_(true),
  pos_(Card::number_of_tcolors, Card::nocardcolor),
  direction_(Card::number_of_tcolors, Direction::down)
{
  // *** should work better, when you can get words of the string
  // *** (expand the string-class)
  string remain_string = value;
  string word; // the word of the string
  Card::TColor tcolor = Card::nocardcolor;

  String::remove_blanks(remain_string);
  for (unsigned i = 0; i < Card::number_of_tcolors; i++) {
    if (remain_string.size() == 0)
      break;
    // get the color
    word = string(remain_string, 0, remain_string.find(' '));
    if (remain_string.find(' ') == string::npos)
      word = remain_string;
    if ((word == to_string(Card::trump))
        || (word == _(Card::trump)))
      tcolor = Card::trump;
    else if ((word == to_string(Card::club))
             || (word == _(Card::club)))
      tcolor = Card::club;
    else if ((word == to_string(Card::spade))
             || (word == _(Card::spade)))
      tcolor = Card::spade;
    else if ((word == to_string(Card::heart))
             || (word == _(Card::heart)))
      tcolor = Card::heart;
    else if ((word == to_string(Card::diamond))
             || (word == _(Card::diamond)))
      tcolor = Card::diamond;
    else if ((i == 0)
             && ((word == "unsorted")
                 || (word == _("Setting::CardsOrder::unsorted"))
                 || (word == "random")
                 || (word == _("Setting::CardsOrder::random"))
                 || (word == "mixed")
                 || (word == _("Setting::CardsOrder::mixed"))
                )
            ) {
      // random order
      this->sorted_ = false;
      break;
    } else {
      cerr << "Unknown TColor \"" << word << "\"" << endl;
      break ;
    }

    if (!this->sorted())
      break;

    Direction direction;

    // remove the used word from the string
    if (remain_string.length() == 0)
      break;
    remain_string = string(remain_string, word.length(), std::string::npos);
    String::remove_blanks(remain_string);

    // get the direction
    word = string(remain_string, 0, remain_string.find(' '));
    if (remain_string.find(' ') == string::npos)
      word = remain_string;
    if ((word == to_string(Direction::up))
        || (word == _(Direction::up))
        || (word == to_string(Direction::up) + ",")
        || (word == _(Direction::up) + ","))
      direction = Direction::up;
    else if ((word == to_string(Direction::down))
             || (word == _(Direction::down))
             || (word == to_string(Direction::down) + ",")
             || (word == _(Direction::down) + ","))
      direction = Direction::down;
    else {
      cerr << "Unknown direction \"" << word << "\"" << endl;
      break ;
    }

    // removed the used word from the string
    remain_string = string(remain_string, word.length(), std::string::npos);
    String::remove_blanks(remain_string);

    // set the position of the tcolor and the direction
    this->pos_[tcolor] = i;
    this->direction_[tcolor] = direction;
  } // for (i < Card::number_of_tcolors)

  this->ensure_validity();
} // Setting::CardsOrder::CardsOrder(string value)

/** copy constructor
 **
 ** @param    cards_order   cards order to copy
 **/
Setting::CardsOrder::CardsOrder(CardsOrder const& cards_order) = default;

/** copy operator
 ** all but the settings reference
 **
 ** @param    cards_order   cards order to copy
 **/
Setting::CardsOrder&
Setting::CardsOrder::operator=(CardsOrder const& cards_order)
{
  this->sorted_ = cards_order.sorted_;
  this->pos_ = cards_order.pos_;
  this->direction_ = cards_order.direction_;
  return *this;
} // CardsOrder& CardsOrder::operator=(CardsOrder cards_order)

/** ensure the validity of the card order, t.i. each color has a different position
 **
 ** @return   whether the sorting was valid and each direction is set
 **/
bool
Setting::CardsOrder::ensure_validity()
{
  bool result = true;
  { // check for double position
    for (unsigned c = 0; c < Card::number_of_tcolors; c++) {
      if (   (this->tcolor(c) == Card::nocardcolor)
          || (this->pos(this->tcolor(c)) != c) ) {
        result = false;
        break;
      }
    }
    if (!result) {
      this->pos_[Card::trump]   = 0;
      this->pos_[Card::club]    = 1;
      this->pos_[Card::heart]   = 2;
      this->pos_[Card::spade]   = 3;
      this->pos_[Card::diamond] = 4;
      this->direction_[Card::trump]   = Direction::up;
      this->direction_[Card::club]    = Direction::down;
      this->direction_[Card::spade]   = Direction::down;
      this->direction_[Card::heart]   = Direction::down;
      this->direction_[Card::diamond] = Direction::down;
    }
  } // check for double position
  { // check directions
    if (this->direction(Card::trump) == Direction::none) {
      this->direction_set(Card::trump, Direction::up);
      result = false;
    }
    if (this->direction(Card::club) == Direction::none) {
      this->direction_set(Card::club, Direction::down);
      result = false;
    }
    if (this->direction(Card::spade) == Direction::none) {
      this->direction_set(Card::spade, Direction::down);
      result = false;
    }
    if (this->direction(Card::heart) == Direction::none) {
      this->direction_set(Card::heart, Direction::down);
      result = false;
    }
    if (this->direction(Card::diamond) == Direction::none) {
      this->direction_set(Card::diamond, Direction::down);
      result = false;
    }
  } // check directions
  return result;
} // bool Setting::CardsOrder::ensure_validity()

/** @return   whether the cards are sorted
 **/
Setting::CardsOrder::operator bool() const
{
  return this->sorted_;
} // Setting::CardsOrder::operator bool()

/** @return   the order as a string
 **/
Setting::CardsOrder::operator string() const
{
  string value;

  if (this->sorted_) {
    for (unsigned c = 0; c < Card::number_of_tcolors; c++) {
      value += to_string(this->tcolor(c)) + " " + to_string(this->direction(c));
      if (c + 1 < Card::number_of_tcolors)
        value += ", ";
    } // for (c < Card::number_of_tcolors))
  } else { // if !(this->sorted())
    value = "mixed";
  } // if !(this->sorted())

  return value;
} // Setting::CardsOrder::operator string()

/** -> result
 **
 ** @param    order   order to compare with
 **
 ** @return   whether 'this' is equal to 'order'
 **/
bool
Setting::CardsOrder::operator==(CardsOrder const& order) const
{
  if (this->sorted() != order.sorted())
    return false;

  for(unsigned i = 0; i < Card::number_of_tcolors; i++)
    if ( (this->tcolor(i) != order.tcolor(i))
        || (this->direction(i) != order.direction(i)) )
      return false;

  return true;
} // bool Setting::CardsOrder::operator==(CardsOrder order) const

/** @return   whether the cards are sorted
 **/
bool
Setting::CardsOrder::sorted() const
{
  return this->sorted_;
} // bool Setting::CardsOrder::sorted() const

/** @return   whether the cards are mixed
 **/
bool
Setting::CardsOrder::mixed() const
{
  return !this->sorted_;
} // bool Setting::CardsOrder::mixed() const

/** -> result
 **
 ** @param    a   first card
 ** @param    b   second card
 **
 ** @return   the relative position of b to a,
 **            t.i. -1, if a < b, 0 if a == b, 1, if a > b
 **/
int
Setting::CardsOrder::relative_position(HandCard const& a, HandCard const& b) const
{
  if (this->pos(a.tcolor()) < this->pos(b.tcolor()))
    return -1;
  if (this->pos(a.tcolor()) > this->pos(b.tcolor()))
    return 1;

  // the two cards are the same color
  if (a == b)
    return 0;
  if (a.less(b))
    return ((this->direction(a.tcolor()) == Direction::up) ? -1 :  1);
  else
    return ((this->direction(a.tcolor()) == Direction::up) ?  1 : -1);
} // int Setting::CardsOrder::relative_position(HandCard a, HandCard b) const

/** -> result
 **
 ** @param    tcolor   trump color
 **
 ** @return   the direction of the tcolor
 **/
Setting::CardsOrder::Direction
Setting::CardsOrder::direction(Card::TColor const tcolor) const
{
  return this->direction_[tcolor];
} // Setting::CardsOrder::Direction CardsOrder::direction(Card::TColor tcolor) const

/** -> result
 **
 ** @param    pos_a   position of the color
 **
 ** @return   the direction of the color at position 'pos_a'
 **/
Setting::CardsOrder::Direction
Setting::CardsOrder::direction(unsigned const pos_a) const
{
  for (unsigned tcolor = 0; tcolor < Card::number_of_tcolors; tcolor++)
    if (this->pos_[tcolor] == pos_a) {
      return this->direction_[tcolor];
    }

  return Direction::none;
} // Setting::CardsOrder::Direction CardsOrder::direction(unsigned pos_a) const

/** -> result
 **
 ** @param    tcolor   trump color
 **
 ** @return   the position of the tcolor
 **/
unsigned
Setting::CardsOrder::pos(const Card::TColor tcolor) const
{
  if (tcolor == Card::unknowncardcolor)
    return 101;
  return this->pos_[tcolor];
} // unsigned Setting::CardsOrder::pos(Card::TColor tcolor) const

/** -> result
 **
 ** @param    pos_a   position
 **
 ** @return   the tcolor at the position 'pos_a'
 **/
Card::TColor
Setting::CardsOrder::tcolor(unsigned const pos_a) const
{
  for (unsigned tcolor = 0; tcolor < 5; tcolor++)
    if (this->pos_[tcolor] == pos_a) {
      return Card::TColor(tcolor);
    }

  return Card::nocardcolor;
} // Card::TColor Setting::CardsOrder::tcolor(unsigned pos_a) const

/** toggle the direction of 'tcolor'
 **
 ** @param    tcolor   tcolor to toggle the direction of
 **/
void
Setting::CardsOrder::direction_toggle(const Card::TColor tcolor)
{
  auto const old_value = *this;

  this->direction_[tcolor] = ( (direction(tcolor) == Direction::up)
                              ? Direction::down
                              : Direction::up);

  // resort the hands
  if (game_status & GameStatus::game)
    ::party->game().players().sort_hands();

  this->setting_->signal_changed_(Type::cards_order, &old_value);
} // void Setting::CardsOrder::direction_toggle(Card::TColor tcolor)

/** toggle the direction of the color at 'pos_a'
 **
 ** @param    pos_a   position
 **/
void
Setting::CardsOrder::direction_toggle(unsigned const pos_a)
{
  auto const old_value = *this;

  for (unsigned tcolor = 0; tcolor < 5; tcolor++)
    if (this->pos_[tcolor] == pos_a)
      this->direction_[tcolor] = ( (direction(pos_a) == Direction::up)
                                  ? Direction::down
                                  : Direction::up);

  // resort the hands
  if (game_status & GameStatus::game)
    ::party->game().players().sort_hands();

  this->setting_->signal_changed_(Type::cards_order, &old_value);
} // void Setting::CardsOrder::direction_toggle(unsigned const pos_a)

/** sets the direction of the given color
 **
 ** @param    tcolor      color
 ** @param    direction   new direction
 **
 ** @return   direction of the color on the given position
 **/
void
Setting::CardsOrder::direction_set(Card::TColor const tcolor,
                                   Direction const direction)
{
  auto const old_value = *this;

  if (this->direction(tcolor) == direction)
    return ;

  this->direction_[tcolor] = direction;

  // resort the hands
  if (game_status & GameStatus::game)
    ::party->game().players().sort_hands();

  this->setting_->signal_changed_(Type::cards_order, &old_value);
} // void Setting::CardsOrder::direction_set(Card::TColor tcolor, Direction direction)

/** sets the direction at the given position
 **
 ** @param    pos         position
 ** @param    direction   new direction
 **
 ** @return   new direction
 **/
void
Setting::CardsOrder::direction_set(unsigned const pos,
                                   Direction const direction)
{
  this->direction_set(this->tcolor(pos), direction);
} // Setting::CardsOrder::direction_set(unsigned pos, Direction direction)

/** pushes the color on 'p' one position to the left
 **
 ** @param    pos_a   position
 **
 ** @return   the moved color
 **/
void
Setting::CardsOrder::pos_to_left(unsigned const pos_a)
{
  auto const old_value = *this;

  unsigned c;
  for (c = 0; c < Card::number_of_tcolors; c++) {
    if (this->pos(Card::TColor(c)) == pos_a)
      this->pos_[c] = (pos_a - 1 + Card::number_of_tcolors) % Card::number_of_tcolors;
    else if ( ((pos(Card::TColor(c)) + Card::number_of_tcolors)
               % Card::number_of_tcolors)
             == (pos_a - 1 + Card::number_of_tcolors) % Card::number_of_tcolors)
      this->pos_[c] = pos_a;
  }

  // resort the hands
  if (game_status & GameStatus::game)
    ::party->game().players().sort_hands();

  this->setting_->signal_changed_(Type::cards_order, &old_value);
} // void Setting::CardsOrder::pos_to_left(unsigned pos_a)

/** pushes the color on 'p' one position to the right
 **
 ** @param    pos_a   position
 **
 ** @return   the moved color
 **/
void
Setting::CardsOrder::pos_to_right(unsigned const pos_a)
{
  this->pos_to_left((pos_a + 1) % Card::number_of_tcolors);
} // Setting::CardsOrder::pos_to_right(unsigned pos_a)

/** set whether the cards are to be sorted
 **
 ** @param    sorted   whether the cards are to be sorted
 **
 ** @return   whether the cards are sorted
 **/
void
Setting::CardsOrder::sorted_set(bool const sorted)
{
  if (this->sorted_ == sorted)
    return ;

  auto const old_value = *this;

  this->sorted_ = sorted;

  if (this == &(::setting(Type::cards_order))) {
    // resort the hands
    if (game_status & GameStatus::game) {
      auto& players = ::party->game().players();
      if (this->sorted())
        players.sort_hands();
      else
        players.mix_hands();
    }

    this->setting_->signal_changed_(Type::cards_order, &old_value);
  } // if (this == &(::setting(Type::cards_order)))
} // void Setting::CardsOrder::sorted_set(bool sorted)

/** mix the cards
 **/
void
Setting::CardsOrder::mix()
{
  auto const old_value = *this;

  this->sorted_ = false;

  if (this == &(::setting(Type::cards_order))) {
    // resort the hands
    if (game_status & GameStatus::game)
      if (!this->sorted())
        ::party->game().players().mix_hands();

    this->setting_->signal_changed_(Type::cards_order, &old_value);
  } // if (this == &(::setting(Type::cards_order)))
} // void Setting::CardsOrder::mix()

/** -> result
 **
 ** @param    direction   the direction
 **
 ** @return   name of the direction
 **/
string
to_string(Setting::CardsOrder::Direction const direction)
{
  switch(direction) {
  case Setting::CardsOrder::Direction::none:
    return "none";
    (void)_("Direction::none");
  case Setting::CardsOrder::Direction::up:
    return "up";
    (void)_("Direction::up");
  case Setting::CardsOrder::Direction::down:
    return "down";
    (void)_("Direction::down");
  } // switch(direction)

  return "";
} // string to_string(Setting::CardsOrder::Direction direction)

/** -> result
 **
 ** @param    direction   direction to translate
 **
 ** @return   translation of the direction
 **/
string
gettext(Setting::CardsOrder::Direction const direction)
{
  return gettext("Direction::" + to_string(direction));
} // string gettext(Setting::CardsOrder::Direction direction);

/** print the cards order
 **
 ** @param    ostr   stream to write into
 **/
void
Setting::CardsOrder::write(ostream& ostr) const
{
  if (this->sorted()) {
    ostr << this->tcolor(0) << " " << to_string(this->direction(0));
    for (unsigned i = 1; i < Card::number_of_tcolors; i++)
      ostr << ", " << this->tcolor(i) << " " << to_string(this->direction(i));
  } else { // if !(this->sorted())
    ostr << "mixed";
  } // if !(this->sorted())
} // void Setting::CardsOrder::write(ostream& ostr) const

/** @param    ostr          stream to write into
 ** @param    cards_order   cards order to write
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Setting::CardsOrder const& cards_order)
{
  cards_order.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, Setting::CardsOrder cards_order)
