/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "gameplay.h"

#include "../party/party.h"
#include "../party/rule.h"
#include "../game/game.h"
#include "../game/game_summary.h"
#include "../card/trick.h"

namespace OS_NS {

/** constructor
 **
 ** @param    ostr   output stream
 **/
Gameplay::Gameplay(ostream* ostr) :
  ostr(ostr)
{
  Party::signal_open().connect(*this, &Gameplay::party_open, this->disconnector_);
  Game::signal_open().connect(*this, &Gameplay::game_open, this->disconnector_);
}

/** destructor
 **/
Gameplay::~Gameplay() = default;

/** the party is opened
 **
 ** @param    party   party that is opened
 **/
void
Gameplay::party_open(Party const& party)
{
  this->party_ = &party;
  ::party->signal_start().connect(*this, &Gameplay::party_start, this->disconnector_);
  ::party->signal_finish().connect(*this, &Gameplay::party_finish, this->disconnector_);
  ::party->signal_close().connect(*this, &Gameplay::party_close, this->disconnector_);
  *this->ostr << "party open\n";
} // void Gameplay::party_open(Party party)

/** the party is started
 **/
void
Gameplay::party_start()
{ 
  *this->ostr << "party start\n";
  //*this->ostr << ::party.rule();
} // void Gameplay::party_start()

/** the party is finished
 **/
void
Gameplay::party_finish()
{
  *this->ostr << "party finish\n";
} // void Gameplay::party_finish()

/** the party is closed
 **/
void
Gameplay::party_close()
{
  this->party_ = {};
  *this->ostr << "party close\n";
} // void Gameplay::party_close()

/** the game is opened
 **
 ** @param    game   game that is opened
 **/
void
Gameplay::game_open(Game const& game)
{
  this->game_ = &game;
  {
    auto& game = ::party->game();
    game.signal_cards_distributed().connect(*this, &Gameplay::game_cards_distributed, this->disconnector_);
    game.signal_redistribute().connect(*this, &Gameplay::game_redistribute, this->disconnector_);
    game.signal_reservation_got().connect(*this, &Gameplay::reservation_got, this->disconnector_);
    game.signal_start().connect(*this, &Gameplay::game_start, this->disconnector_);
    game.signal_finish().connect(*this, &Gameplay::game_finish, this->disconnector_);
    game.signal_close().connect(*this, &Gameplay::game_close, this->disconnector_);
    game.tricks().signal_trick_open().connect(*this, &Gameplay::trick_open, this->disconnector_);
    game.tricks().signal_trick_full().connect(*this, &Gameplay::trick_full, this->disconnector_);
    game.signal_card_played().connect(*this, &Gameplay::card_played, this->disconnector_);
    game.announcements().signal_announcement_made().connect(*this, &Gameplay::announcement_made, this->disconnector_);
    game.swines().signal_swines_announced().connect(*this, &Gameplay::swines_announced, this->disconnector_);
    game.swines().signal_hyperswines_announced().connect(*this, &Gameplay::hyperswines_announced, this->disconnector_);
    game.poverty().signal_shift().connect(*this, &Gameplay::poverty_shift, this->disconnector_);
    game.poverty().signal_take_denied_by_all().connect(*this, &Gameplay::poverty_take_denied_by_all, this->disconnector_);
    game.poverty().signal_take_accepted().connect(*this, &Gameplay::poverty_take_accepted, this->disconnector_);
    game.marriage().signal_marriage().connect(*this, &Gameplay::marriage, this->disconnector_);
    game.marriage().signal_genscher().connect(*this, &Gameplay::genscher, this->disconnector_);
  }
  *this->ostr << "game open: " << this->party_->gameno() << '\n';
  *this->ostr << "seed: " << game.seed() << '\n';
} // void Gameplay::game_open(Game game)

/** the cards are distributed
 **/
void
Gameplay::game_cards_distributed()
{
  *this->ostr << "cards distributed\n";
} // void Gameplay::game_cards_distributed()

/** the game is redistributed
 **/
void
Gameplay::game_redistribute()
{
  *this->ostr << "game redistribute\n";
} // void Gameplay::game_redistribute()

/** a reservation has been made
 **/
void
Gameplay::reservation_got(Reservation const& reservation,
                          Player const& player)
{
  *this->ostr << "reservation: "
    << player.no() << " " << player.name() << ": "
    << reservation;
} // void Gameplay::reservation_got(Reservation reservation, Player player)

/** the game is started
 **
 ** @param    game   game that is started
 **/
void
Gameplay::game_start()
{
  auto const& game = *this->game_;
  *this->ostr << "game start: " << game.type() << "\n";
  if (is_solo(game.type())) {
    *this->ostr << "Soloplayer: "
      << game.players().soloplayer().no()
      << "\n";
  }
} // void Gameplay::game_start()

/** the game is finished
 **
 ** @param    game   game that is finished
 **/
void
Gameplay::game_finish()
{ 
  *this->ostr << "game finish\n";
  *this->ostr << this->party_->last_game_summary();
} // void Gameplay::game_finish()

/** the game is closed
 **
 ** @param    game   game that is closed
 **/
void
Gameplay::game_close()
{
  this->game_ = {};
  *this->ostr << "game close\n";
} // void Gameplay::game_close()

/** 'player' has played 'card'
 **
 ** @param    card   the played card
 **/
void
Gameplay::card_played(HandCard const& card)
{ 
  *this->ostr << this->game_->tricks().current().actcardno() << ". "
    << card.player().name() << ": "
    << card << "\n";
} // void Gameplay::card_played(HandCard const& card)

/** the announcement 'announcement' has been made by player 'player'
 **
 ** @param    announcement   the announcement
 ** @param    player      the player, who has made the announcement
 **/
void
Gameplay::announcement_made(Announcement const announcement,
                            Player const& player)
{
  *this->ostr << "announcement: "
    << "player " << player.no() << ": "
    << announcement
    << "\n";
} // void Gameplay::announcement_made(Announcement announcement, Player player)

/** the player has swines
 **
 ** @param    player      the player with the swines
 **/
void
Gameplay::swines_announced(Player const& player)
{
  *this->ostr << "swines announced: "
    << "player " << player.no()
    << "\n";
} // void Gameplay::swines_announced(Player player)

/** the player has hyperswines
 **
 ** @param    player      the player with the swines
 **/
void
Gameplay::hyperswines_announced(Player const& player)
{
  *this->ostr << "hyperswines announced: "
    << "player " << player.no()
    << "\n";
} // void Gameplay::hyperswines_announced(Player player)

/** 'player' shifts 'cardno' cards
 **
 ** @param    player   the player who shifts the cards
 ** @param    cardno   the number of cards that are shifted
 **/
void
Gameplay::poverty_shift(Player const& player, unsigned const cardno)
{ 
  *this->ostr << "poverty: "
    << "player " << player.no()
    << " shifts " << cardno << " cards.\n";
} // void Gameplay::poverty_shift(Player player, unsigned cardno)

/** all players have denied to take the cards
 **/
void
Gameplay::poverty_take_denied_by_all()
{ 
  *this->ostr << "poverty: "
    << "denied by all players\n";
} // void Gameplay::poverty_take_denied_by_all()

/** 'player' accepts to take the shifted cards
 ** and returns 'cardno' cards with 'trumpno' trumps
 **
 ** @param    player   the player who has denied to take the cards
 ** @param    cardno   number of cards that are given back
 ** @param    trumpno   number of trumps of the cards
 **/
void
Gameplay::poverty_take_accepted(Player const& player,
                                unsigned const cardno,
                                unsigned const trumpno)
{ 
  *this->ostr << "poverty: "
    << "player " << player.no() << " accepts poverty and returns "
    << cardno << " cards with " << trumpno << " trumps.\n";
} // void Gameplay::poverty_take_accepted(Player player, unsigned cardno, unsigned trumpno)

/** marriage: Information of the new team
 **
 ** @param    bridegroom   the player with the marriage
 ** @param    bride      the bride
 **/
void
Gameplay::marriage(Player const& bridegroom,
                   Player const& bride)
{
  *this->ostr << "Marriage: "
    << "player " << bridegroom.no()
    << " with player " << bridegroom.no()
    << '\n';
} // void Gameplay::marriage(Player bridegroom, Player bride)

/** genscher: Information of the new team
 **
 ** @param    player   genscher player
 ** @param    partner   partner of the gensher player
 **/
void
Gameplay::genscher(Player const& player,
                   Player const& partner)
{
  *this->ostr << "Genscher: "
    << "player " << player.no()
    << " with player " << player.no()
    << '\n';
} // void Gameplay::genscher(Player player, Player partner)

/** 'trick' has been opened
 **
 ** @param    trick   opened trick
 **/
void
Gameplay::trick_open(Trick const& trick)
{
  *this->ostr << "\n"
    << "Trick: " << this->game_->tricks().current_no() << "\n";
} // void Gameplay::trick_open(Trick trick)

/** the trick is full
 **
 ** @param    trick   full trick
 **/
void
Gameplay::trick_full(Trick const& trick)
{
  *this->ostr << "Winner: "
    << trick.winnerplayer().name()
    << "\n";
} // void Gameplay::trick_full(Trick trick)

} // namespace OS_NS
