/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#pragma once

#include "../basetypes.h"

#include "players.h"
#include "../player/player.h"
class Rule;
class GamePointsTable;
class GameSummary;
class Human;


class Party;
extern unique_ptr<Party> party;

using PartyPlayer = Player;

class Party {
  friend class Game;
  public:
  using Player = PartyPlayer;
  using Players = PartyPlayers;

  Party();
  Party(Party const&) = delete;
  Party& operator=(Party const&) = delete;

  void init();

  // signals
  static Signal<void(Party&)>& signal_open() noexcept;
  Signal<void()>& signal_get_settings() noexcept;
  Signal<void()>& signal_loaded() noexcept;
  Signal<void()>& signal_start() noexcept;
  Signal<void(unsigned)>& signal_start_round() noexcept;
  Signal<void()>& signal_finish() noexcept;
  Signal<void()>& signal_close() noexcept;
  Signal<void()>& signal_seed_changed() noexcept;
  Signal<void()>& signal_startplayer_changed() noexcept;

  Rule const& rule() const noexcept;
  Rule& rule() noexcept;

  Players const& players() const noexcept;
  Players& players() noexcept;

  bool random_seed() const noexcept;
  Seed seed() const noexcept;
  void set_seed(Seed seed) noexcept;
  void set_random_seed() noexcept;
  Seed next_seed() noexcept;

  bool random_startplayer() const noexcept;
  unsigned startplayer() const noexcept;
  void set_startplayer(unsigned const startplayer) noexcept;
  void set_random_startplayer() noexcept;

  int points() const noexcept;

  Game const& game() const noexcept;
  Game& game() noexcept;

  void set_replay_game() noexcept;
  bool is_replayed_game() const noexcept;

  // whether the party can be configured (seed, startplayer)
  bool in_configuration() const noexcept;

  unsigned roundno() const noexcept;
  unsigned remaining_rounds() const noexcept;
  unsigned remaining_normal_games() const noexcept;
  unsigned remaining_games() const noexcept;
  unsigned gameno() const noexcept;
  unsigned finished_games() const noexcept;
  bool is_duty_soli_round() const noexcept;
  bool is_duty_soli_round(unsigned round) const noexcept;
  bool is_last_game() const noexcept;
  bool is_finished() const noexcept;

  // the remaining points for the party
  int remaining_points() const noexcept;

  bool starts_new_round(unsigned const gameno) const noexcept;
  unsigned round_of_game(unsigned const gameno) const noexcept;
  unsigned round_startgame(unsigned const roundno) const noexcept;

  void add_game_summary(unique_ptr<GameSummary> game_summary);
  GameSummary const& last_game_summary() const noexcept;
  GameSummary const& game_summary(unsigned const gameno) const;

  vector<unsigned> const& bock_multipliers() const noexcept;
  // the bock multiplier for the current game
  unsigned current_bock_multiplier() const;
  // the bock multiplier for the game gameno
  unsigned bock_multiplier(unsigned const gameno) const noexcept;


  int pointsum() const;
  int pointsum_till_game(unsigned const gameno) const;
  int pointsum_till_round(unsigned const roundno) const;
  int pointsum_in_round(unsigned const roundno) const;
  int pointsum(Player const& player) const;
  int pointsum(unsigned const playerno) const;
  int pointsum_till_game(unsigned const gameno, unsigned const playerno) const;
  int pointsum_till_round(unsigned const roundno, unsigned const playerno) const;
  int pointsum_in_round(unsigned const roundno, unsigned const playerno) const;

  unsigned rang(unsigned const playerno) const;
  unsigned rang(Player const& player) const;

  unsigned played_duty_soli(unsigned const playerno) const;
  unsigned played_duty_soli(Player const& player) const;
  unsigned played_duty_color_soli(unsigned const playerno) const;
  unsigned played_duty_color_soli(Player const& player) const;
  unsigned played_duty_picture_soli(unsigned const playerno) const;
  unsigned played_duty_picture_soli(Player const& player) const;
  unsigned played_duty_free_soli(unsigned const playerno) const;
  unsigned played_duty_free_soli(Player const& player) const;
  unsigned played_lust_soli(unsigned const playerno) const;
  unsigned played_lust_soli(Player const& player) const;

  unsigned remaining_duty_soli() const;
  unsigned remaining_duty_soli(unsigned const playerno) const;
  unsigned remaining_duty_soli(Player const& player) const;
  unsigned remaining_duty_color_soli(unsigned const playerno) const;
  unsigned remaining_duty_color_soli(Player const& player) const;
  unsigned remaining_duty_picture_soli(unsigned const playerno) const;
  unsigned remaining_duty_picture_soli(Player const& player) const;
  unsigned remaining_duty_free_soli(unsigned const playerno) const;
  unsigned remaining_duty_free_soli(Player const& player) const;

  void write(ostream& ostr) const;
  void write_pointstable(ostream& ostr) const;
  bool read(istream& istr);

  // save the party
  bool save(string const& filename) const;
  // load the party
  bool load(string const& filename);
  // load the players
  bool load_players();


  // open the party
  void open();
  // reset the party
  void reset();
  // play the party (main program flow)
  void play();
  // play a game
  void play_game();
  // close the party
  void close();

  // change of the rules
  void rule_changed(int const type, void const* const old_value);

  private:
  // save the party automatically
  bool auto_save() const;
  public:
  void remove_auto_save() const;
  // save the changes of the party (t.i. rules and players)
  void save_changes() const;

  private:
  // increment the round
  void increment_round();

  private:
  // rules for the party
  unique_ptr<Rule> rule_;
  // players
  unique_ptr<Players> players_;
  // starting seed
  bool random_seed_ = true;
  Seed first_seed_ = 0;
  // starting startplayer
  bool random_startplayer_ = false;
  unsigned first_startplayer_ = 0;
  // the game in which the round starts
  vector<unsigned> round_startgame_;
  // Anzahl der gespielten normalen Spiele;
  unsigned finished_normal_games_ = 0;

  // the number of the duty soli round
  optional<unsigned> duty_soli_round_;

  // the points for the party
  int points_;
  // current game
  unique_ptr<Game> game_;
  // the current seed
  Seed seed_ = 0;
  // the current startplayer
  unsigned startplayer_ = 0;
  // whether to replay the last game
  bool replay_game_ = false;
  // whether this is a replayed game
  bool is_replayed_game_ = false;

  // game summaries
  vector<unique_ptr<GameSummary>> game_summaries_;
  // the multipliers for bock games
  vector<unsigned> bock_multipliers_;

  // set_ai_players
  // partie typ (tutorial...)
  // pflichtstsoli
  // punkte topf
  // tunier status

  // Signals
  static Signal<void(Party&)> signal_open_;
  Signal<void()> signal_get_settings_;
  Signal<void()> signal_loaded_;
  Signal<void()> signal_start_;
  Signal<void(unsigned)> signal_start_round_;
  Signal<void()> signal_finish_;
  Signal<void()> signal_close_;
  Signal<void()> signal_seed_changed_;
  Signal<void()> signal_startplayer_changed_;
}; // class Party

ostream& operator<<(ostream& ostr, Party const& p);
