/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"
#include "players.h"
#include "party.h"
#include "rule.h"

#include "../player/human/human.h"
#include "../ui/ui.h"
#include "../misc/setting.h"
#include "../utils/string.h"

/** constructor:
 ** load the players from the private data directory
 **
 ** @param    party   corresponding party
 **/
PartyPlayers::PartyPlayers(Party& party) :
  party_(&party)
{
  (void)party_;
  this->players_.push_back(Player::create(Player::Type::human));
  while (this->size() < party.rule()(Rule::Type::number_of_players)) {
    this->players_.push_back(Player::create(Player::Type::ai));
  }
} // PartyPlayers::PartyPlayers()

/** constructor:
 ** load the players from the stream
 **
 ** @param    party   corresponding party
 ** @param    istr    stream to read from
 **/
PartyPlayers::PartyPlayers(Party& party, istream& istr) :
  party_(&party)
{
  while (istr.good()
         && istr.peek() != '}') {
    auto player = Player::create(istr);
    if (!player)
      throw ReadException("Party Players read: read of player failed");

  this->players_.push_back(move(player));

    // overread empty lines
    while (   (istr.peek() == '\n')
           || (istr.peek() == '\r'))
      istr.get();
  } // while (istr.peek() != '}')
} // PartyPlayers::PartyPlayers(Party& party, istream& istr)

/** @return   signal: two players have swapped
 **/
Signal<void(Player const&, Player const&)>&
PartyPlayers::signal_players_swapped()
{ return this->signal_players_swapped_; }

/** @return   signal: a player has changed
 **/
Signal<void(Player const&)>&
PartyPlayers::signal_player_changed()
{ return this->signal_player_changed_; }

/** @return   whether there is no player
 **/
bool
PartyPlayers::empty() const
{
  return this->players_.empty();
} // bool PartyPlayers::empty() const

/** @return   the number of players
 **/
unsigned
PartyPlayers::size() const
{
  return this->players_.size();
} // unsigned PartyPlayers::size() const

/** @return   iterator to first player
 **/
PartyPlayers::iterator
PartyPlayers::begin()
{
  return PartyPlayers::iterator(*this, 0);
} // iterator PartyPlayers::begin()

/** @return   iterator behind the last player
 **/
PartyPlayers::iterator
PartyPlayers::end()
{
  return PartyPlayers::iterator(*this, this->size());
} // iterator PartyPlayers::end()

/** @return   iterator to first player
 **/
PartyPlayers::const_iterator
PartyPlayers::begin() const
{
  return PartyPlayers::const_iterator(*this, 0);
} // iterator PartyPlayers::begin() const

/** @return   iterator behind the last player
 **/
PartyPlayers::const_iterator
PartyPlayers::end() const
{
  return PartyPlayers::const_iterator(*this, this->size());
} // iterator PartyPlayers::end() const

/** -> result
 **
 ** @param    p   number of the player
 **
 ** @return   the player with the number 'p'
 **/
Player&
PartyPlayers::player(unsigned const p)
{
  DEBUG_ASSERTION(p < this->players_.size(),
                  "PartyPlayers::player(p):\n"
                  "'p' is not valid (" << p << ")");

  return *this->players_[p];
} // Player& PartyPlayers::player(unsigned p)

/** -> result
 **
 ** @param    p   number of the player
 **
 ** @return   the player with the number 'p'
 **/
Player const&
PartyPlayers::player(unsigned const p) const
{
  DEBUG_ASSERTION(p < this->players_.size(),
                  "PartyPlayers::player(p):\n"
                  "'p' is not valid (" << p << ")");

  return *this->players_[p];
} // Player const& PartyPlayers::player(unsigned const p) const

/** -> result
 **
 ** @param    p   number of the player
 **
 ** @return   the player with the number 'p'
 **/
Player&
PartyPlayers::operator[](unsigned const p)
{
  DEBUG_ASSERTION(p < this->players_.size(),
                  "PartyPlayers::player(p):\n"
                  "'p' is not valid (" << p << ")");

  return *this->players_[p];
} // Player& PartyPlayers::operator[](unsigned p)

/** -> result
 **
 ** @param    p   number of the player
 **
 ** @return   the player with the number 'p'
 **/
Player const&
PartyPlayers::operator[](unsigned const p) const
{
  DEBUG_ASSERTION(p < this->players_.size(),
                  "PartyPlayers::player(p):\n"
                  "'p' is not valid (" << p << ")");

  return *this->players_[p];
} // Player const& PartyPlayers::operator[](unsigned const p) const

/** -> result
 **
 ** @param    player   the player
 **
 ** @return   the position of the player, UINT_MAX if not found
 **/
unsigned
PartyPlayers::no(Player const& player) const
{
  for (unsigned p = 0; p < this->size(); ++p)
    if (&this->player(p) == &player)
      return p;
  return UINT_MAX;
} // unsigned PartyPlayers::no(Player player) const

/** @return   number of humans in the party
 **/
unsigned
PartyPlayers::count_humans() const
{
  return count_if(this->players_,
                  [](auto const& player) { return player->type() == Player::Type::human; });
} // unsigned PartyPlayers::count_humans() const

/** @return   human player if there exists exactly one,
 **            else 'nullptr'
 **/
Human*
PartyPlayers::human_player()
{
  if (this->count_humans() != 1)
    return {};

  return dynamic_cast<Human*>(find_if(this->players_,
                                      [](auto const& player) { return player->type() == Player::Type::human; }
                                     )->get());
} // Human* PartyPlayers::human_player()

/** @return   human player if there exists exactly one,
 **            else 'nullptr'
 **/
Human const*
PartyPlayers::human_player() const
{
  if (this->count_humans() != 1)
    return nullptr;

  return dynamic_cast<Human*>(find_if(this->players_,
                                      [](auto const& player) { return player->type() == Player::Type::human; }
                                     )->get());
} // Human const* PartyPlayers::human_player() const

/** Sets the player in the party
 **
 ** @param    n   the number of the player
 ** @param    player   the player to be set
 **/
void
PartyPlayers::set(unsigned const n, unique_ptr<Player> player)
{
  DEBUG_ASSERTION(n < this->players_.size(),
                  "PartyPlayers::set(n, player):\n"
                  "'n' is not valid");

  DEBUG_ASSERTION(!(::game_status & GameStatus::game),
                  "PartyPlayers::set_player(n, player):\n"
                  "  game_status is in the game.");

  this->players_[n] = move(player);

  ::ui->player_changed(this->player(n));
  this->signal_player_changed_(this->player(n));
} // void PartyPlayers::set(unsigned const n, unique_ptr<Player> player)

/** swaps the position of the players
 **
 ** @param    n0   the first player
 ** @param    n1   the second player
 **/
void
PartyPlayers::swap(unsigned const n0, unsigned const n1)
{
  DEBUG_ASSERTION(n0 < this->players_.size(),
                  "PartyPlayers::swap_players(n0, n1):\n"
                  "'n0' is not valid");

  DEBUG_ASSERTION(n1 < this->players_.size(),
                  "PartyPlayers::swap_players(n0, n1):\n"
                  "'n1' is not valid");

  DEBUG_ASSERTION(!(::game_status & GameStatus::game),
                  "PartyPlayers::swap_players(n0, n1):\n"
                  "  game_status is in the game.");

  // first set the names, so that empty names are replaced by the default names
  this->player(n0).set_name(this->player(n0).name());
  this->player(n1).set_name(this->player(n1).name());
  std::swap(this->players_[n0], this->players_[n1]);
  ::ui->players_switched(this->player(n0), this->player(n1));
  this->signal_players_swapped_(this->player(n0), this->player(n1));
} // void PartyPlayers::swap(unsigned n0, unsigned n1)

/** sets the type of the player 'player_no' to 'new_type'
 **
 ** @param    player   the player
 ** @param    new_type   the new type
 **
 ** @todo   keep the aiconfig
 **/
void
PartyPlayers::set_type(Player& player,
                       Player::Type const new_type)
{
  this->set_type(this->no(player), new_type);
} // void PartyPlayers::set_type(Player& player, Player::Type new_type)

/** sets the type of the player 'player_no' to 'new_type'
 **
 ** @param    player_no   the number of the player
 ** @param    new_type   the new type
 **
 ** @todo   keep the aiconfig
 **/
void
PartyPlayers::set_type(unsigned const player_no,
                       Player::Type const new_type)
{
  DEBUG_ASSERTION(player_no < this->players_.size(),
                  "PartyPlayers::set_type(): "
                  << "  Illegal Value for player number: "
                  << player_no); 

  DEBUG_ASSERTION(this->players_[player_no],
                  "PartyPlayers::set_type(): "
                  << "  player " << player_no << " does not exists");

  auto const& player = this->player(player_no);

  if (player.type() == new_type)
    return ;

  bool const default_name
    = (   Player::default_name(player.type(), player_no)
       == player.name());
  // create the new player
  this->set(player_no, Player::create(new_type, player));

  if (default_name)
    this->player(player_no).set_default_name();
} // void PartyPlayers::set_type(unsigned const player_no, Player::Type const new_type)

/** write the players into the output stream
 **
 ** @param    ostr   output stream
 **/
void
PartyPlayers::write(ostream& ostr) const
{
  ostr << "players\n"
    << "{\n";
  for (auto const& player : this->players_)
    ostr << "{\n" << *player << "}\n\n";
  ostr << "}\n";
} // void PartyPlayers::write(ostream& ostr) const

/** write the players in short form (name, type) in the output stream
 **
 ** @param    ostr   output stream
 **/
void
PartyPlayers::write_short(ostream& ostr) const
{
  ostr << "party players:\n";
  for (unsigned p = 0; p < this->players_.size(); ++p) {
    ostr << p << ": ";
    this->player(p).write_short(ostr);
    ostr << '\n';
  }
} // void PartyPlayers::write_players_short(ostream& ostr) const

/** load the players
 ** loads the files 'player.?' from the FreeDoko private directory
 **
 ** @return   whether all players could be loaded
 **/
bool
PartyPlayers::load()
{
  bool loaded_fail = false;
  for (unsigned p = 0; p < this->size(); ++p) {
    auto player = Player::create(::setting(Setting::Type::private_data_directory)
                                 + "/player."
                                 + String::to_string(p + 1));
    if (player)
      this->set(p, move(player));
    else
      loaded_fail = true;
  } // for (p < this->size())

  return !loaded_fail;
} // bool PartyPlayers::load()

/** save the players
 ** save in the files 'player.?' in the FreeDoko private directory
 **/
void
PartyPlayers::save() const
{
  unsigned n = 0;
  for (auto const& p : this->players_) {
    ++n;
    p->save(::setting(Setting::Type::private_data_directory)
            + "/player." + String::to_string(n));
  } // for (p : this->players_)
} // void PartyPlayers::save() const
