/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *   
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "rule.h"

#include "../utils/file.h"

#include "party.h"
#include "../ui/ui.h"
#include "../misc/setting.h"
#include "../class/readconfig/readconfig.h"


/** constructor
 **/
Rule::Rule() :
  bool_(Type::bool_number),
  unsigned_(Type::unsigned_number)
{
  this->reset_to_hardcoded();
} // Rule::Rule()

/** copy constructor
 ** The values are copied, but not the signal
 **
 ** @param    rule   rule to copy
 **/
Rule::Rule(Rule const& rule) :
  bool_(rule.bool_),
  unsigned_(rule.unsigned_),
  counting_(rule.counting_)
{ }

/** copy operator
 ** The values are copied, but not the signal
 **
 ** @param    rule   rule to copy
 **/
Rule&
Rule::operator=(Rule const& rule)
{
  this->bool_ = rule.bool_;
  this->unsigned_ = rule.unsigned_;
  this->counting_ = rule.counting_;
  return *this;
} // Rule& Rule::operator=(Rule const& rule)

/** @return   signal: a rule is changed
 **/
Signal<void(int, void const*)>& 
Rule::signal_changed()
{ return this->signal_changed_; }

/** resets the rules to the default values
 **/
void
Rule::reset_to_hardcoded()
{
  this->set_to(Ruleset::freedoko);
} // void Rule::reset_to_hardcoded()

/** resets the rules to a hardcoded ruleset
 **
 ** @param    ruleset   ruleset to set to
 **/
void
Rule::set_to(Rule::Ruleset const& ruleset)
{
  // first set to the ddv rules, then change according to the ruleset
  //
  this->set(Type::mutate,				false);

  this->set(Type::with_nines,				true);

  this->set(Type::bock,				false);
  this->set(Type::bock_multiplier,			2);
  this->set(Type::bock_append,			false);
  this->set(Type::bock_120,				false);
  this->set(Type::bock_solo_lost,			false);
  this->set(Type::bock_re_lost,			false);
  this->set(Type::bock_heart_trick,			false);
  this->set(Type::bock_black,				false);

  this->set(Type::throw_with_nines,			false);
  this->set(Type::min_number_of_throwing_nines,	5);
  this->set(Type::throw_with_kings,			false);
  this->set(Type::min_number_of_throwing_kings,	5);
  this->set(Type::throw_with_nines_and_kings,		false);
  this->set(Type::min_number_of_throwing_nines_and_kings,	7);
  this->set(Type::throw_with_richness,		false);
  this->set(Type::min_richness_for_throwing,	        70);
  this->set(Type::force_redistribute,	               	true);

  this->set(Type::poverty,				false);
  this->set(Type::poverty_shift,			false);

  this->set(Type::poverty_shift_only_trump,		false);
  this->set(Type::poverty_fox_do_not_count,		false);
  this->set(Type::poverty_fox_shift_extra,		false);
  this->set(Type::poverty_fox_shift_open,		false);
  this->set(Type::throw_with_one_trump,		false);

  this->set(Type::throw_when_fox_highest_trump,	false);

  this->set(Type::marriage_determination,		12);
  this->set(Type::marriage_first_foreign,		true);
  this->set(Type::marriage_first_color,		false);
  this->set(Type::marriage_first_single_color,	false);
  this->set(Type::marriage_first_trump,		false);
  this->set(Type::marriage_first_one_decides,		false);
  this->set(Type::marriage_before_poverty,		false);

  this->set(Type::genscher,				false);

  this->set(Type::solo,				true);
  this->set(Type::solo_jack,				true);
  this->set(Type::solo_queen,				true);
  this->set(Type::solo_king,				true);
  this->set(Type::solo_queen_jack,			true);
  this->set(Type::solo_king_jack,			true);
  this->set(Type::solo_king_queen,			true);
  this->set(Type::solo_koehler,			true);
  this->set(Type::solo_color,				true);
  this->set(Type::solo_meatless,			true);

  this->set(Type::throwing_before_solo,		false);

  this->set(Type::dollen,				true);
  this->set(Type::dollen_second_over_first,		false);
  this->set(Type::dollen_contrary_in_last_trick,	false);
  this->set(Type::dollen_first_over_second_with_swines,	false);

  this->set(Type::swines,				false);
  this->set(Type::swines_announcement_begin,		true);
  this->set(Type::swine_only_second,			false);
  this->set(Type::swines_in_solo,			true);
  this->set(Type::swines_in_poverty,			true);
  this->set(Type::hyperswines,			false);
  this->set(Type::hyperswines_announcement_begin,	true);
  this->set(Type::hyperswines_in_solo,		true);
  this->set(Type::hyperswines_in_poverty,		true);
  this->set(Type::swines_and_hyperswines_joint,	false);

  this->set(Type::extrapoint_catch_fox,		false);
  this->set(Type::extrapoint_catch_fox_last_trick,	false);
  this->set(Type::extrapoint_fox_last_trick,		false);
  this->set(Type::extrapoint_double_fox_last_trick,	false);

  this->set(Type::extrapoint_charlie,			false);
  this->set(Type::extrapoint_catch_charlie,		false);
  this->set(Type::extrapoint_double_charlie,		false);
  this->set(Type::extrapoint_catch_double_charlie,	false);
  this->set(Type::extrapoint_catch_charlie_only_with_diamond_queen, false);

  this->set(Type::extrapoint_dolle_jabs_dolle,	false);
  this->set(Type::extrapoint_heart_trick,		false);

  this->set(Type::knocking,	                        false);
  this->set(Type::announcement_limit_only_for_current,	false);
  this->set(Type::announcement_till_full_trick,	false);
  this->set(Type::announcement_first_trick_thirty_points,	false);
  this->set(Type::announcement_first_trick_thirty_points_only_first,	false);
  this->set(Type::announcement_first_trick_thirty_points_only_re_contra,	true);
  this->set(Type::announcement_first_trick_thirty_points_in_marriage,	true);
  this->set(Type::announcement_no_120,		11);
  this->set(Type::announcement_no_90,			 9);
  this->set(Type::announcement_no_60,			 6);
  this->set(Type::announcement_no_30,			 3);
  this->set(Type::announcement_no_0,			 3);
  this->set(Type::announcement_last,			 3);
  this->set(Type::announcement_reply,			false);
  this->set(Type::announcement_strict_limit,	        false);
  this->set(Type::announcement_re_doubles,		false);
  this->set(Type::announcement_contra_doubles_against_re,	false);

  this->set(Type::number_of_rounds_limited,		true);
  this->set(Type::number_of_rounds,			20);
  this->set(Type::points_limited,			false);
  this->set(Type::points,				40);
  this->set(Type::number_of_duty_soli,		0);
  this->set(Type::number_of_duty_color_soli,		0);
  this->set(Type::number_of_duty_picture_soli,	0);
  this->set(Type::offer_duty_solo,	        	true);
  this->set(Type::lustsolo_player_leads,		false);
  this->set(Type::solo_always_counts_triple,		false);

  // *** this->set(Type::party_points_SUMMING,	"plus");

  this->set(Type::number_of_players,			4);
  this->set(Type::number_of_players_in_game,		4);

  this->set(Type::counting,				Counting::plus);

  switch (ruleset) {
  case Ruleset::ddv:
    break;
  case Ruleset::freedoko:
    this->set(Type::bock,				true);
    this->set(Type::bock_120,				true);
    this->set(Type::throw_with_nines,			true);
    this->set(Type::poverty,				true);
    this->set(Type::poverty_shift,			true);
    this->set(Type::marriage_first_color,		true);
    this->set(Type::marriage_first_single_color,	true);
    this->set(Type::marriage_first_trump,		true);
    this->set(Type::dollen_second_over_first,		true);
    this->set(Type::swines,				true);
    this->set(Type::extrapoint_catch_fox,		true);
    this->set(Type::extrapoint_fox_last_trick,		true);
    this->set(Type::extrapoint_charlie,			true);
    this->set(Type::extrapoint_catch_charlie,		true);
    this->set(Type::announcement_first_trick_thirty_points,	false);
    this->set(Type::announcement_re_doubles,		true);
    this->set(Type::lustsolo_player_leads,		true);
    break;
  } // switch (ruleset)
} // void Rule::set_to(Rule::Ruleset ruleset)


/** resets the rules to the values for the seed statistics
 **/
void
Rule::reset_to_seed_statistics()
{
  this->set(Type::throw_with_nines,			true);
  this->set(Type::min_number_of_throwing_nines,	5);
  this->set(Type::throw_with_kings,			true);
  this->set(Type::min_number_of_throwing_kings,	5);
  this->set(Type::throw_with_nines_and_kings,		false);
  this->set(Type::min_number_of_throwing_nines_and_kings,	7);
  this->set(Type::throw_with_richness,		false);
  this->set(Type::min_richness_for_throwing,	        70);
  this->set(Type::force_redistribute,                	true);

  this->set(Type::poverty,				true);
  this->set(Type::poverty_shift,			true);
  this->set(Type::poverty_fox_do_not_count,		true);
  this->set(Type::poverty_fox_shift_extra,		true);
  this->set(Type::throw_when_fox_highest_trump,	true);

  this->set(Type::solo,				true);
  this->set(Type::solo_jack,				true);
  this->set(Type::solo_queen,				true);
  this->set(Type::solo_king,				true);
  this->set(Type::solo_queen_jack,			true);
  this->set(Type::solo_king_jack,			true);
  this->set(Type::solo_king_queen,			true);
  this->set(Type::solo_koehler,			true);
  this->set(Type::solo_color,				true);
  this->set(Type::solo_meatless,			true);

  this->set(Type::throwing_before_solo,		false);

  this->set(Type::swines,				true);
  this->set(Type::hyperswines,			true);
} // void Rule::reset_to_seed_statistics()

/** -> result
 **
 ** @param    type   rule type
 **
 ** @return   the value of 'type'
 **/
bool
Rule::operator()(Type::Bool const type) const
{
  return (this->dependencies(type) && this->value(type));
} // bool Rule::operator()(Type::Bool type) const

/** -> result
 **
 ** @param    type   rule type
 **
 ** @return   the value of 'type'
 **/
unsigned
Rule::operator()(Type::Unsigned const type) const
{
  switch(type) {
  case Type::number_of_duty_soli:
    if (this->dependencies(type))
      return this->value(type);
    else
      return 0;
  default:
    return this->value(type);
  } // switch(type)

  DEBUG_ASSERT(false);
  return 0;
} // unsigned Rule::operator()(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   rule type
 **
 ** @return   the value of 'type'
 **/
unsigned
Rule::operator()(Type::UnsignedExtra const type) const
{
  switch(type) {
  case Type::number_of_same_cards:
    return ((*this)(Type::number_of_players_in_game) / 2);
  case Type::number_of_card_colors:
    return Card::number_of_colors;
  case Type::number_of_card_tcolors:
    return Card::number_of_tcolors;
  case Type::number_of_card_values:
    return Card::number_of_values + ((*this)(Type::with_nines) ? 0 : -1);
  case Type::number_of_cards:
    return ((*this)(Type::number_of_same_cards)
            * (*this)(Type::number_of_card_colors) * (*this)(Type::number_of_card_values));
  case Type::number_of_teams:
    return 2;
  case Type::number_of_players_per_team:
    return (*this)(Type::number_of_players_in_game) / (*this)(Type::number_of_teams);
  case Type::number_of_games_per_round:
    return (*this)(Type::number_of_players);
  case Type::max_number_of_tricks_in_game:
    return 2 * Card::number_of_values * Card::number_of_colors / 4;
  case Type::number_of_tricks_in_game:
    return (*this)(Type::number_of_cards) / (*this)(Type::number_of_players_in_game);
  case Type::points_for_doppelkopf:
    return 40;
  case Type::max_number_of_poverty_trumps:
    return 3;
  } // switch(type)

  DEBUG_ASSERTION(false,
                  "Rule::operator()(" << type << ")\n"
                  "  type = " << type << " not in switch");
  return 0;
} // unsigned Rule::operator()(Type::UnsignedExtra type) const

/** -> result
 **
 ** @param    type   rule type
 **
 ** @return   the value of 'type'
 **/
Counting
Rule::operator()(Type::Counting const type) const
{
  return this->counting();
} // Counting Rule::operator()(Type::Counting type) const

/** -> result
 **
 ** @param    game_type   gametype
 **
 ** @return   whether the rule 'game_typ' is valid
 **/
bool
Rule::operator()(GameType const game_type) const
{
  switch(game_type) {
  case GameType::normal:
    return true;
  case GameType::thrown_nines:
    return (*this)(Type::throw_with_nines);
  case GameType::thrown_kings:
    return (*this)(Type::throw_with_kings);
  case GameType::thrown_nines_and_kings:
    return (*this)(Type::throw_with_nines_and_kings);
  case GameType::thrown_richness:
    return (*this)(Type::throw_with_richness);
  case GameType::redistribute:
    return (*this)(Type::force_redistribute);
  case GameType::fox_highest_trump:
    return (*this)(Type::throw_when_fox_highest_trump);
  case GameType::poverty:
    return (*this)(Type::poverty);
  case GameType::genscher:
    return (*this)(Type::genscher);
  case GameType::marriage:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
    return true;
  case GameType::solo_jack:
    return (*this)(Type::solo_jack);
  case GameType::solo_queen:
    return (*this)(Type::solo_queen);
  case GameType::solo_king:
    return (*this)(Type::solo_king);
  case GameType::solo_queen_jack:
    return (*this)(Type::solo_queen_jack);
  case GameType::solo_king_jack:
    return (*this)(Type::solo_king_jack);
  case GameType::solo_king_queen:
    return (*this)(Type::solo_king_queen);
  case GameType::solo_koehler:
    return (*this)(Type::solo_koehler);
  case GameType::solo_club:
  case GameType::solo_spade:
  case GameType::solo_heart:
  case GameType::solo_diamond:
    return (*this)(Type::solo_color);
  case GameType::solo_meatless:
    return (*this)(Type::solo_meatless);
  } // switch(game_type)

  DEBUG_ASSERT(false);
  return false;
} // bool Rule::operator()(GameType game_type) const

/** -> result
 **
 ** @param    marriage_selector
 **
 ** @return   whether the 'marriage_selector' is valid
 **/
bool
Rule::operator()(MarriageSelector const marriage_selector) const
{
  switch(marriage_selector) {
  case MarriageSelector::silent:
  case MarriageSelector::team_set:
    return false;
  case MarriageSelector::first_foreign:
    return (*this)(Type::marriage_first_foreign);
  case MarriageSelector::first_trump:
    return (*this)(Type::marriage_first_trump);
  case MarriageSelector::first_color:
    return (*this)(Type::marriage_first_color);
  case MarriageSelector::first_club:
  case MarriageSelector::first_spade:
  case MarriageSelector::first_heart:
    return (*this)(Type::marriage_first_single_color);
  } // switch(marriage_selector)

  DEBUG_ASSERT(false);
  return false;
} // bool Rule::operator()(MarriageSelector marriage_selector) const

/** -> result
 **
 ** @param    special_point   specialpoint
 **
 ** @return   whether the rule 'special_point' is valid
 **/
bool
Rule::operator()(Specialpoint::Type const special_point) const
{
  switch(special_point) {
  case Specialpoint::Type::nospecialpoint:
    return true;
  case Specialpoint::Type::caught_fox:
    return (*this)(Type::extrapoint_catch_fox);
  case Specialpoint::Type::caught_fox_last_trick:
    return (*this)(Type::extrapoint_catch_fox_last_trick);
  case Specialpoint::Type::fox_last_trick:
    return (*this)(Type::extrapoint_fox_last_trick);
  case Specialpoint::Type::charlie:
    return (*this)(Type::extrapoint_charlie);
  case Specialpoint::Type::caught_charlie:
    return (*this)(Type::extrapoint_catch_charlie);
  case Specialpoint::Type::dolle_caught_dolle:
    return (*this)(Type::extrapoint_dolle_jabs_dolle);
  case Specialpoint::Type::heart_trick:
    return (*this)(Type::extrapoint_heart_trick);
  case Specialpoint::Type::doppelkopf:
    return true;

  case Specialpoint::Type::won:
  case Specialpoint::Type::no90:
  case Specialpoint::Type::no60:
  case Specialpoint::Type::no30:
  case Specialpoint::Type::no0:
  case Specialpoint::Type::no120_said:
  case Specialpoint::Type::no90_said:
  case Specialpoint::Type::no60_said:
  case Specialpoint::Type::no30_said:
  case Specialpoint::Type::no0_said:
  case Specialpoint::Type::no90_said_120_got:
  case Specialpoint::Type::no60_said_90_got:
  case Specialpoint::Type::no30_said_60_got:
  case Specialpoint::Type::no0_said_30_got:
  case Specialpoint::Type::contra_won:
  case Specialpoint::Type::solo:
  case Specialpoint::Type::no120_reply:
  case Specialpoint::Type::no90_reply:
  case Specialpoint::Type::no60_reply:
  case Specialpoint::Type::no30_reply:
  case Specialpoint::Type::no0_reply:
    return true;

  case Specialpoint::Type::bock:
    return (*this)(Type::bock);
  } // switch(special_point)

  DEBUG_ASSERT(false);
  return false;
} // bool Rule::operator()(Specialpoint::Type special_point) const

/** -> result
 **
 ** @param    announcement
 **
 ** @return   how many remaining cards there must be to make the announcement
 **/
unsigned
Rule::remaining_cards(Announcement const announcement) const
{
  switch(announcement) {
  case Announcement::noannouncement:
  case Announcement::reply:
  case Announcement::no120_reply:
  case Announcement::no90_reply:
  case Announcement::no60_reply:
  case Announcement::no30_reply:
  case Announcement::no0_reply:
    return UINT_MAX;
  case Announcement::no120:
    return (*this)(Type::announcement_no_120);
  case Announcement::no90:
    return (*this)(Type::announcement_no_90);
  case Announcement::no60:
    return (*this)(Type::announcement_no_60);
  case Announcement::no30:
    return (*this)(Type::announcement_no_30);
  case Announcement::no0:
    return (*this)(Type::announcement_no_0);
  } // switch(announcement)

  DEBUG_ASSERT(false);
  return 0;
} // unsigned Rule::remaining_cards(Announcement announcement) const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   whether the dependencies of 'type' are met
 **/
bool
Rule::dependencies(Type::Bool const type) const
{
  switch(type) {
  case Type::mutate:
#ifdef RELEASE
    return false;
#else
    return true;
#endif

  case Type::with_nines:
    return true;

  case Type::bock:
    return true;
  case Type::bock_append:
    return (*this)(Type::bock);
  case Type::bock_120:
    return (*this)(Type::bock);
  case Type::bock_solo_lost:
    return (*this)(Type::bock);
  case Type::bock_re_lost:
    return (   (*this)(Type::bock)
            && !(*this)(Type::knocking));
  case Type::bock_heart_trick:
    return (*this)(Type::bock);
  case Type::bock_black:
    return (*this)(Type::bock);

  case Type::throw_with_nines:
    return (*this)(Type::with_nines);
  case Type::throw_with_kings:
    return true;
  case Type::throw_with_nines_and_kings:
    return (*this)(Type::with_nines);
  case Type::throw_with_richness:
    return true;
  case Type::force_redistribute:
    return true;

  case Type::poverty:
    return true;
  case Type::poverty_shift:
    return (*this)(Type::poverty);
  case Type::poverty_shift_only_trump:
  case Type::poverty_fox_shift_open:
  case Type::throw_with_one_trump:
    return (*this)(Type::poverty_shift);

  case Type::poverty_fox_do_not_count:
    return (*this)(Type::poverty);

  case Type::poverty_fox_shift_extra:
    return (   (*this)(Type::poverty_fox_do_not_count)
            && !(*this)(Type::poverty_shift_only_trump) );

  case Type::throw_when_fox_highest_trump:
    return true;

  case Type::marriage_first_foreign:
  case Type::marriage_first_color:
  case Type::marriage_first_single_color:
  case Type::marriage_first_trump:
    return true;
  case Type::marriage_first_one_decides:
    return true;
  case Type::marriage_before_poverty:
    return (*this)(Type::poverty);

  case Type::genscher:
    return !(!(*this)(Type::with_nines)
             && (*this)(Type::hyperswines));

  case Type::solo:
    return true;
  case Type::solo_jack:
  case Type::solo_queen:
  case Type::solo_king:
  case Type::solo_queen_jack:
  case Type::solo_king_jack:
  case Type::solo_king_queen:
  case Type::solo_koehler:
  case Type::solo_color:
  case Type::solo_meatless:
    return (*this)(Type::solo);

  case Type::throwing_before_solo:
    return (   (*this)(Type::solo)
            && (   (*this)(Type::throw_with_nines)
                || (*this)(Type::throw_with_kings)
                || (*this)(Type::throw_with_nines_and_kings)
                || (*this)(Type::throw_with_richness)
                || (*this)(Type::throw_with_one_trump)
                || (*this)(Type::throw_when_fox_highest_trump)
                || (    (*this)(Type::poverty)
                    && !(*this)(Type::poverty_shift) )
                || (*this)(Type::force_redistribute)
               ) );

  case Type::dollen:
    return true;
  case Type::dollen_second_over_first:
    return (*this)(Type::dollen);
  case Type::dollen_contrary_in_last_trick:
    return (*this)(Type::dollen);
  case Type::dollen_first_over_second_with_swines:
    return ((*this)(Type::dollen)
            && (*this)(Type::swines)
            && ( (*this)(Type::dollen_second_over_first)
                || (*this)(Type::dollen_contrary_in_last_trick)));

  case Type::swines:
    return true;
  case Type::swines_announcement_begin:
    return ((*this)(Type::swines)
            && !(this->value(Type::swine_only_second)));
  case Type::swine_only_second:
    return ((*this)(Type::swines)
            && !(this->value(Type::swines_announcement_begin))
            && !(this->value(Type::hyperswines)));
  case Type::swines_in_solo:
    return ((*this)(Type::swines)
            && (*this)(Type::solo_color));
  case Type::swines_in_poverty:
    return ((*this)(Type::swines)
            && (*this)(Type::poverty_shift));
  case Type::hyperswines:
    return ((*this)(Type::swines)
            && !(*this)(Type::swine_only_second));
#ifdef WORKAROUND
    // this activated leads to an infinte recursion
    // && !(!(*this)(Type::with_nines)
    // && (*this)(Type::genscher)));
#endif
  case Type::hyperswines_announcement_begin:
    return ((*this)(Type::hyperswines)
            && (*this)(Type::swines_announcement_begin));
  case Type::hyperswines_in_solo:
    return ((*this)(Type::hyperswines)
            && (*this)(Type::swines_in_solo));
  case Type::hyperswines_in_poverty:
    return ((*this)(Type::hyperswines)
            && (*this)(Type::swines_in_poverty));
  case Type::swines_and_hyperswines_joint:
    return (*this)(Type::hyperswines);

  case Type::extrapoint_catch_fox:
    return true;

  case Type::extrapoint_catch_fox_last_trick:
    return (*this)(Type::extrapoint_catch_fox);

  case Type::extrapoint_fox_last_trick:
    return true;

  case Type::extrapoint_double_fox_last_trick:
    return (*this)(Type::extrapoint_fox_last_trick);

  case Type::extrapoint_charlie:
    return true;
  case Type::extrapoint_catch_charlie:
  case Type::extrapoint_double_charlie:
    return (*this)(Type::extrapoint_charlie);
  case Type::extrapoint_catch_double_charlie:
    return (*this)(Type::extrapoint_double_charlie);
  case Type::extrapoint_catch_charlie_only_with_diamond_queen:
    return (*this)(Type::extrapoint_catch_charlie);

  case Type::extrapoint_dolle_jabs_dolle:
    return ((*this)(Type::dollen)
            && ( (*this)(Type::dollen_second_over_first)
                || (*this)(Type::dollen_contrary_in_last_trick) ));
  case Type::extrapoint_heart_trick:
    return ((*this)(Type::dollen)
            && !(*this)(Type::with_nines));

  case Type::knocking:
  case Type::announcement_limit_only_for_current:
  case Type::announcement_till_full_trick:
  case Type::announcement_first_trick_thirty_points:
    return true;
  case Type::announcement_reply:
    return !(*this)(Type::knocking);
  case Type::announcement_strict_limit:
    return true;

  case Type::announcement_first_trick_thirty_points_only_first:
    return (*this)(Type::announcement_first_trick_thirty_points);

  case Type::announcement_first_trick_thirty_points_only_re_contra:
    return (*this)(Type::announcement_first_trick_thirty_points);

  case Type::announcement_first_trick_thirty_points_in_marriage:
    return (*this)(Type::announcement_first_trick_thirty_points);

  case Type::announcement_re_doubles:
    return true;

  case Type::announcement_contra_doubles_against_re:
    return (*this)(Type::announcement_re_doubles);

  case Type::number_of_rounds_limited:
    return true;

  case Type::points_limited:
    return ((*this)(Type::counting) != Counting::plusminus);

  case Type::offer_duty_solo:
    return (*this)(Type::number_of_duty_soli);

  case Type::lustsolo_player_leads:
    return (*this)(Type::solo);

  case Type::solo_always_counts_triple:
    return (   (*this)(Type::solo)
            && ((*this)(Type::counting) != Counting::plusminus) );
  } // switch(type)

  DEBUG_ASSERT(false);
  return false;
} // bool Rule::dependencies(Type::Bool) const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   whether the dependencies of 'type' are met
 **/
bool
Rule::dependencies(Type::Unsigned const type) const
{
  switch(type) {
  case Type::min_number_of_throwing_nines:
    return (*this)(Type::throw_with_nines);
  case Type::min_number_of_throwing_kings:
    return (*this)(Type::throw_with_kings);
  case Type::min_number_of_throwing_nines_and_kings:
    return (*this)(Type::throw_with_nines_and_kings);
  case Type::min_richness_for_throwing:
    return (*this)(Type::throw_with_richness);

  case Type::bock_multiplier:
    return (*this)(Type::bock);

  case Type::marriage_determination:
  case Type::announcement_no_120:
  case Type::announcement_no_90:
  case Type::announcement_no_60:
  case Type::announcement_no_30:
  case Type::announcement_no_0:
  case Type::announcement_last:
    return true;

  case Type::number_of_rounds:
    return (*this)(Type::number_of_rounds_limited);

  case Type::points:
    return (*this)(Type::points_limited);

  case Type::number_of_duty_soli:
    return (   (*this)(Type::solo)
            && (   (*this)(Type::solo_jack)
                || (*this)(Type::solo_queen)
                || (*this)(Type::solo_king)
                || (*this)(Type::solo_queen_jack)
                || (*this)(Type::solo_king_jack)
                || (*this)(Type::solo_king_queen)
                || (*this)(Type::solo_koehler)
                || (*this)(Type::solo_color)
                || (*this)(Type::solo_meatless))
            && (   (*this)(Type::number_of_rounds_limited)
                || (*this)(Type::points_limited)));

  case Type::number_of_duty_color_soli:
    return (   ((*this)(Type::number_of_duty_soli) > 0)
            && (*this)(Type::solo_color));

  case Type::number_of_duty_picture_soli:
    return (   ((*this)(Type::number_of_duty_soli) > 0)
            && (   (*this)(Type::solo_jack)
                || (*this)(Type::solo_queen)
                || (*this)(Type::solo_king)
                || (*this)(Type::solo_queen_jack)
                || (*this)(Type::solo_king_jack)
                || (*this)(Type::solo_king_queen)
                || (*this)(Type::solo_koehler) ) );

  case Type::number_of_players:
  case Type::number_of_players_in_game:
    return true;
  } // switch(type)

  DEBUG_ASSERT(false);
  return false;
} // bool Rule::dependencies(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   the value of 'type'
 **/
bool
Rule::value(Type::Bool const type) const
{
  return this->bool_[type - Type::bool_first];
} // bool Rule::value(Type::Bool type) const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   the value of 'type'
 **/
unsigned
Rule::value(Type::Unsigned const type) const
{
  return this->unsigned_[type - Type::unsigned_first];
} // unsigned Rule::value(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   the value of 'type'
 **/
Counting
Rule::value(Type::Counting const type) const
{
  return this->counting_;
} // Counting Rule::value(Type::Counting type) const

/** @return   the counting
 **/
Counting const&
Rule::counting() const
{
  return this->counting_;
} // Counting const& Rule::counting() const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   the minimal value for 'type'
 **/
unsigned
Rule::min(Type::Unsigned const type) const
{
  switch(type) {
  case Type::bock_multiplier:
    return 2;

  case Type::min_number_of_throwing_nines:
  case Type::min_number_of_throwing_kings:
    return ( ( ((*this)(Type::number_of_same_cards) * (*this)(Type::number_of_card_colors))
              / (*this)(Type::number_of_players_in_game))
            + 1);

  case Type::min_number_of_throwing_nines_and_kings:
    return ( ( 2 * ((*this)(Type::number_of_same_cards)
                    * (*this)(Type::number_of_card_colors))
              / (*this)(Type::number_of_players_in_game))
            + 1);

  case Type::min_richness_for_throwing:
    return ( (  (*this)(Type::number_of_same_cards)
              * (*this)(Type::number_of_card_colors)
              * (  Card::ace
                 + Card::ten
                 + Card::king
                 + Card::queen
                 + Card::jack
                 + Card::nine)
             ) / (*this)(Type::number_of_players_in_game)
            + 1);


  case Type::marriage_determination:
    return 1;

  case Type::announcement_no_120:
  case Type::announcement_no_90:
  case Type::announcement_no_60:
  case Type::announcement_no_30:
  case Type::announcement_no_0:
  case Type::announcement_last:
    return 1;

  case Type::number_of_rounds:
    return 1;

  case Type::points:
    return 1;

  case Type::number_of_duty_soli:
    return ((*this)(Type::number_of_duty_color_soli)
            + (*this)(Type::number_of_duty_picture_soli));

  case Type::number_of_duty_color_soli:
    return 0;

  case Type::number_of_duty_picture_soli:
    return 0;

  case Type::number_of_players:
    return 4;

  case Type::number_of_players_in_game:
    return 4;
  } // switch(type)

  DEBUG_ASSERT(false);
  return 0;
} // unsigned Rule::min(Type::Unsigned type) const

/** -> result
 **
 ** @param    type   type
 **
 ** @return   the maximal value for 'type'
 **/
unsigned
Rule::max(Type::Unsigned const type) const
{
  switch(type) {
  case Type::bock_multiplier:
    return 256;

  case Type::min_number_of_throwing_nines:
  case Type::min_number_of_throwing_kings:
    return ((*this)(Type::number_of_same_cards) * (*this)(Type::number_of_card_colors));

  case Type::min_number_of_throwing_nines_and_kings:
    return 2 * ((*this)(Type::number_of_same_cards) * (*this)(Type::number_of_card_colors));

  case Type::min_richness_for_throwing:
    return (8 * Card::ace + 4 * Card::ten);

  case Type::marriage_determination:
    return (*this)(Type::max_number_of_tricks_in_game);

  case Type::announcement_no_120:
  case Type::announcement_no_90:
  case Type::announcement_no_60:
  case Type::announcement_no_30:
  case Type::announcement_no_0:
  case Type::announcement_last:
    return (*this)(Type::max_number_of_tricks_in_game);

  case Type::number_of_rounds:
    return UINT_MAX - 1;

  case Type::points:
    return UINT_MAX - 1;

  case Type::number_of_duty_soli:
    return ((*this)(Type::number_of_rounds)
            * (*this)(Type::number_of_players));

  case Type::number_of_duty_color_soli:
    return ((*this)(Type::number_of_duty_soli)
            - (*this)(Type::number_of_duty_picture_soli));

  case Type::number_of_duty_picture_soli:
    return ((*this)(Type::number_of_duty_soli)
            - (*this)(Type::number_of_duty_color_soli));

  case Type::number_of_players:
    return 4;

  case Type::number_of_players_in_game:
    return (((*this)(Type::number_of_players) / 2) * 2);
  } // switch(type)

  DEBUG_ASSERT(false);
  return 0;
} // unsigned Rule::max(Type::Unsigned type) const

/** @return   all card colors
 **/
vector<Card::Color> const&
Rule::card_colors() const
{
  static const vector<Card::Color> colors = {Card::club, Card::spade, Card::heart, Card::diamond};

  DEBUG_ASSERTION((colors.size() == (*this)(Type::number_of_card_colors)),
                  "Rule::card_colors():\n"
                  "  Number of card colors is not valid");

  return colors;
} // vector<Card::Color> const& Rule::card_colors() const

/** @return   all card values (depending on rule 'with nines')
 **/
vector<Card::Value> const&
Rule::card_values() const
{
  static const vector<Card::Value> values_without_nines = {Card::ace, Card::king, Card::queen, Card::jack, Card::ten};
  static const vector<Card::Value> values_with_nines = {Card::ace, Card::king, Card::queen, Card::jack, Card::ten, Card::nine};

  vector<Card::Value> const& values
    = ( (*this)(Type::with_nines)
       ? values_with_nines
       : values_without_nines );

  DEBUG_ASSERTION((values.size() == (*this)(Type::number_of_card_values)),
                  "Rule::card_values():\n"
                  "  Number of card values is not valid");

  return values;
} // vector<Card::Value> const& Rule::card_values() const

/** @return   all cards (depending on rule 'with nines')
 **/
vector<Card> const&
Rule::cards() const
{
  static vector<Card> const cards_without_nines
    = { Card::club_ace,
      Card::club_king,
      Card::club_queen,
      Card::club_jack,
      Card::club_ten,
      Card::spade_ace,
      Card::spade_king,
      Card::spade_queen,
      Card::spade_jack,
      Card::spade_ten,
      Card::heart_ace,
      Card::heart_king,
      Card::heart_queen,
      Card::heart_jack,
      Card::heart_ten,
      Card::diamond_ace,
      Card::diamond_king,
      Card::diamond_queen,
      Card::diamond_jack,
      Card::diamond_ten,
    };
  static vector<Card> const cards_with_nines
    = { Card::club_ace,
      Card::club_king,
      Card::club_queen,
      Card::club_jack,
      Card::club_ten,
      Card::spade_ace,
      Card::spade_king,
      Card::spade_queen,
      Card::spade_jack,
      Card::spade_ten,
      Card::heart_ace,
      Card::heart_king,
      Card::heart_queen,
      Card::heart_jack,
      Card::heart_ten,
      Card::diamond_ace,
      Card::diamond_king,
      Card::diamond_queen,
      Card::diamond_jack,
      Card::diamond_ten,
      Card::club_nine,
      Card::spade_nine,
      Card::heart_nine,
      Card::diamond_nine,
    };

  vector<Card> const& cards
    = (  (*this)(Type::with_nines)
       ? cards_with_nines
       : cards_without_nines );

  DEBUG_ASSERTION((cards.size()
                   == ( (*this)(Type::number_of_card_colors)
                       * (*this)(Type::number_of_card_values) ) ),
                  "Rule::cards():\n"
                  "  Number of cards is not valid:\n"
                  "  'cards.size() = " << cards.size()
                  << " != "
                  <<  (*this)(Type::number_of_card_colors)
                  << " * " << (*this)(Type::number_of_card_values) );

  return cards;
} // vector<Card> Rule::cards() const

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   type
 ** @param    value   value
 **/
void
Rule::set(Type::Bool type, bool value)
{
  if (this->value(type) == value)
    return ;

#ifdef OUTDATED
  // loading in a game leads now to problems
#ifdef RELEASE
  DEBUG_ASSERTION((::game_status < GameStatus::party_play),
                  "Rule::set(Type::Bool, value):\n"
                  "  'game_status' is not valid: " << ::game_status);
#endif
#endif

  bool const old_value = this->bool_[type - Type::bool_first];
  this->bool_[type - Type::bool_first] = value;

  switch (type) {
  case Type::marriage_first_foreign:
  case Type::marriage_first_color:
  case Type::marriage_first_single_color:
  case Type::marriage_first_trump:
    // if none is set, set 'first color'
    if (   !this->value(Type::marriage_first_foreign)
        && !this->value(Type::marriage_first_color)
        && !this->value(Type::marriage_first_single_color)
        && !this->value(Type::marriage_first_trump)) {
      type = Type::marriage_first_color;
      value = true;
      this->set(type, value);
    }
    break;
  default:
    break;
  } // switch(type)

  this->signal_changed_(type, &old_value);
  // the set-function has to be called again with the same values,
  // because else the ui could change the values
  this->set(type, value);
} // void Rule::set(Type::Bool, bool);

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   type
 ** @param    value   value in text form
 **/
void
Rule::set(Type::Bool const type, string const& value)
{
  if (   (value != "true")
      && (value != "false")
      && (value != "yes")
      && (value != "no")
      && (value != "0")
      && (value != "1"))
    cerr << "illegal value '" << value << "' for '" << type << "', "
      << "must be a boolean ('true' or 'false' or 'yes' or 'no' or '1' or '0'. "
      << "Ignoring"
      << endl;
  else
    this->set(type, (((value == "false")
                      || (value == "no")
                      || (value == "0"))
                     ? false
                     : true));
} // void Rule::set(Type::Bool type, string value)

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   type
 ** @param    value   value
 **/
void
Rule::set(Type::Unsigned const type, unsigned const value)
{
  if (this->value(type) == value)
    return ;

#ifdef OUTDATED
  // loading in a game leads now to problems
#ifdef RELEASE
  DEBUG_ASSERTION((::game_status < GameStatus::party_play),
                  "Rule::set(Type::Unsigned, value):\n"
                  "  'game_status' is not valid: " << ::game_status);
#endif
#endif

  unsigned const old_value = this->unsigned_[type - Type::unsigned_first];
  this->unsigned_[type - Type::unsigned_first] = value;

  switch(type) {
  case Type::announcement_no_120:
    if (value < this->value(Type::announcement_no_90))
      this->set(Type::announcement_no_90, value);
    break;
  case Type::announcement_no_90:
    if (value > this->value(Type::announcement_no_120))
      this->set(Type::announcement_no_120, value);
    if (value < this->value(Type::announcement_no_60))
      this->set(Type::announcement_no_60, value);
    break;
  case Type::announcement_no_60:
    if (value > this->value(Type::announcement_no_90))
      this->set(Type::announcement_no_90, value);
    if (value < this->value(Type::announcement_no_30))
      this->set(Type::announcement_no_30, value);
    break;
  case Type::announcement_no_30:
    if (value > this->value(Type::announcement_no_60))
      this->set(Type::announcement_no_60, value);
    if (value < this->value(Type::announcement_no_0))
      this->set(Type::announcement_no_0, value);
    break;
  case Type::announcement_no_0:
    if (value > this->value(Type::announcement_no_30))
      this->set(Type::announcement_no_30, value);
    if (value < this->value(Type::announcement_last))
      this->set(Type::announcement_last, value);
    break;
  case Type::announcement_last:
    if (value > this->value(Type::announcement_no_0))
      this->set(Type::announcement_no_0, value);
    break;
  default:
    break;
  } // switch(type)

  this->signal_changed_(type, &old_value);
  // the set-function has to be called again with the same values,
  // because else the ui could change the values
  this->set(type, value);
} // void Rule::set(Type::Unsigned type, unsigned value)

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   type
 ** @param    value   value in text form
 **/
void
Rule::set(Type::Unsigned const type, string const& value)
{
  char* end_ptr;
  auto const number = strtoul(value.c_str(), &end_ptr, 0);
  if (*end_ptr != '\0') {
    cerr << "illegal value '" << value << "' for '" << type << "', "
      << "must be a digit.\n"
      << "Ignoring"
      << endl;
    return ;
  }

  this->set(type, number);
} // void Rule::set(Type::Unsigned type, string value)

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   type
 ** @param    value   counting
 **/
void
Rule::set(Type::Counting const type, Counting const value)
{
  if (this->counting_ == value)
    return ;

  auto const old_value = this->counting();

  this->counting_ = value;

  this->signal_changed_(type, &old_value);
  // the set-function has to be called again with the same values,
  // because else the ui could change the values
  this->set(type, value);
} // void Rule::set(Type::Counting type, Counting value)

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   type
 ** @param    value   value in text form
 **/
void
Rule::set(Type::Counting const type, string const& value)
{
  if (value == to_string(Counting::plus))
    this->set(type, Counting::plus);
  else if (value == to_string(Counting::minus))
    this->set(type, Counting::minus);
  else if (value == to_string(Counting::plusminus))
    this->set(type, Counting::plusminus);
#ifdef WORKAROUND
  // 0.7.0 name changed
  else if (value == to_string(Counting::plus))
    this->set(type, Counting::plus);
  else if (value == to_string(Counting::minus))
    this->set(type, Counting::minus);
  else if (value == to_string(Counting::plusminus))
    this->set(type, Counting::plusminus);
#endif
  else
    cerr << "illegal value '" << value << "' for '" << type << "', "
      << "must be "
      << "'" << to_string(Counting::plus) << "', "
      << "'" << to_string(Counting::minus) << "' or "
      << "'" << to_string(Counting::plusminus) << ".\n"
      << "Ignoring"
      << endl;

  return ;
} // void Rule::set(Type::Counting type, string value)

/** sets the rule 'type' to the value 'value'
 **
 ** @param    type   name of the type
 ** @param    value   value in text form
 **
 ** @return   whether the type was found
 **/
bool
Rule::set(string const& type, string const& value)
{
  unsigned i;

  for (i = Type::bool_first; i <= Type::bool_last; i++) {
    if (type == to_string(Rule::Type::Bool(i))) {
      this->set(Rule::Type::Bool(i), value);
      return true;
    }
  }
  for (i = Type::unsigned_first; i <= Type::unsigned_last; i++) {
    if (type == to_string(Rule::Type::Unsigned(i))) {
      this->set(Rule::Type::Unsigned(i), value);
      return true;
    }
  }

  if (type == to_string(Type::counting)) {
    this->set(Type::counting, value);
    return true;
  }

#ifndef OUTDATED
  // 0.7.2b -> 0.7.3
  if (type == "throw with five nines") {
    this->set(Type::throw_with_nines, value);
    this->set(Type::min_number_of_throwing_nines, 5);
    return true;
  }
  if (type == "throw with six kings") {
    this->set(Type::throw_with_kings, value);
    this->set(Type::min_number_of_throwing_kings, 6);
    return true;
  }
#endif

  return false;
} // bool Rule::set(string type, string value)

/** mutate the rule
 ** change one rule (bool: flip; unsigned: increase/decrease)
 **/
void
Rule::mutate()
{
  int const r = rand() % (Type::unsigned_last + 1);

  // do not mutate all rules
  if (   (r == Type::number_of_players)
      || (r == Type::number_of_players_in_game)
      || (r == Type::number_of_rounds_limited)
      || (r == Type::lustsolo_player_leads)
      || (r == Type::points_limited)
      || (r == Type::mutate)
      || (   (r >= Type::solo)
          && (r <= Type::solo_meatless))
     )
    return ;

  if (   (r >= Type::bool_first)
      && (r <= Type::bool_last) ) {
    Type::Bool const type = static_cast<Type::Bool>(r);
    cout << "mutate " << type
      << ": " << this->value(type)
      << " (" << (*this)(type) << ")";
    this->set(type, !this->value(type));
    cout << " → " << this->value(type)
      << " (" << (*this)(type) << ")"
      << endl;
  } else if (   (r >= Type::unsigned_first)
             && (r <= Type::unsigned_last) ) {
    Type::Unsigned const type = static_cast<Type::Unsigned>(r);
    cout << "mutate " << type
      << ": " << this->value(type)
      << " (" << (*this)(type) << ")";
    unsigned value = this->value(type);
    if (value != UINT_MAX) {
      if (rand() % 2 == 0) {
        if (value > this->min(type))
          value -= 1;
      } else {
        if (value < this->max(type))
          value += 1;
      }
    }
    this->set(type, value);
    cout << " → " << this->value(type)
      << " (" << (*this)(type) << ")"
      << endl;
  } else {
    cout << "mutate " << r << endl;
  }

  return ;
} // void Rule::mutate()

/** load the rules from the given file
 ** (or the file in the rules directory)
 **
 ** @param    file   file to load the rules from
 **
 ** @return   whether the rules could be loaded
 **/
bool
Rule::load(string const& filename)
{
  {
    ifstream istr(filename.c_str());
    if (istr.good())
      return this->read(istr);
  }
  {
    ifstream istr((::setting(Setting::Type::private_data_directory)
                   + "/" + ::setting(Setting::Type::rules_directory)
                   + "/" + filename).c_str());
    if (istr.good())
      return this->read(istr);
  }
  {
    ifstream istr((::setting(Setting::Type::public_data_directory)
                   + "/" + ::setting(Setting::Type::rules_directory)
                   + "/" + filename).c_str());
    if (istr.good())
      return this->read(istr);
  }

  return false;
} // bool Rule::load(string filename)

/** load the rules from the given stream
 **
 ** @param    istr   stream to load the rules from
 **
 ** @return   whether the rules could be loaded
 **/
bool
Rule::read(istream& istr)
{
  this->reset_to_hardcoded();

  unsigned depth = 0;

  // load the rules
  while (istr.good()) {
    Config config;
    istr >> config;

    // finished with the config file
    if ((config.name == "") && (config.value == ""))
      break;

    if (config.separator) {
      // a setting
      if (!this->set(config.name, config.value)) {
        cerr << "rule file:	"
          << "ignoring unknown rule '" << config.name << "'."
          << endl;
      }
    } else { // if (config.separator)
      // a setting
      // if the value is in parentencies, remove both
      if(config.name == "!end") {
        // ignore the rest of the file
        break;
      } else if(config.name == "!stdout") {
        // output of the data to 'stdout'
        cout << config.value << endl;
      } else if(config.name == "!stderr") {
        // output of the data to 'stderr'
        cerr << config.value << endl;
      } else if(config.name == "{") {
        depth += 1;
      } else if(config.name == "}") {
        if (depth == 0) {
          cerr << "Rulefile: found a '}' without a '{' before.\n"
            << "Finish reading the the file."
            << endl;
          break;
        } // if (depth == 0)
        depth -= 1;
        if (depth == 0)
          break;
      } else if(config.name == "") {
        cerr << "Rule file: "
          << "Ignoring line \'" << config.value << "\'.\n";
      } else {
        cerr << "Rule file: "
          << "Rule '" << config.name << "' unknown.\n"
          << "Ignoring it.\n";
      } // if (config.name == .)
    } // config.separator
  } // while (istr.good())

  return true;
} // bool Rule::read(istream& istr)

/** saves the rules in the given file
 **
 ** @param    filename   file to save the rules into
 **
 ** @return   whether the rules could be saved
 **/
bool
Rule::save(string const& filename) const
{
  auto filename_tmp = filename + ".tmp";
  ofstream ostr(filename_tmp.c_str());
  if (!ostr.good()) {
    ::ui->information(_("Error::Rule::save: Error opening temporary file %s. Aborting saving.", filename_tmp), InformationType::problem);
    return false;
  }

  ostr << "# FreeDoko rules (version " << *::version << ")\n"
    << '\n';
  this->write(ostr);

  if (!ostr.good()) {
    ::ui->information(_("Error::Rule::save: Error saving in temporary file %s. Keeping temporary file (for bug tracking).", filename_tmp), InformationType::problem);
    return false;
  }
  ostr.close();

#ifdef WINDOWS
  unlink(filename.c_str());
#endif
  if (rename(filename_tmp.c_str(), filename.c_str())) {
    ::ui->information(_("Error::Rule::save: Could not rename temporary file %s to requested file %s. Keeping temporary file.", filename_tmp, filename), InformationType::problem);
    return false;
  }

  return true;
} // bool Rule::save(string file) const

/** puts the rules in the ouput stream
 **
 ** @param    ostr   output stream
 **/
void
Rule::write(ostream& ostr) const
{
  std::ios_base::fmtflags const flags = ostr.flags();
  ostr << std::boolalpha;

  if (this->value(Type::mutate))
    ostr << Type::mutate << " = "
      << this->value(Type::mutate) << '\n'
      << '\n';

  ostr << Type::number_of_rounds_limited << " = "
    << this->value(Type::number_of_rounds_limited) << '\n'
    << "  "
    << Type::number_of_rounds << " = "
    << this->value(Type::number_of_rounds) << '\n'

    << Type::points_limited << " = "
    << this->value(Type::points_limited) << '\n'
    << "  "
    << Type::points << " = "
    << this->value(Type::points) << '\n'

    << Type::number_of_duty_soli << " = "
    << this->value(Type::number_of_duty_soli) << '\n'
    << "  "
    << Type::number_of_duty_color_soli << " = "
    << this->value(Type::number_of_duty_color_soli) << '\n'
    << "  "
    << Type::number_of_duty_picture_soli << " = "
    << this->value(Type::number_of_duty_picture_soli) << '\n'
    << Type::offer_duty_solo << " = "
    << this->value(Type::offer_duty_solo) << '\n'
    << Type::lustsolo_player_leads << " = "
    << this->value(Type::lustsolo_player_leads) << '\n'
    << Type::solo_always_counts_triple << " = "
    << this->value(Type::solo_always_counts_triple) << '\n'

    // *** << Type::party_points_SUMMING << " = "
    // *** << this->value(Type::party_points_SUMMING) << '\n'

    << Type::number_of_players << " = "
    << this->value(Type::number_of_players) << '\n'

    << Type::number_of_players_in_game << " = "
    << this->value(Type::number_of_players_in_game) << '\n'

    << '\n'

    << Type::with_nines << " = "
    << this->value(Type::with_nines) << '\n'

    << '\n'

    << Type::bock << " = "
    << this->value(Type::bock) << '\n'
    << "  "
    << Type::bock_multiplier << " = "
    << this->value(Type::bock_multiplier) << '\n'
    << "  "
    << Type::bock_append << " = "
    << this->value(Type::bock_append) << '\n'
    << "  "
    << Type::bock_120 << " = "
    << this->value(Type::bock_120) << '\n'
    << "  "
    << Type::bock_solo_lost << " = "
    << this->value(Type::bock_solo_lost) << '\n'
    << "  "
    << Type::bock_re_lost << " = "
    << this->value(Type::bock_re_lost) << '\n'
    << "  "
    << Type::bock_heart_trick << " = "
    << this->value(Type::bock_heart_trick) << '\n'
    << "  "
    << Type::bock_black << " = "
    << this->value(Type::bock_black) << '\n'

    << '\n'


    << Type::throw_with_nines << " = "
    << this->value(Type::throw_with_nines) << '\n'

    << Type::min_number_of_throwing_nines << " = "
    << this->value(Type::min_number_of_throwing_nines) << '\n'

    << Type::throw_with_kings << " = "
    << this->value(Type::throw_with_kings) << '\n'

    << Type::min_number_of_throwing_kings << " = "
    << this->value(Type::min_number_of_throwing_kings) << '\n'

    << Type::throw_with_nines_and_kings << " = "
    << this->value(Type::throw_with_nines_and_kings) << '\n'

    << Type::min_number_of_throwing_nines_and_kings << " = "
    << this->value(Type::min_number_of_throwing_nines_and_kings) << '\n'

    << Type::throw_with_richness << " = "
    << this->value(Type::throw_with_richness) << '\n'

    << Type::min_richness_for_throwing << " = "
    << this->value(Type::min_richness_for_throwing) << '\n'

    << Type::force_redistribute << " = "
    << this->value(Type::force_redistribute) << '\n'

    << '\n'

    << Type::poverty << " = "
    << this->value(Type::poverty) << '\n'
    << "  "
    << Type::poverty_shift << " = "
    << this->value(Type::poverty_shift) << '\n'
    << "  "
    << Type::poverty_shift_only_trump << " = "
    << this->value(Type::poverty_shift_only_trump) << '\n'
    << "  "
    << Type::poverty_fox_do_not_count << " = "
    << this->value(Type::poverty_fox_do_not_count) << '\n'
    << "  "
    << Type::poverty_fox_shift_extra << " = "
    << this->value(Type::poverty_fox_shift_extra) << '\n'
    << "  "
    << Type::poverty_fox_shift_open << " = "
    << this->value(Type::poverty_fox_shift_open) << '\n'
    << "  "
    << Type::throw_with_one_trump << " = "
    << this->value(Type::throw_with_one_trump) << '\n'
    << "  "
    << Type::throw_when_fox_highest_trump << " = "
    << this->value(Type::throw_when_fox_highest_trump) << '\n'

    << '\n'

    << Type::marriage_determination << " = "
    << this->value(Type::marriage_determination) << '\n'
    << "  "
    << Type::marriage_first_foreign << " = "
    << this->value(Type::marriage_first_foreign) << '\n'
    << "  "
    << Type::marriage_first_color << " = "
    << this->value(Type::marriage_first_color) << '\n'
    << "  "
    << Type::marriage_first_single_color << " = "
    << this->value(Type::marriage_first_single_color) << '\n'
    << "  "
    << Type::marriage_first_trump << " = "
    << this->value(Type::marriage_first_trump) << '\n'
    << "  "
    << Type::marriage_first_one_decides << " = "
    << this->value(Type::marriage_first_one_decides) << '\n'
    << "  "
    << Type::marriage_before_poverty << " = "
    << this->value(Type::marriage_before_poverty) << '\n'

    << '\n'

    << Type::genscher << " = "
    << this->value(Type::genscher) << '\n'

    << '\n'

    << Type::solo << " = "
    << this->value(Type::solo) << '\n'
    << "  "
    << Type::solo_jack << " = "
    << this->value(Type::solo_jack) << '\n'
    << "  "
    << Type::solo_queen << " = "
    << this->value(Type::solo_queen) << '\n'
    << "  "
    << Type::solo_king << " = "
    << this->value(Type::solo_king) << '\n'
    << "  "
    << Type::solo_queen_jack << " = "
    << this->value(Type::solo_queen_jack) << '\n'
    << "  "
    << Type::solo_king_jack << " = "
    << this->value(Type::solo_king_jack) << '\n'
    << "  "
    << Type::solo_king_queen << " = "
    << this->value(Type::solo_king_queen) << '\n'
    << "  "
    << Type::solo_koehler << " = "
    << this->value(Type::solo_koehler) << '\n'
    << "  "
    << Type::solo_color << " = "
    << this->value(Type::solo_color) << '\n'
    << "  "
    << Type::solo_meatless << " = "
    << this->value(Type::solo_meatless) << '\n'
    << "  "
    << Type::throwing_before_solo << " = "
    << this->value(Type::throwing_before_solo) << '\n'

    << '\n'

    << Type::dollen << " = "
    << this->value(Type::dollen) << '\n'
    << "  "
    << Type::dollen_second_over_first << " = "
    << this->value(Type::dollen_second_over_first) << '\n'
    << "  "
    << Type::dollen_contrary_in_last_trick << " = "
    << this->value(Type::dollen_contrary_in_last_trick) << '\n'
    << "  "
    << Type::dollen_first_over_second_with_swines << " = "
    << this->value(Type::dollen_first_over_second_with_swines) << '\n'

    << '\n'

    << Type::swines << " = "
    << this->value(Type::swines) << '\n'
    << "  "
    << Type::swines_in_solo << " = "
    << this->value(Type::swines_in_solo) << '\n'
    << "  "
    << Type::swines_in_poverty << " = "
    << this->value(Type::swines_in_poverty) << '\n'
    << "  "
    << Type::swines_announcement_begin << " = "
    << this->value(Type::swines_announcement_begin) << '\n'
    << "  "
    << Type::swine_only_second << " = "
    << this->value(Type::swine_only_second) << '\n'

    << Type::hyperswines << " = "
    << this->value(Type::hyperswines) << '\n'
    << "  "
    << Type::hyperswines_in_solo << " = "
    << this->value(Type::hyperswines_in_solo) << '\n'
    << "  "
    << Type::hyperswines_in_poverty << " = "
    << this->value(Type::hyperswines_in_poverty) << '\n'
    << "  "
    << Type::hyperswines_announcement_begin << " = "
    << this->value(Type::hyperswines_announcement_begin) << '\n'
    << "  "
    << Type::swines_and_hyperswines_joint << " = "
    << this->value(Type::swines_and_hyperswines_joint) << '\n'

    << '\n'

    << "# Extrapoints:\n"

    << Type::extrapoint_catch_fox << " = "
    << this->value(Type::extrapoint_catch_fox) << '\n'
    << Type::extrapoint_catch_fox_last_trick << " = "
    << this->value(Type::extrapoint_catch_fox_last_trick) << '\n'
    << Type::extrapoint_fox_last_trick << " = "
    << this->value(Type::extrapoint_fox_last_trick) << '\n'
    << Type::extrapoint_double_fox_last_trick << " = "
    << this->value(Type::extrapoint_double_fox_last_trick) << '\n'

    << Type::extrapoint_dolle_jabs_dolle << " = "
    << this->value(Type::extrapoint_dolle_jabs_dolle) << '\n'
    << Type::extrapoint_heart_trick << " = "
    << this->value(Type::extrapoint_heart_trick) << '\n'

    << Type::extrapoint_charlie << " = "
    << this->value(Type::extrapoint_charlie) << '\n'
    << Type::extrapoint_catch_charlie << " = "
    << this->value(Type::extrapoint_catch_charlie) << '\n'
    << Type::extrapoint_double_charlie << " = "
    << this->value(Type::extrapoint_double_charlie) << '\n'
    << Type::extrapoint_catch_double_charlie << " = "
    << this->value(Type::extrapoint_catch_double_charlie) << '\n'
    << Type::extrapoint_catch_charlie_only_with_diamond_queen << " = "
    << this->value(Type::extrapoint_catch_charlie_only_with_diamond_queen) << '\n'


    << '\n'

    << Type::announcement_first_trick_thirty_points << " = "
    << this->value(Type::announcement_first_trick_thirty_points) << '\n'

    << Type::announcement_first_trick_thirty_points_only_first << " = "
    << this->value(Type::announcement_first_trick_thirty_points_only_first) << '\n'

    << Type::announcement_first_trick_thirty_points_only_re_contra << " = "
    << this->value(Type::announcement_first_trick_thirty_points_only_re_contra) << '\n'

    << Type::announcement_first_trick_thirty_points_in_marriage << " = "
    << this->value(Type::announcement_first_trick_thirty_points_in_marriage) << '\n'

    << Type::announcement_re_doubles << " = "
    << this->value(Type::announcement_re_doubles) << '\n'

    << Type::announcement_contra_doubles_against_re << " = "
    << this->value(Type::announcement_contra_doubles_against_re) << '\n'

    << '\n'

    << "# till when to make the announcements:\n"

    << Type::knocking << " = "
    << this->value(Type::knocking) << '\n'

    << Type::announcement_no_120 << " = "
    << this->value(Type::announcement_no_120) << '\n'
    << Type::announcement_no_90 << " = "
    << this->value(Type::announcement_no_90) << '\n'
    << Type::announcement_no_60 << " = "
    << this->value(Type::announcement_no_60) << '\n'
    << Type::announcement_no_30 << " = "
    << this->value(Type::announcement_no_30) << '\n'
    << Type::announcement_no_0 << " = "
    << this->value(Type::announcement_no_0) << '\n'
    << Type::announcement_last << " = "
    << this->value(Type::announcement_last) << '\n'

    << Type::announcement_limit_only_for_current << " = "
    << this->value(Type::announcement_limit_only_for_current) << '\n'
    << Type::announcement_till_full_trick << " = "
    << this->value(Type::announcement_till_full_trick) << '\n'

    << Type::announcement_reply << " = "
    << this->value(Type::announcement_reply) << '\n'

    << Type::announcement_strict_limit << " = "
    << this->value(Type::announcement_strict_limit) << '\n'

    << '\n'

    << "# "
    << Type::number_of_same_cards << " = "
    << (*this)(Type::number_of_same_cards) << '\n'
    << "# "
    << Type::number_of_card_colors << " = "
    << (*this)(Type::number_of_card_colors) << '\n'
    << "# "
    << Type::number_of_card_tcolors << " = "
    << (*this)(Type::number_of_card_tcolors) << '\n'
    << "# "
    << Type::number_of_card_values << " = "
    << (*this)(Type::number_of_card_values) << '\n'
    << "# "
    << Type::number_of_cards << " = "
    << (*this)(Type::number_of_cards) << '\n'
    << "# "
    << Type::number_of_teams << " = "
    << (*this)(Type::number_of_teams) << '\n'
    << "# "
    << Type::number_of_players_per_team << " = "
    << (*this)(Type::number_of_players_per_team) << '\n'
    << "# "
    << Type::number_of_tricks_in_game << " = "
    << (*this)(Type::number_of_tricks_in_game) << '\n'

    << '\n'

    << Type::counting << " = "
    << (*this)(Type::counting) << '\n'
    ;

  ostr.flags(flags);
} // void Rule::write(ostream& ostr) const

/** prints the rule in 'ostr'
 **
 ** @param    ostr   output stream
 ** @param    rule   rule to print
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Rule const& rule)
{
  rule.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, Rule rule)

/** -> result
 **
 ** @param    rule1   first rule that is compared
 ** @param    rule2   second rule that is compared
 **
 ** @return   whether the two rules are equal (according to value)
 **/
bool
operator==(Rule const& rule1, Rule const& rule2)
{
  for (int i = Rule::Type::bool_first; i <= Rule::Type::bool_last; ++i)
    if (rule1.value(static_cast<Rule::Type::Bool>(i))
        != rule2.value(static_cast<Rule::Type::Bool>(i)))
      return false;

  for (int i = Rule::Type::unsigned_first; i <= Rule::Type::unsigned_last; ++i)
    if (rule1.value(static_cast<Rule::Type::Unsigned>(i))
        != rule2.value(static_cast<Rule::Type::Unsigned>(i)))
      return false;

  return true;
} // bool operator==(Rule rule1, Rule rule2)

/** -> result
 **
 ** @param    rule1   first rule that is compared
 ** @param    rule2   second rule that is compared
 **
 ** @return   whether the two rules are different (according to value)
 **/
bool
operator!=(Rule const& rule1, Rule const& rule2)
{
  return !(rule1 == rule2);
} // bool operator!=(Rule rule1, Rule rule2)
