/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "ai.h"

#include "../../game/game.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

/** constructor
 **/
AiDummy::AiDummy() :
Player(Player::Type::ai_dummy)
{ }

/** Construktor
 **
 ** @param    istr   stream with the infos
 **/
AiDummy::AiDummy(istream& istr) :
Player(Player::Type::ai_dummy)
{
  (void)istr;
  DEBUG_ASSERTION(false,
		  "AiDummy::AiDummy(istr):\n"
		  "  Contructor not implemented");
} // AiDummy::AiDummy(istream& istr)

/** copy constructor
 **
 ** @param    player   player to copy
 **/
AiDummy::AiDummy(Player const& player) :
  Player(player),
  Aiconfig()
{
  this->set_type(Player::Type::ai_dummy);
} // AiDummy::AiDummy(Player const& player)

/** copy constructor
 **
 ** @param    player     player to copy
 ** @param    aiconfig   configuration to copy
 **/
AiDummy::AiDummy(Player const& player, Aiconfig const& aiconfig) :
  Player(player),
  Aiconfig(aiconfig)
{
  this->set_type(Player::Type::ai_dummy);
} // AiDummy::AiDummy(Player const& player, Aiconfig const& aiconfig)

/** clone the player
 **
 ** @return   pointer of a clone
 **/
unique_ptr<Player>
AiDummy::clone() const
{  
  return make_unique<AiDummy>(*this);
} // virtual unique_ptr<Player> AiDummy::clone() const

/** -> result
 ** (test of poverty, marriage, swines, hyperswines)
 **
 ** @param    duty_solo   whether the player has to play a solo
 **    (default: false)
 **
 ** @return   simple Reservation
 **/
Reservation const&
AiDummy::reservation_get(bool const duty_solo)
{
  this->reservation() = this->Player::reservation_get_default();

  if (duty_solo) {
    for (auto const t : game_type_list) {
      if (   is_real_solo(t)
          && this->game().rule()(t) ) {
        this->reservation().game_type = t;
        break;
      }
    } // for (t)
  } // if (duty_solo)

  return this->reservation();
} // Reservation const& AiDummy::reservation_get(bool duty_solo = false)

/** -> result
 ** non trumps, color aces and lowest color cards
 **
 ** @return   cards to shift
 **/
HandCards
AiDummy::poverty_shift()
{
  // the number of cards to shift
  unsigned const numberofpovertycards = this->hand().count_poverty_cards();
  // the cards to shift
  HandCards cards_to_shift;

  // first add all poverty cards
  for (auto const& c : this->hand().cards())
    if (c.istrump())
      cards_to_shift.push_back(c);

  // second add all color cards in the order: ace, nine, king, ten
  vector<Card::Value> values = {Card::ace, Card::nine, Card::king, Card::ten};
  for (auto v = values.begin();
       (   (v != values.end())
        && (cards_to_shift.size() < numberofpovertycards));
       ++v)
    for (auto c = this->hand().cards().begin();
         (   (c != this->hand().cards().end())
          && (cards_to_shift.size() < numberofpovertycards));
         ++c)
      if (!c->istrump())
        if (c->value() == *v)
          cards_to_shift.push_back(*c);

  this->sorted_hand().remove(cards_to_shift);

  return cards_to_shift;
} // HandCards AiDummy::poverty_shift()

/** -> result
 ** whether to accept the poverty
 **
 ** @param    cardno   the number of cards that are shifted
 **
 ** @return   false (do not accept)
 **/
bool
AiDummy::poverty_take_accept(unsigned cardno)
{
  return false;
} // bool AiDummy::poverty_take_accept(unsigned cardno)

/** change the poverty cards
 ** should not be called
 **
 ** @param    cards   the cards the player takes
 **
 ** @return   cards to return
 **/
HandCards
AiDummy::poverty_cards_change(HandCards const& cards)
{
  DEBUG_ASSERTION(false,
                  "AiDummy::poverty_cards_change(cards)\n"
                  "  call of this function althoug the dummy ai should reject any poverty");

  return cards;
} // HandCards AiDummy::poverty_cards_change(HandCards cards)

/** the poverty player gets 'cards'
 **
 ** @param    cards    cards the player gets back
 **/
void
AiDummy::poverty_cards_get_back(HandCards const& cards)
{
  this->sorted_hand().add(cards);
} // void AiDummy::poverty_cards_get_back(HandCards cards)

/** -> result:
 ** if the dummy ai can get the trick, it takes its best card
 ** (also, if its the first card in the trick)
 ** else (if the dummy ai has no better than the winning card of the trick)
 ** the worst card is taken
 **
 ** @return   valid card
 **/
HandCard
AiDummy::card_get()
{
  // the first card -- play the highest
  // if the ai cannot get the trick: play the lowest
  // if the ai can get the trick: play the highest

  HandCard card;
  bool can_win = false;

  for (auto const& c : this->hand().cards()) {
    if (this->game().tricks().current().isvalid(c, this->hand())) {
      if (card.is_empty()) {
        card = c;
      } else { // if !(!card)
        bool const is_winnercard
          = (this->game().tricks().current().isempty()
             ? true
             : this->game().tricks().current().winnercard().less(c)
            );
        can_win |= is_winnercard;

        if (this->game().tricks().current().isempty()) {
          // take the smallest card
          if (card.less(c))
            card = c;
        } else if (this->game().tricks().current().islastcard()) {
          // if the player can win the trick take the smallest winnercard,
          // else take the smallest card.

          if (can_win
              && is_winnercard
              && c.less(card))
            card = c;
          if (!can_win
              && !is_winnercard
              && c.less(card))
            card = c;
        } else { // if !(is first or last card in trick)
          // if the player can win the trick take the greatest winnercard,
          // else take the smallest card.

          if (can_win
              && is_winnercard
              && card.less(c))
            card = c;
          if (!can_win
              && !is_winnercard
              && c.less(card))
            card = c;
        } // if !(is first or last card in trick)
      } // if !(!card)
    } // if (this->game().tricks().current().isvalid(c, this->hand()))
  } // for (c : this->hand().cards())

  // valid_cards.card(RAND(valid_cards.cardsnumber()))
  return card;
} // HandCard AiDummy::card_get()

/** @return   announcement of the player
 **/
Announcement
AiDummy::announcement_request() const
{
  // points of the own team
  unsigned points = 0;

  // count the points, the own team has made 
  for (auto const& player : this->game().players()) {
    if (this->game().teaminfo().get(player) == this->team())
      points += this->game().points_of_player(player);
  }

  if (points == 240)
    return Announcement::no0;
  if (points > 240 - 30)
    return Announcement::no30;
  if (points > 240 - 60)
    return Announcement::no60;
  if (points > 240 - 90)
    return Announcement::no90;
  if (points > 240 - 120)
    return Announcement::no120;

  return Announcement::noannouncement;
} // Announcement AiDummy::announcement_request() const

/** genscher: -> result
 ** take the player with the most points
 **
 ** @return   pointer to the new teammate,
 **   NULL if the genscher is not to be announced
 **
 ** @todo       take current trick into account
 **/
Player const*
AiDummy::genscher_partner()
{ 
  // player with the most points
  unsigned best_player = UINT_MAX;
  // most points
  unsigned points_max = 0;
  for (unsigned p = 0; p < this->game().playerno(); ++p) {
    if (this->game().player(p) == *this)
      continue;

    if (   (this->game().points_of_player(p) > points_max)
        || (   (this->game().points_of_player(p) > points_max)
            && (this->game().teaminfo().get(this->game().player(p)) == this->team()) )
       ) {
      points_max = this->game().points_of_player(p);
      best_player = p;
    }
  } // for (p < this->game().playerno())

  if (   (best_player == UINT_MAX)
      || (this->game().teaminfo().get(this->game().players().player(best_player)) == this->team()) )
    return {};
  else
    return &this->game().player(best_player);
} // Player const* AiDummy::genscher_partner()
