/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#pragma once

#include "aiDb.h"

#include "../player.h"
#include "../aiconfig.h"
#include "../../game/game.h"

#include "../../basetypes.h"

#include "cards_information.h"
#include "team_information.h"
#include "heuristic.h"

class VirtualGamesInterface;
class HeuristicInterface;

class Ai: public Player, public Aiconfig {
  friend class WVirtualGames;
  friend class Gametree;
  friend struct TrickWeighting;
  friend struct Heuristics;
  friend class ::Heuristic;

  friend class VirtualGamesAi;
  friend class HeuristicAi;
  public:

  Ai();
  Ai(Aiconfig const& aiconfig);
  Ai(Player const& player);
  Ai(Player const& player, Aiconfig const& aiconfig);
  Ai(istream& istr);
  Ai(Ai const& ai);
  Ai& operator=(Ai const& ai) = delete;
  // clone the player
  unique_ptr<Player> clone() const override;

  ~Ai() override;

  virtual operator HeuristicInterface const&() const;
  virtual operator VirtualGamesInterface const&() const;
  TeamInformation const& team_information() const;
  TeamInformation& team_information();
  CardsInformation const& cards_information() const;
  CardsInformation& cards_information();
  AiDb const* db() const;
  AiDb* db();

  // write the ai in the stream
  ostream& write(ostream& ostr) const override;

  // read the config 'config' from 'istr'
  bool read(Config const& config, istream& istr) override;

  // compares the ai
  virtual bool isequal(Ai const& ai) const;

  AiType const& aitype() const;
  unsigned future_limit() const;
  Rating::Type const& rating() const;

  // set the cards information
  void set_cards_information(CardsInformation const& cards_information);

  // set the game
  void set_game(Game& game) override;
  Hand const& set_hand(Hand const& hand) override;

  // h hands of players 0 till number_of_players number of player for Weighting
  void new_game(Game& game) override;
  Reservation const& reservation_get(bool isDuty = false ) override;
  void rule_changed(int type, void const* old_value) override;
  void game_open(Game& game) override;
  void game_start() override;
  void game_close(Game const& game) override;
  HandCard card_get() override;
  virtual HandCard card_suggestion();
  void card_played(HandCard const& card) override;
  void check_swines_announcement_at_game_start() override;
  void swines_announced(Player const& player) override;
  void hyperswines_announced(Player const& player) override;
  void marriage(Player const& bridegroom, Player const& bride) override;
  void genscher(Player const& genscher, Player const& partner) override;
  Card nextcard(Trick const& trick);

  Heuristic last_heuristic() const;
  ::Heuristic::Rationale const& last_heuristic_rationale() const;
  void set_last_heuristic_to_manual();
  void set_last_heuristic_to_bug_report();

  // the teaminfo of 'player'
  Team teaminfo(Player const& player) const; 
  // update the teaminfo
  // (compare with the teaminfo of the game and check the current trick
  void teaminfo_update() override;
  // the partner of the player
  Player const* partner(Player const& player) const;
  // the points the team has made
  unsigned points_of_team() const;

  // set the teams
  virtual void set_teams(vector<Team> const& teams);

  // the trick is opened
  void trick_open(Trick const& trick) override;
  // the trick is full
  void trick_full(Trick const& trick) override;

  Announcement announcement_request() const override;
  void check_and_make_first_announcement(HandCard const& handcard);
  void announcement_made(Announcement const& announcement,
                         Player const& player) override;

  bool colorjabbed(Card::Color c) const;

  unsigned color_runs(Card::Color color) const;
  unsigned color_runs(Trick const& trick) const;

  bool jabbedbyownteam(Card::Color c) const;

  bool jabbedbyotherteam(Card::Color c) const;


  bool next_low() const;

  void check_low_high(HandCard const& card);

  bool single_ace() const;

  protected:
  unsigned const& last_trick_to_calculate() const;
  void set_last_trick_to_calculate(unsigned last_trick_to_calculate);

  // 'player' shifts cards
  HandCards poverty_shift() override;
  // returns whether 'player' accepts the poverty
  bool poverty_take_accept(unsigned cardno) override;
  // the player changes the cards of the poverty
  HandCards poverty_cards_change(HandCards const& cards) override;
  // the poverty player gets 'cards'
  void poverty_cards_get_back(HandCards const& cards) override;

  public:
  Card trump_card_limit() const;
  Card lowest_trump_card_limit() const;

  private:

  MarriageSelector get_Marriage() const;

  public:
  Player const* genscher_partner() override;

  private:

  int calculate_meatless_color( Card::Color co );

  int checksolo(GameType gametype) const;
  int checksolo_single_picture(GameType gametype) const;
  int checksolo_double_picture(GameType gametype) const;
  int checksolo_koehler(GameType gametype) const;
  int checksolo_color(GameType gametype) const; 
  int checksolo_meatless(GameType gametype) const;

  private:
  unique_ptr<TeamInformation> team_information_;
  unique_ptr<CardsInformation> cards_information_;
  // Team of all players known by this player
  vector<Team> teaminfo_;
  vector<Trick> trick_; // last played tricks
  unsigned trickno_ = 0;
  vector<vector<bool> > colorjabbed_; // remembering for each player if he has jabbed this color
  bool low_ = false; // next card should be low or high trump
  Card::Color meatlesscolor_ = Card::nocardcolor;
  vector<bool> trumpless_; // players has no trumps

  unsigned last_trick_to_calculate_ = 12;

  Aiconfig::Heuristic last_heuristic_ = Aiconfig::Heuristic::no_heuristic;
  ::Heuristic::Rationale last_heuristic_rationale_;

  public:
  Hand const& handofplayer(Player const& player) const;
  Team teamofplayer(Player const& player) const; 

  protected:
  unique_ptr<VirtualGamesInterface> vgi_;
  unique_ptr<HeuristicInterface> hi_;

  bool silent_marriage_ = false;
}; // class Ai
