/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 **********************************************************************/

#include "constants.h"

#include "aiDb.h"

#include <sstream>

/** constructor
 **/
AiDb::AiDb() = default;

/** constructor
 **/
AiDb::AiDb(istream& istr) :
  PlayersDb()
{
  this->read(istr);
}

/** constructor
 **/
AiDb::AiDb(PlayersDb const& players_db) :
  PlayersDb(players_db)
{}

/** copy constructor
 **/
AiDb::AiDb(AiDb const& aidb) = default;

/** destructor
 **/
AiDb::~AiDb() = default;

/** @return   a number for the group 'general'
 **/
unsigned
AiDb::heuristics_group_general() const
{
  unsigned num = 0;

  for (auto const heuristic : aiconfig_heuristic_list) {
    if (is_general(heuristic)) {
      num += this->heuristic(heuristic);
    }
  }

  return num;
} // unsigned AiDb::heuristics_group_general() const

/** @return   a number for the group 'poverty'
 **/
unsigned
AiDb::heuristics_group_poverty() const
{
  unsigned num = 0;

  for (auto const heuristic : aiconfig_heuristic_list) {
    if (for_poverty(heuristic)) {
      num += this->heuristic(heuristic);
    }
  }

  return num;
} // unsigned AiDb::heuristics_group_poverty() const

/** @return   a number for the group 'solo'
 **/
unsigned
AiDb::heuristics_group_solo() const
{
  unsigned num = 0;

  for (auto const heuristic : aiconfig_heuristic_list) {
    if (for_solo(heuristic)) {
      num += this->heuristic(heuristic);
    }
  }

  return num;
} // unsigned AiDb::heuristics_group_solo() const

/** @return   a number for the group 'solo color'
 **/
unsigned
AiDb::heuristics_group_solo_color() const
{
  unsigned num = 0;

  for (auto const heuristic : aiconfig_heuristic_list) {
    if (for_color_solo(heuristic)) {
      num += this->heuristic(heuristic);
    }
  }

  return num;
} // unsigned AiDb::heuristics_group_solo_color() const

/** @return   a number for the group 'solo picture'
 **/
unsigned
AiDb::heuristics_group_solo_picture() const
{
  unsigned num = 0;

  for (auto const heuristic : aiconfig_heuristic_list) {
    if (for_picture_solo(heuristic)) {
      num += this->heuristic(heuristic);
    }
  }

  return num;
} // unsigned AiDb::heuristics_group_solo_picture() const

/** @return   a number for the group 'solo meatless'
 **/
unsigned
AiDb::heuristics_group_solo_meatless() const
{
  unsigned num = 0;

  for (auto const heuristic : aiconfig_heuristic_list) {
    if (for_meatless_solo(heuristic)) {
      num += this->heuristic(heuristic);
    }
  }

  return num;
} // unsigned AiDb::heuristics_group_solo_meatless() const

/** writes the database to 'ostr'
 **
 ** @param    ostr   output stream
 **
 ** @return   output stream
 **/
ostream&
AiDb::write(ostream& ostr) const
{
  this->PlayersDb::write(ostr);

  ostr << '\n';

  ostr << "heuristics\n"
    << "{\n";
  for (auto const heuristic : aiconfig_heuristic_list) {
    ostr << setw(32) << heuristic << " =\t"
      << this->heuristic(heuristic) << '\n';
  }
  ostr <<"}\n";

  return ostr;
} // ostream& AiDb::write(ostream& ostr) const

/** reads the group 'group' from 'istr'
 **
 ** @param    group   name of the group to read
 ** @param    istr   stream with the group info
 **
 ** @return   whether the group 'group' is known
 **/
bool
AiDb::read_group(string const& group, istream& istr)
{
  if (group == "heuristics")
    this->read_group_heuristics(istr);
#ifndef OUTDATED
  // 0.7.4   2006-09-14
  else if (group == "Heuristics")
    this->read_group_heuristics(istr);
#endif
  else
    return this->PlayersDb::read_group(group, istr);

  return true;
} // bool AiDb::read_group(string const& group, istream& istr)

/** reads the heuristics group from 'istr'
 **
 ** @param    istr   stream with the heuristics info
 **/
void
AiDb::read_group_heuristics(istream& istr)
{
  unsigned depth = 0;

  // load the configuration
  while (istr.good()) {
    Config config;
    istr >> config;

    if (config.separator) {
      auto const heuristic = Aiconfig::Heuristic_from_name(config.name);
      if (heuristic != Aiconfig::Heuristic::error) {
        istringstream istr_value(config.value);
        istr_value >> this->heuristics_[heuristic];
      } else {
        cerr << "ai database: group 'heuristics': "
          << "unknown type '" << config.name << "'."
          << endl;
      }
    } else { // if (config.separator)
      if(config.name == "{") {
        depth += 1;
      } else if(config.name == "}") {
        if (depth == 0) {
          cerr << "ai database: group 'heuristics': "
            << "found a '}' without a '{' before.\n"
            << "Finish reading the the file."
            << endl;
          break;
        } // if (depth == 0)
        depth -= 1;
        if (depth == 0)
          break;
      } else if(config.name == "") {
        cerr << "ai database: group 'heuristics': "
          << "Ignoring line \'" << config.value << "\'.\n";
      } else {
        cerr << "ai database: group 'heuristics': "
          << "type '" << config.name << "' unknown.\n"
          << "Ignoring it.\n";
      } // if (config.name == .)
    } // config.separator
  } // while (istr.good())

  return ;
} // void AiDb::read_group_heuristics(istream& istr)


/** @return   how often each statistic was used
 **/
unsigned
AiDb::heuristic(Aiconfig::Heuristic const heuristic) const
{
  return this->heuristics_[heuristic];	
}

/** increase the counter for heuristic
 **/
void
AiDb::increase(Aiconfig::Heuristic const heuristic)
{
  this->heuristics_[heuristic] += 1;
}

/** clears the database
 **/
void
AiDb::clear()
{
  for (auto const h : aiconfig_heuristic_list)
    this->heuristics_[h] = 0;
  this->PlayersDb::clear();
} // void AiDb::clear()
