/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "heuristic.h"
#include "ai.h"
#include "../player.h"
#include "team_information.h"
#include "cards_information.h"

#include "heuristics.h"
#include "heuristics/call_function.h"
#include "heuristics/play_highest_color.h"
#include "heuristics/play_trump.h"
#include "heuristics/play_to_jab_later.h"
#include "heuristics/play_bad_color.h"
#include "heuristics/jab_color_trick_with_dolle.h"
#include "heuristics/save_dolle.h"
#include "heuristics/serve_trump_trick.h"
#include "heuristics/pfund_in_first_color_run.h"
#include "heuristics/start_with_low_trump.h"
#include "heuristics/jab_color_over_fox.h"
#include "heuristics/jab_for_doppelkopf.h"
#include "heuristics/try_for_doppelkopf.h"
#include "heuristics/draw_trump.h"
#include "heuristics/best_winning_card.h"

#include "heuristics/meatless/meatless_play_highest_color.h"
#include "heuristics/picture_solo/picture_second_best_trump.h"
#include "heuristics/color_solo/color_jab_for_ace.h"


/** @return   a heuristic object
 **/
unique_ptr<Heuristic>
Heuristic::create(Aiconfig::Heuristic const heuristic,
		  Ai const& ai)
{
  switch (heuristic) {
  case Aiconfig::Heuristic::error:
  case Aiconfig::Heuristic::no_heuristic:
  case Aiconfig::Heuristic::manual:
  case Aiconfig::Heuristic::bug_report:
  case Aiconfig::Heuristic::only_valid_card:
  case Aiconfig::Heuristic::choose_best_card:
    DEBUG_ASSERTION(false,
                    "Heuristic::create(" << heuristic << ")\n"
                    "  heuristic '" << heuristic << "' is not valid");
    return {};
  case Aiconfig::Heuristic::play_to_marry:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_to_marry);
  case Aiconfig::Heuristic::play_to_get_married:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_to_get_married);
  case Aiconfig::Heuristic::choose_ace:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_ace);
  case Aiconfig::Heuristic::choose_ten:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_ten);
  case Aiconfig::Heuristic::choose_single_ace:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_single_ace);
  case Aiconfig::Heuristic::choose_for_color_trick:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_for_color_trick);
  case Aiconfig::Heuristic::pfund_in_first_color_run:
    return make_unique<Heuristic::PfundInFirstColorRun>(ai);
  case Aiconfig::Heuristic::jab_color_over_fox:
    return make_unique<Heuristic::JabColorOverFox>(ai);
  case Aiconfig::Heuristic::start_with_color:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::start_with_color);
  case Aiconfig::Heuristic::start_with_low_color:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::start_with_low_color);
  case Aiconfig::Heuristic::start_with_low_trump:
    return make_unique<Heuristic::StartWithLowTrump>(ai);
  case Aiconfig::Heuristic::retry_color:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::retry_color);
  case Aiconfig::Heuristic::play_color_for_partner:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_color_for_partner);
  case Aiconfig::Heuristic::try_color_for_partner:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::try_color_for_partner);
  case Aiconfig::Heuristic::play_color_for_partner_ace:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_color_for_partner_ace);
  case Aiconfig::Heuristic::play_bad_color:
    return make_unique<Heuristic::PlayBadColor>(ai);
  case Aiconfig::Heuristic::jab_color_trick_with_dolle:
    return make_unique<Heuristic::JabColorTrickWithDolle>(ai);
  case Aiconfig::Heuristic::save_dolle:
    return make_unique<Heuristic::SaveDolle>(ai);
  case Aiconfig::Heuristic::serve_color_trick:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::serve_color_trick);
  case Aiconfig::Heuristic::serve_trump_trick:
    return make_unique<Heuristic::ServeTrumpTrick>(ai);
  case Aiconfig::Heuristic::choose_pfund:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_pfund);
  case Aiconfig::Heuristic::choose_pfund_poverty:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_pfund);
  case Aiconfig::Heuristic::choose_pfund_before_partner:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::choose_pfund_before_partner);
  case Aiconfig::Heuristic::jab_for_ace:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::jab_for_ace);
  case Aiconfig::Heuristic::create_fehl:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::create_fehl);
  case Aiconfig::Heuristic::best_winning:
    return make_unique<Heuristic::BestWinningCard>(ai);
  case Aiconfig::Heuristic::low_high:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_low_high);
  case Aiconfig::Heuristic::play_trump:
    return make_unique<Heuristic::PlayTrump>(ai);
  case Aiconfig::Heuristic::play_for_team:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_for_team);
  case Aiconfig::Heuristic::jab_fox:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::jab_fox);
  case Aiconfig::Heuristic::jab_for_doppelkopf:
    return make_unique<Heuristic::JabForDoppelkopf>(ai);
  case Aiconfig::Heuristic::try_for_doppelkopf:
    return make_unique<Heuristic::TryForDoppelkopf>(ai);
  case Aiconfig::Heuristic::play_for_partner_worries:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_for_partner_worries);
  case Aiconfig::Heuristic::partner_backhand_draw_trump:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::partner_backhand_draw_trump);
  case Aiconfig::Heuristic::draw_trump:
    return make_unique<Heuristic::DrawTrump>(ai);
  case Aiconfig::Heuristic::play_to_jab_later:
    return make_unique<Heuristic::PlayToJabLater>(ai);
  case Aiconfig::Heuristic::play_highest_color_card_in_game:
    return make_unique<Heuristic::PlayHighestColor>(ai);
  case Aiconfig::Heuristic::get_trick_for_announcement:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::get_trick_for_keeping_announcement);
  case Aiconfig::Heuristic::grab_trick:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::grab_trick);
  case Aiconfig::Heuristic::let_trick_pass:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::let_trick_pass);
  case Aiconfig::Heuristic::poverty_special_play_pfund:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_special_play_pfund);
  case Aiconfig::Heuristic::poverty_special_give_no_points:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_special_give_no_points);
  case Aiconfig::Heuristic::poverty_special_offer_pfund:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_special_offer_pfund);
  case Aiconfig::Heuristic::poverty_re_trump_color_trick_high:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_re_trump_color_trick_high);
  case Aiconfig::Heuristic::poverty_re_play_trump:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_re_play_trump);
  case Aiconfig::Heuristic::poverty_contra_play_color:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_contra_play_color);
  case Aiconfig::Heuristic::poverty_contra_trump_color_trick_high:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_contra_trump_color_trick_high);
  case Aiconfig::Heuristic::poverty_leave_to_partner:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_leave_to_partner);
  case Aiconfig::Heuristic::poverty_overjab_re:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_overjab_re);
  case Aiconfig::Heuristic::poverty_best_winning_card:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::poverty_best_winning_card);
  case Aiconfig::Heuristic::play_color_in_solo:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_color_in_solo);
  case Aiconfig::Heuristic::meatless_playhighestcolor:
    return make_unique<Heuristic::MeatlessPlayHighestColor>(ai);
  case Aiconfig::Heuristic::picture_secondbesttrump:
    return make_unique<Heuristic::PictureSecondBestTrump>(ai);
  case Aiconfig::Heuristic::color_jab_for_ace:
    return make_unique<Heuristic::ColorJabForAce>(ai);
  case Aiconfig::Heuristic::color_low_high:
    return make_unique<Heuristic::CallFunction>(ai, heuristic,
                                                &Heuristics::play_low_high);
  }
  return {};
}

/** constructor
 **/
Heuristic::Heuristic(Ai const& ai, Aiconfig::Heuristic const heuristic) :
  ai_(&ai),
  heuristic_(heuristic),
  name_(to_string(heuristic))
{ }

/** destructor
 **/
Heuristic::~Heuristic() = default;

/** @return   the heuristic (enum)
 **/
Aiconfig::Heuristic
Heuristic::heuristic() const
{
  return this->heuristic_;
}

/** @return   the name of the heuristic
 **/
string const&
Heuristic::name() const
{
  return this->name_;
}

/** @return   the rationale of the heuristic
 **/
Heuristic::Rationale const&
Heuristic::rationale() const
{
  return this->rationale_;
}

/** set the player
 **/
void
Heuristic::set_player(Player const& player)
{
  this->player_ = &player;
}

/** @return   the player
 **/
Player const&
Heuristic::player() const
{
  return *this->player_;
}

/** @return   the hand of the player
 **/
Hand const&
Heuristic::hand() const
{
  return this->player_->hand();
}

/** @return   the team of the player
 **/
Team
Heuristic::team() const
{ return this->player_->team(); }

/** @return  whether only the own team is behind
 **/
bool
Heuristic::only_own_team_behind(Trick const& trick) const
{
  auto const playerno = trick.game().players().size();
  // take a look if all players coming in this trick are of mine team
  for (unsigned i = trick.actcardno() + 1; i < playerno; ++i)
    if (!this->guessed_same_team(trick.player_of_card(i)))
      return false;

  return true;
}

/** @return   whether the soloplayer is behind
 **/
bool
Heuristic::soloplayer_behind(Trick const& trick) const
{
  auto const& game = trick.game();
  return (   game.is_solo()
          && !trick.has_played(game.players().soloplayer()));
}

/** @return  whether only the own team is behind
 **/
bool
Heuristic::same_team(Player const& player) const
{
  if (!!player.game().marriage().selector())
    return false;
  return (this->team_information().team(player) == this->team());
}

/** @return  whether only the own team is behind
 **/
bool
Heuristic::guessed_same_team(Player const& player) const
{
  if (!!player.game().marriage().selector())
    return false;
  return (this->guessed_team(player) == this->team());
}

/** @return   the number of color runs
 **/
unsigned
Heuristic::color_runs(Card::TColor const color) const
{
  return this->cards_information().color_runs(color);
}

/** @return   aiconfig value
 **/
bool
Heuristic::value(Aiconfig::Type::Bool const type) const
{
  return this->ai_->value(type);
}

/** @return   aiconfig value
 **/
int
Heuristic::value(Aiconfig::Type::Int const type) const
{
  return this->ai_->value(type);
}

/** @return   aiconfig value
 **/
Card
Heuristic::value(Aiconfig::Type::Card const type) const
{
  return this->ai_->value(type);
}

/** @return   aiconfig value
 **/
Card
Heuristic::trump_card_limit() const
{
  if (this->ai_->silent_marriage_)
    return this->value(Aiconfig::Type::trumplimit_solocolor);

  switch (this->player_->game().type()) {
  case GameType::solo_club:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_diamond:
    return this->value(Aiconfig::Type::trumplimit_solocolor);
  case GameType::solo_jack:
    return this->value(Aiconfig::Type::trumplimit_solojack);
  case GameType::solo_queen:
    return this->value(Aiconfig::Type::trumplimit_soloqueen);
  case GameType::solo_king:
    return this->value(Aiconfig::Type::trumplimit_soloking);
  case GameType::solo_queen_jack:
    return this->value(Aiconfig::Type::trumplimit_solojackqueen);
  case GameType::solo_king_jack:
    return this->value(Aiconfig::Type::trumplimit_solojackking);
  case GameType::solo_king_queen:
    return this->value(Aiconfig::Type::trumplimit_soloqueenking);
  case GameType::solo_koehler:
    return this->value(Aiconfig::Type::trumplimit_solokoehler);
  case GameType::solo_meatless:
    return this->value(Aiconfig::Type::trumplimit_meatless);
  default:
    return this->value(Aiconfig::Type::trumplimit_normal);
  }
}

/** @return   the guessed hand of the player
 **/
TeamInformation const&
Heuristic::team_information() const
{
  return this->ai_->team_information();
}

/** @return   the guessed hand of the player
 **/
Team
Heuristic::guessed_team(Player const& player) const
{
  if (this->value(Aiconfig::Type::teams_known))
    return player.team();

  if (player.game().type() == GameType::marriage)
    return player.team();

  return this->team_information().guessed_team(player);
}

/** @return   the guessed hand of the player
 **/
CardsInformation const&
Heuristic::cards_information() const
{
  return this->ai_->cards_information();
}

/** @return   the guessed hand of the player
 **/
Hand const&
Heuristic::guessed_hand(Player const& player) const
{
  auto const& game = player.game();
  if (   !game.isvirtual()
      && (player == this->player()))
    return this->player_->hand();
  if (this->value(Aiconfig::Type::hands_known))
    return player.hand();

  // If it is the last trick and the player has already played a card,
  // his hand is empty.
  if (   (::game_status == GameStatus::game_play)
      && (game.tricks().real_remaining_no() == 1)
      && (game.tricks().current().cardno_of_player(player)
          < game.tricks().current().actcardno()))
    return player.hand();

#ifndef RELEASE
  Hand const& hand = this->cards_information().estimated_hand(player);
  hand.self_check();
  return hand;
#else
  return this->cards_information().estimated_hand(player);
#endif
}


/** @return   lowest trump card on hand, but gives only fuchs or ten back if there are better cards to keep
 **/
Card
Heuristic::lowest_best_trump_card(Trick const& trick) const
{
  if (   !trick.isempty()
      && !trick.startcard().istrump()
      && this->hand().contains(trick.startcard().tcolor())) {
    return Card::empty;
  }
  auto const trumps = this->hand().trumps();
  auto const& card = trumps.lowest_trump();
  if (card.value() < 10) {
    return card;
  }

  auto card2 = trumps.next_higher_card(card);
  while (   card2
         && card2.value() >= 10) {
    card2 = trumps.next_higher_card(card2);
  }
  if (!card2)
    return card;

  if (card2.less(this->trump_card_limit()))
    return card2;

  return card;
}

/** @return   whether this is a solo game
 **/
bool
Heuristic::condition_solo_meatless(Trick const& trick)
{
  if (trick.game().type() == GameType::solo_meatless) {
    this->rationale_.add(_("Heuristic::condition::meatless solo"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no meatless solo"));
    return false;
  }
}

/** @return   whether this is a picture solo game
 **/
bool
Heuristic::condition_solo_picture(Trick const& trick)
{
  if (is_picture_solo(trick.game().type())) {
    this->rationale_.add(_("Heuristic::condition::picture solo"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no picture solo"));
    return false;
  }
}

/** @return   whether this is a color solo game
 **/
bool
Heuristic::condition_solo_color(Trick const& trick)
{
  if (is_color_solo(trick.game().type())) {
    this->rationale_.add(_("Heuristic::condition::color solo"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no color solo"));
    return false;
  }
}

/** @return   whether this is a solo game
 **/
bool
Heuristic::condition_solo_game(Trick const& trick)
{
  if (is_solo(trick.game().type())) {
    this->rationale_.add(_("Heuristic::condition::solo game"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no solo game"));
    return false;
  }
}

/** @return   whether a soloplayer is behind
 **/
bool
Heuristic::condition_soloplayer_behind(Trick const& trick)
{
  if (this->soloplayer_behind(trick)) {
    this->rationale_.add(_("Heuristic::condition::solo player behind"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::solo player has already played"));
    return false;
  }
}

/** @return   whether this is the startcard of the trick
 **/
bool
Heuristic::condition_startcard(Trick const& trick)
{
  if (trick.isstartcard()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no empty trick"));
    return false;
  }
}

/** @return   whether this is the second card in the trick
 **/
bool
Heuristic::condition_second_card(Trick const& trick)
{
  if (trick.actcardno() == 2) {
    this->rationale_.add(_("Heuristic::condition::second card"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::not second card"));
    return false;
  }
}

/** @return   whether this is the second last card in the trick
 **/
bool
Heuristic::condition_second_last_card(Trick const& trick)
{
  if (trick.remainingcardno() == 2) {
    this->rationale_.add(_("Heuristic::condition::second last card"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::not second last card"));
    return false;
  }
}

/** @return   whether this is the last card of the trick
 **/
bool
Heuristic::condition_last_card(Trick const& trick)
{
  if (trick.islastcard()) {
    this->rationale_.add(_("Heuristic::condition::last card"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no last card"));
    return false;
  }
}

/** @return   whether this is the first color run
 **/
bool
Heuristic::condition_first_color_run(Trick const& trick)
{
  if (trick.isstartcard()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  } else if (trick.startcard().istrump()) {
    this->rationale_.add(_("Heuristic::condition::trump trick"));
    return false;
  } else {
    auto const color = trick.startcard().tcolor();
    if (this->ai_->cards_information().color_runs(color)) {
      this->rationale_.add(_("Heuristic::condition::color was already played"));
      return false;
    } else {
      this->rationale_.add(_("Heuristic::condition::first color run"));
      return true;
    }
  }
}

/** @return   whether this is a color trick
 **/
bool
Heuristic::condition_color_trick(Trick const& trick)
{
  if (trick.isstartcard()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  } else if (trick.startcard().istrump()) {
    this->rationale_.add(_("Heuristic::condition::color trick"));
    return false;
  } else {
    this->rationale_.add(_("Heuristic::condition::trump trick"));
    return true;
  }
}

/** @return   whether this is a pure color trick
 **/
bool
Heuristic::condition_pure_color_trick(Trick const& trick)
{
  if (trick.isstartcard()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }

  for (auto const& card : trick) {
    if (card.istrump()) {
      this->rationale_.add(_("Heuristic::condition::trump in trick"));
      return false;
    }
  }
  this->rationale_.add(_("Heuristic::condition::pure color trick"));
  return true;
}

/** @return   whether the startcard is 'card'
 **/
bool
Heuristic::condition_startcard_is(Trick const& trick, Card const card)
{
  if (trick.isstartcard()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }
  if (trick.startcard() == card) {
    this->rationale_.add(_("Heuristic::condition::startcard is %s", _(card)));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::startcard is not %s", _(card)));
    return false;
  }
}

/** @return   whether the player has color on the hand
 **/
bool
Heuristic::condition_partner_guessed(Trick const& trick)
{
  if (this->ai_->team_information().guessed_partner()) {
    this->rationale_.add(_("Heuristic::condition::partner guessed"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no partner guessed"));
    return false;
  }
}

/** @return   whether the player has color on the hand
 **/
bool
Heuristic::condition_last_player_guessed_same_team(Trick const& trick)
{
  if (this->guessed_same_team(trick.lastplayer())) {
    this->rationale_.add(_("Heuristic::condition::last player of the same team"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::last player not of the same team"));
    return false;
  }
}

/** @return   whether the opposite team has made an announcement
 **/
bool
Heuristic::condition_opposite_team_announcement()
{
  auto const team = this->team();
  if (!is_real(this->team())) {
    this->rationale_.add(_("Heuristic::condition::team not set"));
    return false;
  }
  if (!!this->player().game().announcements().last(opposite(team))) {
    this->rationale_.add(_("Heuristic::condition::announcement of the opposite team"));
    return true;
  }
  this->rationale_.add(_("Heuristic::condition::no announcement of the opposite team"));
  return false;
}

/** @return   whether the player has color on the hand
 **/
bool
Heuristic::condition_has_color()
{
  if (this->hand().hascolor()) {
    this->rationale_.add(_("Heuristic::condition::have color"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::only trump on the hand"));
    return false;
  }
}

/** @return   whether the player has trump on the hand
 **/
bool
Heuristic::condition_has_trump()
{
  if (this->hand().hastrump()) {
    this->rationale_.add(_("Heuristic::condition::have trump"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::no trump on the hand"));
    return false;
  }
}

/** @return   whether the player has the color on the hand
 **/
bool
Heuristic::condition_has(Card::Color const color)
{
  if (this->hand().contains(color)) {
    this->rationale_.add(_("Heuristic::condition::have %s", _(color)));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::have not %s", _(color)));
    return false;
  }
}

/** @return   whether the player has color on the hand
 **/
bool
Heuristic::condition_has_ten_points_in_trump()
{
  if (this->hand().contains_ten_points_in_trump()) {
    this->rationale_.add(_("Heuristic::condition::have ten points in trump"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::have no ten points in trump"));
    return false;
  }
}

/** @return   whether the player has a dolle
 **/
bool
Heuristic::condition_has_dolle()
{
  if (this->hand().count_dolle()) {
    this->rationale_.add(_("Heuristic::condition::have dolle"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::have no dolle"));
    return false;
  }
}

/** @return   whether an opposite player is behind
 **/
bool
Heuristic::condition_opposite_team_behind(Trick const& trick)
{
  if (trick.islastcard()) {
    this->rationale_.add(_("Heuristic::condition::last card"));
    return true;
  }

  auto const& game = trick.game();
  if (   (game.type() == GameType::marriage)
      && game.marriage().is_undetermined()
      && (this->player_->team() != Team::re)) {
    this->rationale_.add(_("Heuristic::condition::undetermined marriage"));
    return true;
  }

  auto const playerno = game.players().size();
  for (unsigned i = trick.actcardno() + 1; i < playerno; ++i) {
    auto const& p = trick.player_of_card(i);
    if (!this->guessed_same_team(p)) {
      this->rationale_.add(_("Heuristic::condition::opposite team behind"));
      return true;
    }
  }

  this->rationale_.add(_("Heuristic::condition::only own team behind"));
  return false;
}

/** @return   whether an opposite player has a swine or hyperswine
 **/
bool
Heuristic::condition_opposite_swine_or_hyperswine(Trick const& trick)
{
  auto const& cards_information = this->ai_->cards_information();
  for (auto const& p : trick.game().players()) {
    if (!this->guessed_same_team(p)) {
      auto const& h = cards_information.estimated_hand(p);
      if (h.count_swines()) {
        this->rationale_.add(_("Heuristic::condition::%s has a swine", p.name()));
        return true;
      } else if (h.count_hyperswines()) {
        this->rationale_.add(_("Heuristic::condition::%s has a hyperswine", p.name()));
        return true;
      }
    }
  }

  this->rationale_.add(_("Heuristic::condition::no opposite swine or hyperswine"));
  return false;
}

/** @return   whether an opposite player behind has swine or hyperswine
 **/
bool
Heuristic::condition_opposite_swine_or_hyperswine_behind(Trick const& trick)
{
  auto const& cards_information = this->ai_->cards_information();
  auto const playerno = trick.game().players().size();
  for (unsigned i = trick.actcardno() + 1; i < playerno; ++i) {
    auto const& p = trick.player_of_card(i);
    if (!this->guessed_same_team(p)) {
      auto const& h = cards_information.estimated_hand(p);
      if (h.count_hyperswines()) {
        this->rationale_.add(_("Heuristic::condition::%s has a swine", p.name()));
        return true;
      } else if (h.count_hyperswines()) {
        this->rationale_.add(_("Heuristic::condition::%s has a hyperswine", p.name()));
        return true;
      }
    }
  }

  this->rationale_.add(_("Heuristic::condition::no opposite swine or hyperswine behind"));
  return false;
}

/** @return   whether the card can be jabbed by the opposite team
 **/
bool
Heuristic::condition_can_be_jabbed_by_opposite_team(Trick const& trick,
                                                    Card const card)
{
  if (trick.islastcard()) {
    this->rationale_.add(_("Heuristic::condition::last card"));
    return false;
  }

  auto const playerno = trick.game().players().size();
  bool only_own_team_behind = true;
  for (unsigned i = trick.actcardno() + 1; i < playerno; ++i) {
    auto const& p = trick.player_of_card(i);
    if (!this->guessed_same_team(p)) {
      only_own_team_behind = false;
      Trick trick2 = trick;
      trick2.add(HandCard(this->player_->hand(), card));
      Hand const& hand2 = this->cards_information().estimated_hand(p);
      if (hand2.higher_card_exists(trick2.winnercard())) {
        this->rationale_.add(_("Heuristic::condition::opposite team behind can jab %s", _(card)));
        return true;
      }
    }
  }

  if (only_own_team_behind) {
    this->rationale_.add(_("Heuristic::condition::only own team behind"));
    return false;
  }
  this->rationale_.add(_("Heuristic::condition::no opposite player behind can jab the trick"));
  return false;
}

/** @return   whether the opposite team has trump
 **/
bool
Heuristic::condition_opposite_team_has_trump()
{
  auto const& players = this->player().game().players();
  for (auto const& p : players) {
    if (!guessed_same_team(p)) {
      if (this->guessed_hand(p).hastrump()) {
        this->rationale_.add(_("Heuristic::condition::the other team has trumps"));
        return true;

      }
    }
  }
  this->rationale_.add(_("Heuristic::condition::the other team has no trumps"));
  return false;
}

/** @return   whether the trick is a trump trick
 **/
bool
Heuristic::condition_trump_trick(Trick const& trick)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }
  if (trick.startcard().istrump()) {
    this->rationale_.add(_("Heuristic::condition::trump trick"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::color trick"));
    return false;
  }
}

/** @return   whether the winnerplayer is of the same team as the player
 **/
bool
Heuristic::condition_winnerplayer_same_team(Trick const& trick)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }
  auto const& game = trick.game();
  auto const& winnerplayer = trick.winnerplayer();
  if (   (game.type() == GameType::marriage)
      && game.marriage().is_undetermined()) {
    if (   (this->player_->team() != Team::re)
        || !is_selector(trick.startcard().tcolor(),
                        game.marriage().selector())) {
      this->rationale_.add(_("Heuristic::condition::undetermined marriage"));
      return false;
    }
  }
  if (this->same_team(winnerplayer)) {
    this->rationale_.add(_("Heuristic::condition::winnerplayer same team"));
    return true;
  } else if (this->guessed_same_team(winnerplayer)) {
    this->rationale_.add(_("Heuristic::condition::winnerplayer guessed same team"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::winnerplayer not of same team"));
    return false;
  }
}

/** @return   whether the winnercard is trump
 **/
bool
Heuristic::condition_jabbed_with_trump(Trick const& trick)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  } else if (trick.winnercard().istrump()) {
    this->rationale_.add(_("Heuristic::condition::jabbed with trump"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::not jabbed with trump"));
    return false;
  }
}

/** @return   whether the own team has jabbed with at least trump card limit
 **/
bool
Heuristic::condition_own_team_jabbed_with_trump_card_limit(Trick const& trick)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  } else if (!this->guessed_same_team(trick.winnerplayer())) {
    this->rationale_.add(_("Heuristic::condition::winnerplayer not of own team"));
    return false;
  } else if (!trick.winnercard().istrump()) {
    this->rationale_.add(_("Heuristic::condition::winnercard is color card"));
    return false;
  } else if (trick.winnercard().less(this->trump_card_limit())) {
    this->rationale_.add(_("Heuristic::condition::winnercard below trump card limit"));
    return false;
  } else {
    this->rationale_.add(_("Heuristic::condition::winnercard at least trump card limit"));
    return true;
  }
}

/** @return   whether the player can jab
 **/
bool
Heuristic::condition_can_jab(Trick const& trick)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }

  if (this->hand().can_jab(trick)) {
    this->rationale_.add(_("Heuristic::condition::can jab"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::cannot jab"));
    return false;
  }
}

/** @return   whether the player can jab with trump
 **/
bool
Heuristic::condition_can_jab_with_trump(Trick const& trick)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }

  auto const color = trick.startcard().tcolor();
  if (   (color != Card::TColor::trump)
      && this->hand().contains(color)) {
    this->rationale_.add(_("Heuristic::condition::cannot jab with trump"));
    return false;
  } else {
    this->rationale_.add(_("Heuristic::condition::can jab with trump"));
    return true;
  }
}

/** @return   whether the player can jab with trump
 **/
bool
Heuristic::condition_can_jab_with(Trick const& trick, Card const card)
{
  if (trick.isempty()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }

  if (!trick.isvalid(card, this->hand())) {
    this->rationale_.add(_("Heuristic::condition::%s not valid", _(card)));
    return false;
  } else if (trick.isjabbed(HandCard(this->hand(), card))) {
    this->rationale_.add(_("Heuristic::condition::trick is jabbed by %s", _(card)));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::trick is not jabbed by %s", _(card)));
    return false;
  }
}

/** @return   whether there are enough points for a doppelkopf in the trick so far
 **/
bool
Heuristic::condition_enough_trick_points_for_doppelkopf(Trick const& trick)
{
  if (trick.points() >= 10 * trick.actcardno()) {
    this->rationale_.add(_("Heuristic::condition::enough trick points so far (%u points for %u players)", trick.points(), trick.actcardno()));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::not enough points"));
    return false;
  }
};
