/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "heuristics.h"

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"


#ifndef RELEASE
// whether to print some weighting informations
//#define ANNOUNCE_DEBUG
//#define ANNOUNCE_DEBUG_DETAIL  //CalcHandValue
#endif

/** value of cards for announcement
 **
 ** @param    hi        heuristic interface
 **
 ** @return   the value of the hand
 **/
int
Heuristics::CalcHandValue( HeuristicInterface const& hi)
{
  return Heuristics::CalcHandValue(hi, hi.game());
} // int Heuristics::CalcHandValue( HeuristicInterface hi)

/** value of cards for announcement
 **
 ** @param    hi        heuristic interface
 ** @param    g         current game
 **
 ** @return   the value of the hand
 **/
int
Heuristics::CalcHandValue( HeuristicInterface const& hi, const Game& g )
{
  // @heuristic::name   ?  calc hand value  ?
  // @heuristic::idea calculates the value of a hand for decision making of announcements, soli decisions,...
  int value = 0;

  unsigned v_queen = 0;
  unsigned v_jack = 0;
  unsigned v_highest = 0;
  unsigned v_king = 0;
  unsigned v_ace = 0;
  unsigned v_fehl = 0;

  switch (g.type()) {
  case GameType::normal:
    { v_queen = 2; v_jack = 1; v_highest = 3; v_king = 0; v_ace = 1; v_fehl = 2;

      if (hi.value(Aiconfig::Type::aggressive))
        v_highest += 1;
      if (hi.team() == Team::re)
      {
        value = 1;
      } else
      {
        value = 0;
      }
    }
    break;

  case GameType::poverty:
    { v_queen = 2; v_jack = 1; v_highest = 3; v_king = 0; v_ace = 0; v_fehl = 2;
      value = -3;
    }
    break;

  case GameType::marriage:
    { v_queen = 2; v_jack = 1; v_highest = 3; v_king = 0; v_ace = 1; v_fehl = 2; }
    break;

  case GameType::solo_jack:
    { v_queen = 0; v_jack = 3; v_highest = 0; v_king = 0; v_ace = 2; v_fehl = 1;
      if (hi.team() == Team::re) {
        value = -6;
      } else {
        value = -2;
      }
    }
    break;

  case GameType::solo_queen:
    { v_queen = 3; v_jack = 0; v_highest = 0; v_king = 0; v_ace = 2; v_fehl = 1;
      if (hi.team() == Team::re) {
        value = -6;
      } else {
        value = -2;
      }
    }
    break;

  case GameType::solo_king:
    { v_queen = 0; v_jack = 0; v_highest = 0; v_king = 3; v_ace = 2; v_fehl = 1;
      if (hi.team() == Team::re) {
        value = -6;
      } else {
        value = -2;
      }
    }
    break;

  case GameType::solo_queen_jack:
    { v_queen = 3; v_jack = 2; v_highest = 0; v_king = 0; v_ace = 1; v_fehl = 1;
      if (hi.team() == Team::re) {
        value = -6;
      } else {
        value = -3;
      }
    }
    break;

  case GameType::solo_king_jack:
    { v_queen = 0; v_jack = 2; v_highest = 0; v_king = 3; v_ace = 1; v_fehl = 1;
      if (hi.team() == Team::re) {
        value = -6;
      } else {
        value = -3;
      }
    }
    break;

  case GameType::solo_king_queen:
    { v_queen = 2; v_jack = 0; v_highest = 0; v_king = 3; v_ace = 1; v_fehl = 1;
      if (hi.team() == Team::re) {
        value = -6;
      } else {
        value = -3;
      }
    }
    break;

  case GameType::solo_koehler:
    { v_queen = 2; v_jack = 1; v_highest = 0; v_king = 3; v_ace = 0; v_fehl = 0;
      if (hi.team() == Team::re) {
        value = -5;
      } else {
        value = -2;
      }
    }
    break;

  case GameType::solo_club:
  case GameType::solo_heart:
  case GameType::solo_spade:
  case GameType::solo_diamond:
    { v_queen = 2; v_jack = 1; v_highest = 3; v_king = 0; v_ace = 1; v_fehl = 2;
      if (hi.team() == Team::re) {
        value = -7;
      } else {
        value = 0;
      }
    }
    break;


  case GameType::solo_meatless:
    {
      v_queen = 0; v_jack = 0; v_highest = 0; v_king = 0; v_ace = 4; v_fehl = 1; // missing aces
      value = -2;
    }
    break;

  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::thrown_richness:
  case GameType::fox_highest_trump:
  case GameType::redistribute:
  case GameType::genscher:
  case GameType::marriage_solo:
  case GameType::marriage_silent:
                break;
  }

  if ( !hi.game().rule()(Rule::Type::with_nines) ) {
    v_ace= max(0,static_cast<int>(v_ace)-1);
    value += 1; // compensate for reduced ace value
  }

#ifdef ANNOUNCE_DEBUG_DETAIL
  COUT << "CalcHandValue(1)\t" << hi.player().name() << "\t" << value << endl;
#endif


  Player const& player = hi.player();

  for (auto const card : hi.hand()) {

    if(   card.isdolle()
       || card.isswine()
       || card.possible_swine()
       || card.ishyperswine()
       || card.possible_hyperswine()
      ) {
      value += v_highest;
    } else if( card.value()==Card::queen
              || card.possible_genscher()  ) {
      value += v_queen;
      if( g.type()==GameType::solo_queen &&
         card.color() == Card::club )
        value +=1;
    } else if( card.value()==Card::jack ) {
      value += v_jack;
      if( g.type()==GameType::solo_jack &&
         card.color() == Card::club )
        value +=1;
    } else if( card.value()==Card::king ) {
      value += v_king;
      if( g.type()==GameType::solo_king &&
         card.color() == Card::club )
        value +=1;
    } else if(    card.value()==Card::ace
              && !card.istrump()
              && !hi.jabbedbyotherteam(card.color())) {
      value += v_ace;
      if (::in_running_game()
          && (hi.game().tricks().current().cardno_of_player( player ) == 0 ) ) {
        if (hi.hand().count( card.color() ) == 1) {
          value += v_ace;
          if( hi.color_runs( card.color() ) == 0 ) {
            value += v_ace;
          }
        } else {
          value += 1;
          if( hi.color_runs( card.color() ) == 0 ) {
            value += v_ace;
          }
        }

      }
      if ( card.color() == Card::heart
          && hi.hand().count( card.color() ) > 1
          && hi.game().rule()(Rule::Type::dollen)) {
        value -=1;
      }
    }

  }
#ifdef ANNOUNCE_DEBUG_DETAIL
  COUT << "CalcHandValue(2)\t" << hi.player().name() << "\t" << value << endl;
#endif

  for (auto const color : hi.game().rule().card_colors()) {
    if (   !Card( color, Card::ace ).istrump( hi.game() )
        && !hi.hand().contains(color)
        && !hi.jabbedbyotherteam(color)
       )
      if( ::in_running_game() ) {
        if(     hi.game().tricks().current().actcardno()>0
           && hi.color_runs( color ) == 0
           && (hi.game().tricks().current().startcard().tcolor()
               != color) )
          value += v_fehl;
      } else {
        value += v_fehl;
      }
  }

  if( g.type() == GameType::solo_meatless ) {
    for (auto const color : hi.game().rule().card_colors()) {
      if( hi.hand().contains(color) )
        value -=  (v_fehl
                   * ( 2 - hi.hand().count(color, Card::ace ) ));
    }
    if( hi.is_soloplayer()) {
      unsigned longColor = 0;
      for (auto const color : hi.game().rule().card_colors()) {
        if(   hi.hand().count(color, Card::ace)  == 2
           && hi.hand().count(color) > longColor )
          longColor = hi.hand().count(color);
      }
      value += longColor;
    }
  } else  {

    for (auto const color : hi.game().rule().card_colors()) {

      if( color == hi.game().cards().trumpcolor() )
        continue;

      // reduction for long colors
      if (         hi.hand().count( color ) > 2
          && !hi.jabbedbyownteam(color) )
        value -= hi.hand().count( color ) - (2 + hi.hand().count( color, Card::ace ));

      // including this trick the player does not have the color
      if (    !hi.hand().contains( color )
          && !(    ::in_running_game()
               && ( hi.game().tricks().current().actcardno()
                   > hi.game().tricks().current().cardno_of_player( player ) )
               && ( hi.game().tricks().current().card_of_player( player ).tcolor()
                   == color )
              )
         )
        continue;

      if (  ( hi.hand().count( color, Card::ace ) == 0 )
          && !hi.jabbedbyownteam(color) )
      {
        // previous        value -= v_ace;
        value -= v_fehl;
        //unprotected tens on hand
        if (    hi.hand().count( color, Card::ten ) > 0
            &&  hi.hand().count( color, Card::ten ) == hi.hand().count( color ) )
          value -= v_fehl;
      }

      if( (   hi.hand().count( color )
           + hi.color_runs( color )
           + (   ::in_running_game()
              && hi.game().tricks().current().actcardno() > 0
              && ( hi.game().tricks().current().startcard().color()
                  == color
                 ) ? 1 : 0 )
           > ( hi.game().cards().count_unique(color) - 1
              - ( hi.game().rule()(Rule::Type::with_nines) ? 0 : 1 ) )
          ) && !hi.jabbedbyownteam(color) )
        value -= v_fehl;

      if ( (   hi.hand().count( color )
            + hi.color_runs( color )
            + ( (   ::in_running_game()
                 && hi.game().tricks().current().actcardno()>0
                 && ( hi.game().tricks().current().startcard().color()
                     == color)
                ) ? 1 : 0)
            >= 3)
          && ( hi.hand().count( Card( color, Card::ace ) ) < 2 )
          && !hi.jabbedbyownteam(color) )
        value -= v_ace;

    } // (for color)
  } // if (...)
#ifdef ANNOUNCE_DEBUG_DETAIL
  COUT << "CalcHandValue(3)\t" << hi.player().name() << "\t" << value << endl;
#endif

  if( g.type() == GameType::solo_koehler ) {
    for (auto color : hi.game().rule().card_colors()) {
      if (   hi.hand().contains(color)
          && !hi.jabbedbyownteam(color) )
        value -= 5;
    }
  } // if( g.type() == GameType::solo_koehler )


  if (   (   hi.game().swines().swines_announced()
          && hi.game().swines().swines_owner() != hi.player() )
      || hi.game().swines().swines_announcement_valid( hi.player() )) {
    Player const& swines_player = (  hi.game().swines().swines_announced()
                                   ? hi.game().swines().swines_owner()
                                   : hi.player() );
    if (   hi.teamofplayer(swines_player) == hi.team()
        || hi.game().swines().swines_announcement_valid(swines_player)) {
      if (swines_player != hi.player() ) {
        value += v_highest;
      } else {
        value += 1;
      }
    } else {
      value -= 2* v_highest;
    }
  }

#ifdef ANNOUNCE_DEBUG_DETAIL
  COUT << "CalcHandValue(4)\t" << hi.player().name() << "\t" << value << endl;
#endif

  if(  (   hi.game().swines().hyperswines_announced()
        && hi.game().swines().hyperswines_owner() != hi.player()
       )
     || hi.game().swines().hyperswines_announcement_valid(hi.player())
    )
  {
    Player const& hyperswines_player = (  hi.game().swines().hyperswines_announced()
                                        ? hi.game().swines().hyperswines_owner()
                                        : hi.player() );
    if (   hi.teamofplayer(hyperswines_player) == hi.team()
        || hi.game().swines().hyperswines_announcement_valid( hyperswines_player )) {
      if (hyperswines_player != hi.player()) {
        value += v_highest;
      } else {
        value += 1;
      }
    } else {
      value -= 2 * v_highest;
    }
  }

#ifdef ANNOUNCE_DEBUG_DETAIL
  COUT << "CalcHandValue(5)\t" << hi.player().name() << "\t" << value << endl;
#endif


  if( !is_solo( hi.game().type() )
     || hi.is_soloplayer())
  {
    if (hi.hand().count_trumps()
        >= ( (hi.cards_information().remaining_trumps_others() / 3)
            + round(  (::in_running_game()
                       ? hi.game().tricks().remaining_no()
                       : hi.game().tricks().max_size()
                      )/ 3)
           ) )
      value += 2;
  }

  if (::in_running_game()
      && (hi.game().tricks().current().cardno_of_player( player ) == 0 ) )
  { // count aces like trumps, if I serve
    unsigned cardno = hi.hand().count_trumps();

    for (auto const color : hi.game().rule().card_colors()) {

      if( color == hi.game().cards().trumpcolor() )
        continue;

      if(  ((hi.color_runs( color ) == 0)
            || hi.jabbedbyownteam(color) )
         && (hi.hand().count( color ) <= 1 + hi.hand().count( Card( color, Card::ace ) ) )
        )
      {
        cardno += hi.hand().count( Card( color, Card::ace ) );

      }

    }

    if (cardno >= 9 )
      value += 1;
    if (cardno >= hi.game().tricks().remaining_no() - 1 )
      value += 1;
    if (cardno >= hi.game().tricks().remaining_no() )
      value += 2;
    if (cardno == hi.hand().cardsnumber() )
      value += 2;


    if ( g.is_solo()
        && cardno >= 9 )
      value += 2; // reference 096567

  } else
  {
    if (hi.hand().count_trumps() >= 9 )
      value += 1;
    if (hi.hand().count_trumps() >= hi.game().tricks().remaining_no() - 1 )
      value += 2;
    if (hi.hand().count_trumps() >= hi.game().tricks().remaining_no() )
      value += 2;
    if (hi.hand().count_trumps() == hi.hand().cardsnumber() )
      value += 2;

    if ( g.is_solo()
        && hi.hand().count_trumps() >= 9 )
      value += 1;
  }



  if (hi.hand().count_trumps() -1 < (hi.hand().cardsnumber()+1) / 2 )
    value -= 1;
  if (hi.hand().count_trumps() -1 < (hi.hand().cardsnumber()+1) / 3 )
    value -= 1;
  if (hi.hand().count_trumps() -1 < (hi.hand().cardsnumber()+1) / 4 )
    value -= 1;

  if ( g.type() == GameType::normal ) // remove a value if a lot of small trumps are on hand
  {
    value -= (  hi.hand().count_trumps() //2do
              - hi.hand().count( Card::queen )
              - hi.hand().count( Card::jack )
              - hi.hand().count_swines()
              - hi.hand().count_hyperswines()
              - hi.hand().count_dolle()) / 4;
  }

  if ( g.type() == GameType::normal ) // remove a value for two heart aces
  {
    if(    hi.game().rule()(Rule::Type::dollen)
       && hi.game().cards().trumpcolor() != Card::heart
       && hi.hand().count(Card::heart,Card::ace) == 2)
      value -=1;
  }

  if (    (::game_status > GameStatus::game_start)
      && (hi.game().type() == GameType::poverty)
      && (hi.game().players().poverty_partner() == hi.player()) )
    value -= 2;

  if ( (   g.type() == GameType::normal
        || g.type() == GameType::poverty )
      && hi.team() == Team::re )
  {
    if( hi.hand().count( Card::spade_queen ) == 2 )
      value += 2;
    if( hi.hand().count( Card::heart_queen ) == 2 )
      value += 1;
  }

  if (   ( hi.hand().count( Card::queen ) < 3 ) // check for some queens on hand
      ||(    hi.team() == Team::contra
         && hi.hand().count( Card::queen ) < 4 )
     )
  {
    value -= v_queen;
  }


  if( ::in_running_game() )
  {
    if( (   hi.game().tricks().current().isfull()  // lost trick
         && maybe_to_team(hi.teamofplayer(hi.game().tricks().current().winnerplayer())) != hi.team()
        ) ||
       (    !hi.game().tricks().current().isempty() // lost color trick
        && !hi.game().tricks().current().startcard().istrump()
        && !hi.is_winnerplayer(hi.game().tricks().current())
       )
      )
      value -= 1;
    if(   hi.game().tricks().current().isempty()
       &&  hi.same_player(hi.game().tricks().current().startplayer()) )
      value += 1;

    //first tricks, other player serves
    if ( hi.game().tricks().current().no() < 2
        &&  !hi.same_player(hi.game().tricks().current().startplayer()) )
      value -= 1;
  }


#ifdef ANNOUNCE_DEBUG_DETAIL
  COUT << "CalcHandValue(6)\t" << hi.player().name() << "\t" << value << endl;
#endif

  return value;
} // Heuristics::CalcHandValue( HeuristicInterface hi, Game g )


/** checks if an announcement should be made
 **
 ** @param    hi        heuristic interface
 **
 ** @return   returns true if an announcement should be made
 **/
bool
Heuristics::make_announcement( HeuristicInterface const& hi)
{
  // @heuristic::name   ?  make announcement  ?
  // @heuristic::idea announcement decision bases on Heuristics::CalcHandValue
  auto const& game = hi.game();
  auto const& player = hi.player();
  auto const& current_trick = game.tricks().current();

  if (hi.player().next_announcement() == Announcement::noannouncement) {
    return false;
  }

  if (max_points_for_opposite_team(hi) < 120)
    return true;

  int const opp_p = calcPointsOfOppositeTeam(hi);
  int const own_p = calcPointsOfOwnTeam(hi);
  if (own_p > 120)
    return true;
  if (opp_p >= 120)
    return false;

  int value = 0;

  value = CalcHandValue( hi );

  value += own_p / 20; // previous 25

  if (   !game.is_solo()
      && !game.swines().swines_announced()
      && !game.swines().hyperswines_announced()) {
    value += ((hi.hand().count_club_queens()+1)*hi.hand().count_dolle())*calcPointsOfOwnTeam(hi) / 25;
  }

  value -= opp_p / 15; // previous 20


  if(   (hi.team() == Team::contra)
     && (hi.hand().count_dolle() == 0)
     && !hi.hand().has_swines()
     && !hi.hand().has_hyperswines()
    ) {
  value -=1;
  }

  if( game.announcements().last( opposite( hi.team() ) ).announcement
     != Announcement::noannouncement  ) {
    value -= 1;
    if (   is_solo(game.type())
        && (hi.team() == Team::re) ) {
      value -= 6;

    }

  }
  if (   is_solo(game.type())
      && (hi.team() == Team::re) )
  {
    value -= 2;

    if (game.type() == GameType::solo_meatless)
      value -=2;

    if (   game.swines().swines_announced()
        && game.players().soloplayer() != game.swines().swines_owner() ) {
      value -=5;

    }
    if (   game.swines().hyperswines_announced()
        && game.players().soloplayer() != game.swines().hyperswines_owner() ) value -=2;
  }

  if (   !is_solo(game.type())
      && hi.team() == Team::contra
      && hi.hand().count_dolle() == 2)
    value +=1; // reference 235643

  // as startplayer color aces to play increase the value
  if (   current_trick.no() == 0
      && current_trick.isempty()
      && !hi.game().is_solo()) {
    auto const& current_trick = hi.game().tricks().current();
    if (current_trick.startplayer() == hi.player()) {
      auto const& hand = hi.hand();
      auto const& game = hi.game();
      auto const& cards_information = hi.cards_information();
      int modifier = 0;
      for (auto const color : game.cards().colors()) {
        auto const ace = Card(color, Card::ace);
        if (   !ace.istrump(game)
            && hand.contains(ace)) {
          if (cards_information.remaining_others(color) == 4) {
            value += 2 + modifier;
            modifier -= 1;
          } else if (cards_information.remaining_others(color) > 4) {
            value += 4 + modifier;
            modifier -= 1;
          }
        }
      }
    }
  }


  // (trumps - no trumps) = trumps - (cards on hand - trumps ) = 2* trumps - cards on hand
  if (2 * hi.hand().count_trumps() > hi.hand().cardsnumber()) {

    if ( hi.value(Aiconfig::Type::aggressive)) {
      value += (2 * hi.hand().count_trumps() - hi.hand().cardsnumber() ) / 6;
    } else {
      value += (2 * hi.hand().count_trumps() - hi.hand().cardsnumber() ) / 9;
    }
  }


  Trick const t = (game_status < GameStatus::game_play
                   ? Trick(game.startplayer())
                   : current_trick);
  if (   hi.teamofplayer(t.winnerplayer()) != hi.team()
      ||  oppositeTeamCanWinTrick( t, hi ) ) {
    value -= 1;
    if ( is_solo( game.type() )
        && ( hi.team() == Team::re ) )
      value -= 5;
  }
  if (   hi.teamofplayer(t.winnerplayer()) == hi.team()
      &&  !oppositeTeamCanWinTrick( t, hi ) )
    value += 2;

  if (   (game.type() == GameType::marriage)
      && (game.marriage().selector() == MarriageSelector::team_set)
      && ( hi.team() == Team::re ) ) {
    if ( hi.is_soloplayer())
      value += 1;
    else
      value += 3; // previous 6 // 2do should be 3*numberofqueens still in game
    if( hi.is_winnerplayer(t))
      value +=  hi.hand().count( Card::club, Card::ace )
        + hi.hand().count( Card::spade, Card::ace )
        + hi.hand().count( Card::heart, Card::ace );

    if (   game.swines().swines_announced()
        && game.players().soloplayer() == game.swines().swines_owner())
    {
      value +=2 + 2*hi.hand().count_dolle();
      if (!game.swines().hyperswines_announced())
        value +=2; // reference 040594
    }
    if (   game.swines().hyperswines_announced()
        && game.players().soloplayer() == game.swines().hyperswines_owner() )
      value +=2;
    if (   !game.swines().swines_announced()
        && !game.swines().hyperswines_announced()) {
      value += 2 + 2 * hi.hand().count_dolle();
    }
  } else {
    if ( hi.team() == Team::re ) {
      if ( hi.hand().has_swines() ) {
        value +=2 * hi.hand().count_dolle();
      }
      if (   !game.swines().swines_announced()
          && !game.swines().hyperswines_announced()) {
        value +=2*hi.hand().count_dolle();
      }
    }

  }

  if (   (game.type() == GameType::poverty)
      && (player == game.players().poverty_partner()) )
  {
    value -= 15 - hi.hand().count_dolle() - hi.hand().count_club_queens(); // previously: 7, 8, 13
    // 13 Reference 243666


    if( current_trick.isfull() )
      value -= 6; // reference 243666
  }

  if (   (game.type() == GameType::normal)
      || (game.type() == GameType::marriage) ) {

    if( hi.hand().count_trumps() < 1.5* hi.hand().cardsnumber() )
      value -=  ( hi.hand().cardsnumber() - hi.hand().count_trumps() - 1);
  }

  if (::in_running_game()) {
    if( (   current_trick.isfull()  // trick won
         && hi.is_winnerplayer(current_trick)
         && current_trick.points() > 15
        )
      )
      value += 1;

    if (   current_trick.isfull()  // trick lost
        && !hi.is_winnerplayer(current_trick) )
      value -= 1;

    if( current_trick.startplayer() != player
       && current_trick.isempty() )
    {
      value -= 1;

    }
  }


#ifdef ANNOUNCE_DEBUG
  if( !game.isvirtual() )
    cout << "re/CO: " << player.name() << "\t" << value << "\t"
      << (int)hi.value(Aiconfig::Type::announcelimit) << " + "
      << (int)hi.value(Aiconfig::Type::announceconfig) << std::endl;
#endif

  return ( value >=  (static_cast<int>(hi.value(Aiconfig::Type::announcelimit))
                      + static_cast<int>(hi.value(Aiconfig::Type::announceconfig))));
} // bool Heuristics::make_announcement( HeuristicInterface hi)





/** checks if No90 should be said
 **
 ** @param    hi        heuristic interface
 **
 ** @return   true if hi should say No90
 **/
bool
Heuristics::say_no90( HeuristicInterface const& hi )
{
  // @heuristic::name   ?  say no 90  ?
  // @heuristic::idea decision to make the annoncement no 90 depending on Heuristics::CalcHandValue
  if (max_points_for_opposite_team(hi) < 90)
    return true;

  int value = 0;
  Game const& g = hi.game();

  int own_p = calcPointsOfOwnTeam(hi);
  int opp_p = calcPointsOfOppositeTeam(hi);

  if( own_p > 150 )
    return true;

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     > Announcement::no120 )
    return false;

  value = CalcHandValue( hi )
    - max( 0, static_cast<int>(g.tricks().remaining_no()) - 7)
    * static_cast<int>(hi.value(Aiconfig::Type::announcelimitdec));

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     != Announcement::noannouncement  )
    value -= 2;

  if  (   hi.team_information().guessed_partner()
       && (hi.team_information().guessed_partner()->announcement()
           != Announcement::noannouncement)
       && (hi.team() == Team::re) ) {
    if (   !hi.game().swines().swines_announced()
        && !hi.game().swines().hyperswines_announced()) {
      value += 8*hi.hand().count_dolle(); // 7: reference 123343, 8: 139722_1
    }
    opp_p -= 21* hi.hand().count_dolle(); // reference 237268
    value += 2; // reference 139722_1,
  }

  if (   (hi.game().type() == GameType::poverty )
      && (hi.game().players().poverty_partner() == hi.player()))
  {
    value -= 2;
    value -= (2-hi.hand().count_dolle());
    value -= 5 - hi.hand().count_dolle() - hi.hand().count_club_queens();
    value -= 7 - hi.hand().count(Card::queen);
  }

  value -= (2-hi.hand().count_dolle());

  if (   !hi.game().swines().swines_announced()
      && !hi.game().swines().hyperswines_announced())
  { // reference 101732
    value += hi.hand().count_dolle()*hi.hand().count_club_queens();
    if( (hi.cards_information().played(Card::club_queen) == 0)
       && (hi.team() == Team::re)
       && (!is_solo( hi.game().type() ))
      )
      value += 1 * hi.hand().count_dolle(); // reference 139722
  }

  if( (   hi.game().type() == GameType::marriage
       && hi.game().marriage().selector() == MarriageSelector::team_set
       &&  hi.team() == Team::re ) )

  {
    if (   !hi.value(Aiconfig::Type::aggressive) )
    {
      value -=3;
    }

    if( hi.cards_information().played(Card::club_queen) == 0
       && hi.hand().count_trumps() >= g.tricks().remaining_no() - 1
      )

    {

      value += 1 + hi.hand().count_dolle();
      if (hi.hand().count_club_queens() == 2)
        value -=3;
    }
  }
  for (auto const color : hi.game().rule().card_colors()) {
    if( color == hi.game().cards().trumpcolor() )
      continue;

    if ( !hi.jabbedbyownteam(color)
        && hi.hand().count( color ) > 0 )
    {
      if( ( hi.hand().count( color, Card::ace ) == 0 ) )
      {
        value -= (hi.color_runs( color )+hi.hand().count( color ));
      } else
        value -= 1;
    }

  } // (for color)

  if(    !g.tricks().has_trick(::opposite(hi.team())) )
    value += 2 * static_cast<int>(hi.value(Aiconfig::Type::announcelimitdec));

  opp_p+= 3 * g.tricks().remaining_no(); //previous: 5,4
  own_p+= 8 * g.tricks().remaining_no(); // previous: 6, 7

  own_p += static_cast<int>(1.9 * hi.hand().count_trumps()
                            * hi.hand().count_swines());
  own_p += static_cast<int>(1.4 * hi.hand().count_trumps()
                            * hi.hand().count_hyperswines());


  if( is_solo( hi.game().type() ) )
  {
    own_p += 20 * hi.hand().count_trumps();
    opp_p += 20 * hi.cards_information().remaining_trumps_others();
  }



  Trick const t = (game_status < GameStatus::game_play
                   ? Trick(g.startplayer())
                   : g.tricks().current());

  if (   hi.teamofplayer(t.winnerplayer()) != hi.team()
      ||  oppositeTeamCanWinTrick( t, hi ) )
    value -= 2;

  if (   hi.teamofplayer(t.winnerplayer()) == hi.team()
      &&  !oppositeTeamCanWinTrick( t, hi ) )
  {
    value += 1;

  }

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     != Announcement::noannouncement  )
  {
    value -= 3;
  }

  if (!hi.value(Aiconfig::Type::aggressive))
  {
    value -= 1;
    if( hi.team() == Team::contra)
      value -= 1;
  }




#ifdef ANNOUNCE_DEBUG
  if( !g.isvirtual() )
    cout << "no90: " << hi.player().name() << "\t" << value << " + "
      << 2 * (int)hi.value(Aiconfig::Type::announcelimitdec ) << "\t"
      << (int)hi.value(Aiconfig::Type::announcelimit) << " + "
      << 2 * (int)hi.value(Aiconfig::Type::announceconfig) << std::endl;
  cout << "\t" << own_p << "\t" << 1.3 * opp_p << std::endl;
#endif
  return (   (    value + 2 * hi.value(Aiconfig::Type::announcelimitdec )
              >     hi.value(Aiconfig::Type::announcelimit)
              + 2*hi.value(Aiconfig::Type::announceconfig)
             )
          && (   (hi.game().played_points() <= 60)
              || (own_p > 1.3 * opp_p) )  // previous: 2, 1.5
          && (opp_p < 70)
          && (   opp_p > 0
              || g.announcements().last_chance_to_announce(Announcement::no90,
                                                           hi.player())
             )
         );
} // bool Heuristics::say_no90( HeuristicInterface hi )


/** checks if No60 should be said
 **
 ** @param    hi        heuristic interface
 **
 ** @return   true if hi should say No60
 **/
bool
Heuristics::say_no60( HeuristicInterface const& hi)
{
  // @heuristic::name   ?  say no 60  ?
  // @heuristic::idea decision to make the annoncement no 60 depending on Heuristics::CalcHandValue
  if (max_points_for_opposite_team(hi) < 60)
    return true;

  int value = 0;
  Game const& g = hi.game();

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     > Announcement::no120 )
    return false;



  value = CalcHandValue( hi )
    - max( 0, static_cast<int>(g.tricks().remaining_no()) - 5)
    * static_cast<int>(hi.value(Aiconfig::Type::announcelimitdec));

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     != Announcement::noannouncement  )
    value -= 2;

  int own_p = calcPointsOfOwnTeam(hi);
  int opp_p = calcPointsOfOppositeTeam(hi);


  if( own_p > 180 )
    return true;

  if (   (hi.game().type() == GameType::poverty
          && hi.game().players().poverty_partner() == hi.player()) )
  {
    value -= 7;

  }

  if (    is_solo(hi.game().type()) )
  {
    value -= 6;
  }

  value -= g.tricks().remaining_no() / 4;


  opp_p+= 3 * g.tricks().remaining_no(); //previous: 5, 4
  opp_p-=4; // reference 233652
  own_p+= 6 * g.tricks().remaining_no(); // previous: 5

  own_p += 3 * hi.hand().count_trumps() * hi.hand().count_swines();
  own_p += 3 * hi.hand().count_trumps() * hi.hand().count_hyperswines();

  Trick const t = (game_status < GameStatus::game_play
                   ? Trick(g.startplayer())
                   : g.tricks().current());

  if (   hi.teamofplayer(t.winnerplayer()) != hi.team()
      ||  oppositeTeamCanWinTrick( t, hi ) )
    value -= 2;

  if (   hi.teamofplayer(t.winnerplayer()) == hi.team()
      &&  !oppositeTeamCanWinTrick( t, hi ) )
    value += 1;

  if (   (hi.game().type() == GameType::poverty)
      && (hi.player() == hi.game().players().poverty_partner()) )
  {
    value -= 4 - hi.hand().count_dolle() - hi.hand().count_club_queens();
  }
  if ( hi.game().rule()(Rule::Type::dollen) )
  {
    value -= 5 - 2*hi.hand().count_dolle();
    if ( hi.cards_information().played(Card::dolle)+hi.hand().count_dolle() < 2 )
      value -=2;
  }
#ifdef ANNOUNCE_DEBUG
  if( !g.isvirtual() )
    cout << "no60: " << hi.player().name() << "\t" << value << " + "
      << 3 * (int)hi.value(Aiconfig::Type::announcelimitdec ) << "\t"
      << (int)hi.value(Aiconfig::Type::announcelimit) << " + "
      << 3 * (int)hi.value(Aiconfig::Type::announceconfig) << std::endl;
  cout << own_p << "\t" << opp_p << std::endl;
#endif
  return (   (  value + 3 * hi.value(Aiconfig::Type::announcelimitdec)
              >   hi.value(Aiconfig::Type::announcelimit)
              + 3 * hi.value(Aiconfig::Type::announceconfig)
             )
          && own_p > 1.7 * opp_p  // previous: 3, 2 (1.7 reference:233652)
          && opp_p < 60
          && (   opp_p > 0
              || g.announcements().last_chance_to_announce(Announcement::no60,
                                                           hi.player())
             )
         );
} // bool Heuristics::say_no60( HeuristicInterface hi)


/** checks if No30 should be said
 **
 ** @param    hi        heuristic interface
 **
 ** @return   true if hi should say No30
 **/
bool
Heuristics::say_no30( HeuristicInterface const& hi)
{
  // @heuristic::name   ?  say no 30  ?
  // @heuristic::idea decision to make the annoncement no 30 depending on Heuristics::CalcHandValue
  if (max_points_for_opposite_team(hi) < 30)
    return true;

  int value = 0;
  Game const& g = hi.game();

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     > Announcement::no120 )
    return false;
  if( g.announcements().last( hi.team() ) != Announcement::no60 ) // do not skip an announcement
    return false;


  value = CalcHandValue( hi )
    - max( 0, static_cast<int>(g.tricks().remaining_no()) - 3)
    * static_cast<int>(hi.value(Aiconfig::Type::announcelimitdec));

  if( g.announcements().last( opposite( hi.team() ) ).announcement
     != Announcement::noannouncement  )
    value -= 2;

  int own_p = calcPointsOfOwnTeam(hi);
  int opp_p = calcPointsOfOppositeTeam(hi);

  if( own_p > 210 )
    return true;


  opp_p+= 4 * g.tricks().remaining_no();
  own_p+= 5 * g.tricks().remaining_no();

  value -=   hi.cards_information().remaining_others(Card::dolle);

#ifdef ANNOUNCE_DEBUG
  if( !g.isvirtual() )
    cout << "no30: " << hi.player().name() << "\t" << value << " + "
      << 6 * (int)hi.value(Aiconfig::Type::announcelimitdec ) << "\t"
      << (int)hi.value(Aiconfig::Type::announcelimit) << " + "
      << (int)hi.value(Aiconfig::Type::announceconfig) << std::endl;
#endif
  return (  (   value + 6 * hi.value(Aiconfig::Type::announcelimitdec)
             >   hi.value(Aiconfig::Type::announcelimit)
             + hi.value(Aiconfig::Type::announceconfig)
            )
          && own_p > 7 * opp_p
          && opp_p < 30
          && (    opp_p > 0
              || g.announcements().last_chance_to_announce(Announcement::no30,
                                                           hi.player())
             )
         );
} // bool Heuristics::say_no30( HeuristicInterface hi)


/** test whether a 'no 0' could be announced
 ** checks, that the other team does not have a trick
 ** and that the hand value is big enoug.
 **
 ** @param    hi     heuristic interface
 **
 ** @return   whether to announce 'no 0'
 **/
bool
Heuristics::say_no0( HeuristicInterface const& hi)
{
  // @heuristic::name   ?  say no 0  ?
  // @heuristic::idea decision to make the annoncement no 0 depending on Heuristics::CalcHandValue
  Game const& game = hi.game();

  if( game.announcements().last( opposite( hi.team() ) ).announcement
     != Announcement::noannouncement )
    return false;

  if( game.announcements().last( hi.team() ) != Announcement::no30 ) // do not skip an announcement
    return false;

  // make sure, that the other team does not have a trick
  for (unsigned p = 0; p < game.playerno(); ++p)
    if (   (hi.teamofplayer(game.player(p)) != hi.team())
        && game.tricks().has_trick(game.player(p)))
      return false;

  // as long as the current trick is not closed, it has to be taken into account
  if (   game.tricks().current().isfull()
      && (hi.teamofplayer(game.tricks().current().winnerplayer())
          != hi.team()) )
    return false;

  // wait for the last possible moment
  if (!game.announcements().last_chance_to_announce(Announcement::no0,
                                                    hi.player()))
    return false;

  if( highest_trump_of_opposite_team(hi).less(hi.hand().lowest_trump())
     && hi.hand().count_trumps()
     >= hi.game().tricks().remaining_no())
    return true;

  // ToDo: check that no opposite player has the highest trump card
  // Test: force a player by 'announcelimitdec' high



  int const value = CalcHandValue( hi )
    - max( 0, static_cast<int>(game.tricks().remaining_no()) )
    * static_cast<int>(hi.value(Aiconfig::Type::announcelimitdec));

#ifdef ANNOUNCE_DEBUG
  if( !game.isvirtual() )
    cout << "no0: " << hi.player().name() << "\t" << value << " + "
      << 4 * (int)hi.value(Aiconfig::Type::announcelimitdec ) << "\t"
      << (int)hi.value(Aiconfig::Type::announcelimit) << " + "
      << 3 * (int)hi.value(Aiconfig::Type::announceconfig) << std::endl;
#endif

  return (value + 4 * static_cast<int>(hi.value(Aiconfig::Type::announcelimitdec))
          > static_cast<int>(hi.value(Aiconfig::Type::announcelimit)
                             + 3 * hi.value(Aiconfig::Type::announceconfig))
         );
} // bool Heuristics::say_no0( HeuristicInterface hi)


/** checks if an announcement should be replied with an other announcement
 **
 ** @param    hi        heuristic interface
 **
 ** @return   Announcement::Reply if wanted
 **/
Announcement
Heuristics::make_reply( HeuristicInterface const& hi)
{
  // @heuristic::name   ?  make reply  ?
  // @heuristic::idea announcement reply decision bases on Heuristics::CalcHandValue

  auto const& game = hi.game();
  auto const team = hi.team();
  auto const announcement_opp
    = game.announcements().last(opposite(team)).announcement ;
  if( announcement_opp == Announcement::noannouncement)
    return Announcement::noannouncement;

  unsigned const opp_p = calcPointsOfOppositeTeam(hi);
  unsigned const own_p = calcPointsOfOwnTeam(hi);
  // The opposite team has not fulfilled its announcement -> make replay
  if (announcement_opp == Announcement::no0) {
    return Heuristics::make_reply_no0(hi);
  }

  if (own_p >= point_limit_for_opposite_team(announcement_opp)) {
    return Announcement::reply;
  }
  // Still 30 points to make for the announcement -> no reply
  if (opp_p + 30 >= needed_points(announcement_opp))
    return Announcement::noannouncement;
  // no 0 -> no reply
  if (announcement_opp == Announcement::no0)
    return Announcement::noannouncement;

  if (hi.hand().count_trumps() - 1
      <= (game.tricks().remaining_no() + 1) / 2 )
    return Announcement::noannouncement;

  if (announcement_opp == Announcement::no120) {
    // Sonderfall: Bei Gegenansage wird mindestens eine Karte größer Kreuz Dame benötigt
    if (   !hi.game().is_solo()
        && !hi.hand().higher_card_exists(Card::club_queen)) {
      return Announcement::noannouncement;
    }
  }


  {
    int value = 0;

    value = CalcHandValue( hi );


    if (   (game.type() == GameType::marriage)
        && (game.marriage().selector() == MarriageSelector::team_set)
        && ( team != Team::re ) )
    {
      value -=5; // previous 2
    }


    if (   game.is_solo()
        && hi.is_soloplayer())
      value -= 3;

    switch (announcement_opp) {
    case Announcement::no90:
      value += 4 * hi.value(Aiconfig::Type::announceconfigreply);
      break;
    case Announcement::no60:
      value += 3 * hi.value(Aiconfig::Type::announceconfigreply);
      break;
    case Announcement::no30:
      value += 2 * hi.value(Aiconfig::Type::announceconfigreply);
      break;
    case Announcement::no0:
      value += hi.value(Aiconfig::Type::announceconfigreply);
      break;
    default:
      break;
    } // switch (announcement_opp)

    auto const& current_trick = game.tricks().current();
    if (   hi.teamofplayer(current_trick.winnerplayer()) != team
        ||  oppositeTeamCanWinTrick( current_trick, hi ) ) {
      value -= 1;
    }

    for (auto const& card : hi.hand()) {
      if(   card.isdolle()
         || card.isswine()
         || card.possible_swine()
         || card.ishyperswine()
         || card.possible_hyperswine() )
        value += hi.value(Aiconfig::Type::announceconfigreply);
    }

    if (   (game.type() == GameType::poverty)
        && (hi.player() == game.players().poverty_partner()) )
    {
      value -= 6 - hi.hand().count_dolle() - hi.hand().count_club_queens();
      if( game.announcements().last(opposite(team)).announcement
         != Announcement::noannouncement  )
      {
        value -= 5;
      }
    }

#ifdef ANNOUNCE_DEBUG
    if( !game.isvirtual() )
      cout << "reply: " << hi.player().name() << "\t" << value << "\t"
        << hi.value(Aiconfig::Type::announcelimitreply) << std::endl;
#endif
    if( value >  static_cast<int>(hi.value(Aiconfig::Type::announcelimitreply)) )
      return Announcement::reply;
  }

  return Announcement::noannouncement;
} // Announcement Heuristics::make_reply( HeuristicInterface hi)

/** @return   whether a reply to no 0 should be made
 **/
Announcement
Heuristics::make_reply_no0(HeuristicInterface const& hi)
{
  auto const& game = hi.game();
  auto const team = hi.team();
  if (game.tricks().has_trick(team))
    return Announcement::reply;

  auto const& hand = hi.hand();
  auto const& cards_information = hi.cards_information();
  auto const& highest_card = hand.highest_trump();
  auto const highest_opposite_card = cards_information.highest_remaining_trump_of_others();
  if (!highest_card)
    return Announcement::noannouncement;
  if (!highest_opposite_card)
    return Announcement::reply;
  if (   highest_opposite_card.less(highest_card)
      && !highest_card.less(highest_opposite_card))
    return Announcement::reply;

  return Announcement::noannouncement;
}

/** @return   the maximal points the opposite team can make
 **/
unsigned
Heuristics::max_points_for_opposite_team(HeuristicInterface const& hi)
{
  auto const& game = hi.game();

  if (game.is_solo()
      && game.players().is_soloplayer(hi.player())) {
    return max_points_for_opposite_team_soloplayer(hi);
  }


  if (is_real(hi.team())) {
    return 240 - game.points_of_team(hi.team());
  } else {
    return 240;
  }
}

/** @return   the maximal points the opposite team can make
 **/
unsigned
Heuristics::max_points_for_opposite_team_soloplayer(HeuristicInterface const& hi)
{
  auto const& game = hi.game();

  // Stiche zählen, die nicht an den Solospieler gehen
  unsigned res = game.points_of_team(Team::contra);
  auto const& player = hi.player();
  auto const& cards_information = hi.cards_information();
  auto const& trick = game.tricks().current();

  auto hand = hi.hand();
  if (trick.has_played(player)) {
    hand.add(trick.card_of_player(player));
  }

  { // prüfen, ob alle Trümpfe gezogen werden können
    auto const card = cards_information.highest_remaining_trump_of_others();
    if (card) {
      auto const n = hand.count_ge(card);
      if (n < cards_information.remaining_trumps_others()) {
        // Es können nicht alle Trümpfe gezogen werden
        return 240;
      }
    }
  }

  // Für jede Farbe zählen, wie viele Stiche sicher sind. Der Rest gibt 3 * 11 + Punkte der Karte für den Gegner
  for (auto const color : game.rule().card_colors()) {
    auto const card = cards_information.highest_remaining_card_of_others(color);
    if (!card)
      continue;
    auto const cards = hand.cards(color);
    auto const n = cards.count_ge(card);
    if (n == cards.size())
      continue;
    if (n >= cards_information.remaining_others(color))
      continue;
    for (auto const c : cards.cards_not_ge(card)) {
      res += c.points() + 3 * 11;
    }
  }

  return res;
}
