/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "heuristics.h"

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

/** check if own team can still win this trick.
 **
 ** @param    trick   the current trick
 ** @param    hi      heuristic interface
 **
 ** @return   returns true if the own team can win the trick
 **/
bool
Heuristics::ownTeamCanWinTrick( Trick const &t, HeuristicInterface const& hi )
{
  if (t.winnerplayer().team() == hi.team())
    return true;

  // take a look if all players coming in this trick of own team can still win the trick
  for( unsigned i = t.actcardno(); i < hi.game().playerno(); i++ ) {
    if (   !hi.same_player(t.player_of_card(i))
        && ( maybe_to_team(hi.teamofplayer(t.player_of_card( i ))) == hi.team() ) )
    {
      if (hi.handofplayer(t.player_of_card( i )
                         ).validcards(t).higher_card_exists( t.winnercard() ))
        return true;
    } // if( maybe_to_team(hi.teamofplayer(t.player_of_card(i))) != hi.team() )
  } // for (i)

  return false;
} // bool Heuristics::ownTeamCanWinTrick ( Trick t, HeuristicInterface hi )

/** check if opposite team can still win this trick.
 **
 ** @param    trick   the current trick
 ** @param    hi      heuristic interface
 **
 ** @return   returns true if the opposite team can win the trick
 **
 ** @todo      use highest_opposite_card_behind for oppositeTeamCanWinTrick
 **/
bool
Heuristics::oppositeTeamCanWinTrick( Trick const &t, HeuristicInterface const& hi )
{
  if (!hi.guessed_same_team(t.winnerplayer()))
    return true;

  // take a look if all players coming in this trick of other team can still win the trick
  for( unsigned i = t.actcardno(); i < hi.game().playerno(); i++ )
    if (   !hi.guessed_same_team(t.player_of_card(i))
        && hi.handofplayer(t.player_of_card(i)
                          ).higher_card_exists( t.winnercard() ))
      return true;

  return false;
} // bool Heuristics::oppositeTeamCanWinTrick( Trick t, HeuristicInterface hi )

/** check if the card will win the trick
 **
 ** @param    trick   the current trick
 ** @param    hi      heuristic interface
 ** @param    card    the card to check
 **
 ** @return   returns true if the card will win the trick
 **/
bool
Heuristics::card_will_win_trick(Trick const& trick,
                                HeuristicInterface const& hi,
                                HandCard const& card)
{
  return (   trick.winnercard().less(card)
          && !card.less(highest_card_behind_of_opposite_team(trick, hi)));
} // bool Heuristics::card_will_win_trick(Trick t, HeuristicInterface hi)

/** @return   number of no trump cards per color
 **/
unsigned
Heuristics::number_of_no_trumps_per_color()
{
  auto const& game = ::party->game();
  auto const c = (   (game.cards().trumpcolor() == Card::spade)
                  ? Card::club
                  : Card::spade);

  return 2 * count_if(game.rule().card_values(),
                      [&game, c](auto v) { return Card(c, v).istrump(game); });
} // unsigned Heuristics::number_of_no_trumps_per_color()

/** finds lowest card with value cv
 **
 ** @param    h       the hand to search for a card
 ** @param    cv      the value to search for
 ** @param    hi      heuristic interface
 **
 ** @return   lowest card with the value cv (or 'Card()' if no card was found)
 **/
Card
Heuristics::lowest_card( Hand const& h, Card::Value const cv, HeuristicInterface const& hi )
{
  Card c;

  unsigned i;

  for( i = 0; i < h.cardsnumber(); i++ )
  {
    // find any card
    if( h.card(i).value() == cv )
    {
      c = h.card(i);
      break;
    } // if ( h.card(i).value() == cv )
  } // for ( i < h.cardsnumber() )

  if (c.is_empty())
    return c;

  for( i = 0; i < h.cardsnumber(); i++ )
  {
    // find lowest card
    if(   h.card(i).value() == cv
       && h.card(i).less(c) )
      c = h.card(i);
  } // for ( i < h.cardsnumber() )

  return c;
} // Card Heuristics::lowest_card( Hand h, Card::Value cv, HeuristicInterface hi )


/** finds highest card with value cv
 **
 ** @param    h       the hand to search for a card
 ** @param    cv      the value to search for
 ** @param    hi      heuristic interface
 **
 ** @return   highest card with the value cv (or 'Card()' if no card was found)
 **/
Card
Heuristics::highest_card( Hand const& h, Card::Value const cv, HeuristicInterface const& hi )
{
  Card c;

  unsigned i;

  for( i = 0; i < h.cardsnumber(); i++ )
  {
    // find any card
    if( h.card(i).value() == cv )
    {
      c = h.card(i);
      break;
    } // if ( h.card(i).value() == cv )
  } // for ( i < h.cardsnumber() )

  if (c.is_empty())
    return Card::empty;

  for( i = 0; i < h.cardsnumber(); i++ )
  {
    // find highest card
    if(   h.card(i).value() == cv
       && !h.card(i).less(c) )
      c=h.card(i);
  } // for ( i < h.cardsnumber() )

  return c;
} // Card Heuristics::highest_card( Hand h, Card::Value cv, HeuristicInterface hi )


/** finds lowest jack
 **
 ** @param    h       the hand to search for a card
 ** @param    hi      heuristic interface
 **
 ** @return   lowest jack on hand (or 'Card()' if no card was found)
 **/
Card
Heuristics::lowest_jack( Hand const& h, HeuristicInterface const& hi )
{
  return lowest_card( h, Card::jack, hi );
} // Card Heuristics::lowest_jack ( Hand h, HeuristicInterface hi )


/** finds highest queen
 **
 ** @param    h       the hand to search for a card
 ** @param    hi      heuristic interface
 **
 ** @return   highest queen on hand (or 'Card()' if no card was found)
 **/
Card
Heuristics::highest_queen( Hand const& h, HeuristicInterface const& hi )
{
  return highest_card( h, Card::queen, hi );
} // Card Heuristics::highest_queen( Hand h, HeuristicInterface hi )


/** finds lowest queen
 **
 ** @param    h       the hand to search for a card
 ** @param    hi      heuristic interface
 **
 ** @return   lowest queen on hand (or 'Card()' if no card was found)
 **/
Card
Heuristics::lowest_queen( Hand const& h, HeuristicInterface const& hi )
{
  return lowest_card( h, Card::queen, hi );
} // Card Heuristics::lowest_queen ( Hand h, HeuristicInterface hi )

/** checks if all remaining player in the trick are of my team
 **
 ** @param    trick   trick to check
 ** @param    hi      Heuristic interface
 **
 ** @return   returns true if all remaining players are in my team
 **/
bool
Heuristics::checkAllMyTeam( Trick const& t, HeuristicInterface const& hi )
{
  // take a look if all players coming in this trick are of mine team
  for( unsigned i = t.actcardno() + 1; i < hi.game().playerno(); i++ )
    if (!hi.guessed_same_team(t.player_of_card(i)))
      return false;

  return true;
} // bool Heuristics::checkAllMyTeam( Trick t, HeuristicInterface hi )

/** checks if all remaining player in the trick are of the opposite team
 **
 ** @param    trick   trick to check
 ** @param    hi      Heuristic interface
 **
 ** @return   returns true if all remaining players are in the opposite team
 **/
bool
Heuristics::checkAllOppositeTeam(Trick const& t, HeuristicInterface const& hi)
{
  // take a look if all players coming in this trick are of the opposite team
  for (unsigned i = t.actcardno() + 1; i < hi.game().playerno(); ++i)
    if (!hi.guessed_opposite_team(t.player_of_card(i)))
      return false;

  return true;
} // bool Heuristics::checkAllOppositeTeam(Trick t, HeuristicInterface hi)


/** -> result
 **
 ** @param    trick   trick to check
 ** @param    hi      heuristic interface
 **
 ** @return   whether the opposite team has a trump behind
 **
 ** @todo      check for 'must have' of the color of the trick
 **/
bool
Heuristics::opposite_trump_behind(Trick const& trick, HeuristicInterface const& hi)
{
  for (unsigned i = trick.actcardno() + 1; i < hi.game().playerno(); ++i) {
    if (hi.team() == hi.teamofplayer(trick.player_of_card(i)))
      continue;

    // check for 'must have' of the color of the trick
    if (hi.cards_information().of_player(trick.player_of_card(i)).can_have(Card::trump))
      return true;
  } // for (i < hi.game().playerno())

  return false;
} // bool Heuristics::opposite_trump_behind(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick   trick to check
 ** @param    hi      heuristic interface
 **
 ** @return   the highest card of the playes of the opposite team behind
 **
 ** @todo      check for 'must have' of the color of the trick
 **/
HandCard const&
Heuristics::highest_card_behind_of_opposite_team(Trick const& trick,
                                                 HeuristicInterface const& hi)
{
  HandCard const* highest_card = &HandCard::empty;

  // take a look if all players coming in this trick are of mine team
  for (unsigned i = trick.actcardno() + 1; i < hi.game().playerno(); ++i) {
    if (hi.game().marriage().is_undetermined()) {
      if (::is_selector((  trick.isempty()
                         ? Card::trump
                         : trick.startcard().tcolor()),
                        hi.game().marriage().selector())) {
        if (   (hi.team() != Team::re)
            && (hi.game().players().soloplayer() == trick.player_of_card(i)) ) {
          // in a selector trick, the re player is seen as of the own team
          continue;
        }
      }
    } else { // if (normal game)
      if (hi.team() == hi.teamofplayer(trick.player_of_card(i)))
        continue;
    }

    if (   (hi.game().type() == GameType::poverty)
        && (trick.player_of_card(i) == hi.game().players().soloplayer()) ) {
      if (   !trick.isempty()
          && !trick.startcard().istrump()) {
        HandCard const& c
          = hi.handofplayer(trick.player_of_card(i)).highest_card(trick.startcard().tcolor());
        if (highest_card->less(c))
          highest_card = &c;
        continue;
      }
    }
    // check for 'must have' of the color of the trick
    HandCard const& c = hi.handofplayer(trick.player_of_card(i)).highest_card();
    if (highest_card->less(c))
      highest_card = &c;
  } // for (i < hi.game().playerno())

  return *highest_card;
} // HandCard const& Heuristics::highest_card_behind_of_opposite_team(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick   trick to check
 ** @param    hi      heuristic interface
 **
 ** @return   the cards of the playes of the opposite team behind
 **
 ** @todo      check for 'must have' of the color of the trick
 **/
HandCards
Heuristics::valid_cards_behind_of_opposite_team(Trick const& trick,
                                                HeuristicInterface const& hi)
{
  HandCards cards;

  // take a look if all players coming in this trick are of mine team
  for (unsigned i = trick.actcardno() + 1; i < hi.game().playerno(); ++i) {
    Player const& p = trick.player_of_card(i);
    if (hi.game().marriage().is_undetermined()) {
      if (::is_selector((  trick.isempty()
                         ? Card::trump
                         : trick.startcard().tcolor()),
                        hi.game().marriage().selector())) {
        if (   (hi.team() != Team::re)
            && (hi.game().players().soloplayer() == p) ) {
          // in a selector trick, the re player is seen as of the own team
          continue;
        }
      }
    } else { // if (normal game)
      if (hi.team() == hi.teamofplayer(p))
        continue;
    }

    cards.add(hi.handofplayer(p).validcards(trick));
    // special case: second color run
    if (   !trick.isempty()
        && !trick.startcard().istrump()
        && (hi.cards_information().of_player(p).must_have(trick.startcard().tcolor()) == 0)
        && (hi.color_runs(trick.startcard().tcolor()) > 0))
      cards.add(hi.handofplayer(p).trumps());
  } // for (i < hi.game().playerno())

  return cards;
} // HandCards Heuristics::valid_cards_behind_of_opposite_team(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick   trick to check
 ** @param    hi      heuristic interface
 **
 ** @return   the highest card of the playes of the same team behind
 **
 ** @todo      check for 'must have' of the color of the trick
 **/
HandCard const&
Heuristics::highest_card_behind_of_same_team(Trick const& trick,
                                             HeuristicInterface const& hi)
{
  HandCard const* highest_card = &HandCard::empty;

  // take a look if all players coming in this trick are of mine team
  for (unsigned i = trick.actcardno() + 1; i < hi.game().playerno(); ++i) {

    if (hi.game().marriage().is_undetermined()) {
      if (::is_selector((  trick.isempty()
                         ? Card::trump
                         : trick.startcard().tcolor()),
                        hi.game().marriage().selector())) {
        if (   (hi.team() != Team::re)
            && (hi.game().players().soloplayer() != trick.player_of_card(i)) ) {
          // in a selector trick, the contra players are seen as of the opposite team
          continue;
        }
      }
    } else { // if (normal game)
      if (hi.team() != hi.teamofplayer(trick.player_of_card(i)))
        continue;
    }

    if (   (hi.game().type() == GameType::poverty)
        && hi.game().players().is_soloplayer(trick.player_of_card(i) )) {
      if (   !trick.isempty()
          && !trick.startcard().istrump()) {
        HandCard const& c
          = hi.handofplayer(trick.player_of_card(i)).highest_card(trick.startcard().tcolor());
        if (highest_card->less(c))
          highest_card = &c;
        continue;
      }
    }
    // check for 'must have' of the color of the trick
    HandCard const& c = hi.handofplayer(trick.player_of_card(i)).highest_card();
    if (highest_card->less(c))
      highest_card = &c;
  } // for (i < hi.game().playerno())

  return *highest_card;
} // HandCard const& Heuristics::highest_card_behind_of_same_team(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    hi      heuristic interface
 **
 ** @return   the lowest trump the player has which gets over the of the opposite team, or the highest trump; empty, if the trick can not be jabbed
 **/
HandCard
Heuristics::lowest_highest_trump(Trick const& trick, HeuristicInterface const& hi)
{
  HandCard lh_trump = hi.hand().highest_trump();
  if (   (lh_trump == Card::empty)
      || !trick.winnercard().less(lh_trump))
    return HandCard::empty;

  return Heuristics::optimized_card_to_play(lh_trump, trick, hi);
} // Card Heuristics::lowest_highest_trump(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    hi      heuristic interface
 **
 ** @return   the highest trump of the playes of the opposite team
 **/
HandCard const&
Heuristics::highest_trump_of_opposite_team(HeuristicInterface const& hi)
{
  auto const* highest_trump = &HandCard::empty;

  for (unsigned p = 0; p < hi.game().playerno(); ++p) {
    if (hi.team() == hi.teamofplayer(p))
      continue;

    auto const& c = hi.handofplayer(hi.game().player(p)).highest_trump();
    if (highest_trump->less(c))
      highest_trump = &c;
  } // for (p < hi.game().playerno())

  return *highest_trump;
} // HandCard const& Heuristics::highest_trump_of_opposite_team(HeuristicInterface hi)

/** -> result
 **
 ** @param    hi      heuristic interface
 **
 ** @return   the number of tricks the team will get
 **/
unsigned
Heuristics::secure_tricks_of_team(HeuristicInterface const& hi)
{
  HandCard const& highest_trump = highest_trump_of_opposite_team(hi);
  if (highest_trump.is_empty()) {
    // the opposite team has no more trump
    for (unsigned p = 0; p < hi.game().playerno(); ++p)
      if (   (hi.teamofplayer(p) == hi.team())
          && (   hi.handofplayer(p).count_trumps()
              == hi.handofplayer(p).cardsnumber()))
        return hi.game().tricks().remaining_no();
  } // if (!highest_trump)

  unsigned result = 0;
  for (unsigned p = 0; p < hi.game().playerno(); ++p)
    if (hi.teamofplayer(p) == hi.team())
      result = max(result,
                   hi.cards_information().of_player(p).must_have_cards().higher_cards_no(highest_trump));

  return result;
} // unsigned Heuristics::secure_tricks_of_team(HeuristicInterface const& hi)

/** finds lowest trump card on Hand h
 **
 ** @param    trick     trick to check
 ** @param    h         hand to choose the card from
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::lowest_trump_card(Trick const& t, Hand const& h)
{
  Card c;
  HandCards const ha = h.validcards( t );
  unsigned i;

  // find any card that's allowed for this trick
  for( i = 0; i < ha.cardsnumber(); i++ ) {
    if( ha.card( i ).istrump() )
    {
      c = ha.card( i );
      break;
    }
  } // for ( i < ha.cardsnumber() )

  if (c.is_empty())
    return Card::empty;

  // find a better card
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    if(   ha.card(i).less( c )
       && ha.card(i).istrump()
      )
      c = ha.card( i );
  } // for ( i < ha.cardsnumber() )

  return c;
} // Card Heuristics::lowest_trump_card( Trick t, Hand h )


/** finds highest trump card on Hand h
 **
 ** @param    trick     trick to check
 ** @param    h         hand to choose the card from
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::highest_trump_card( Trick const& t, Hand h )
{
  Card c;
  HandCards const ha = h.validcards( t );
  unsigned i;

  // find any card that's allowed for this trick
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    if( ha.card( i ).istrump() )
    {
      c = ha.card( i );
      break;
    }
  } // for ( i < ha.cardsnumber() )

  if (c.is_empty())
    return Card::empty;

  // find a better card
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    if(   c.less( ha.card(i) )
       && ha.card(i).istrump()
      )
      c = ha.card( i );
  } // for ( i < ha.cardsnumber() )

  return c;
} // Card Heuristics::highest_trump_card( Trick t, Hand h )


/** finds lowest trump card on Hand h, but gives only fuchs or ten back
 ** if there are better cards to keep
 **
 ** @param    trick     trick to check
 ** @param    h         hand to choose the card from
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::lowest_best_trump_card( Trick const& t, Hand const& h,
                                   HeuristicInterface const& hi )
{
  // @heuristic::name   ?  lowest best trump  ?
  // @heuristic::idea   find smallest best trump card (avoiding for example a fox in most situations)
  // @heuristic::action  play the smallest trump, but no fox/ten

  HandCard c( h );
  HandCards const ha = hi.hand().validcards( t );

  // find any card that's allowed for this trick
  for(unsigned i = 0; i < ha.cardsnumber(); i++ ) {
    if ( ha.card( i ).istrump() ) {
      c = ha.card( i );
      break;
    }
  } // for( i < ha.cardsnumber() )

  if (c.is_empty())
    return Card::empty;

  // find a better card
  for(unsigned i = 0; i < ha.cardsnumber(); i++ ) {
    if(   ha.card( i ).istrump()
       && (   (c.value() >= Card::ten)
           || ha.card( i ).less( c ) )
       && !ha.card( i ).isfox()
       && (   ha.card(i).value() < Card::ten
           || c.isfox()
          )
      ) {
      c = ha.card( i );
    }
  } // for( i = 0; i < ha.cardsnumber(); i++ )

  if( !c.less( hi.trump_card_limit() ) ) {
    // maybe a diamond ten isn't all that bad
    for(unsigned i = 0; i < ha.cardsnumber(); i++ ) {
      if(   ha.card( i ).less( c )
         && ha.card( i ).istrump()
         && !ha.card( i ).isfox()
        ) {
        c = ha.card( i );
      }
    }
  } // if( !c.less( hi.trump_card_limit() ) )

  if( !c.less( hi.trump_card_limit() ) ) {
    // maybe a diamond Ace isn't all that bad
    for(unsigned i = 0; i < ha.cardsnumber(); i++ ) {
      if(   ha.card( i ).less( c )
         && ha.card(i).istrump()
        ) {
        c = ha.card( i );
      }
    }
  } // if( !c.less( hi.trump_card_limit() ) )

  return c;
} // Card Heuristics::lowest_best_trump_card( Trick t, Hand h, HeuristicInterface hi )


/** best pfund on hand: a trump Ass, color ten or Ace, trump ten, a king,
 ** a notrump queen or jack, nine
 **
 ** @param    trick     trick to check
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::choose_pfund_card(Trick const& trick, HeuristicInterface const& hi)
{
  // @heuristic::name   ?  choose pfund card  ?
  // @heuristic::idea   find a pfund card
  // @heuristic::action   play a pfund

  Trick t = trick;

  DEBUG_ASSERTION( !t.isempty(),
                  "Heuristics::choose_pfund_card():\n"
                  "  Called with empty trick");


  bool allmyteam = (checkAllMyTeam( t, hi ) && ownTeamCanWinTrick(t, hi));
  allmyteam = (   allmyteam
               || (   hi.game().type() == GameType::poverty
                   && hi.is_soloplayer()
                   && trick.cardno_of_player(hi.game().players().poverty_partner()) == 3 ) );
  // marriage determination for the bride
  if (   (hi.game().type() == GameType::marriage)
      && is_selector(t.startcard().tcolor(),
                     hi.game().marriage().selector())
      && hi.is_soloplayer() )
    allmyteam = true;

  if (   hi.guessed_same_team(t.winnerplayer())
      && !oppositeTeamCanWinTrick(t, hi))
    allmyteam = true;

  // first color run
  if (   !t.startcard().istrump()
      && (hi.color_runs(t.startcard().color()) == 0)
      && (   (t.winnercard().value() == Card::ace)
          || t.winnercard().istrump()))
    allmyteam |= hi.guessed_same_team(t.winnerplayer());

  Card c;

  HandCards const ha = hi.hand().validcards( t );

  // first color trick in a solo, soloplayer wins
  bool solo_check =
    (   is_solo( hi.game().type() )
     && hi.color_runs(t.startcard().color()) == 0
     && !t.startcard().istrump()
     && t.winnerplayer() == hi.game().players().soloplayer()
     && t.winnercard().less( hi.trump_card_limit() )
     && !t.islastcard() ) ;

  if (   !hi.game().is_solo()
      && !(   hi.same_team(t.winnerplayer())
           && t.islastcard()
           && t.cardno_of_player( t.winnerplayer() ) == 0
           && !t.winnercard().istrump() ) //only use trump ace when partner is not winning already in first position
      && !(   hi.game().rule()(Rule::Type::genscher)
           && (hi.game().type() == GameType::normal)
           && !hi.cards_information().played(Card(Card::diamond,
                                                  Card::king))
           && (ha.count_trumps() >= 6) // *Value*
          )
     ) {
    // then find first trump ace
    for( unsigned i = 0; i < ha.cardsnumber(); i++ ) {
      if(   ha.card(i).value() == Card::ace
         && ha.card(i).istrump()
         && (   t.islastcard()
             || allmyteam
             || ha.count_trumps() < 3
             || (   hi.color_runs(t.startcard().color()) == 0
                 && !t.startcard().istrump()
                )
             || (   ! hi.cards_information().higher_card_exists( t.winnercard())
                 //!t.winnercard().less( Card::club_queen )
                 && hi.same_team(t.winnerplayer())
                )
             || (   hi.game().marriage().is_undetermined()
                 && hi.same_player(hi.game().players().soloplayer())
                 && hi.game().marriage().selector() == MarriageSelector::first_foreign
                )
            )
         && !ha.card(i).possible_swine()
        )
      {
        t.add(ha.card( i ));
        if (   (hi.teamofplayer( t.winnerplayer() )== hi.team())
            || (   (   hi.value(Aiconfig::Type::trusting)
                    || (hi.team_information().certainty() >= 30) ) // **Value**
                && hi.guessed_same_team(t.winnerplayer()) )
            || (   allmyteam
                && !hi.trump_card_limit().less(t.winnercard()) ) ) {
          return ha.card(i);
        }
      }
    } // for (i < ha.cardsnumber)
  } // if partner is start player and with none trump card.

  { // choose a single ten of a color
    Card best_ten;
    for (auto
         color = hi.game().rule().card_colors().begin();
         color != hi.game().rule().card_colors().end();
         ++color) {
      Card const local_ten(*color, Card::ten);
      if (   !hi.hand().contains(local_ten)
          || local_ten.istrump(hi.game())
          || !t.isvalid(local_ten, hi.hand()))
        continue;

      if (ha.count(local_ten.color()) == 1) {
        if (   best_ten.is_empty()
            || (   hi.cards_information().remaining_others(*color)
                && (hi.cards_information().remaining_others(*color)
                    < hi.cards_information().remaining_others(best_ten.color())) )
           ) {
          best_ten = local_ten;
        }
      } // if (single ten)
    } // for (color \in card_colors)
    if(!best_ten.is_empty()) {
      return best_ten;
    }
  } // choose a single ten of a color

  if (   (   hi.teamofplayer(t.winnerplayer()) == hi.team()
          || solo_check
          || allmyteam )
      // do not pfund an ace in a meatless solo, if the soloplayer is behind
      && !(   (hi.game().type() == GameType::solo_meatless)
           && (   !t.has_played(hi.game().players().soloplayer())
               || (t.cardno_of_player(hi.game().players().soloplayer())
                   < t.winnercardno()) ) ) )
  {
    { // choose a ten of color if the corresponding ace is on the hand
      Card best_ten;
      for (auto
           color = hi.game().rule().card_colors().begin();
           color != hi.game().rule().card_colors().end();
           ++color) {
        Card const local_ten( *color, Card::ten );
        if (   !hi.hand().contains(local_ten)
            || local_ten.istrump( hi.game() )
            || !t.isvalid(local_ten, hi.hand())
            || !hi.hand().contains(Card(*color, Card::ace))
            || hi.cards_information().remaining_others(Card(*color, Card::ace))
            || (   hi.hand().contains(Card(*color, Card::ace))
                && !hi.cards_information().remaining_others(Card(*color, Card::ace)))
           )
          continue;
        if (   best_ten.is_empty()
            || (   hi.cards_information().remaining_others(best_ten.color())
                && !(   (hi.hand().count(*color) < hi.hand().count(best_ten.color()))
                     || (hi.cards_information().remaining_others(*color) == 0))
               )
           ) {
          best_ten = local_ten;
        }
      } // for (color \in card_colors)
      if(!best_ten.is_empty()) {
        if (hi.cards_information().remaining_others(best_ten))
          return best_ten;
        else
          return Card(best_ten.color(), Card::ace);
      }
    } // choose a ten of color if the corresponding ace is on the hand
    { // choose an ace of a color 
      Card best_ace;
      // choose an ace of a color 
      // * if the ace is single
      for (auto
           color = hi.game().rule().card_colors().begin();
           color != hi.game().rule().card_colors().end();
           ++color) {
        Card const local_ace( *color, Card::ace );
        if (   !hi.hand().contains(local_ace)
            || local_ace.istrump( hi.game() )
            || !t.isvalid(local_ace, hi.hand()))
          continue;

        // if the ace is single, take it
        if (hi.hand().count(*color) == 1) {
          if(   best_ace.is_empty()
             || (hi.cards_information().remaining_others(*color)
                 && (hi.cards_information().remaining_others(*color)
                     < hi.cards_information().remaining_others(best_ace.color())) )
            ) {
            best_ace = local_ace;
          }
        }
      } // for (color)
      if(!best_ace.is_empty()) {
        return best_ace;
      }

      // choose an ace of a color 
      // * if the player has two aces
      for (auto
           color = hi.game().rule().card_colors().begin();
           color != hi.game().rule().card_colors().end();
           ++color) {
        Card const local_ace( *color, Card::ace );
        if (   !hi.hand().contains(local_ace)
            || local_ace.istrump( hi.game() )
            || !t.isvalid(local_ace, hi.hand()))
          continue;

        // if the player has two aces, take it
        if (   hi.hand().count( local_ace )
            == hi.game().rule()(Rule::Type::number_of_same_cards)) {
          if(   best_ace.is_empty()
             || (hi.cards_information().remaining_others(*color)
                 < hi.cards_information().remaining_others(best_ace.color()))
            ) {
            best_ace = local_ace;
          }
        }
      } // for (color)
      if(!best_ace.is_empty()) {
        return best_ace;
      }

      // choose an ace of a color 
      // * if there are not enough remaining color cards and the player has not the ten or no other player has a ten
      for (auto
           color = hi.game().rule().card_colors().begin();
           color != hi.game().rule().card_colors().end();
           ++color) {
        Card const local_ace( *color, Card::ace );
        if (   !hi.hand().contains(local_ace)
            || local_ace.istrump( hi.game() )
            || !t.isvalid(local_ace, hi.hand())
            ||   (   !Card(*color, Card::ten).istrump(hi.game())
                  && hi.cards_information().remaining_others(*color, Card::ten))
           )
          continue;

        if (   (   (hi.game().type() == GameType::poverty)
                && (hi.is_soloplayer() ) )
            || (   (hi.hand().count(local_ace) == 1)
                && (hi.cards_information().remaining_others(*color)
                    < hi.game().playerno() - 1
                    + ( (   (t.startcard().tcolor() == *color)
                         && !t.winnercard().istrump())
                       ? (t.remainingcardno() - 1)
                       : 0))
                && (hi.game().type() != GameType::solo_meatless)
               )
           ) {
          if (best_ace.is_empty()) {
            best_ace = local_ace;
          } else if (hi.cards_information().remaining_others(*color)
                     > hi.cards_information().remaining_others(best_ace.color())) {
            best_ace = local_ace;
          }
        }
      } // for (color)
      if(!best_ace.is_empty()) {
        return best_ace;
      }
    } // choose an ace of a color 

    { // choose a ten of color
      Card best_ten;
      for (auto
           color = hi.game().rule().card_colors().begin();
           color != hi.game().rule().card_colors().end();
           ++color) {
        Card const local_ten( *color, Card::ten );
        if (   !hi.hand().contains(local_ten)
            || local_ten.istrump( hi.game() )
            || !t.isvalid(local_ten, hi.hand()))
          continue;
        if (best_ten.is_empty())
          best_ten = local_ten;
        else if (hi.cards_information().remaining_others(*color) == 0) {
          if (hi.cards_information().remaining_others(best_ten.color()) == 0) {
            if (hi.hand().count(*color) < hi.hand().count(best_ten.color()))
              best_ten = local_ten;
          }
        } else if (hi.cards_information().remaining_others(best_ten.color()) == 0)
          ;
        else if (hi.hand().count(*color) < hi.hand().count(best_ten.color()))
          best_ten = local_ten;
      } // for (color \in card_colors)

      if(!best_ten.is_empty()) {
        return best_ten;
      }
    } // choose a ten of color

    { // choose a color ace from the longest color
      Card best_ace;
      // do not pfund an ace in a meatless solo, if the soloplayer is behind
      if (!(   (hi.game().type() == GameType::solo_meatless)
            && (   !t.has_played(hi.game().players().soloplayer())
                || (t.cardno_of_player(hi.game().players().soloplayer())
                    < t.winnercardno()) ) ) ) {
        for( auto
            color = hi.game().rule().card_colors().begin();
            color != hi.game().rule().card_colors().end();
            ++color ) {
          Card const local_ace( *color, Card::ace );
          if (   !hi.hand().contains(local_ace)
              || local_ace.istrump( hi.game() )
              || !t.isvalid(local_ace, hi.hand()))
            continue;
          if (  (hi.hand().count(local_ace) == 1)
              && (  (   (   (hi.cards_information().remaining_others(*color)
                             < hi.game().playerno() - 1
                             + ( (   (t.startcard().tcolor() == *color)
                                  && !t.winnercard().istrump())
                                ? (t.remainingcardno() - 1)
                                : 0))
                         || !(   hi.hand().contains(*color, Card::ten)
                              && !HandCard(hi.hand(), *color, Card::ten).is_special())
                        )
                     && (hi.game().type() != GameType::solo_meatless))
                  || (hi.hand().count(*color) == 1)
                  || (trick.points() + 11 >= 40)
                 )
              || !hi.hand().contains(*color, Card::ten)
             )
          {
            if(   best_ace.is_empty()
               || (hi.cards_information().remaining_others(*color)
                   > hi.cards_information().remaining_others(best_ace.color()))
              ) {
              best_ace = local_ace;
            }
          }
        } // for (color \in card_colors)

        if(!best_ace.is_empty()) {
          return best_ace;
        }
      } // if (no meatless solo with solo player behind)
    } // choose a color ace
  } // if (trick goes to own team)

  bool const to_fat_solo
    = (   is_solo( hi.game().type() )
       && (static_cast<int>(t.points()) > hi.value( Aiconfig::Type::limitqueen ))
       && (t.player_of_card( 3 ) == hi.game().players().soloplayer()));
  // or an ace of color
  for( unsigned i = 0; i < ha.cardsnumber(); i++ ) {
    int cardsInGame
      = hi.cards_information().remaining_others(ha.card(i).color());

    if(   (ha.card(i) == Card(t.startcard().color(), Card::ace))
       && (hi.color_runs(ha.card(i).color()) == 0)
       && (hi.hand().count(t.startcard().color()) >= 3)
       && !ha.card(i).istrump()
       /// @todo   see 'choose pfund', check that enough cards still remain
       && (   cardsInGame >= 2
           || ( hi.game().type() == GameType::poverty
               && hi.is_soloplayer() )
          )
       && (   hi.teamofplayer(t.winnerplayer()) == hi.team()
           || solo_check
           || allmyteam
          )
       && !(hi.game().type() == GameType::solo_meatless)
      ) {
      return ha.card(i);
    }

    if(   ha.card(i).value()==Card::ace
       && (  hi.color_runs(ha.card(i).color()) > 0
           || (   t.winnercard().istrump()
               && allmyteam )
          )
       && cardsInGame >= 2
       && !ha.card(i).istrump()
       && t.isvalid( ha.card(i),hi.hand() )
       && !to_fat_solo
       && (   (   hi.game().type()!=GameType::solo_jack // with not much trumps
               && hi.game().type()!=GameType::solo_queen // it is good to
               && hi.game().type()!=GameType::solo_king  // keep a second ace
              )
           || hi.color_runs(t.startcard().color()) > 0
          )
      )
    {
      // prefer a ten to an ace
      if (   hi.hand().contains(ha.card(i).color(), Card::ten)
          && !HandCard(hi.hand(), ha.card(i).color(), Card::ten).istrump()
          && hi.cards_information().remaining_others(Card(ha.card(i).color(), Card::ten)))
        return Card(ha.card(i).color(), Card::ten);
      return ha.card(i);
    }
  } // for ( i < ha.cardsnumber() )

  // or an ace of trump
  for( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(   ha.card(i).value() == Card::ace
       && ha.card(i).istrump()
       && !ha.card(i).possible_swine()
      )
    {
      return ha.card(i);
    }

  // or a ten of trump
  for( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(   ha.card(i).value() == Card::ten
       && ha.card(i).istrump()
       && !ha.card(i).isdolle()
      )
    {
      return ha.card(i);
    }


  // or a King of color
  for ( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(   ha.card(i).value() == Card::king
       && !ha.card(i).istrump()
      )
    {
      return ha.card(i);
    }


  // or a king of trump
  for( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(   (ha.card(i).value() == Card::king)
       && ha.card(i).istrump()
       && !ha.card(i).possible_hyperswine()
       && !is_picture_solo(hi.game().type())
      )
    {
      return ha.card(i);
    }


  // or Queen if no trump
  for ( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(    ha.card(i).value()==Card::queen
       && !ha.card(i).istrump()
       && hi.game().type() != GameType::solo_queen
       && hi.game().type() != GameType::solo_king_queen
       && hi.game().type() != GameType::solo_queen_jack
       && hi.game().type() != GameType::solo_koehler
      )
    {
      return ha.card(i);
    }


  // or Jack if no trump
  for( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(   ha.card(i).value() == Card::jack
       && !ha.card(i).istrump()
       && hi.game().type() != GameType::solo_jack
       && hi.game().type() != GameType::solo_queen_jack
       && hi.game().type() != GameType::solo_king_jack
       && hi.game().type() != GameType::solo_koehler
      )
    {
      return ha.card(i);
    }


  // or a nine of color
  for( unsigned i = 0; i < ha.cardsnumber(); i++ )
    if(   ha.card(i).value() == Card::nine
       && !ha.card(i).istrump()
      )
    {
      return ha.card(i);
    }


  { // or a small jack of trump
    Card best_jack;
    for ( unsigned i = 0; i < ha.cardsnumber(); i++ ) {
      if(   ha.card(i).value() == Card::jack
         && (ha.card(i).color() <= Card::heart)
         && ha.card(i).istrump()
        ) {
        if (best_jack.is_empty()
            || ha.card(i).less(best_jack))
          best_jack = ha.card(i);
      }
    } // for (i)
    if (!best_jack.is_empty())
      return best_jack;
  } // or a small jack of trump

  return c;
} // Card Heuristics::choose_pfund_card( Trick trick, HeuristicInterface hi )


/** checks if a solo player sits behind the player
 **
 ** @param    t    trick
 ** @param    hi   heuristics interface
 **
 ** @return   whether the solo player sits behind
 **/
bool
Heuristics::soloPlayerBehind( Trick const& t, HeuristicInterface const& hi )
{
  return (   hi.game().is_solo()
          && !t.has_played(hi.game().players().soloplayer()));
} // bool Heuristics::soloPlayerBehind( Trick const& t, HeuristicInterface const& hi )

/** -> result
 **
 ** @param    trick   current trick
 ** @param    hand    hand of the player
 **
 ** @return   lowest card to play in the trick,
 **            Card::empty, if none is found
 **/
Card
Heuristics::lowest_card(Trick const& trick,
                        Hand const& hand)
{
  if (trick.isstartcard())
    return Heuristics::lowest_free_card(hand);

  if (!hand.contains(trick.startcard().tcolor()))
    return Heuristics::lowest_free_card(hand);

  if (trick.startcard().istrump())
    return Heuristics::lowest_trump_card(hand);
  else
    return Heuristics::lowest_color_card(trick.startcard().tcolor(), hand);
} // static Card Heuristics::lowest_card(Trick trick, Hand hand)


/** -> result
 ** decision order:
 ** 1) normal trump with less than 10 points
 ** 2) normal trump
 ** 3) special trump
 **
 ** @param    hand    hand of the player
 **
 ** @return   lowest trump to play
 **/
Card
Heuristics::lowest_trump_card(Hand const& hand)
{
  DEBUG_ASSERTION(hand.contains(Card::trump),
                  "Heuristics::lowest_trump_card(hand)\n"
                  "  hand does not contain any trump");

  HandCard const& lowest_trump = hand.lowest_trump();
  Game const& game = hand.game();

  if (   !game.cards().is_special(lowest_trump)
      && (lowest_trump.value() < Card::ten) )
    return lowest_trump;

  DEBUG_ASSERTION(!is_picture_solo(game.type()),
                  "Heuristics::lowest_trump_card(hand)\n"
                  "  picture solo: " << game.type());


  vector<Card> const& trumps = game.cards().trumps();
  // search lowest trump with less then 10 points
  for (auto const trump : trumps)
    if (hand.contains(trump))
      if (   !game.cards().is_special(HandCard(hand, trump))
          && (trump.value() < Card::ten))
        return trump;

  // search lowest trump but no special card
  for (auto const trump : trumps)
    if (hand.contains(trump))
      if (!game.cards().is_special(HandCard(hand, trump)))
        return trump;
  // search lowest trump
  Card c;
  for (auto const trump : trumps)
    if (hand.contains(trump)
        && (trump.less(HandCard(hand, c))))
      c = trump;

  return c;
} // static Card Heuristics::lowest_trump_card(Hand hand)


/** -> result
 **
 ** @param    tcolor   tcolor to serve
 ** @param    hand     hand of the player
 **
 ** @return   lowest card to play the tcolor
 **/
Card
Heuristics::lowest_color_card(Card::TColor const& tcolor,
                              Hand const& hand)
{
  DEBUG_ASSERTION((tcolor != Card::trump),
                  "Heuristics::lowest_color_card(tcolor, hand)\n"
                  "  tcolor ist trump");
  DEBUG_ASSERTION(hand.contains(tcolor),
                  "Heuristics::lowest_color_card(tcolor, hand)\n"
                  "  hand does not contain the color " << tcolor);

  return hand.lowest_card(tcolor);
} // static Card Heuristics::lowest_color_card(Card::TColor tcolor, Hand hand)

/** finds lowest color card on Hand h, which is the best in a meatless solo
 **
 ** @param    trick     trick to check
 ** @param    h         hand to choose the card from
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::lowest_color_card_meatless(Trick const& t, Hand const& h)
{
  Card c;
  HandCards const ha = h.validcards( t );
  unsigned i;

  // find any card that's allowed for this trick
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    if( !ha.card( i ).istrump() )
    {
      c = ha.card( i );
      break;
    }
  } // for ( i < ha.cardsnumber() )

  if (c.is_empty())
    return Card::empty;

  // find a better card
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    if(   !ha.card( i ).istrump()
       && c.value() <= ha.card(i).value()
       && (   ha.count( ha.card(i).color() ) >  ha.count( c.color() )
           || (   ha.count( ha.card(i).color(), Card::ten ) == 0
               && ha.count( ha.card(i).color(), Card::ace ) == 0
              )
          )
      )
      c = ha.card(i);
  } // for ( i < ha.cardsnumber() )

  return c;
} // Card Heuristics::lowest_color_card_meatless( Trick t, Hand h )


/** -> result
 **
 ** @param    hand     hand of the player
 **
 ** @return   lowest card to play (free color selection),
 **            Card::empty, if none is found
 **
 ** @todo      replace with Heuristics::choose_best_prefiltered_card()
 **/
Card
Heuristics::lowest_free_card(Hand const& hand)
{
  if (hand.count_trumps() == hand.cardsnumber())
    return Heuristics::lowest_trump_card(hand);

  // search the lowest value
  Card::Value lowest_value = Card::ace;
  for (auto const c : hand.cards())
    if (c.value() < lowest_value)
      lowest_value = c.value();

  // the card colors with the lowest value and not trump
  vector<Card::Color> card_colors;
  for (auto
       c = hand.game().rule().card_colors().begin();
       c != hand.game().rule().card_colors().end();
       ++c)
    if (!Card(*c, lowest_value).istrump(hand.game()))
      card_colors.push_back(*c);

  if (card_colors.empty())
    return Card::empty;

  if (lowest_value < Card::ace) {
    // search single card with lowest value
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c) {
      if (   (hand.count(*c) == 1)
          && (hand.contains(Card(*c, lowest_value)))) {
        return Card(*c, lowest_value);
      }
    }
  } // if (lowest_value < Card::ace)

  if (lowest_value < Card::ten) {
    // search lowest card of the smallest color without ace and ten
    Card::Color best_color = Card::nocardcolor;
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c)
      if (   (hand.count(Card(*c, Card::ace)) == 0)
          && (hand.count(Card(*c, Card::ten)) == 0))
        if (   (hand.count(*c) > 0)
            && (   (best_color == Card::nocardcolor)
                || (hand.count(*c) < hand.count(best_color)) ) )
          best_color = *c;
    if (best_color != Card::nocardcolor) {
      return hand.lowest_card(best_color);
    }
  } // if (lowest_value < Card::ten)

  if (lowest_value == Card::ace) {
    // search double ace
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c)
      if (hand.count(*c) == 2) {
        return Card(*c, lowest_value);
      }
  } // if (lowest_value == Card::ace)

  if (lowest_value == Card::ten) {
    // search only tens
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c)
      if (   hand.contains(Card(*c, lowest_value))
          && (hand.count(*c)
              == hand.count(Card(*c, lowest_value))) ) {
        return Card(*c, lowest_value);
      }

    // search biggest color
    Card::Color best_color = Card::nocardcolor;
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c) {
      if (hand.contains(Card(*c, lowest_value)))
        if (   (best_color == Card::nocardcolor)
            || (hand.count(*c) > hand.count(best_color)))
          best_color = *c;
    } // for (c \in card_colors())
    if (best_color != Card::nocardcolor) {
      return hand.lowest_card(best_color);
    }
  } // if (lowest_value == Card::ten)

  { // search lowest card of the biggest color
    Card::Color best_color = Card::nocardcolor;
    unsigned best_cards_no = 2;
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c)
      if (hand.count(*c) > best_cards_no) {
        best_color = *c;
        best_cards_no = hand.count(*c);
      }
    if (best_color != Card::nocardcolor) {
      return hand.lowest_card(best_color);
    }
  } // search lowest card of the biggest color

  { // search lowest card
    Card::Color best_color = Card::nocardcolor;
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c)
      if (   hand.contains(Card(*c, lowest_value))
          && (   (best_color == Card::nocardcolor)
              || (hand.count(*c) > hand.count(best_color))))
        best_color = *c;
    if (best_color != Card::nocardcolor) {
      return hand.lowest_card(best_color);
    }
  } // search lowest card

  return Card::empty;
} // Card Heuristics::lowest_free_card(Hand hand)

/** -> result
 **
 ** @param    hi       heuristic interface
 ** @param    hand     hand of the player
 ** @param    cards    cards to choose from
 **
 ** @return   lowest card to play from 'cards'
 **/
Card
Heuristics::choose_best_prefiltered_card(Trick const& trick,
                                         HeuristicInterface const& hi,
                                         Hand const& hand,
                                         HandCards const& cards)
{
  if (cards.cardsnumber() == 0)
    return Card::empty;
  if (cards.cardsnumber() == 1)
    return cards[0];

  if (cards.count_trumps() < cards.cardsnumber())
    return Heuristics::choose_best_prefiltered_color_card(trick, hi,
                                                          hand, cards);
  else
    return Heuristics::choose_best_prefiltered_trump_card(trick, hi,
                                                          hand, cards);
} // Card Heuristics::choose_best_prefiltered_card(Trick trick, HeuristicsInterface hi, Hand hand, HandCards cards)

/** -> result
 **
 ** @param    hi       heuristic interface
 ** @param    hand     hand of the player
 ** @param    cards    cards to choose from
 **
 ** @return   lowest card to play from 'cards',
 **            Card::empty, if none is found
 **/
Card
Heuristics::choose_best_prefiltered_color_card(Trick const& trick,
                                               HeuristicInterface const& hi,
                                               Hand const& hand,
                                               HandCards const& cards)
{
  DEBUG_ASSERTION(cards.count_trumps() < cards.cardsnumber(),
                  "Heuristics::choose_best_prefilterd_color_card(hand, cards)\n"
                  "  'cards' only contains trumps");
  if (cards.cardsnumber() == 0)
    return Card::empty;
  if (cards.cardsnumber() == 1)
    return cards[0];

  Game const& game = hand.game();
  Rule const& rule = game.rule();

  // whether the trick goes to the own team
  bool const own_team = (   !trick.isempty() 
                         && (trick.winnerteam() == hi.team()));

  if (own_team) {
    // search a single ace
    for (auto const c : rule.card_colors()) {
      if (   !Card(c, Card::ace).istrump(game)
          && (hand.count(c) == 1)
          && (cards.contains(c, Card::ace))) {
        // ToDo: consider two colors with the same number of cards
        return Card(c, Card::ace);
      }
    } // for (c \in rule.card_colors())
    // search single ten
    for (auto const c : rule.card_colors()) {
      if (   !Card(c, Card::ten).istrump(game)
          && (hand.count(c) == 1)
          && (cards.contains(c, Card::ten))) {
        // ToDo: consider two colors with the same number of cards
        return Card(c, Card::ten);
      }
    } // for (c \in rule.card_colors())
  } // if (own team)

  // search the lowest / highest value
  Card::Value best_value = (own_team ? Card::nine : Card::ace);
  for (auto const card : cards) {
    if (   !card.istrump()
        && (own_team
            ? (card.value() > best_value)
            : (card.value() < best_value) )
       ) {
      best_value = card.value();
    }
  } // for (c)

  // the card colors with the best value and not trump
  vector<Card::Color> card_colors;
  for (auto const c : rule.card_colors()) {
    if (   !Card(c, best_value).istrump(hand.game())
        && cards.contains(c, best_value) ) {
      card_colors.push_back(c);
    }
  }

  if (card_colors.empty())
    return Card::empty;


  if (best_value < Card::ace) {
    // search single card with best value
    for (auto const c : card_colors) {
      if (   (hand.count(c) == 1)
          && (cards.contains(c, best_value))) {
        // ToDo: consider two colors with the same number of cards
        return Card(c, best_value);
      }
    } // for (c : card_colors)
  } // if (best_value < Card::ace)

  if (best_value < Card::ten) {
    // search best card of the smallest color without ace and ten
    Card::Color best_color = Card::nocardcolor;
    for (auto const c : card_colors) {
      if (   cards.contains(c)
          && (hand.count(Card(c, Card::ace)) == 0)
          && (hand.count(Card(c, Card::ten)) == 0))
        if (   (best_color == Card::nocardcolor)
            || (cards.count(c) < hand.count(best_color)))
          // ToDo: improve selection -- card value comparison, compare remaining cards of the color, compare jabbing, ...
          best_color = c;
    } // for (c : card_colors)
    if (best_color != Card::nocardcolor) {
      return (own_team
              ? cards.highest_card(best_color)
              : cards.lowest_card(best_color));
    }
  } // if (best_value < Card::ten)

  if (best_value == Card::ace) {
    // search double ace
    for (auto const c : card_colors) {
      if (   cards.contains(c, best_value)
          && (hand.count(c, best_value) == 2) ) {
        return Card(c, best_value);
      }
    } // for (c : card_colors)
  } // if (best_value == Card::ace)

  if (best_value == Card::ten) {
    // search only tens
    for (auto const c : card_colors) {
      if (   cards.contains(c, best_value)
          && (hand.count(c)
              == hand.count(c, best_value)) ) {
        return Card(c, best_value);
      }
    }
  } // if (best_value == Card::ten)

  { // search best card of the biggest color with min 3 cards
    Card::Color best_color = Card::nocardcolor;
    unsigned best_cards_no = 2;
    for (auto const c : card_colors) {
      if (   (hand.count(c) > best_cards_no)
          && (cards.contains(c)) ) {
        best_color = c;
        best_cards_no = hand.count(c);
      }
    } // for (c : card_colors)
    if (   (best_color != Card::nocardcolor)
        && cards.contains(hand.lowest_card(best_color)) ) {
      // ToDo: take smallest card from cards
      return (own_team
              ? cards.highest_card(best_color)
              : cards.lowest_card(best_color));
    }
  } // search best card of the biggest color

  { // search best card of the largest color
    Card::Color best_color = Card::nocardcolor;
    for (vector<Card::Color>::const_iterator
         c = card_colors.begin();
         c != card_colors.end();
         ++c)
      if (   cards.contains(Card(*c, best_value))
          && (   (best_color == Card::nocardcolor)
              || (hand.count(*c) > hand.count(best_color))))
        best_color = *c;
    if (best_color != Card::nocardcolor) {
      return (own_team
              ? cards.highest_card(best_color)
              : cards.lowest_card(best_color));
    }
  } // search best card

  return Card::empty;
} // static Card Heuristics::choose_best_prefiltered_color_card(Trick trick, HeuristicsInterface hi, Hand hand, HandCards cards)


/** -> result
 **
 ** @param    hi       heuristic interface
 ** @param    hand     hand of the player
 ** @param    cards    cards to choose from
 **
 ** @return   lowest card to play from 'cards'
 **/
Card
Heuristics::choose_best_prefiltered_trump_card(Trick const& trick,
                                               HeuristicInterface const& hi,
                                               Hand const& hand,
                                               HandCards const& cards)
{
  DEBUG_ASSERTION(cards.count_trumps() == cards.cardsnumber(),
                  "Heuristics::choose_best_prefilterd_trump_card(hand, cards)\n"
                  "  'cards' does contain color cards");
  if (cards.cardsnumber() == 0)
    return Card::empty;
  if (cards.cardsnumber() == 1)
    return cards[0];

  // whether the trick goes to the own team
  bool const own_team = (trick.winnerteam() == hi.team());

  Game const& game = hand.game();
  if (own_team
      && (game.cards().trumpcolor() != Card::nocardcolor)) {
    if (cards.contains(Card(game.cards().trumpcolor(), Card::ace)))
      return Card(game.cards().trumpcolor(), Card::ace);
    if (cards.contains(Card(game.cards().trumpcolor(), Card::ten)))
      return Card(game.cards().trumpcolor(), Card::ten);
    if (cards.contains(Card(game.cards().trumpcolor(), Card::king)))
      return Card(game.cards().trumpcolor(), Card::king);
  } // if (own_team) {

  HandCard card(hand);

  // find any card that's allowed for this trick
  for(unsigned i = 0; i < cards.cardsnumber(); i++ ) {
    if ( cards.card( i ).istrump() ) {
      card = cards.card( i );
      break;
    }
  } // for( i < ha.cardsnumber() )

  if( card.is_empty())
    return Card::empty;

  // find a better card
  for(unsigned i = 0; i < cards.cardsnumber(); i++ ) {
    if(   cards.card( i ).less(card)
       && cards.card( i ).istrump()
       && !cards.card( i ).isfox()
       && (   cards.card(i).value() != Card::ten
           || card.isfox()
          )
      )
      card = cards.card( i );
  } // for( i = 0; i < cards.cardsnumber(); i++ )

  if( !card.less( hi.trump_card_limit() ) ) {
    // maybe a diamond ten isn't all that bad
    for(unsigned i = 0; i < cards.cardsnumber(); i++ ) {
      if(   cards.card( i ).less( card )
         && cards.card( i ).istrump()
         && !cards.card( i ).isfox()
        )
        card = cards.card( i );
    }
  } // if( !card.less( cards.trump_card_limit() ) )

  if( !card.less( hi.trump_card_limit() ) ) {
    // maybe a diamond Ace isn't all that bad
    for(unsigned i = 0; i < cards.cardsnumber(); i++ ) {
      if(   cards.card( i ).less( card )
         && cards.card(i).istrump()
        )
        card = cards.card( i );
    }
  } // if( !c.less( hi.trump_card_limit() ) )

  return card;
} // Card Heuristics::choose_best_prefiltered_trump_card(Trick trick, HeuristicsInterface hi, Hand hand, HandCards cards)

/** calculates all points of own team in Game g
 **
 ** @param    hi        heuristic interface
 ** @param    with_current_trick   whether to take the current trick into account
 **
 ** @return   points of own team in the game (minimum estimator)
 **/
unsigned
Heuristics::calcPointsOfOwnTeam( HeuristicInterface const& hi,
                                bool const with_current_trick)
{
  if (!::in_running_game())
    return 0;

  Game const& g = hi.game();
  unsigned points = 0;
  unsigned players = 0;

  for (unsigned p = 0; p < g.playerno(); ++p) {
    if (hi.guessed_same_team(p)) {
      points += g.points_of_player(p);
      players += 1;
    }
  } // for (p < g.playerno())

  if (   (players == 1)
      && !hi.game().is_solo()) {
    // add the smallest points of the unknown players
    unsigned min_points = UINT_MAX;
    for (unsigned p = 0; p < g.playerno(); ++p) {
      if (   !hi.guessed_same_team(p)
          && !hi.guessed_opposite_team(p)
          && (g.points_of_player(p) <= min_points))
        min_points = g.points_of_player(p);
    } // for (p < g.playerno())
    if (min_points < UINT_MAX)
      points += min_points;
  } // if (players == 1)

  Trick const& tr = g.tricks().current();
  if (tr.isfull()) {
    if (hi.guessed_same_team(tr.winnerplayer()))
      points += tr.points();
  } else if (with_current_trick) {
    if (   checkAllMyTeam(tr, hi)
        && (   hi.guessed_same_team(tr.winnerplayer())
            || (   !tr.has_played(g.player(hi.player().no()))
                && tr.winnercard().less(hi.hand().validcards(tr).highest_card()))
           )
       )
      points += tr.points() + hi.hand().validcards(tr).lowest_value();
  }

  return points;
} // unsigned Heuristics::calcPointsOfOwnTeam( HeuristicInterface hi, bool with_current_trick = true )


/** calculates all points of opposite team in Game g
 **
 ** @param    hi        heuristic interface
 ** @param    g         current game
 ** @param    with_current_trick   whether to take the current trick into account
 **
 ** @return   points of other team in game (maximum estimator)
 **/
unsigned
Heuristics::calcPointsOfOppositeTeam(HeuristicInterface const& hi,
                                     bool const with_current_trick)
{
  if (!::in_running_game())
    return 0;

  Game const& g = hi.game();
  Trick const& tr = g.tricks().current();
  // calculation
  // * without current trick:
  //   played points
  //   - points got by own team (without current trick)
  //   - points of the current trick
  // * with current trick:
  //   played points
  //   - points got by own team (with current trick)
  //   + highest value of the player (worst case), if he has not played in the current trick
  if (with_current_trick)
    return (g.played_points() // all points in the game
            - calcPointsOfOwnTeam(hi, with_current_trick) // points potentially in opposite game
            + (!tr.has_played(hi.player()) // if player has not played, add card.
               ? hi.hand().validcards(tr).highest_value() : 0) // worst case: most point in opposite team
           );
  else
    return (g.played_points() // all points in the game
            - calcPointsOfOwnTeam(hi, with_current_trick) // points potentially in opposite game
            - tr.points() // without current trick remove points of this trick => still remaining
           );

#ifdef OUTDATED // minimum estimator
  // 0.7.10
  unsigned points = 0;
  unsigned players = 0;

  for (unsigned p = 0; p < g.playerno(); ++p) {
    if (hi.guessed_opposite_team(p)) {
      points += g.points_of_player(p);
      players += 1;
    }
  } // for (p < g.playerno())

  if (   (players == 1)
      && !hi.game().is_solo()) {
    // add the smallest points of the unknown players
    unsigned min_points = UINT_MAX;
    for (unsigned p = 0; p < g.playerno(); ++p) {
      if (   !hi.guessed_opposite_team(p)
          && !hi.guessed_same_team(p)
          && (g.points_of_player(p) <= min_points))
        min_points = g.points_of_player(p);
    } // for (p < g.playerno())
    if (min_points < UINT_MAX)
      points += min_points;
  } // if (players == 1)

  Trick const& tr = g.tricks().current();
  if (tr.isfull()) {
    if (hi.guessed_opposite_team(tr.winnerplayer()))
      points += tr.points();
  } else if (with_current_trick) {
    if (   checkAllOppositeTeam(tr, hi)
        || hi.guessed_opposite_team(tr.winnerplayer()))
      points += tr.points() + hi.hand().validcards(tr).lowest_value();
  }
  return points;
#endif
} // unsigned Heuristics::calcPointsOfOppositeTeam( HeuristicInterface hi, bool with_current_trick = true)

/** -> result
 **
 ** @param    old_card        best card till now
 ** @param    test_card        card to test
 ** @param    hi                hi (with the configuration)
 **
 ** @return   whether the card 'test_card' is to preferred to 'old_card'
 **                that is has more points
 **/
bool
Heuristics::better_points_optimize(HandCard const& old_card,
                                   HandCard const& test_card,
                                   HeuristicInterface const& hi )
{
  return (  (  static_cast<int>(hi.game().tricks().current_no())
             >= hi.value(Aiconfig::Type::first_trick_for_trump_points_optimization)
            )
          &&   ( old_card.value() < test_card.value() )
          || ( test_card.less( old_card ) && old_card.value() == test_card.value() )
         );
} // bool Heuristics::better_points_optimize(HandCard old_card, HandCard test_card, Ai hi)


/** -> result
 **
 ** @param    card               card to optimize
 ** @param    trick              current trick
 ** @param    hi                 heuristics interface
 ** @param    only_same_points   whether only cards with the same points should be checked (default: false)
 **
 ** @return   the given card or a lower one, if they are equal in the sense of jabbing
 **/
HandCard
Heuristics::optimized_card_to_play(Card const& c,
                                   Trick const& trick,
                                   HeuristicInterface const& hi,
                                   bool const only_same_points)
{
  if (c == Card::empty)
    return HandCard::empty;

  Hand const& hand = hi.hand();
  Game const& game = trick.game();
  Player const& player = hi.player();
  HandCard const card(hand, c);

  // do only optimize trump cards
  if (!card.istrump())
    return card;

  // do not optimize club queen in normal games
  if (   (game.type() == GameType::normal)
      && (card == Card::club_queen)
      && (game.teaminfo().get(player) != Team::re))
    return card;

  // optimize to a lower card, of there is no partner before the opposite team
  bool optimize_to_lower = false;
  // last or second last in trump
  if (trick.remainingcardno() <= 2)
    optimize_to_lower = true;
  // own team in the back
  if (hi.guessed_same_team(trick.lastplayer()))
    optimize_to_lower = true;
  // only opposite team behind
  if (checkAllOppositeTeam(trick, hi))
    optimize_to_lower = true;
  // I play the highest card in the game
  if (!hi.cards_information().higher_card_exists(card))
    optimize_to_lower = true;
  // in a solo: the soloplayer is following directly
  if (   (optimize_to_lower == false)
      && is_solo(game.type())
      && (hi.teamofplayer(game.players().following(trick.actplayer())) == Team::re)
     )
    optimize_to_lower = true;

  if (optimize_to_lower) {
    Card lower_card = card;
    HandCards const opposite_cards
      = valid_cards_behind_of_opposite_team(trick, hi);
    bool const winnercard = trick.winnercard().less(card);

    for (Card c = card;
         c != Card::empty;
         c = game.cards().next_lower_card(c)) {
      if (   !card.is_special()
          && hand.contains(c)
          && HandCard(hi.hand(), c).is_special())
        continue;
      if (   only_same_points
          && c.value() != lower_card.value())
        break;
      // only optimize in jack/queen/king/special cards
      if (   (c.value() != Card::jack)
          && (c.value() != Card::queen)
          && !(   (game.type() == GameType::solo_koehler)
               && (c.value() == Card::king))
          && !HandCard(hi.hand(), c).is_special())
        break;
      if (   winnercard
          && !trick.winnercard().less(c))
        break;
      if (   opposite_cards.contains(c)
          && HandCard(hi.hand(), c).less(c)) {
        break;
      }
      if (hand.contains(c)) {
        // save a charlie
        if (!(   (c == Card::charlie)
              && game.rule()(Rule::Type::extrapoint_charlie)
              && (game.type() == GameType::normal) ) )
          lower_card = c;
      }
      if (opposite_cards.contains(c)) {
        // special case: club queen, but partner unknown
        if (!(   (game.type() == GameType::normal)
              && (game.teaminfo().get(player) == Team::re)
              && (c == Card::club_queen) ) )
          break;
      }
    } // for (c)

    return HandCard(hi.hand(), lower_card);
  } // if (optimize_to_lower)

  return card;
} // Card Heuristics::optimized_card_to_play(Card card, Trick trick, HeuristicInterface hi, bool only_same_points = false);

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   the lowest card to serve the trick
 **/
Card
Heuristics::lowest_serving_card(Trick const& trick, HeuristicInterface const& hi)
{
  if (trick.isempty())
    return Card::empty;

  Card card = lowest_card(trick, hi.hand());
  if (!card.is_empty())
    return card;

  if (hi.hand().contains(trick.startcard().tcolor())) {
    return hi.hand().lowest_card(trick.startcard().tcolor());
  } else {
    for (auto
         c = hi.game().rule().card_colors().begin();
         c != hi.game().rule().card_colors().end();
         ++c) {
      Card::Color const& color = *c;
      if (hi.hand().contains(color)) {
        if (   card.is_empty()
            || (   (hi.hand().lowest_card(color).value()
                    < card.value())
                || (   (hi.hand().lowest_card(color).value()
                        == card.value())
                    && (hi.hand().count(color)
                        < hi.hand().count(card.color())) ) ) )
          card = hi.hand().lowest_card(color);
      }
    }
    if (!card.is_empty())
      return card;
  }

  card = hi.hand().lowest_trump();
  if (!card.is_empty())
    return card;

  return hi.hand().card(0);
} // Card Heuristics::lowest_serving_card(Trick trick, HeuristicInterface hi)

/** -> result
 **
 ** @param    trick     current trick
 ** @param    hi        heuristic interface
 **
 ** @return   a card to jab the trick
 **/
Card
Heuristics::jabbing_card(Trick const& trick, HeuristicInterface const& hi)
{
  if (trick.isempty())
    return Card::empty;

  {
    Card const card = best_winning_card(trick, hi);
    if (!card.is_empty())
      return card;
  }

  {
    Card const card = hi.hand().next_higher_card(trick.winnercard());
    if (   !card.is_empty()
        && trick.isvalid(card, hi.hand()))
      return card;
  }

  return Card::empty;
} // Card Heuristics::jabbing_card(Trick trick, HeuristicInterface hi)


/** returns the second highest trump (useful for color soli)
 **
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **/
Card
Heuristics::SecondBestTrump(HeuristicInterface const& hi)
{
  // @heuristic::name   ?  second best trump  ?
  // @heuristic::idea   play the seconmd highest trump (especially useful for solo player in picture soli)
  Hand h = hi.hand();

  Card c;
  Card bestcard;

  // there is a trump to draw
  if( hi.cards_information().remaining_trumps_others() == 0 )
    return c;

  // first find highest trump on hand
  unsigned i;
  unsigned best = 0;
  bestcard = h.card( best );
  for (i = 1; i < h.cardsnumber(); ++i) {
    if(   h.card(i).istrump()
       && bestcard.less( h.card(i) )
      ) {
      bestcard = h.card(i);
      best = i;
    }
  } // for ( i < h.cardsnumber() )

  // find second best card
  for (i = 0; i < h.cardsnumber(); ++i) {
    if (   h.card(i).istrump()
        && (  h.card(i).less(bestcard)
            || h.card(i) == bestcard
           )
        && c.less( h.card(i) )
        && i != best
       ) {
      c = h.card( i );
    }
  } // for ( i < h.cardsnumber() )

  DEBUG_ASSERT(c);
  return c;
} // Card Heuristics::SecondBestTrump(HeuristicInterface const& hi)

/** -> result
 **
 ** @param    trick       current trick
 ** @param    hi          heuristic interface
 ** @param    forceFehl   whether to force a fehl card
 **                        default: false
 **
 ** @return   low fehl card with many cards of the color remaining
 **/
Card
Heuristics::choose_best_fehl(Trick const& trick,
                             HeuristicInterface const& hi,
                             bool const forceFehl)
{
  // the best card
  Card best_card;
  // the best number of remaining cards
  // There should be at lest so as many cards remaining
  // as there are other players
  unsigned best_remaining = 0;
  if (!forceFehl)
    best_remaining = (hi.game().playerno() - 1);

  // all valid cards
  HandCards const cards = hi.hand().validcards(trick);

  for (auto const& card : cards) {
    auto const remaining
      = hi.cards_information().remaining_others(card.color());
    if(   !card.istrump()
       && (remaining > best_remaining)
       && (hi.hand().count(card.color()) <= 1)
       && (   card.value() == Card::king
           || card.value() == Card::nine )
      ) {
      best_card = card;
      best_remaining = remaining;
    } // if (found better card)
  } // for (card \in cards)

  return best_card;
} // Card Heuristics::choose_best_fehl( Trick trick, HeuristicInterface hi, bool forceFehl = false )
