/*********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "heuristics.h"

#include "heuristic_interface.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "../../game/game.h"
#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../card/trick.h"

#define DEBUG_HEURISTIC \
  if (   (   ::debug.filter("heuristics") \
          || ::debug.filter(name)) \
      && !hi.game().isvirtual()) \
cout << __FILE__ << '#' << __LINE__ << " (" << trick.actcardno() << ')' << hi.player().name() << ": " << name << " – "

/** finds best card on Hand of hi, which wins this trick,
 ** non trump ace or lowest trump
 ** greater than jack, only in lastcard lower trumps are allowed
 ** the card will be chosen taking the value of the trick in account
 **
 ** @param    trick     trick to check
 ** @param    v         value of the card
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::best_winning_card_single_picture(Trick const& t,
                                             Card::Value const v,
                                             HeuristicInterface const& hi)
{
  // @heuristic::name   ?  best winning card single picture  ?
  // @heuristic::idea   Only valid for soli with only one picture as trump: Tries to find the best card to win a trick depending on different values like Aiconfig::limitqueen
  // @heuristic::action  jab the trick

  if ( t.isstartcard() )
    return Card::empty;

  Card c;
  HandCards const ha = hi.hand().validcards( t );
  unsigned i;

  bool allmyteam = checkAllMyTeam( t, hi );

  // find any card that wins this trick greater or equal to lowest_trump_card_limit
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    if(   t.isvalid( ha.card(i), hi.hand() )
       && t.isjabbed(ha.card(i))
       && (   hi.lowest_trump_card_limit().less( ha.card( i ) )
           || allmyteam
          )
      )
    {
      c = ha.card(i);
      break;
    }
  } // for ( i < ha.cardsnumber() )

  // if can't win this trick and I'am not the last player: that's it
  if (c.is_empty() && !t.islastcard())
    return Card::empty;

  // or this is a color trick of a first color run with my card the first trump
  allmyteam =  (  allmyteam
                || (   t.actcardno() > 1
                    && hi.color_runs(t.startcard().color()) == 0
                    && !t.contains_trump()
                   ) );

  // find a better card
  for( i = 0; i < ha.cardsnumber(); i++ )
  {

    if(   t.isvalid( ha.card( i ), hi.hand() )
       && t.isjabbed(ha.card(i))
      )
    {
      // non trump ace is always the best choice if color is not already jabbed
      if(   ha.card(i).value() == Card::ace
         && !ha.card(i).istrump()
         && (   !hi.colorjabbed( ha.card(i).color() )
             || hi.jabbedbyownteam( ha.card(i).color() )
             || allmyteam
            )
        )
      {
        c=ha.card(i);
        break;
      }

      if ( ha.card(i).istrump() )
      {

        // find lowest trump
        if( (    hi.lowest_trump_card_limit().less( ha.card(i) )
             || t.islastcard()
             || allmyteam
            )
           && ha.card(i).less( c )
          )
        {

          c = ha.card(i);
          continue;
        }

      }// if card istrump
    } // if winneris
  } // for ( i < ha.cardsnumber() )


  // if last card or first_color_run of trick let's check for a better card
  // Ace, ten, king, nine

  if(   t.islastcard()
     || allmyteam
     ||(    !t.isstartcard()
        && !t.startcard().istrump()
        && hi.color_runs(t.startcard().color()) < 1
        && (   t.startcard().color()!= Card::heart
            || !HandCard( hi.hand(), Card::heart, Card::ten ).isdolle()
           )
       )
    )
  {

    for( i = 0; i < ha.cardsnumber(); i++ )
    {
      if(   t.isvalid( ha.card( i ), hi.hand() )
         && t.isjabbed(ha.card(i))
        )
      {

        // ace is always the best card
        if(  !ha.card(i).isswine()
           && ha.card(i).value()==Card::ace
           && (    allmyteam
               || !ha.card( i ).istrump()
               || t.islastcard()
              )
          )
        {
          c = ha.card(i);
          break;
        }

        if(   ha.card(i).value() == Card::ten
           && !ha.card(i).isdolle()
           && (     t.islastcard()
               || allmyteam
               || (  hi.color_runs( t.startcard().color() ) == 0
                   && !t.startcard().istrump()
                  )
              )
          )
        {
          c = ha.card(i);
          if ( t.islastcard()
              || allmyteam )
            break;
          continue;
        }

        if(   ha.card( i ).value() != v
           && ha.card( i ).value() < Card::ten )
        {
          c = ha.card(i);
          continue;
        }

      } // if( t.isvalid( ha.card( i ), hi.hand() ) && t.isjabbed(ha.card(i))
    } // for (i < ha.cardsnumber())
  } // if ( t.islastcard() || allmyteam || ... )

  return c;
} // Card Heuristics::best_winning_card_single_picture(Trick t, Card::Value v, HeuristicInterface hi)

/** finds best card on Hand of hi, which wins this trick,
 ** non trump ace or lowest trump
 ** greater than jack, only in lastcard lower trumps are allowed
 ** the card will be chosen taking the value of the trick in account
 **
 ** @param    trick     trick to check
 ** @param    v1        value of the first card
 ** @param    v2        value of the second card
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::best_winning_card_double_picture(Trick const& t,
                                             Card::Value const v1,
                                             Card::Value const v2,
                                             HeuristicInterface const& hi)
{
  // @heuristic::name   ?  best winning card double picture  ?
  // @heuristic::idea   Only valid for soli with only tow picutes as trump: Tries to find the best card to win a trick depending on different values like Aiconfig::limitqueen
  // @heuristic::action  jab the trick

  if ( t.isstartcard() )
    return Card::empty;

  Trick tr;
  Card c;
  HandCards const ha = hi.hand().validcards(t);
  unsigned i;

  bool allmyteam = checkAllMyTeam( t, hi );;


  // find any card that wins this trick greater or equal to lowest_trump_card_limit
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    tr = t;
    tr.add(ha.card(i));
    if(   hi.is_winnerplayer(tr)
       && t.isvalid( ha.card(i),hi.hand() )
       && (   hi.lowest_trump_card_limit().less( ha.card(i) )
           || allmyteam
          )
      )
    {
      c = ha.card(i);
      break;
    }
  } // for ( i < ha.cardsnumber() )

  // if can't win this trick and I'am not the last player: that's it
  if (c.is_empty() && !t.islastcard() )
    return Card::empty;

  // or this is a color trick of a first color run with my card the first trump
  allmyteam =    allmyteam
    ||(   t.actcardno() > 1
       && hi.color_runs(t.startcard().color()) == 0
       && !t.contains_trump()
      );

  // find a better card
  for ( i = 0; i < ha.cardsnumber(); i++ )
  {
    tr = t;
    tr.add(ha.card( i ));
    if(   hi.is_winnerplayer(tr)
       && t.isvalid( ha.card(i),hi.hand() )
      )
    {
      // non trump ace is always the best choice if color is not already jabbed
      if(   ha.card(i).value() == Card::ace
         && !ha.card(i).istrump()
         && (   !hi.colorjabbed( ha.card(i).color() )
             || hi.jabbedbyownteam(ha.card(i).color())
             || allmyteam
            )
        )
      {
        c = ha.card( i );
        break;
      }

      if( ha.card(i).istrump() )
      {

        // find lowest trump
        if( (   hi.lowest_trump_card_limit().less( ha.card(i) )
             || t.islastcard()
             || allmyteam
            )
           && ha.card(i).less(c)
          )
        {
          c = ha.card(i);
          continue;
        }

      }// if card istrump

    } // if winneris
  } // for ( i < ha.cardsnumber() )

  // if last card or first_color_run of trick let's check for a better card
  // Ace, ten, king, nine
  if(   t.islastcard()
     || allmyteam
     ||(   !t.isstartcard()
        && !t.startcard().istrump()
        && hi.color_runs(t.startcard().color())<1
        && (   t.startcard().color()!= Card::heart
            || !HandCard( hi.hand(), Card::heart, Card::ten ).isdolle()
           )
       )
    )
  {

    for( i = 0; i < ha.cardsnumber(); i++ )
    {
      tr = t;
      tr.add(ha.card(i));

      if(   hi.is_winnerplayer(tr)
         && t.isvalid(ha.card(i),hi.hand())
        )
      {
        // ace is always the best card
        if(   !ha.card(i).isswine()
           &&  ha.card(i).value()==Card::ace
           && (   allmyteam
               || !ha.card(i).istrump()
               || t.islastcard()
              )
          )
        {
          c = ha.card(i);
          break;
        }

        if(   ha.card(i).value() == Card::ten
           && !ha.card(i).isdolle()
           && (   t.islastcard()
               || allmyteam
               || (   hi.color_runs(t.startcard().color()) == 0
                   && !t.startcard().istrump()
                  )
              )
          )
        {
          c = ha.card(i);
          if ( t.islastcard()
              || allmyteam )
            break;

          continue;
        }

        if(    ha.card(i).value() != v1
           && ha.card(i).value() != v2
           && ha.card(i).value() < Card::ten )
        {
          c = ha.card(i);
          continue;
        }

      } // if (hi.is_winnerplayer(tr))
    } // for ( i < ha.cardsnumber() )
  } // if ( t.islastcard() || allmyteam || (....) )

  return c;
} // Card Heuristics::best_winning_card_double_picture( ... )


/** finds best card on Hand of hi, which wins this trick,
 ** non trump ace or lowest trump
 ** greater than jack, only in lastcard lower trumps are allowed
 ** the card will be chosen taking the value of the trick in account
 **
 ** @param    trick     trick to check
 ** @param    hi        heuristic interface
 **
 ** @return   the card to play ( or 'Card()' if no card was found )
 **/
Card
Heuristics::best_winning_card_triple_picture(Trick const& t,
                                             HeuristicInterface const& hi)
{
  // @heuristic::name   ?  best winning card triple picture  ?
  // @heuristic::idea Only valid for soli with only three pictures as trump: Tries to find the best card to win a trick depending on different values like Aiconfig::limitqueen
  // @heuristic::action  jab the trick

  if ( t.isstartcard() )
    return Card::empty;

  Trick tr;
  Card c;
  HandCards const ha = hi.hand().validcards( t );
  unsigned i;

  bool allmyteam = checkAllMyTeam( t, hi );;

  // find any card that wins this trick greater or equal to lowest_trump_card_limit
  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    tr = t;
    tr.add(ha.card( i ));

    if(   hi.is_winnerplayer(tr)
       && t.isvalid( ha.card(i), hi.hand() )
       && (  hi.lowest_trump_card_limit().less( ha.card(i) )
           || allmyteam
          )
      )
    {
      c = ha.card(i);
      break;
    }
  } // for ( i < ha.cardsnumber() )
  // if can't win this trick and I'am not the last player: that's it
  if( c.is_empty() && !t.islastcard() )
    return Card::empty;

  // or this is a color trick of a first color run with my card the first trump
  allmyteam =    allmyteam
    || (   t.actcardno() > 1
        && hi.color_runs(t.startcard().color()) == 0
        && !t.contains_trump()
       );

  // find a better card

  for( i = 0; i < ha.cardsnumber(); i++ )
  {
    tr = t;
    tr.add(ha.card(i));
    if(   hi.is_winnerplayer(tr)
       && t.isvalid( ha.card( i ), hi.hand() ) )
    {
      // non trump ace is always the best choice if color is not already jabbed
      if(   ha.card(i).value() == Card::ace
         && !ha.card(i).istrump()
         && (   !hi.colorjabbed( ha.card( i ).color() )
             || hi.jabbedbyownteam( ha.card( i ).color() )
             || allmyteam
            )
        )
      {
        c = ha.card( i );
        break;
      }

      if( ha.card(i).istrump() )
      {

        // find lowest trump
        if( (   hi.lowest_trump_card_limit().less( ha.card(i) )
             || t.islastcard()
             || allmyteam )
           && ha.card(i).less(c) )
        {

          c = ha.card( i );
          continue;
        }

      }// if card istrump

    } // if winneris
  } // for ( i < ha.cardsnumber() )



  // if last card or first_color_run of trick let's check for a better card
  // Ace, ten, king, nine
  if(   t.islastcard()
     || allmyteam
     ||(   !t.isstartcard()
        && !t.startcard().istrump()
        && hi.color_runs( t.startcard().color() ) < 1
        && (   t.startcard().color()!= Card::heart
            || !HandCard( hi.hand(), Card::heart, Card::ten ).isdolle()
           )
       )
    )
  {

    for( i = 0; i < ha.cardsnumber(); i++ )
    {
      tr = t;
      tr.add(ha.card( i ));

      if(   hi.is_winnerplayer(tr)
         && t.isvalid( ha.card( i ), hi.hand() ) )
      {

        // ace is always the best card
        if(   !ha.card(i).isswine()
           && ha.card(i).value() == Card::ace
           && (   allmyteam
               || !ha.card(i).istrump()
               || t.islastcard()
              )
          )
        {
          c = ha.card(i);
          break;
        }

        if(    ha.card(i).value() == Card::ten
           && !ha.card(i).isdolle()
           && (   t.islastcard()
               || allmyteam
               || (   hi.color_runs( t.startcard().color()) == 0
                   && !t.startcard().istrump()
                  )
              )
          )
        {
          c = ha.card( i );
          if ( t.islastcard()
              || allmyteam )
            break;
          continue;
        }

        if( ha.card(i).value() == Card::nine )
        {
          c = ha.card(i);
          continue;
        }
      } // if ( hi.is_winnerplayer(tr) && t.isvalid( ... ) )
    } // for ( i < ha.cardsnumber() )
  } // if ( t.islastcard() || allmyteam || ( ...... ) )

  return c;
} // Card Heuristics::best_winning_card_triple_picture( Trick t, HeuristicInterface hi)

/** Play shortest color in game for solo contra in hope solo player and partner
 ** can jab
 **
 ** @param    trick        current trick
 ** @param    hi        heuristic interface
 **
 ** @return   card to play, 'Card()' if the heuristic does not match
 **
 ** @todo         take card value into account
 **/
Card
Heuristics::play_color_in_solo(Trick const& trick,
                               HeuristicInterface const& hi)
{
  // @heuristic::name        play color in solo
  // @heuristic::idea        Play shortest color in game for solo contra in hope solo player and partner can jab.

  // @heuristic::condition   start card in the trick
  if (!trick.isstartcard())
    return Card::empty;

  // @heuristic::condition   contra player in a picture or color solo
  if (hi.team() != Team::contra)
    return Card::empty;

  if (!(   is_picture_solo(hi.game().type())
        || is_color_solo(hi.game().type()) )
     )
    return Card::empty;

  unsigned best_remaining = UINT_MAX;
  Card best_card;

  // @heuristic::action      play color card with the least cards remaining
  // 2do color with the minimum remaining points??? (if the re player is last player or all players behind the re players have to serve the color)
  for( auto
      color = hi.game().rule().card_colors().begin();
      color != hi.game().rule().card_colors().end();
      ++color ) {
    if(   (hi.cards_information().remaining_others(*color) < best_remaining)
       && hi.hand().contains(*color)
       && (*color != hi.game().cards().trumpcolor()) ) {
      best_remaining = hi.cards_information().remaining_others(*color);
      best_card = hi.hand().lowest_card(*color);
    }
  } // for (color \in card_colors)


  return best_card;
} // Card Heuristics::play_color_in_solo(Trick trick, HeuristicInterface hi)

