/**********************************************************************
 *
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "jab_for_doppelkopf.h"

#include "../team_information.h"
#include "../../../card/trick.h"
#include "../../../game/game.h"
#include "../../../party/rule.h"

#include "../ai.h"
#include "../heuristics.h"

/** constructor
 **/
Heuristic::JabForDoppelkopf::JabForDoppelkopf(Ai const& ai) :
  Heuristic(ai, Aiconfig::Heuristic::jab_for_doppelkopf)
{ }

/** destructor
 **/
Heuristic::JabForDoppelkopf::~JabForDoppelkopf() = default;

/** @return  card to play or empty card
 **/
Card
Heuristic::JabForDoppelkopf::get_card(Trick const& trick)
{
  this->rationale_.clear();
  if (!(   !this->condition_solo_game(trick)
        && this->condition_last_card(trick)
        && this->condition_can_jab_with_trump(trick)
        && this->condition_enough_trick_points_for_doppelkopf(trick)
        && this->condition_has_ten_points_in_trump()
       ))
    return {};

  auto const& game = trick.game();
  auto const& hand = this->hand();

  { // non special cards
    auto const try_cards = {
      Card(game.cards().trumpcolor(), Card::ace),
      Card(game.cards().trumpcolor(), Card::ten),
    };
    for (auto const card : try_cards) {
      if (!hand.contains(card))
        continue;
      if (!trick.isjabbed(card))
        continue;

      if (HandCard(hand, card).is_special())
        continue;

      this->rationale_.add(_("Heuristic::return::no special card - %s", _(card)));
      return card;
    }
  }

  { // special cards
    auto const try_cards = {
      Card(game.cards().trumpcolor(), Card::ten),
      Card(game.cards().trumpcolor(), Card::ace),
      Card::dolle
    };

    for (auto const card : try_cards) {
      if (!hand.contains(card))
        continue;
      if (!trick.isjabbed(card))
        continue;

      if (!HandCard(hand, card).is_special())
        continue;

      if (   !this->guessed_same_team(trick.winnerplayer())
          && !trick.winnercard().less(this->trump_card_limit())) {
        this->rationale_.add(_("Heuristic::return::special card needed for winning the trick - %s", _(card)));
        return card;
      }

      if (Heuristics::secure_tricks_of_team(*this->ai_)
          == game.tricks().remaining_no()) {
        this->rationale_.add(_("Heuristic::return::special card not needed for an announcement - %s", _(card)));
        return card;
      }

      auto const announcement = game.announcements().last(this->team());
      if (  (announcement == Announcement::no60)
          ? Heuristics::say_no30(*this->ai_)
          : (announcement == Announcement::no90)
          ? Heuristics::say_no60(*this->ai_)
          : (announcement == Announcement::no120)
          ? Heuristics::say_no90(*this->ai_)
          : is_reply(announcement)
          ? Heuristics::say_no90(*this->ai_)
          : (announcement == Announcement::noannouncement)
          ? Heuristics::make_announcement(*this->ai_)
          : false) {
        this->rationale_.add(_("Heuristic::return::announcement can be improved - %s", _(card)));
        return card;
      }
    } // for (card)
  }

  this->rationale_.add(_("Heuristic::reject::no card found to jab"));
  return Card::empty;
}

