/**********************************************************************
 *
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "pfund_in_first_color_run.h"

#include "../ai.h"
#include "../team_information.h"
#include "../heuristics.h"
#include "../../../game/game.h"
#include "../../../party/rule.h"


/** constructor
 **/
Heuristic::PfundInFirstColorRun::PfundInFirstColorRun(Ai const& ai) :
  Heuristic(ai, Aiconfig::Heuristic::pfund_in_first_color_run)
{ }

/** destructor
 **/
Heuristic::PfundInFirstColorRun::~PfundInFirstColorRun() = default;

/** @return  card to play or empty card
 **/
Card
Heuristic::PfundInFirstColorRun::get_card(Trick const& trick)
{
  this->rationale_.clear();
  if (!(   !this->condition_startcard(trick)
        && this->condition_first_color_run(trick))) {
    return {};
  }
  if (this->condition_winnerplayer_same_team(trick)) {
    if (!this->condition_winnercard_ace_or_trump(trick)) {
      return {};
    }
  } else {
    if (!(   !this->condition_last_card(trick)
          && !this->condition_winnercard_ace_or_trump(trick)
          && !this->condition_startcard_is(trick, Card(trick.startcard().color(), Card::ten))
          && !this->condition_opposite_team_behind(trick))) {
      return {};
    }
  }

  {
    auto const card = Heuristics::choose_best_fehl(trick, *this->ai_, false);
    auto const& hand = this->hand();
    auto const& cards_information = this->ai_->cards_information();
    if (   card
        && (hand.count(card.color()) == 1)
        && (cards_information.color_runs(card.color()) == 0) ) {
      this->rationale_.add(_("Heuristic::return::create fehl: %s", _(card)));
      return card;
    }
  }

  {
    auto const card = Heuristics::choose_pfund_card(trick, *this->ai_);
    this->rationale_.add(_("Heuristic::return::pfund: %s", _(card)));
    return card;
  }
}

/** @return   whether the winnercard is a color ace or a trump
 **/
bool
Heuristic::PfundInFirstColorRun::condition_winnercard_ace_or_trump(Trick const& trick)
{
  if (trick.isstartcard()) {
    this->rationale_.add(_("Heuristic::condition::empty trick"));
    return false;
  }
  auto const& startcard = trick.startcard();
  if (startcard.istrump()) {
    this->rationale_.add(_("Heuristic::condition::trump trick"));
    return false;
  }
  auto const color = startcard.color();
  auto const& winnercard = trick.winnercard();
  auto const& cards_information = this->ai_->cards_information();
  if (winnercard.istrump()) {
    this->rationale_.add(_("Heuristic::condition::winnercard is trump"));
    return true;
  } else if (   (winnercard.value() == Card::ace)
             && (winnercard.color() == color) ) {
    if (cards_information.played(color) > trick.cards().count(color)) {
      this->rationale_.add(_("Heuristic::condition::%s played before", _(color)));
      return false;
    }
    if (cards_information.remaining_others(color) + 2 <= trick.remainingcardno()) {
      this->rationale_.add(_("Heuristic::condition::not enough remaining cards of %s", _(color)));
      return false;
    }
    this->rationale_.add(_("Heuristic::condition::winnercard is %s", _(winnercard)));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::winnercard is neither trump nor %s", _(Card(color, Card::ace))));
    return false;
  }
}
