/**********************************************************************
 *
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "serve_trump_trick.h"

#include "../ai.h"
#include "../team_information.h"
#include "../heuristics.h"
#include "../../../game/game.h"
#include "../../../party/rule.h"


/** constructor
 **/
Heuristic::ServeTrumpTrick::ServeTrumpTrick(Ai const& ai) :
  Heuristic(ai, Aiconfig::Heuristic::serve_trump_trick)
{ }

/** destructor
 **/
Heuristic::ServeTrumpTrick::~ServeTrumpTrick() = default;

/** @return  card to play or empty card
 **/
Card
Heuristic::ServeTrumpTrick::get_card(Trick const& trick)
{
  this->rationale_.clear();
  if (!(   this->condition_trump_trick(trick)
        && !this->condition_can_jab(trick)
        && !(   this->condition_solo_game(trick)
             && this->condition_soloplayer_behind(trick))
       ))
    return {};

  if (this->hand().hastrump()) {
    return this->serve_with_trump(trick);
  } else {
    return this->serve_with_color(trick);
  }
}

/** @return  trump to play or empty card
 **/
Card
Heuristic::ServeTrumpTrick::serve_with_trump(Trick const& trick)
{
  auto const& game = this->player().game();
  auto const& hand = this->hand();
  auto const card = hand.lowest_trump();
  if (card) {
    if (   card.points() < 10
        && !card.possible_genscher()) {
      this->rationale_.add(_("Heuristic::return::play low trump %s", _(card)));
      return card;
    }

    for (auto c : {Card::diamond_jack, Card::heart_jack, Card::spade_jack}) {
      if (   hand.contains(c)
          && c.istrump(game)) {
        this->rationale_.add(_("Heuristic::return::play low trump %s", _(c)));
        return c;
      }
    }
    this->rationale_.add(_("Heuristic::reject::no suitable trump to serve found"));
    return {};
  }

  this->rationale_.add(_("Heuristic::reject::no suitable trump to serve found"));
  return {};
}

/** @return  color card to play or empty card
 **/
Card
Heuristic::ServeTrumpTrick::serve_with_color(Trick const& trick)
{
  auto const& colors = this->player().game().cards().colors();
  auto const& hand = this->hand();
  for (auto const v : {Card::nine, Card::jack, Card::queen, Card::king}) {
    for (auto const c : colors) {
      auto const card = Card(c, v);
      if (   hand.contains(card)
          && hand.count(c) == 1) {
        this->rationale_.add(_("Heuristic::return::single %s", _(card)));
        return card;
      } 
    }
  }
  for (auto const c : colors) {
    auto const card = Card(c, Card::nine);
    if (   hand.contains(card)
        && hand.count(c) == hand.count(card)) {
      this->rationale_.add(_("Heuristic::return::low color card: %s", _(card)));
      return card;
    } 
  }
  for (auto const v : {Card::nine, Card::jack, Card::queen, Card::king}) {
    for (auto const c : colors) {
      auto const card = Card(c, v);
      if (   hand.contains(card)
          && hand.highest_card(c).points() <= 4) {
        this->rationale_.add(_("Heuristic::return::low color card: %s", _(card)));
        return card;
      } 
    }
  }
  this->rationale_.add(_("Heuristic::reject::no suitable color card to serve found"));
  return {};
}
