/**********************************************************************
 *
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 - 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "try_for_doppelkopf.h"

#include "../team_information.h"
#include "../../../card/trick.h"
#include "../../../game/game.h"
#include "../../../party/rule.h"

#include "../ai.h"
#include "../heuristics.h"

/** constructor
 **/
Heuristic::TryForDoppelkopf::TryForDoppelkopf(Ai const& ai) :
  Heuristic(ai, Aiconfig::Heuristic::try_for_doppelkopf)
{ }

/** destructor
 **/
Heuristic::TryForDoppelkopf::~TryForDoppelkopf() = default;

/** @return  card to play or empty card
 **/
Card
Heuristic::TryForDoppelkopf::get_card(Trick const& trick)
{
  this->rationale_.clear();
  if (!(   !this->condition_solo_game(trick)
        && this->condition_second_last_card(trick)
        && this->condition_can_jab_with_trump(trick)
        && this->condition_enough_trick_points_for_doppelkopf(trick)
        && this->condition_has_ten_points_in_trump()
        && this->condition_last_player_has_10_points(trick)
        && (   this->condition_last_player_guessed_same_team(trick)
            || this->condition_last_player_must_play_10_points(trick))
        && this->condition_can_announce_no_120()
       ))

    return {};

  auto const& game = trick.game();
  auto const& hand = this->hand();

  { // non special cards
    auto const try_cards = {
      Card(game.cards().trumpcolor(), Card::ace),
      Card(game.cards().trumpcolor(), Card::ten),
    };
    for (auto const card : try_cards) {
      if (!hand.contains(card))
        continue;
      if (!trick.isjabbed(card))
        continue;

      if (HandCard(hand, card).is_special())
        continue;

      this->rationale_.add(_("Heuristic::return::no special card - %s", _(card)));
      return card;
    }
  }

  { // special cards
    auto const try_cards = {
      Card(game.cards().trumpcolor(), Card::ten),
      Card(game.cards().trumpcolor(), Card::ace),
      Card::dolle
    };

    for (auto const card : try_cards) {
      if (!hand.contains(card))
        continue;
      if (!trick.isjabbed(card))
        continue;

      if (!HandCard(hand, card).is_special())
        continue;

      if (   !this->guessed_same_team(trick.winnerplayer())
          && !trick.winnercard().less(this->trump_card_limit())) {
        this->rationale_.add(_("Heuristic::return::special card needed for winning the trick - %s", _(card)));
        return card;
      }

      if (Heuristics::secure_tricks_of_team(*this->ai_)
          == game.tricks().remaining_no()) {
        this->rationale_.add(_("Heuristic::return::special card not needed for an announcement - %s", _(card)));
        return card;
      }

      auto const announcement = game.announcements().last(this->team());
      if (  (announcement == Announcement::no60)
          ? Heuristics::say_no30(*this->ai_)
          : (announcement == Announcement::no90)
          ? Heuristics::say_no60(*this->ai_)
          : (announcement == Announcement::no120)
          ? Heuristics::say_no90(*this->ai_)
          : is_reply(announcement)
          ? Heuristics::say_no90(*this->ai_)
          : (announcement == Announcement::noannouncement)
          ? Heuristics::make_announcement(*this->ai_)
          : false) {
        this->rationale_.add(_("Heuristic::return::announcement can be improved - %s", _(card)));
        return card;
      }
    } // for (card)
  }

  this->rationale_.add(_("Heuristic::reject::no card found to jab"));
  return Card::empty;
}

/** @return   whether the last player has ten points
 **/
bool
Heuristic::TryForDoppelkopf::condition_last_player_has_10_points(Trick const& trick)
{
  auto const& hand = this->guessed_hand(trick.lastplayer());
  auto const cards = hand.validcards(trick);
  if (cards.highest_value() >= 10) {
    this->rationale_.add(_("Heuristic::condition::last player has 10 points"));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::last player does not have 10 points"));
    return false;
  }
}

/** @return   whether the last player must play 10 points
 **/
bool
Heuristic::TryForDoppelkopf::condition_last_player_must_play_10_points(Trick const& trick)
{
  auto const color = trick.startcard().tcolor();
  auto const& hand = this->guessed_hand(trick.lastplayer());
  auto const cards = hand.validcards(trick);
  auto const lowest_value = cards.lowest_value(color);
  if (lowest_value >= 10) {
    this->rationale_.add(_("Heuristic::condition::last player must play 10 points"));
    return false;
  } else {
    this->rationale_.add(_("Heuristic::condition::last player has other then 10 points to play"));
    return false;
  }
}

/** @return   whether the player can announce no 120 or already has
 **/
bool
Heuristic::TryForDoppelkopf::condition_can_announce_no_120()
{
  auto const& player = this->player();
  auto const team = player.team();
  auto const& game = player.game();
  if (player.announcement() == Announcement::noannouncement) {
    this->rationale_.add(_("Heuristic::condition::the player has announced %s", _(team)));
    return true;
  } else if (game.announcements().is_valid(Announcement::no120, player)
            ) {
    this->rationale_.add(_("Heuristic::condition::the player can announce %s", _(team)));
    return true;
  } else {
    this->rationale_.add(_("Heuristic::condition::the player cannot announce %s", _(team)));
    return false;
  }
}
