/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"
#include <fstream>

#include "w_virtual_games.h"

#include "virtual_games_interface.h"
#include "trickweighting.h"
#include "ai.h"
#include "cards_information.h"
#include "cards_information.of_player.h"
#include "team_information.h"
#include "heuristics.h"
#include "../../game/game.h"
#include "../../game/game_summary.h"
#include "../../game/exception.h"
#include "../../card/trick.h"
#include "../../misc/setting.h"
#include "../../ui/ui.h"

// whether to save the runtime
//#define SAVE_RUNTIME

#ifdef RELEASE
#undef SAVE_RUNTIME
#endif

/** constructor
 **
 ** @param    ai   the corresponding ai
 **/
WVirtualGames::WVirtualGames(Ai const& ai) :
  Weighting(ai),
  future_limit_(ai.future_limit())
{
  this->calculate_end_depth();
}

/** destructor
 **/
WVirtualGames::~WVirtualGames() = default;

/** @return   how many cards to calc into the future
 **/
unsigned
WVirtualGames::future_limit() const
{ return this->future_limit_; }

/** @return   how many tricks to calc into the future
 **/
unsigned
WVirtualGames::end_depth() const
{ return this->end_depth_; }

/** calculates the maximal depth of tricks,
 ** so that no more of 'future limit' tricks are considered
 **/
void
WVirtualGames::calculate_end_depth()
{
  // #combination for all tricks
  unsigned long int alltricks = 1;
  // #combination for one trick
  unsigned long int onetrick = 1;

  Trick const& trick = this->vgi().game().tricks().current();

  // count the number of card-combinations for the current trick
  for (unsigned i = trick.actcardno();
       i < this->vgi().game().playerno();
       ++i)
    onetrick *= this->vgi().handofplayer(trick.player_of_card(i)
                                        ).validcards(trick).cardsnumber();

  alltricks *= onetrick;

  // count the number of card-combinations for the other tricks
  // (since I don't know what color is played, I take all cards of the hand)
  for (this->end_depth_ = 1;
       // '<' because also the current trick is counted in tricks_remaining_no
       (this->end_depth_ < this->vgi().game().tricks().remaining_no());
       this->end_depth_ += 1){

    onetrick = 1;
    for (unsigned i = 0; i < this->vgi().game().playerno(); ++i) {
#ifndef POSTPONED
      if (this->vgi().game().tricks().current().cardno_of_player(this->vgi().game().player(i))
          >= this->vgi().game().tricks().current().actcardno())
        // the player still has to play a card in the current trick
        onetrick *= (this->vgi().handofplayer(this->vgi().game().player(i)).cardsnumber()
                     - this->end_depth_);
      else
        // the player has already played a card, so he has one card less 
        // on the hand
        onetrick *= (this->vgi().handofplayer(this->vgi().game().player(i)).cardsnumber()
                     - (this->end_depth_ - 1));

      DEBUG_ASSERTION((onetrick > 0),
                      "WVirtualGames::end_depth_calculate():\n"
                      "  onetrick == 0\n"
                      << "cardsnumber = " << this->vgi().handofplayer(this->vgi().game().player(i)).cardsnumber() << '\n'
                      << "end_depth = " << this->end_depth_ << '\n'
                      << "trickno = " << this->vgi().game().tricks().current_no() << '\n'
                      << "remaining trickno = " << this->vgi().game().tricks().remaining_no() << '\n'
                     );
#else
      onetrick *= this->vgi().hand().cardsnumber() - this->end_depth_;
#endif
    }

    if (alltricks < this->future_limit() / onetrick) {
      alltricks *= onetrick; 
    } else  {
      break;
    }
  } // for (this->end_depth_)

  this->end_depth_ = min(this->end_depth(),
                              ((this->vgi().last_trick_to_calculate()
                                >= this->vgi().game().tricks().current_no())
                               ? (this->vgi().last_trick_to_calculate()
                                  - this->vgi().game().tricks().current_no())
                               : 0));

#ifdef POSTPONED
  this->end_depth_ = max(unsigned(2), this->end_depth());
#endif


  return ;
} // void VirtualGames::calculate_end_depth()

/** -> result
 **
 ** @param    card_   the card
 **
 ** @return   the weighting for the card 'card'
 **/
int
WVirtualGames::weighting(Card const& card_) const
{
  // the weighting
  int weighting = 0;

  // create the virtual game
  auto players = this->vgi().create_virtual_ais();
  auto const game_virt_p = this->vgi().create_virtual_game(players);
  Game& virt_game = *game_virt_p;

  // the players
  auto& ai_virt = virt_game.players();
  // the copy of the original ai
  Ai* ai_copy = dynamic_cast<Ai*>(&ai_virt.player(this->vgi().no()));
  // the card in the virtual game
  HandCard const card(virt_game.players().current_player().hand(), card_);

  { // play the card 'card'
    // make swines announcements
    if (   card.possible_swine()
        && virt_game.swines().swines_announcement_valid(virt_game.players().current_player())) {
      virt_game.swines().swines_announce(virt_game.players().current_player());
    }
    if (   card.possible_hyperswine()
        && virt_game.swines().hyperswines_announcement_valid(virt_game.players().current_player()))
      virt_game.swines().hyperswines_announce(virt_game.players().current_player());

    // play the card
    // code copied from 'Game::nextcard()'
    virt_game.players().current_player().hand().playcard(card);
    virt_game.tricks().current().add(card);
    virt_game.teaminfo().update();
    virt_game.players().next_current_player();
    for (auto& p : virt_game.players()) {
      p.card_played(card);
    }
  } // play the card 'card'

  try {
    // the other virtual players play a card
    while (!virt_game.tricks().current().isfull()) {

      { // update the cards information for all players
        CardsInformation& cards_information(ai_copy->cards_information());

        // save the hands, because 'cards_information' changes, when the hand of 'ai_copy' is set
        vector<unique_ptr<Hand>> hands;
        for (auto const& player : virt_game.players()) {
          hands.push_back(make_unique<Hand>(ai_copy->handofplayer(player)));
#ifdef WORKAROUND
          // as long as the cards information uses much processing time
          ::ui->update();
#endif
        } // for (player)

        unsigned p = 0;
        for (auto const& player : virt_game.players()) {
          cards_information.set_hand(player, *hands[p]);
          ++p;
        } // for (player)

        p = 0;
        for (auto& ai : ai_virt) {
          dynamic_cast<Ai&>(ai).set_cards_information(cards_information);
          ai.set_hand(cards_information.estimated_hand(p));
#ifdef WORKAROUND
          // as long as the cards information uses much processing time
          ::ui->update();
#endif
          ++p;
        } // for (ai)
      } // update the cards information for all players

      DEBUG_ASSERTION((virt_game.players().current_player().hand().cardsnumber() > 0),
                      "WVirtualGames::weightings_calc():\n"
                      "  current player " << virt_game.players().current_player().name()
                      << " has an empty hand");
      try {
#ifdef DKNOF
        virt_game.players().current_player().self_check();
#endif
        virt_game.nextplayer();
      } catch (InvalidGameException const& e) {
        // ToDo: fix cards information
        // or:   remove last played card,
        //       remove it from the hand of the player
        //       and try again
        //       This is not possible, because the information is missing here. -> possible in gametree
        cerr << "WVirtualGames: invalid game\n"
          << "  line " << __LINE__ << '\n'
          << "  ignoring card"
          << endl;
        throw;
      } // try
    } // while (!virt_game.tricks().current().isfull())

    // first trick is full
    virt_game.evaluatetrick();


    // play tricks till the end depth
    for (unsigned t = 1;
         ((t < this->end_depth())
          && (virt_game.tricks().current_no() < virt_game.tricks().max_size()));
         ++t) {
      if (INFO_W_VIRTUAL_GAMES)
        if (!this->vgi().game().isvirtual())
          cout << "  trick depth " << t
            << " (" << this->end_depth() - 1 << ")" << endl;


      // set the future limit of the ai, so that they don't calculate too much
      for (auto& ai : ai_virt)
        dynamic_cast<Ai&>(ai).set_last_trick_to_calculate(virt_game.tricks().current_no()
                                                          + (this->end_depth()
                                                             - t));

      // start a new trick
      virt_game.tricks().add(Trick(virt_game.players().current_player()));
      for (auto& p : virt_game.players()) {
        p.trick_open(virt_game.tricks().current());
      }

      while (!virt_game.tricks().current().isfull()) {
        { // update the cards information for all players
          CardsInformation& cards_information(ai_copy->cards_information());

          // save the hands, because 'cards_information' changes, when the hand of 'ai_copy' is set
          vector<unique_ptr<Hand>> hands;
          for (auto const& player : virt_game.players()) {
            hands.push_back(make_unique<Hand>(ai_copy->handofplayer(player)));
#ifdef WORKAROUND
            // as long as the cards information uses much processing time
            ::ui->update();
#endif
          } // for (player)

          unsigned p = 0;
          for (auto const& player : virt_game.players()) {
            cards_information.set_hand(player, *hands[p]);
            ++p;
          } // for (player)

          p = 0;
          for (auto& ai : ai_virt) {
            dynamic_cast<Ai&>(ai).set_cards_information(cards_information);
            ai.set_hand(*hands[p]);
#ifdef WORKAROUND
            // as long as the cards information uses much processing time
            ::ui->update();
#endif
            ++p;
          } // for (ai)
        } // update the cards information for all players

        try {
          virt_game.nextplayer();
        } catch (InvalidGameException const& e) {
          // ToDo: fix cards information
          // or:   remove last played card,
          //       remove it from the hand of the player
          //       and try again
          //       This is not possible, because the information is missing here. -> possible in gametree
          cerr << "WVirtualGames: invalid game\n"
            << "  line " << __LINE__ << '\n'
            << "ignoring card\n"
            << endl;
          throw;
        } // try
      }
      virt_game.evaluatetrick();
    } // for (t < this->end_depth())

    // finished with creating tricks,
    // now calc the modi for each trick
    for (unsigned t = this->vgi().game().tricks().current_no();
         t < virt_game.tricks().current_no();
         ++t) {
      // add the modi to the weighting
      weighting
        += TrickWeighting::modi(this->vgi(), virt_game.tricks().trick(t),
                                ai_copy->team(),
                                virt_game.tricks().trick(t).card_of_player(*ai_copy) );
    } // for (t < tricks().current_no())

    if (virt_game.tricks().current_no() < virt_game.tricks().max_size() - 1) {
      // add some points, if the own team is in the back,
      weighting += TrickWeighting::backhand(this->vgi(),
                                            card,
                                            virt_game);
    } // if (virt_game.tricks().current_no() < virt_game.tricks().max_size() - 1)


    // game finished, the main reason is the final result...

    if( virt_game.tricks().remaining_no() == 0 ) {
      virt_game.finish();
      GameSummary const game_summary(virt_game);
      // the correspoding virtual player to the ai
      if (game_summary.winnerteam() == Team::noteam) {
        weighting
          += 10000 * (game_summary.points(ai_copy->team())
                      - game_summary.points(opposite(ai_copy->team())));
      } else { // if !(game_summary.winnerteam() == Team::noteam)
        if ( game_summary.winnerteam() == ai_copy->team())
          weighting += 10000 * game_summary.points();
        else
          weighting -= 10000 * game_summary.points();
      } // if !(game_summary.winnerteam() == Team::noteam)
    }

#if 0
    // output of the modi
    if (this->vgi().game() == ::party.game()) {
      cout << this->hand().card(c)
        << ": "
        << weighting
        << endl;

    } // if (this->vgi().game() == ::party.gam())

#endif // output of the modi

  } catch (InvalidGameException const& e) {
    // if the game is invalid, don't take the card
    weighting = INT_MIN;
#ifndef RELEASE
    if (!this->vgi().game().isvirtual()) {

#ifdef DEBUG_ASSERT
#if 0
      cerr << "cards information\n";
      //cerr << static_cast<Ai const&>(virt_game.players().current_player()).cards_information() << endl;
      cerr << '\n';
      cerr << "hands\n";
      for (vector<Player*>::const_iterator p = virt_game.players_begin();
           p != virt_game.players_end();
           ++p)
        cerr << '\n' << (*p)->name() << '\n'
          << (*p)->hand();
      cerr << '\n';
      cerr << "previous trick = " << virt_game.tricks().current_no() - 1 << endl;
      //cerr << virt_game.tricks().trick(virt_game.tricks().current_no() - 1);
      cerr << '\n';
      cerr << "current trick = " << virt_game.tricks().current_no() << endl;
      cerr << virt_game.tricks().current();
      cerr << '\n';
      cerr << "Hand:\n";
      //cerr << virt_game.players().current_player().hand();
      cerr << '\n';
      cerr << "last heuristic: " << static_cast<Ai const&>(virt_game.players().current_player()).lastHeuristic_ << endl;
      cerr << '\n';
      cerr << __FILE__ << " line " << __LINE__ << '\n';
      cerr << "no valid game found for card '" << card << "'"
        << endl;
#endif
#endif
    }
    DEBUG_ASSERTION(this->vgi().game().isvirtual(),
                    "WVirtualGames::weightings_calc():\n"
                    "  no valid game found for card '" << card
                    << "'");
#endif
    DEBUG_CAUGHT();
  } catch (...) {
    // free the memory
    throw;
  } // try

  if (INFO_W_VIRTUAL_GAMES)
    if (!this->vgi().game().isvirtual())
      cout << "weighting " << card << ": "
        << weighting << "\n";

  return weighting;
} // int WVirtualGames::weighting(Card card_) const
