/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/ 

#include "constants.h"

#include "heuristicsmap.h"
#include "player.h"
#include "../game/game.h"
#include "../party/rule.h"
#include "../basetypes.h"

#include "../misc/bug_report.h"

/** -> result
 **
 ** @param    game   game
 **
 ** @return   group of the gametype
 **/
HeuristicsMap::GameTypeGroup
HeuristicsMap::group(Game const& game)
{
  switch (game.type()) {
  case GameType::normal:
    return GameTypeGroup::normal;
  case GameType::thrown_nines:
  case GameType::thrown_kings:
  case GameType::thrown_nines_and_kings:
  case GameType::thrown_richness:
  case GameType::fox_highest_trump:
  case GameType::redistribute:
    DEBUG_ASSERT(false);
    break;
  case GameType::poverty:
    return GameTypeGroup::poverty;
  case GameType::genscher:
    return GameTypeGroup::normal;
  case GameType::marriage:
    if (game.marriage().selector() == MarriageSelector::team_set)
      return GameTypeGroup::normal;
    else
      return GameTypeGroup::marriage_undetermined;
  case GameType::marriage_solo:
    return GameTypeGroup::soli_color;
  case GameType::marriage_silent:
    return GameTypeGroup::marriage_silent;
  case GameType::solo_jack:
  case GameType::solo_queen:
  case GameType::solo_king:
    return GameTypeGroup::soli_single_picture;
  case GameType::solo_queen_jack:
  case GameType::solo_king_jack:
  case GameType::solo_king_queen:
    return GameTypeGroup::soli_double_picture;
  case GameType::solo_koehler:
    return GameTypeGroup::solo_koehler;
  case GameType::solo_club:
  case GameType::solo_spade:
  case GameType::solo_heart:
  case GameType::solo_diamond:
    return GameTypeGroup::soli_color;
  case GameType::solo_meatless:
    return GameTypeGroup::solo_meatless;
  } // switch(game.type())

  DEBUG_ASSERT(false);

  return GameTypeGroup::normal;
} // HeuristicsMap::GameTypeGroup HeuristicsMap::group(Game game)

/** -> result
 **
 ** @param    player   player
 **
 ** @return   group of the player
 **/
HeuristicsMap::PlayerTypeGroup
HeuristicsMap::group(Player const& player)
{
  PlayerTypeGroup group;

  switch(player.team()) {
  case Team::re:
    group = PlayerTypeGroup::re;
    break;
  case Team::contra:
  case Team::maybe_contra:
    group = PlayerTypeGroup::contra;
    break;
  default:
#ifndef RELEASE
    cerr << "HeuristicsMap::group(player):\n"
      <<"  wrong team " << player.team()
      << endl;
#if 0
    ::bug_report.report("HeuristicsMap::group(player):\n"
			"  wrong team " + ::name(player.team()));
#endif
    group = PlayerTypeGroup::contra;
    break;
#endif
#ifdef POSTPONED
    DEBUG_ASSERTION(false,
		    "HeuristicsMap::group(player):\n"
		    "  wrong team " << player.team());
#else
    group = PlayerTypeGroup::contra;
#endif
    break;
  } // switch(gametype)

  // special types

  // poverty
  if (    (player.game().type() == GameType::poverty)
       && (player == player.game().players().soloplayer()) )
    group = PlayerTypeGroup::special;

  // genscher
  if ( (player.game().type() == GameType::normal)
       && player.game().rule()(Rule::Type::genscher)
       && (player.hand().count_trump_kings()
	   == player.game().rule()(Rule::Type::number_of_same_cards)) )
    group = PlayerTypeGroup::special;

  return group;
} // HeuristicsMap::PlayerTypeGroup HeuristicsMap::group(Player const& player)

/** constructor
 **
 ** @param    player   player
 **
 ** @return   key according to the player
 **/
HeuristicsMap::Key::Key(Player const& player) :
  gametype_group(HeuristicsMap::group(player.game())),
  playertype_group(HeuristicsMap::group(player))
{ }

/** -> result
 **
 ** @param    lhs   first key
 ** @param    rhs   second key
 **
 ** @return   whether 'lhs' is equal to 'rhs'
 **/
bool
HeuristicsMap::operator==(HeuristicsMap::Key const lhs,
			 HeuristicsMap::Key const rhs)
{
  return (   (lhs.gametype_group == rhs.gametype_group)
	  && (lhs.playertype_group == rhs.playertype_group));
} // bool HeuristicsMap::operator==(HeuristicsMap::Key lhs, HeuristicsMap::Key rhs)

/** -> result
 **
 ** @param    lhs   first key
 ** @param    rhs   second key
 **
 ** @return   whether 'lhs' is different to 'rhs'
 **/
bool
HeuristicsMap::operator!=(HeuristicsMap::Key const lhs,
                          HeuristicsMap::Key const rhs)
{
  return !(lhs == rhs);
} // bool HeuristicsMap::operator!=(HeuristicsMap::Key lhs, HeuristicsMap::Key rhs)

/** -> result
 **
 ** @param    lhs   first key
 ** @param    rhs   second key
 **
 ** @return   whether 'lhs' is less than 'rhs'
 **/
bool
HeuristicsMap::operator<(HeuristicsMap::Key const lhs,
                         HeuristicsMap::Key const rhs)
{
  return (   (lhs.gametype_group < rhs.gametype_group)
          || (   (lhs.gametype_group == rhs.gametype_group)
              && (lhs.playertype_group < rhs.playertype_group)));
} // bool HeuristicsMap::operator<(HeuristicsMap::Key lhs, HeuristicsMap::Key rhs)

/** -> result
 **
 ** @param    group   group
 **
 ** @return   the name of the group
 **/
string
name(HeuristicsMap::GameTypeGroup const group)
{
  switch (group) {
  case HeuristicsMap::GameTypeGroup::normal:
    return "normal";
    (void)_("HeuristicsMap::GameTypeGroup::normal");
  case HeuristicsMap::GameTypeGroup::poverty:
    return "poverty";
    (void)_("HeuristicsMap::GameTypeGroup::poverty");
  case HeuristicsMap::GameTypeGroup::marriage_undetermined:
    return "undetermined marriage";
    (void)_("HeuristicsMap::GameTypeGroup::undetermined marriage");
  case HeuristicsMap::GameTypeGroup::marriage_silent:
    return "silent marriage";
    (void)_("HeuristicsMap::GameTypeGroup::silent marriage");
  case HeuristicsMap::GameTypeGroup::solo_meatless:
    return "solo meatless";
    (void)_("HeuristicsMap::GameTypeGroup::solo meatless");
  case HeuristicsMap::GameTypeGroup::soli_color:
    return "soli color";
    (void)_("HeuristicsMap::GameTypeGroup::soli color");
  case HeuristicsMap::GameTypeGroup::soli_single_picture:
    return "soli single picture";
    (void)_("HeuristicsMap::GameTypeGroup::soli single picture");
  case HeuristicsMap::GameTypeGroup::soli_double_picture:
    return "soli double picture";
    (void)_("HeuristicsMap::GameTypeGroup::soli double picture");
  case HeuristicsMap::GameTypeGroup::solo_koehler:
    return "solo koehler";
    (void)_("HeuristicsMap::GameTypeGroup::solo koehler");
  } // switch (group)

  return "";
} // string name(HeuristicsMap::GameTypeGroup const group)

/** -> result
 **
 ** @param    name   name of the group
 **
 ** @return   the game type group with the name 'name'
 **/
HeuristicsMap::GameTypeGroup
HeuristicsMap::GameTypeGroup_from_name(string const& name)
{
  for (auto g : game_type_group_all) {
    if (name == ::name(g)) {
      return g;
    }
  }
#ifdef DEPRECATED
  // 2018-01-23
  if (name == "default")
    return GameTypeGroup::normal;
#endif

  DEBUG_ASSERTION(false,
                  "HeuristicsMap::GameTypeGroup_from_name(" << name << ")\n"
                  "  did not find gametype group '" << name << "'");
  return GameTypeGroup::normal;
} // HeuristicsMap::GameTypeGroup HeuristicsMap::GameTypeGroup_from_name(string name)

/** -> result
 **
 ** @param    name   name of the group
 **
 ** @return   the player type group with the name 'name'
 **/
HeuristicsMap::PlayerTypeGroup
HeuristicsMap::PlayerTypeGroup_from_name(string const& name)
{
  for (auto g : player_type_group_all) {
    if (name == ::name(g)) {
      return g;
    }
  }

  DEBUG_ASSERTION(false,
                  "HeuristicsMap::PlayerTypeGroup_from_name(\"" << name << "\")\n"
                  "  did not find playertype group '" << name << "'");
  return PlayerTypeGroup::re;
} // HeuristicsMap::PlayerTypeGroup HeuristicsMap::PlayerTypeGroup_from_name(string name)

/** -> result
 **
 ** @param    group   group
 **
 ** @return   the name of the group
 **/
string
name(HeuristicsMap::PlayerTypeGroup group)
{
  switch (group) {
  case HeuristicsMap::PlayerTypeGroup::contra:
    return "contra";
    (void)_("HeuristicsMap::PlayerTypeGroup::contra");
  case HeuristicsMap::PlayerTypeGroup::re:
    return "re";
    (void)_("HeuristicsMap::PlayerTypeGroup::re");
  case HeuristicsMap::PlayerTypeGroup::special:
    return "special";
    (void)_("HeuristicsMap::PlayerTypeGroup::special");
  } // switch (group)

  return "";
} // string name(HeuristicsMap::PlayerTypeGroup group)

/** translation
 **
 ** @param    gametype_group   type of the game type
 **
 ** return   translation
 **/
string
gettext(HeuristicsMap::GameTypeGroup const gametype_group)
{
  return gettext("HeuristicsMap::GameTypeGroup::"
                 + ::name(gametype_group));
}

/** translation
 **
 ** @param    playertype_group   type of the player type
 **
 ** return   translation
 **/
string
gettext(HeuristicsMap::PlayerTypeGroup const playertype_group)
{
  return gettext("HeuristicsMap::PlayerTypeGroup::"
                 + ::name(playertype_group));
}
