/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 ********************************************************************/

#include "constants.h"

#include "player.h"
#include "playersDb.h"

#include "human/human.h"
#include "ai/ai.h"
#include "ai.dummy/ai.h"
#include "ai.random/ai.h"

#include "../card/hand.h"
#include "../party/party.h"
#include "../party/rule.h"
#include "../game/game_summary.h"
#include "../card/trick.h"
#include "../misc/setting.h"
#include "../ui/ui.h"

// whether to write the databasea
bool Player::write_database = true;

/** -> result
 **
 ** @param    type   playertype
 **
 ** @return   pointer to a new player of the given type
 **/
unique_ptr<Player>
Player::create(Type const type)
{
  switch(type) {
  case Player::Type::unset:
    DEBUG_ASSERTION(false,
                    "Player::create(type):\n"
                    "  type == UNSET");
    return nullptr;
  case Player::Type::human:
    return make_unique<Human>();
  case Player::Type::ai:
    return make_unique<Ai>();
  case Player::Type::ai_dummy:
    return make_unique<AiDummy>();
  case Player::Type::ai_random:
    return make_unique<AiRandom>();
  } // switch(type)

  return {};
} // unique_ptr<Player> Player::create(Type type)

/** -> result
 **
 ** @param    type   playertype
 ** @param    name   name of the player
 **
 ** @return   pointer to a new player of the given type
 **/
unique_ptr<Player>
Player::create(Type const type, string const& name)
{
  auto player = Player::create(type);
  if (player)
    player->set_name(name);

  return player;
} // unique_ptr<Player> Player::create(Type type, string name)

/** -> result
 **
 ** @param    type     playertype
 ** @param    player   the player to be copied
 **
 ** @return   pointer to a new player of the given type and with data
 **      from 'player'
 **/
unique_ptr<Player>
Player::create(Type const type, Player const& player)
{
  unique_ptr<Player> player_new;

  switch(type) {
  case Player::Type::unset:
    DEBUG_ASSERTION(false,
                    "Player::create(type, player):\n"
                    "  type == unset");
    break;
  case Player::Type::human:
#ifdef WORKAROUND
    if (dynamic_cast<Ai const*>(&player))
      player_new = make_unique<Human>(dynamic_cast<Ai const&>(player));
    else if (dynamic_cast<Aiconfig const*>(&player))
      player_new = make_unique<Human>(player, dynamic_cast<Aiconfig const&>(player));
    else
      player_new = make_unique<Human>(player);
#else
    // see 'AI'
    player_new = make_unique<Human>(player);
#endif
    break;

  case Player::Type::ai:
#ifdef WORKAROUND
    if (dynamic_cast<Ai const*>(&player))
      player_new = make_unique<Ai>(dynamic_cast<Ai const&>(player));
    else if (dynamic_cast<Aiconfig const*>(&player))
      player_new = make_unique<Ai>(player, dynamic_cast<Aiconfig const&>(player));
    else
      player_new = make_unique<Ai>(player);
#else
    // if 'p' is of type 'Human', the GNU C++ compiler calls 'Ai(Player)' but
    // not 'Ai(Ai)' although 'Human' is a child of 'Ai'
    player_new = make_unique<Ai>(player);
#endif
    break;

  case Player::Type::ai_dummy:
    if (dynamic_cast<Aiconfig const*>(&player))
      player_new = make_unique<AiDummy>(player, dynamic_cast<Aiconfig const&>(player));
    else
      player_new = make_unique<AiDummy>(player);
    break;

  case Player::Type::ai_random:
    if (dynamic_cast<Aiconfig const*>(&player))
      player_new = make_unique<AiRandom>(player, dynamic_cast<Aiconfig const&>(player));
    else
      player_new = make_unique<AiRandom>(player);
    break;
  } // switch(type)

  return player_new;
} // unique_ptr<Player> Player::create(Type type, Player player)

/** returns a player, created by the infos of the file
 **
 ** @param    filename   file with the player info
 **
 ** @return   pointer to a new player of the given type
 **/
unique_ptr<Player>
Player::create(string const& filename)
{
  ifstream istr(filename.c_str());
  if (istr.fail())
    return {};

#ifdef WORKAROUND
  // test whether the player file is empty
  // Note: if the human player (player 0) cannot be loaded,
  // he is replaced by an ai.
  istr.peek();
  if (istr.eof()) {
    cerr << "could not load player from file '" << filename << "', "
      << "creating new player" << endl;
    return Player::create(Type::ai);
  }
#endif

  return Player::create(istr);
} // unique_ptr<Player> Player::create(string filename)

/** returns a player, created by the infos of the stream
 **
 ** @param    istr   stream with the player info
 **
 ** @return   pointer to a new player of the given type
 **/
unique_ptr<Player>
Player::create(istream& istr)
{
  auto player = make_unique<Player>(Type::unset);

  Config config;
  istr >> config;

  int depth = 0;
  Type type_before = Player::Type::unset;

  if ((config.name == "{")
      && config.value.empty()) {
    depth += 1;
    istr >> config;
  }
  while (istr.good()) {
#ifndef OUTDATED
    // DK: 2005-01-21
    // The player is not included in '{ }', so check, whether the next line
    // is 'Name = ...'
    while ( (istr.peek() == '\n')
           || (istr.peek() == '\r'))
      istr.get();
    if ((istr.peek() == 'N')
        && (depth == 0))
      break;
#endif

    if (istr.eof())
      break;

    if (istr.fail()) {
#ifndef RELEASE
      cerr << "reading player failed" << endl;
#endif
      return nullptr;
    }

    // finished with the config file
    if ((config.name == "") && (config.value == ""))
      continue;

    if ((config.name == "{")
        && config.value.empty())
      depth += 1;

    if ((config.name == "}")
        && config.value.empty()) {
      depth -= 1;
      // finished with the player
      if (depth == -1)
        istr.putback('}');
      if (depth <= 0)
        break;
    }

    if (!player->read(config, istr))
      cerr << "Unknown player line '" << config.name << " = " << config.value
        << '\'' << endl;

    // if the type is set, change the underlying player
    if (player->type() != type_before) {
      unique_ptr<Player> player2;
      switch(player->type()) {
      case Type::unset:
        DEBUG_ASSERTION(false,
                        "Player::create(istr):\n"
                        "  player type is 'UNSET'");
      case Type::human:
        player2 = Player::create(Type::human, *player);
        break;
      case Type::ai:
        player2 = Player::create(Type::ai, *player);
        break;
      case Type::ai_dummy:
        player2 = Player::create(Type::ai_dummy, *player);
        break;
      case Type::ai_random:
        player2 = Player::create(Type::ai_random, *player);
        break;
      } // switch(player->type())
      player = move(player2);
      type_before = player->type();
    } // if (player->type() != UNSET)

    istr >> config;
  } // while (istr.good())

  return player;
} // unique_ptr<Player> Player::create(istream& istr)

/** constructor
 **
 ** @param    type   type of the player
 **/
Player::Player(Type const type)  :
  type_(type),
  sorted_hand_(this->hand_),
  trickpile_(*this),
  db_(make_unique<PlayersDb>())
{ }

/** copy constructor
 **
 ** @param    player  player to copy
 **/
Player::Player(Player const& player) :
  game_(player.game_),
  type_(Player::Type::unset),
  name_(player.name_),
  voice_(player.voice_),
  color_(player.color()),
  team_(player.team_),
  hand_(*this, player.hand_),
  sorted_hand_(player.sorted_hand_, this->hand_),
  reservation_(player.reservation_),
  trickpile_(player.trickpile_, *this),
  db_( (player.db_ == nullptr)
      ? nullptr
      : new PlayersDb(player.db()) ) // *** DK must use a copy!
{ }

/** destructor
 **/
Player::~Player() = default;

/** self check
 ** when an error is found, an ASSERTION is created
 **
 ** @return   wether the self-check was successful (no error)
 **
 ** @todo   check the hand
 ** @todo   check the sorted hand
 **/
bool
Player::self_check() const
{
  DEBUG_ASSERTION(this->db_,
                  "Player::self_check()\n"
                  "  no database");
  if (this->game_) {
    DEBUG_ASSERTION((this->game_->players().no(*this) != UINT_MAX),
                    "Player::self_check()\n"
                    "  player not in game: " << this->name_);


    // ToDo
    // check the hand
    // check the sorted hand

    this->hand().self_check();
  } // if (this->game_)

  return true;
} // virtual bool Player::self_check() const

/** clone the player
 **
 ** @return   pointer of a clone
 **      NULL if the player could get cloned
 **/
unique_ptr<Player>
Player::clone() const
{  
  return make_unique<Player>(*this);
} // virtual unique_ptr<Player> Player::clone() const

/** copy operator
 **
 ** @param    player  player to be copied
 **
 ** @return   copy of player
 **/
Player&
Player::operator=(const Player& player)
{  
  // copy the values
  this->game_ = player.game_;
  this->type_ = player.type_;
  this->name_ = player.name_;
  this->voice_ = player.voice_;
  this->color_ = player.color_;
  this->team_ = player.team_;
  this->hand_ = player.hand_;
  this->hand_.set_player(*this);
  this->sorted_hand_ = SortedHand(player.sorted_hand_, this->hand_);
  this->trickpile_ = player.trickpile_;
  this->db_        = (player.db_
                      ? make_unique<PlayersDb>(*player.db_) // *** DK must use a copy!
                      : nullptr);

  return *this;
} // Player& Player::operator=(const Player& player)

/** -> result
 **
 ** @param    lhs    player 1 to compare
 ** @param    rhs    player 2 to compare
 **
 ** @return   whether the players are equal (they lie in the same memory)
 **
 ** @todo   compare with values
 **/
bool
operator==(Player const& lhs, Player const& rhs)
{
  //return player1.is_equal(player2);
  return (&lhs == &rhs);
} // bool operator==(Player const& lhs, Player const& rhs)

/** -> result
 **
 ** @param    lhs   player 1 to compare
 ** @param    rhs   player 2 to compare
 **
 ** @return   whether the players are not equal
 **              (they don't lie in the same memory)
 **/
bool
operator!=(Player const& lhs, Player const& rhs)
{
  return !(lhs == rhs);
} // bool operator!=(Player const& lhs, Player const& rhs)

/** save the player in the file 'filename'
 **
 ** @param    filename    file the player shall be saved
 **
 ** @return   whether the saving was sucessful
 **/
bool
Player::save(string const& filename) const
{
  string filename_tmp = filename + ".tmp";
  ofstream ostr(filename_tmp.c_str());
  if (!ostr.good()) {
    ::ui->information(_("Error::Player(%s)::save: Error opening temporary file %s. Aborting saving.",
                        this->name(), filename_tmp), InformationType::problem);
    return false;
  }

  ostr << "# FreeDoko player (" << *::version << ")\n"
    << '\n';
  this->write(ostr);

  if (!ostr.good()) {
    ::ui->information(_("Error::Player(%s)::save: Error saving in temporary file %s. Keeping temporary file (for bug tracking).",
                        this->name(), filename_tmp), InformationType::problem);
    return false;
  }
  ostr.close();

#ifdef WINDOWS
  unlink(filename.c_str());
#endif
  if (rename(filename_tmp.c_str(), filename.c_str())) {
    ::ui->information(_("Error::Player(%s)::save: Could not rename temporary file %s to requested file %s. Keeping temporary file.",
                        this->name(), filename_tmp, filename), InformationType::problem);
    return false;
  }

  return true;
} // virtual bool Player::save(string const& filename) const

/** write the player in the stream
 **
 ** @param    ostr    output stream
 **
 ** @return   output stream
 **/
ostream&
Player::write(ostream& ostr) const
{
  ostr << "name = " << this->name_value() << '\n';
  if (this->name_value().empty())
    ostr << "# name = " << this->name() << '\n';
  ostr << "voice = " << this->voice() << '\n'
    << "type = " << this->type() << '\n'
    << "color = " << this->color() << '\n';
  if (Player::write_database) {
    ostr << '\n'
      << "database\n"
      << "{\n"
      << this->db()
      << "}\n";
  } // if (Player::write_database)

  return ostr;
} // virtual ostream& Player::write(ostream& ostr) const

/** output of the player
 **
 ** @param    ostr    output stream
 ** @param    player  player to be put in the stream
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, const Player& player)
{
  player.write(ostr);

  return ostr;
} // ostream& operator<<(ostream& ostr, const Player& player)

/** write the player in the stream in short form (name and type)
 **
 ** @param    ostr    output stream
 **/
void
Player::write_short(ostream& ostr) const
{
  ostr << this->name_value() << " -- "
    << this->type();
  return ;
} // void Player::write_short(ostream& ostr) const

/** read the config of the player
 **
 ** @param    config      configuration to read
 ** @param    istr      input stream
 **
 ** @return   whether the configuration was valid
 **/
bool
Player::read(Config const& config, istream& istr)
{
  if (config.separator) {
#ifndef OUTDATED
    // 0.7.3
    if ((config.name == "name")
        ||(config.name == "Name"))
#else
      if (config.name == "name")
#endif
      {
        this->set_name(config.value);
        return true;
      } else if (config.name == "voice") {
        this->set_voice(config.value);
        return true;
      } else if (config.name == "color") {
        this->set_color(config.value);
        return true;
#ifndef OUTDATED
        // 0.7.3
      } else if ((config.name == "type")
                 ||(config.name == "Type")) {
#else
      } else if (config.name == "Type") {
#endif
        if (this->type() != Type::unset) {
          cerr << "Type already set, ignoring further one" << endl;
          return true;
        }
        if (config.value == ::name(Type::human)) {
          this->set_type(Type::human);
        } else if (config.value == ::name(Type::ai)) {
          this->set_type(Type::ai);
        } else if (config.value == ::name(Type::ai_dummy)) {
          this->set_type(Type::ai_dummy);
        } else if (config.value == ::name(Type::ai_random)) {
          this->set_type(Type::ai_random);
        } else { // if (config.value == 'type')
          DEBUG_ASSERTION(false,
                          "Player::read(config, istr):\n"
                          "  player type '" << config.value << "' not known");
        }
        return true;
        // I know the type, so I can create the player
      }
  } else { // if (config.separator)
#ifndef OUTDATED
    // 0.7.3
    if ( (config.name == "database")
        || (config.name == "Database") )
#else
      if (config.name == "Database")
#endif
      {
        this->db_->read(istr);
        return true;
      }
  } // if (config.separator)

  return false;
} // virtual bool Player::read(Config config, istream& istr)

/** @return   the type of the player
 **/
Player::Type
Player::type() const
{
  return this->type_;
}

/** set the type
 **
 ** @param    type   new type of the player
 **/
void
Player::set_type(Type const type)
{
  this->type_ = type;
}

/** @return   the color of the player
 **/
string
Player::color() const
{
  return this->color_;
}

/** set the color
 **
 ** @param    color   new color of the player
 **/
void
Player::set_color(string const color)
{
  this->color_ = color;
}

/** @return   the team of the player
 **/
Team
Player::team() const
{
  return this->team_;
}

/** set the team
 **
 ** @param    team   new team of the player
 **/
void
Player::set_team(Team const team)
{
  this->team_ = team;
}

/** @return   game the player is playing
 **/
Game&
Player::game()
{
  DEBUG_ASSERTION(this->game_,
                  "Player::game(): 'game_' is a NULL-pointer\n"
                  "  this = " << this);

  return *this->game_;
} // Game& Player::game()

/** @return   game the player is playing
 **/
Game const&
Player::game() const
{
  DEBUG_ASSERTION(this->game_,
                  "Player::game(): 'game_' is a NULL-pointer\n"
                  "  this = " << this);

  return *this->game_;
} // Game const& Player::game() const

/** set the game
 **
 ** @param    game    new game
 **/
void
Player::set_game(Game& game)
{
  this->game_ = &game;
} // void Player::set_game(Game& game)

/** @return   the number of the player in the game, UINT_MAX if not in a game
 **/
unsigned
Player::no() const
{
  if (!this->game_)
    return UINT_MAX;
  return this->game_->players().no(*this);
} // unsigned Player::no() const

/** @return   name of the player
 **              if it is empty, take a default
 **/
string const
Player::name() const
{
  if (this->name_.empty())
    return Player::default_name(this->type(), ::party->players().no(*this));

  return this->name_;
} // string const Player::name() const

/** sets the name to the default one
 **/
void
Player::set_default_name()
{
  this->set_name(Player::default_name(this->type(), ::party->players().no(*this)));
} // void Player::set_default_name()

/** sets the name
 **
 ** @param    name    new name
 **
 ** @return   name
 **/
void
Player::set_name(string const& name)
{
  this->name_ = name;

  if (::ui) {
    // check that this is no virtual player
    if (   (   !this->game_
            && !(::game_status & GameStatus::game)
            && (::party->players().no(*this) != UINT_MAX) )
        || (   this->game_
            && !this->game_->isvirtual()) ) {
      // || (this->game_->no(*this) != UINT_MAX) )
      ::ui->name_changed(*this);
    }
  }

  return ;
} // void Player::set_name(string name)

/** @return   value of the name of the player
 **/
string const&
Player::name_value() const
{
  return this->name_;
} // string const& Player::name_value() const

/** @return   voice of the player
 **              if it is empty, take a default
 **/
string const
Player::voice() const
{
  if (this->voice_.empty())
    return Player::default_voice(this->type(),
                                 ::party->players().no(*this),
                                 this->name());

  return this->voice_;
} // string const Player::voice() const

/** sets the voice to the default one
 **/
void
Player::set_default_voice()
{
  this->set_voice(Player::default_voice(this->type(),
                                        ::party->players().no(*this),
                                        this->name()));
} // void Player::set_default_voice()

/** sets the voice
 **
 ** @param    voice    new voice
 **/
void
Player::set_voice(string const& voice)
{
  this->voice_ = voice;
  ::ui->voice_changed(*this);
} // void Player::set_voice(string voice)

/** @return   value of the name of the player
 **/
string const&
Player::voice_value() const
{
  return this->voice_;
} // string const& Player::voice_value() const

/** -> result
 **
 ** @param    type       the type
 ** @param    playerno   the number of the player
 **
 ** @return   default name of the type
 **/
string
Player::default_name(Type const type, unsigned const playerno)
{
  switch(type) {
  case Type::human:
    return ::setting(Setting::Type::name);
  default:
    switch (playerno) {
    case 0:
      return "Mara";
    case 1:
      return "Gerd";
    case 2:
      return "Erika";
    case 3:
      return "Sven";
    case 4:
      return "Tina";
    case 5:
      return "Alex";
    case UINT_MAX:
      return "";
    } // switch(playerno)
  } // switch(this->type())

  return ::name(type);
} // string Player::default_name(Type type, unsigned playerno)

/** -> result
 **
 ** @param    type       the type
 ** @param    playerno   the number of the player
 ** @param    name       the name of the player
 **
 ** @return   default voice of the type
 **
 ** @todo      take male/female according to the name
 **/
string
Player::default_voice(Type const type,
                      unsigned const playerno,
                      string const& name)
{
  switch(type) {
  case Type::human:
#ifdef POSTPONED
    // todo: add check, that the voice exists
    return ::setting(Setting::Type::name);
#endif
  default:
    switch (playerno) {
    case 0:
      return "female/mbrola"s;
    case 1:
      return "male/mbrola"s;
    case 2:
      return "female/mbrola"s;
    case 3:
      return "male/mbrola"s;
    case 4:
      return "female/mbrola"s;
    case 5:
      return "man/mbrola"s;
    case UINT_MAX:
      return ""s;
    } // switch(playerno)
  } // switch(this->type())

  return ::name(type);
} // string Player::default_voice(Type type, unsigned playerno, string name)

/** @return   the database
 **/
PlayersDb const&
Player::db() const
{
  DEBUG_ASSERTION(this->db_,
                  "Playe::db():\n"
                  "  'this->db_' == NULL");

  return *this->db_;
} // PlayersDb Player::db() const

/** @return   the database
 **/
PlayersDb&
Player::db()
{
  DEBUG_ASSERTION(this->db_,
                  "Playe::db():\n"
                  "  'this->db_' == NULL");

  return *this->db_;
} // PlayersDb& Player::db()

/** -> result
 **
 ** @param    game_type   gametype to check
 **
 ** @return   whether the gametype is a duty solo for this player
 **/
bool
Player::is_remaining_duty_solo(GameType const game_type) const
{
  if (!is_real_solo(game_type))
    return false;
  if (this->remaining_duty_free_soli())
    return true;
  if (is_picture_solo(game_type)
      && this->remaining_duty_picture_soli())
    return true;
  if (is_color_solo(game_type)
      && this->remaining_duty_color_soli())
    return true;
  return false;
} // bool Player::is_remaining_duty_solo(GameType game_type) const

/** @return   the remaining duty solo games to play
 **/
unsigned
Player::remaining_duty_soli() const
{
  DEBUG_ASSERTION(this->game_,
                  "Player::remaining_duty_soli()\n"
                  "not in game");
  return this->game().party().remaining_duty_soli(*this);
} // unsigned Player::remaining_duty_soli() const

/** @return   the remaining duty solo games to play
 **/
unsigned
Player::remaining_duty_color_soli() const
{
  DEBUG_ASSERTION(this->game_,
                  "Player::remaining_duty_color_soli()\n"
                  "not in game");
  return this->game().party().remaining_duty_color_soli(*this);
} // unsigned Player::remaining_duty_color_soli() const

/** @return   the remaining duty solo games to play
 **/
unsigned
Player::remaining_duty_picture_soli() const
{
  DEBUG_ASSERTION(this->game_,
                  "Player::remaining_duty_picture_soli()\n"
                  "not in game");
  return this->game().party().remaining_duty_picture_soli(*this);
} // unsigned Player::remaining_duty_picture_soli() const

/** @return   the remaining duty solo games to play
 **/
unsigned
Player::remaining_duty_free_soli() const
{
  DEBUG_ASSERTION(this->game_,
                  "Player::remaining_duty_free_soli()\n"
                  "not in game");
  return this->game().party().remaining_duty_free_soli(*this);
} // unsigned Player::remaining_duty_free_soli() const

/** @return   hand
 **/
Hand const&
Player::hand() const
{
  return this->hand_;
} // Hand const& Player::hand() const

/** @return   hand
 **/
Hand&
Player::hand()
{
  return this->hand_;
} // Hand& Player::hand()


/** @return   sorted hand
 **/
SortedHand const&
Player::sorted_hand() const
{
  return this->sorted_hand_;
} // SortedHand const& Player::sorted_hand() const

/** @return   hand
 **/
SortedHand&
Player::sorted_hand()
{
  return this->sorted_hand_;
} // SortedHand& Player::hand()


/** sort the hand
 ** only if the player is in a real game
 **
 ** @return   sorted hand
 **/
Hand const&
Player::hand_sort()
{
  if (::setting(Setting::Type::cards_order).sorted()
      && this->game_
      && !(this->game().isvirtual())
      && (this->no() != UINT_MAX)) {
    bool const changed = this->sorted_hand().sort();
    if (changed)
      ::ui->hand_changed(*this);
  } // if (::setting(Setting::Type::cards_order).sorted())

  return this->hand();
} // Hand const& Player::hand_sort()

/** mixes the hand
 **
 ** @return   mixed hand
 **/
Hand const&
Player::hand_mix()
{
  if (!::setting(Setting::Type::cards_order).sorted()) {
    this->sorted_hand().mix();
    if ((this->game_)
        && (!this->game().isvirtual()))
      ::ui->hand_changed(*this);
  } // if (!::setting(Setting::Type::cards_order).sorted())

  return this->hand();
} // Hand const& Player::hand_mix()


/** set the hand
 **
 ** @param    hand    new hand
 **
 ** @return   player with hand h
 **/
Hand const&
Player::set_hand(Hand const& hand)
{
  this->hand_ = hand;
  this->hand_.set_player(*this);

  if (   this->game_
      && !this->game().isvirtual()) {
    // In virtual games the hand is not shown, so it does not have to be sorted.
    if (::setting(Setting::Type::cards_order).sorted())
      this->hand_sort();
    else
      this->hand_mix();
  }

  if (::game_status == GameStatus::game_init)
    this->set_team(  (this->hand().count_club_queens() > 0)
                   ? Team::re
                   : this->hand().contains(Card::unknown)
                   ? Team::unknown
                   : Team::contra );

  return this->hand();
} // Hand const& Player::set_hand(Hand const& hand)


/** @return   whether the player is re
 **/
bool
Player::isre() const
{
  return (this->team() == Team::re);
} // bool Player::isre() const

/** @return   whether the player has swines
 **/
bool
Player::has_swines() const
{
  if (   this->game().swines().swines_announced()
      && (*this == this->game().swines().swines_owner())) {
    if (   (::game_status == GameStatus::game_poverty_shift)
        && (*this == this->game().players().soloplayer()))
      return true;
    // search a swine in all cards of the hand
    for (unsigned c = 0; c < hand().cardsnumber_all(); c++)
      if (hand().card_all(c).isswine())
        return true;
  }

  return false;
} // bool Player::has_swines() const

/** @return   whether the player has hyperswines
 **/
bool
Player::has_hyperswines() const
{
  if (   this->game().swines().hyperswines_announced()
      && (*this == this->game().swines().hyperswines_owner())) {
    // search a hyperswine in all cards of the hand
    for (unsigned c = 0; c < hand().cardsnumber_all(); c++)
      if (hand().card_all(c).ishyperswine())
        return true;
  }

  return false;
} // bool Player::has_hyperswines() const

/** @return   the reservation of the player
 **/
Reservation const&
Player::reservation() const
{
  return this->reservation_;
} // Reservation const& Player::reservation() const

/** @return   the reservation of the player
 **/
Reservation&
Player::reservation()
{
  return this->reservation_;
} // Reservation& Player::reservation()

/** @return   the trickpile of the player
 **/
TrickPile const&
Player::trickpile() const
{
  return this->trickpile_;
} // TrickPile const& Player::trickpile() const

/** update the teaminfo
 **/
void
Player::teaminfo_update()
{ }

/** @return   announcment of this player (taken from 'game')
 **/
Announcement
Player::announcement() const
{
  return this->game().announcements().last(*this);
}

/** @return   announcement of this player (taken from 'game')
 **/
Announcement
Player::next_announcement() const
{
  if (!::is_real(this->team()))
    return Announcement::noannouncement;

  Announcement const last_announcement
    = (this->game().rule()(Rule::Type::knocking)
       ? this->game().announcements().last(*this).announcement
       : this->game().announcements().last(this->team()).announcement);

  Announcement next_announcement = Announcement::noannouncement;

  switch (last_announcement) {
  case Announcement::noannouncement:
    next_announcement = Announcement::no120;
    break;
  case Announcement::no120:
  case Announcement::no90:
  case Announcement::no60:
  case Announcement::no30:
    next_announcement = next(last_announcement);
    break;
  case Announcement::no0:
  default:
    next_announcement = Announcement::noannouncement;
    break;
  } // switch(last_announcement)

  if (!this->game().announcements().is_valid(next_announcement, *this))
    return Announcement::noannouncement;
  return next_announcement;
} // Announcement Player::next_announcement() const

/** @return   the number of tricks the player has (taken from 'game')
 **/
unsigned
Player::number_of_tricks() const
{
  DEBUG_ASSERTION((this->game_),
                  "Player::number_of_tricks():\n"
                  "  'this->game' == NULL");

  return this->game().tricks().number_of_tricks(*this);
} // unsigned Player::number_of_tricks() const

/** @return   the points of the tricks the player has (taken from 'game')
 **/
unsigned
Player::points_in_trickpile() const
{
  DEBUG_ASSERTION((this->game_),
                  "Player::points_in_trickpile():\n"
                  "  'this->game' == NULL");

  return this->game().points_of_player(*this);
} // unsigned Player::points_in_trickpile() const

/** @return   the number of cards the player still has to play in this game
 **/
unsigned
Player::cards_to_play() const
{
  DEBUG_ASSERTION((this->game_),
                  "Player::cards_to_play():\n"
                  "  'this->game' == NULL");

  if (this->game().tricks().empty())
    return this->game().tricks().max_size();

  if (this->game().tricks().current().cardno_of_player(*this)
      < this->game().tricks().current().actcardno())
    return (this->game().tricks().max_size() - this->game().tricks().real_current_no() - 1);
  else
    return (this->game().tricks().max_size() - this->game().tricks().real_current_no());
} // unsigned Player::cards_to_play() const

/** the rule has changed
 **
 ** @param    type        rule type to have changed
 ** @param    old_value   old value of the rule
 **/
void
Player::rule_changed(int const type, void const* const old_value)
{ return ; }

/** the game 'game' is opened
 **
 ** @param    game    opened game
 **/
void
Player::game_open(Game& game)
{
  this->game_ = &game;
  this->hand_ = Hand();
  this->team_ = Team::unknown;
  this->reservation_ = Reservation();

  return ;
} // void Player::game_open(Game& game)

/** the game 'game' is closed
 **
 ** @param    game    closed game
 **/
void
Player::game_close(Game const& game)
{
  // check whether the player is playing in a game
  if (this->game_ == nullptr)
    return;

  DEBUG_ASSERTION((this->game_ == &game)
                  || (::game_status == GameStatus::party_loaded),
                  "Player::game_close(game):\n"
                  "  wrong game is closed: " << this->game_ << " instead of " << &game);

  this->trickpile_.clear();
  this->game_ = nullptr;
} // void Player::game_close(Game const& game)

/** the player can make an announcement
 **
 ** @return   announcement of the player
 **/
Announcement
Player::announcement_request() const
{
  return Announcement::noannouncement;
} // Announcement Player::announcement_request() const

/** the given player has made the given announcement
 **
 ** @param    announcement    the announcement
 ** @param    player          the player, who has made the announcement
 **/
void
Player::announcement_made(Announcement const& announcement,
                          Player const& player)
{ }

/** 'player' shifts 'cardno' cards
 **
 ** @param    player  the player who shifts the cards
 ** @param    cardno  the number of cards that are shifted
 **/
void
Player::poverty_shift(Player const& player, unsigned const cardno)
{  }

/** 'player' denied to take the shifted cards
 **
 ** @param    player  the player who has denied to take the cards
 **/
void
Player::poverty_take_denied(Player const& player)
{  }

/** all players have denied to take the cards
 **/
void
Player::poverty_take_denied_by_all()
{ }

/** 'player' accepts to take the shifted cards
 ** and returns 'cardno' cards with 'trumpno' trumps
 **
 ** @param    player  the player who has accepted to take the cards
 ** @param    cardno  number of cards that are given back
 ** @param    trumpno number of trumps of the cards
 **/
void
Player::poverty_take_accepted(Player const& player,
                              unsigned const cardno, unsigned const trumpno)
{  }

/** returns which cards the player shifts
 **
 ** @return   the cards that are to be shifted
 **/
HandCards
Player::poverty_shift()
{ 
  cerr << "call of Player::poverty shift()"
    << endl;

  return HandCards();
} // HandCards Player::poverty_shift(Player const& player)

/** returns whether 'player' accepts the shifted cards
 **
 ** @param    cardno  the number of shifted cards
 **
 ** @return   whether to accept the cards
 **/
bool
Player::poverty_take_accept(unsigned const cardno)
{ 
  cerr << "call of Player::poverty_take_accept(cardno)"
    << endl;

  return false;
} // void Player::poverty_take_accept(unsigned const cardno)

/** changes the cards from the poverty-player
 **
 ** @param    cards   the cards that are given to the player
 **
 ** @return   the cards that are returned to the poverty-player
 **/
HandCards
Player::poverty_cards_change(HandCards const& cards)
{ 
  cerr << "call of Player::poverty_cards_change(cards)"
    << endl;

  return HandCards();
} // HandCards Player::poverty_cards_change(HandCards const& cards)

/** gets the cards from the partner and add them to the hand
 **
 ** @param    cards   the cards that are given to the player
 **/
void
Player::poverty_cards_get_back(HandCards const& cards)
{ 
  this->sorted_hand().add(cards);
  this->hand_sort();
} // void Player::poverty_cards_get_back(HandCards const& cards)

/** the marriage partner has found a bride
 ** if the bride is the bridegroom, the bridegroom must play a solo
 **
 ** @param    bridegroom   the player with the marriage
 ** @param    bride      the bride
 **/
void
Player::marriage(Player const& bridegroom, Player const& bride)
{ }

/** genscher: -> result
 **
 ** @return   pointer to the new teammate,
 **           nullptr if the genscher is not to be announced
 **/
Player const*
Player::genscher_partner()
{ 
  cerr << "call of Player::genscher_partner()"
    << endl;

  return {};
} // Player const* Player::genscher_partner()

/** genscher: Information of the new team
 **
 ** @param    player   genscher player
 ** @param    partner   partner of the gensher player
 **/
void
Player::genscher(Player const& player,
                 Player const& partner)
{ }

/** a new game has started - init the player:
 ** empty the hand,
 ** set the teaminfo to the teaminfo of the game
 **
 ** @param    game    the game
 **/
void
Player::new_game(Game& game)
{ 
  this->game_ = &game;
  this->hand_ = Hand();

  return ;
} // void Player::new_game()

/** the player makes his reservation
 **
 ** @param    duty_solo   whether the player must play a duty solo
 **
 ** @return   reservation of the player
 **
 ** @todo   duty_solo documentation
 **/
Reservation const&
Player::reservation_get(bool const duty_solo) 
{ 
  DEBUG_ASSERTION(false,
                  "Player::reservation_get():\n"
                  "  call of this function");

  return (this->reservation() = this->reservation_get_default());
} // Reservation Player::reservation_get(bool duty_solo = false)

/** -> result
 ** that is swines and hyperswines, if the player has them,
 ** poverty (nines, kings, trump), marriage
 **
 ** @return   default reservation of the player
 **/
Reservation
Player::reservation_get_default() const
{ 
  Reservation reservation;
  auto const& rule = this->game().rule();
  auto const& hand = this->hand();

  if (hand.count_club_queens()
      == rule(Rule::Type::number_of_same_cards)) {
    reservation.game_type = GameType::marriage;
    if (rule(MarriageSelector::first_foreign))
      reservation.marriage_selector = MarriageSelector::first_foreign;
    else if (rule(MarriageSelector::first_color))
      reservation.marriage_selector = MarriageSelector::first_color;
    else if (rule(MarriageSelector::first_trump))
      reservation.marriage_selector = MarriageSelector::first_trump;
    else if (rule(MarriageSelector::first_club))
      reservation.marriage_selector = MarriageSelector::first_club;
    else if (rule(MarriageSelector::first_spade))
      reservation.marriage_selector = MarriageSelector::first_spade;
    else if (rule(MarriageSelector::first_heart))
      reservation.marriage_selector = MarriageSelector::first_heart;
    else
      DEBUG_ASSERTION(false,
                      "Player::reservation_get_default():\n"
                      "  no marriage selector valid.\n"
                      "  Ruleset:"
                      << rule);
  } // if (possible marriage)

  reservation.swines
    = (   rule(Rule::Type::swines)
       && rule(Rule::Type::swines_announcement_begin)
       && this->game().swines().swines_announcement_valid(*this));

  reservation.hyperswines
    = (   rule(Rule::Type::hyperswines)
       && rule(Rule::Type::hyperswines_announcement_begin)
       && this->game().swines().hyperswines_announcement_valid(*this));


  if (hand.has_poverty()) {
    reservation.game_type = GameType::poverty;
    if (hand.count_trumps()
        > rule(Rule::Type::max_number_of_poverty_trumps))
      reservation.swines = false;
  }

  if (rule(Rule::Type::throw_when_fox_highest_trump)) {
    unsigned c;
    for (c = 0; c < hand.cardsnumber(); ++c)
      // check with diamond ten because of swines!
      if (Card::diamond_ten.less(hand.card(c))
          && !hand.card(c).istrumpace())
        break;

    if (c == hand.cardsnumber())
      reservation.game_type = GameType::fox_highest_trump;
  } // test fox highest trump

  if (rule(Rule::Type::throw_with_richness)
      && (hand.points()
          >= rule(Rule::Type::min_richness_for_throwing)))
    reservation.game_type = GameType::thrown_richness;

  if (rule(Rule::Type::throw_with_nines_and_kings)
      && (hand.count_nines() + hand.count_kings()
          >= rule(Rule::Type::min_number_of_throwing_nines_and_kings)))
    reservation.game_type = GameType::thrown_nines_and_kings;

  if (rule(Rule::Type::throw_with_kings)
      && (hand.count_kings()
          >= rule(Rule::Type::min_number_of_throwing_kings)))
    reservation.game_type = GameType::thrown_kings;

  if (rule(Rule::Type::throw_with_nines)
      && (hand.count_nines()
          >= rule(Rule::Type::min_number_of_throwing_nines)))
    reservation.game_type = GameType::thrown_nines;

  return reservation;
} // Reservation Player::reservation_get_default() const

/** the game is started
 **/
void
Player::game_start() 
{ 
  // last time for swines announcements

  { // update the team
    if (::is_real(this->game().teaminfo().get(*this))) {
      // all solo games:
      this->set_team(this->game().teaminfo().get(*this));
    } else if (this->game().type() == GameType::marriage) {
      // undetermined marriage
      this->set_team(Team::maybe_contra);

    } else {
      // a normal game
      this->set_team((this->hand().count_club_queens() > 0)
                     ? Team::re
                     : this->hand().contains(Card::unknown)
                     ? Team::unknown
                     : Team::contra);
    }
  } // update the team
} // void Player::game_start()

/** the trick is opened
 **
 ** @param    trick   opened trick
 **/
void
Player::trick_open(Trick const& trick)
{  }

/** the player plays a card
 **
 ** @return   card of the player
 **/
HandCard
Player::card_get()
{ 
  DEBUG_ASSERTION(false,
                  "Player::card_get():\n"
                  "  call of this function");

  return this->hand().card(0);
} // HandCard Player::card_get()

/** 'player' has played the card 'card'
 **
 ** @param    card    played card
 **/
void
Player::card_played(HandCard const& card)
{ }

/** the trick is full
 **
 ** @param    trick   full trick
 **/
void
Player::trick_full(Trick const& trick)
{  }

/** the trick is taken
 **/
void
Player::trick_taken()
{  }

/** move the trick in the trickpile
 **
 ** @param    trick   trick
 **/
void
Player::move_in_trickpile(Trick const& trick)
{ 
  DEBUG_ASSERTION((&trick.game() == this->game_),
                  "Player::move_in_trickpile(trick)"
                  "  trick.game != this->game");

  this->trickpile_.add(trick);
} // void Player::move_in_trickpile(Trick const& trick)

/** check a swines announcement at the game start
 **/
void
Player::check_swines_announcement_at_game_start()
{  }

/** 'player' has announce swines
 **
 ** @param    player   player that has announced the swines
 **/
void
Player::swines_announced(Player const& player)
{  }

/** 'player' has announce hyperswines
 **
 ** @param    player   player that has announced the hyperswines
 **/
void
Player::hyperswines_announced(Player const& player)
{  }

/** evaluation of game for playersDb
 **
 ** @param    game_summary   game summary
 **/
void
Player::evaluate(GameSummary const& game_summary)
{
  this->db().evaluate(*this, game_summary);
} // void Player::evaluate(GameSummary game_summary)

/** -> result
 **
 ** @param    type    the type
 **
 ** @return   name of the type
 **/
string
name(Player::Type const type)
{
  return to_string(type);
}

/** -> result
 **
 ** @param    type    the type
 **
 ** @return   name of the type
 **/
string
to_string(Player::Type const type)
{
  switch(type) {
  case Player::Type::unset:
    return "player type unset";
    (void)_("Player::Type::player type unset");
  case Player::Type::human:
    return "human";
    (void)_("Player::Type::human");
  case Player::Type::ai:
    return "ai";
    (void)_("Player::Type::ai");
  case Player::Type::ai_dummy:
    return "dummy ai";
    (void)_("Player::Type::dummy ai");
  case Player::Type::ai_random:
    return "random ai";
    (void)_("Player::Type::random ai");
  } // switch(type)

  return "";
} // string to_string(Player::Type type)

/** -> result
 **
 ** @param    team   team
 **
 ** @return   translation of the type
 **/
string
gettext(Player::Type const type)
{
  return gettext("Player::Type::" + to_string(type));
}

/** output of the name of 'type'
 **
 ** @param    ostr    output stream
 ** @param    type    player type
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Player::Type const& type)
{
  ostr << name(type);
  return ostr;
} // ostream operator<<(ostream& ostr, Player::Type type);
