/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#pragma once


/**
 ** base class for ratings
 ** A rating takes multiple values and returns a rated value.
 ** An example is the 'max' rater, that returns the maximum of the given values.
 **/
class Rating
{
  public:
    // rating types
    enum class Type {
      max,
      min,
      second_max,
      second_min,
      median,
      average,
      average_physical,
      linear,
      linear_reverse
    };
    constexpr static Type type_list[] = {
      Type::max,
      Type::min,
      Type::second_max,
      Type::second_min,
      Type::median,
      Type::average,
      Type::average_physical,
      Type::linear,
      Type::linear_reverse
    };
    static Type type_from_string(string const& name);

    class Max;
    class Min;
    class SecondMax;
    class SecondMin;
    class Median;
    class Average;
    class AveragePhysical;
    class Linear;
    class LinearReverse;

    // returns a rating of the type 'type'
    static Rating* new_(Type type);
    static unique_ptr<Rating> create(Type type);

  public:
    Rating() = delete;
    explicit Rating(Type type);
    virtual ~Rating() = default;

    // write the rating in the stream
    void write(ostream& ostr) const;

    // adds a value
    void add(int value);
    // returns the rated value
    virtual int value() const = 0;
    // deletes the worst 'n' values
    void delete_worst(size_t n);

  protected:
    // sort the values (changes mutable values vector)
    void sort() const;

  protected:
    // all given values
    mutable vector<int> values;

  private:
    // the rating type
    Type type_;
}; // class Rating

// output of the rating
ostream& operator<<(ostream& ostr, Rating const& rating);

// returns the name of the rating type
string to_string(Rating::Type type);
string gettext(Rating::Type type);
ostream& operator<<(ostream& ostr, Rating::Type type);
