/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "full_trick.h"
#include "trick_summary.h"
#include "ui.h"
#include "table.h"
#include "cards.h"

#include "../../game/game.h"
#include "../../card/trick.h"
#include "../../misc/setting.h"

#include <gtkmm/image.h>
#include <gtkmm/main.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    table   the table
 **/
FullTrick::FullTrick(Table* const table) :
  Base(table),
  Gtk::StickyDialog(_("Window::full trick"), *table->ui->main_window, false)
{
  this->set_gravity(Gdk::GRAVITY_STATIC);

  this->signal_realize().connect(sigc::mem_fun(*this, &FullTrick::init));
} // FullTrick::FullTrick(Table* table)

/** destruktor
 **/
FullTrick::~FullTrick() = default;

/** create all subelements
 **/
void
FullTrick::init()
{
  auto close_button = add_close_button(*this);
  close_button->signal_clicked().connect(sigc::mem_fun(*this, &FullTrick::close));

  { // vbox
    this->get_content_area()->set_spacing(1 EX);

    this->trick_summary = Gtk::manage(new TrickSummary(this, false));

    this->get_content_area()->add(*this->trick_summary);

    this->winnercard = this->ui->cards->new_managed_image();
    this->winnercard->set_padding(2 EM, 2 EX);
    this->trick_summary->pack_start(*this->winnercard);
  } // vbox

  this->show_all_children();
  this->trick_summary->update();

  { // move to the top left corner of the window
    int x, y;
    dynamic_cast<Table*>(this->parent)->get_window()->get_origin(x, y);
    this->move(x, y);
  } // move to the top left conrner of the window
} // void FullTrick::init()

/** show the window
 **
 ** @param    trick   trick to show
 **/
void
FullTrick::show_information(::Trick const& trick)
{
  if (!this->get_realized())
    this->realize();
  auto const& game = trick.game();

  bool show = ::setting(Setting::Type::show_full_trick_window);
  if (!show
      && ::setting(Setting::Type::show_full_trick_window_if_special_points)) {
    // test whether the teams of the player of and for are different
    auto const specialpoints = trick.specialpoints();
    auto const& teaminfo = game.teaminfo();
    for (auto const& sp : specialpoints) {
      if (sp.player_of_no == UINT_MAX) {
        show = true;
        break;
      }
      auto const& player_get = game.player(sp.player_get_no);
      auto const& player_of = game.player(sp.player_of_no);
      auto const teaminfo_get = teaminfo.get_human_teaminfo(player_get);
      auto const teaminfo_of = teaminfo.get_human_teaminfo(player_of);
      if (   !is_real(teaminfo_get)
          || (teaminfo_get != teaminfo_of) ) {
        show = true;
        break;
      } // if (trick shall be shown)
    } // for (sp)
  } // if (!show)

  if (show) {
    this->trick_summary->set_trick(trick);
    this->ui->cards->change_managed(this->winnercard,
                                    trick.winnercard());

    this->present();
  }

  sigc::connection timeout_connection;
  if (::setting(Setting::Type::close_full_trick_automatically))
    timeout_connection
      = Glib::signal_timeout().connect(sigc::slot0<bool>(sigc::mem_fun(*this, &FullTrick::on_timeout)),
                                       ::setting(Setting::Type::full_trick_close_delay));

  this->ui->table->mouse_cursor_update();

  try {
    while (::game_status == GameStatus::game_full_trick)
      ::ui->wait();
  } catch (GameStatus const gs) {
    if (gs == GameStatus::game_trick_taken) {
      ::game_status = GameStatus::game_trick_taken;
    } else {
      this->hide();
      throw;
    }
  }

  if (::setting(Setting::Type::close_full_trick_automatically))
    timeout_connection.disconnect();

  if (show)
    this->hide();

  this->ui->cards->change_managed(this->winnercard, Card());
} // void FullTrick::show_information(::Trick trick)

/** close the full trick
 **/
void
FullTrick::close()
{
  DEBUG_ASSERTION(this->ui->game().tricks().current().isfull(),
                  "FullTrick::close()\n"
                  "  the current trick is not full.");
  DEBUG_ASSERTION(::game_status == GameStatus::game_full_trick,
                  "FullTrick::close()\n"
                  "  the game status is not game full trick but " << ::game_status);

#ifdef WORKAROUND
  // problem: this exception is thrown two times
  // It seems that that one time is from the automatic closing
  // and the other from manual closing.
  if (this->ui->thrower.exception_type() == Thrower::Type::gamestatus)
    return ;
#endif
  this->ui->thrower(GameStatus::game_trick_taken, __FILE__, __LINE__);
} // void FullTrick::close()

/** the name of 'player' has changed
 **
 ** @param    player   the player whose name has changed
 **/
void
FullTrick::name_changed(Player const& player)
{
  if (!this->is_visible())
    return ;

  this->trick_summary->update();
} // void FullTrick::name_changed(Player player)

/** the timeout event:
 ** close the window and exit the main loop
 **
 ** @return   false (to stop the timeout event)
 **/
bool
FullTrick::on_timeout()
{
  this->close();

  return false;
} // bool FullTrick::on_timeout()

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
