/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "last_trick.h"
#include "table.h"
#include "trick_summary.h"

#include "ui.h"
#include "main_window.h"

#include "../../game/game.h"
#include "../../party/rule.h"
#include "../../card/trick.h"
#include "../../player/player.h"

#include <gtkmm/box.h>
#include <gtkmm/button.h>
#include <gtkmm/image.h>
#include <gtkmm/label.h>
#include <gtkmm/main.h>
#include <gdk/gdkkeysyms.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    table   the table
 **/
LastTrick::LastTrick(Table* const table) :
  Base(table),
  Gtk::StickyDialog("FreeDoko – " + _("Window::last trick"))
{
  this->signal_realize().connect(sigc::mem_fun(*this, &LastTrick::init));
} // LastTrick::LastTrick(Table* table)

/** destructor
 **/
LastTrick::~LastTrick() = default;

/** create all subelements
 **/
void
LastTrick::init()
{
  add_close_button(*this);

  this->get_content_area()->set_spacing(1 EX);

  this->navigation_container = Gtk::manage(new Gtk::Box());
  this->previous_trick_button = Gtk::manage(new Gtk::Button());
  this->following_trick_button = Gtk::manage(new Gtk::Button());
  {
    auto previous_image = Gtk::manage(new Gtk::Image());
    previous_image->set_from_icon_name("go-previous", Gtk::ICON_SIZE_BUTTON);
    this->previous_trick_button->add(*previous_image);
  }
  {
    auto next_image = Gtk::manage(new Gtk::Image());
    next_image->set_from_icon_name("go-next", Gtk::ICON_SIZE_BUTTON);
    this->following_trick_button->add(*next_image);
  }
  this->previous_trick_button->set_halign(Gtk::ALIGN_START);
  this->navigation_container->pack_start(*this->previous_trick_button,
                                         Gtk::PACK_SHRINK);
  this->previous_trick_button->set_halign(Gtk::ALIGN_END);
  this->navigation_container->pack_end(*this->following_trick_button,
                                       Gtk::PACK_SHRINK);


  this->previous_trick_button->signal_clicked().connect(sigc::mem_fun(*this, &LastTrick::previous_trick));
  this->following_trick_button->signal_clicked().connect(sigc::mem_fun(*this, &LastTrick::following_trick));

  this->get_content_area()->pack_start(*this->navigation_container, Gtk::PACK_SHRINK);

  this->trick_summary = make_unique<TrickSummary>(this);
  //this->trick_summary->set_trick(*this->last_trick);

  this->get_content_area()->pack_start(*this->trick_summary);

  this->show_all_children();
  this->set_trickno(this->trickno());
  this->trick_summary->update();

  this->signal_key_press_event().connect(sigc::mem_fun(*this,
                                                       &LastTrick::key_press));
} // void LastTrick::init()

/** show the last trick
 **
 ** @param    last_trick   the last trick
 **/
void
LastTrick::show_last_trick(::Trick const& last_trick)
{
  this->last_trick_no = last_trick.no();
  auto const& game = last_trick.game();
  if (game.tricks().tricks_in_trickpiles() == 0)
    return ;

  this->realize();

  this->set_trickno(this->last_trick_no);
  this->present();
} // void LastTrick::show_last_trick(::Trick last_trick)

/** hide the window
 **/
void
LastTrick::close()
{
  this->hide();
} // void LastTrick::close()

/** @return   number of the current trick
 **/
unsigned
LastTrick::trickno() const
{
  return this->last_trick_no;
} // unsigned LastTrick::trickno() const

/** set the trickno
 **
 ** @param    trickno   new trickno
 **/
void
LastTrick::set_trickno(unsigned const trickno)
{
  this->last_trick_no = trickno;
  auto const& game = this->ui->game();
  auto const& trick = game.tricks().trick(trickno);
  auto const tricks_in_trickpiles = game.tricks().tricks_in_trickpiles();
  this->trick_summary->set_trick(trick);
  this->previous_trick_button->set_sensitive(trickno > 0);
  this->following_trick_button->set_sensitive(trickno
                                              < tricks_in_trickpiles - 1);

  if (this->trickno() == tricks_in_trickpiles - 1)
    this->set_title(_("Window::last trick"));
  else
    this->set_title(_("Window::%u. trick", trickno + 1));
  if (tricks_in_trickpiles >= 2)
    this->navigation_container->show();
  else
    this->navigation_container->hide();

#ifdef WINDOWS
#ifdef WORKAROUND
  // Sometimes the window is not updated
  this->queue_draw();
#endif
#endif
} // void LastTrick::set_trickno(unsigned const trickno)

/** show the previous trickno
 **/
void
LastTrick::previous_trick()
{
  if (this->trickno() > 0)
    this->set_trickno(this->trickno() - 1);
} // void LastTrick::previous_trick()

/** show the following trickno
 **/
void
LastTrick::following_trick()
{
  auto const& game = this->ui->game();
  auto const trickno = this->last_trick_no;
  if (trickno < game.tricks().tricks_in_trickpiles() - 1)
    this->set_trickno(trickno + 1);
} // void LastTrick::following_trick()

/** the window is hidden
 **/
void
LastTrick::on_hide()
{
  this->Gtk::StickyDialog::on_hide();
  this->last_trick_no = UINT_MAX;
  this->trick_summary->remove_trick();
} // void LastTrick::on_hide()

/** the name of 'player' has changed
 **
 ** @param    player   the player whose name has changed
 **/
void
LastTrick::name_changed(::Player const& player)
{
  if (this->last_trick_no != UINT_MAX)
    this->trick_summary->update();
} // void LastTrick::name_changed(Player const& player)

/** a key has been pressed
 **
 ** @param    key   the key
 **
 ** @return   from 'ui->key_press(key)'
 **
 ** @bug   signal does not get here for '->', '<-' and '\<Esc\>'
 **/
bool
LastTrick::key_press(GdkEventKey* const key)
{
  // whether the key was accepted
  bool accepted = false;

  if ((key->state & GDK_SHIFT_MASK) == 0) {
    switch (key->keyval) {
    case GDK_KEY_Left: // show previous trick
      this->previous_trick();
      accepted = true;
      break;
    case GDK_KEY_Right: // show following trick
      this->following_trick();
      accepted = true;
      break;
    case GDK_KEY_Escape: // close the window
      this->hide();
      accepted = true;
      break;
    } // switch (key->keyval)
  } // if (key->state == 0)

  if (accepted)
    return false;

  return this->ui->key_press(key);
} // bool LastTrick::key_press(GdkEventKey* key)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
