/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "main_window.h"
#include "menu.h"

#include "ui.h"
#include "cards.h"

#include "first_run.h"
#include "program_updated.h"
#include "bug_report.h"

#include <gtkmm/box.h>
#include <gtkmm/grid.h>
#include <gtkmm/entry.h>
namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent   parent object (the ui)
 **/
MainWindow::MainWindow(Base* const parent) :
  Base(parent),
  Window(),
  menu(std::make_unique<Menu>(this)),
  container(Gtk::manage(new Gtk::HBox()))
{
  this->set_icon(this->ui->icon);

  this->set_minimal_size();

  this->set_title(PROGRAM_name + string("   ")
                  + static_cast<string>(*::version));

  auto box = Gtk::manage(new Gtk::VBox(false, 0));
  this->add(*box);
  box->pack_start(*this->menu, false, true);

  box->add(*this->container);
  this->container->show();
  box->show();

#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->signal_window_state_event().connect(sigc::mem_fun(*this, &MainWindow::on_window_state_event));
  this->signal_key_press_event().connect(sigc::mem_fun(*this, &MainWindow::on_key_press_event));
  this->signal_show().connect(sigc::mem_fun(*this, &MainWindow::on_show));
  this->signal_hide().connect(sigc::mem_fun(*this, &MainWindow::on_hide));
#endif
  this->signal_key_press_event().connect(sigc::mem_fun(*this, &MainWindow::on_key_press_event));

  this->ui->bug_report->set_dnd_destination(*this);
} // MainWindow::MainWindow(Base* const parent)

/** destruktor
 **/
MainWindow::~MainWindow() = default;

/** start a new party
 **/
void
MainWindow::start_new_party_event()
{
  this->ui->start_new_party();
} // void MainWindow::start_new_party_event()

/** end a new party
 **/
void
MainWindow::end_party_event()
{
  this->ui->end_party();
} // void MainWindow::end_party_event()

/** quits the program.
 ** 'UI_GTKMM' cannot be used by the signals sinc it is no Gtk object. So the MainWindow is used instead
 **/
void
MainWindow::quit_program_event()
{
  this->ui->quit_program();
} // void MainWindow::quit_program_event()

/** generates an error
 **/
void
MainWindow::generate_error_event()
{
  //this->ui->generate_error(_("BugReport::generated error"));
  this->ui->error(_("BugReport::generated error"));
} // void MainWindow::generate_error_event()

/** a key has been pressed
 **
 ** @param    key   the key
 **
 ** @return   from 'ui->key_press(key)'
 **/
bool
MainWindow::on_key_press_event(GdkEventKey* const key)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  if (this->Gtk::Window::on_key_press_event(key))
    return true;
#endif
  return this->ui->key_press(key);
} // bool MainWindow::on_key_press_event(GdkEventKey* key)

/** the window is shown -- make sure that the start windows stay on top
 **/
void
MainWindow::on_show()
{
  if (::setting(Setting::Type::ui_main_window_width)) {
    this->set_default_size(::setting(Setting::Type::ui_main_window_width),
                           ::setting(Setting::Type::ui_main_window_height));
    this->move(::setting(Setting::Type::ui_main_window_pos_x),
               ::setting(Setting::Type::ui_main_window_pos_y));
  }
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->Gtk::Window::on_show();
#endif
  if (::setting(Setting::Type::ui_main_window_width)) {
    this->move(::setting(Setting::Type::ui_main_window_pos_x),
               ::setting(Setting::Type::ui_main_window_pos_y));
  }
  if (::setting(Setting::Type::ui_main_window_fullscreen))
    this->fullscreen();

  if (this->ui->first_run_window)
    this->ui->first_run_window->raise();
  if (this->ui->program_updated_window)
    this->ui->program_updated_window->raise();
} // void MainWindow::on_show()

/** the window is hidden -- quit the program
 **/
void
MainWindow::on_hide()
{
  this->ui->quit_program();
} // void MainWindow::on_hide()

/** sets the minimal size of the window (depends on the size of the cards)
 **/
void
MainWindow::set_minimal_size()
{
  Gdk::Geometry geometry;
  geometry.min_width = 5 * this->ui->cards->height();
  geometry.min_height = 4 * this->ui->cards->height();

  this->set_geometry_hints(*this, geometry, Gdk::HINT_MIN_SIZE);
} // void MainWindow::set_minimal_size()


/** a change in the window state (p.e. iconify)
 ** track this event in order to iconify/deiconify all windows with the main
 ** window
 **
 ** @param    state    the (new) state
 **
 ** @return   false (let the Gtk::Window class handle the iconification)
 **/
bool
MainWindow::on_window_state_event(GdkEventWindowState* const state)
{
  if (state->new_window_state & GDK_WINDOW_STATE_ICONIFIED)
    this->ui->iconify_all_windows();
  else
    this->ui->deiconify_all_windows();

  ::setting.set(Setting::Type::ui_main_window_fullscreen,
                (this->get_window()->get_state() & Gdk::WINDOW_STATE_FULLSCREEN));

  this->Gtk::Window::on_window_state_event(state);

  return false;
} // bool MainWindow::on_window_state_event(GdkEventWindowState* state)

/** a change in the window size/position
 ** track this event in order to iconify/deiconify all windows with the main
 ** window
 **
 ** @param    state    the (new) state
 **
 ** @return   false;
 **/
bool
MainWindow::on_configure_event(GdkEventConfigure* const event)
{
  if (!(this->get_window()->get_state() & Gdk::WINDOW_STATE_FULLSCREEN)) {
    // Im Vollbild geben get_size() und get_position() die Werte vom Vollbild-Fenster zurück, daher werden die Werte nicht im Vollbild-Modus gespeichert.
    int width, height;
    int pos_x, pos_y;
    this->get_size(width, height);
    this->get_position(pos_x, pos_y);
    ::setting.set(Setting::Type::ui_main_window_pos_x, pos_x);
    ::setting.set(Setting::Type::ui_main_window_pos_y, pos_y);
    ::setting.set(Setting::Type::ui_main_window_width, width);
    ::setting.set(Setting::Type::ui_main_window_height, height);
  }
  this->Gtk::Window::on_configure_event(event);
  return false;
} // bool MainWindow::on_configure_event(GdkEventConfigure* event)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
