/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "rules.h"
#include "ui.h"
#include "icons.h"

#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../misc/setting.h"
#include "../../utils/string.h"

#include "widgets/labelscalespinbutton.h"
#include "widgets/filemenu.h"
#include <gtkmm/main.h>
#include <gtkmm/label.h>
#include <gtkmm/checkbutton.h>
#include <gtkmm/radiobutton.h>
#include <gtkmm/fixed.h>
#include <gtkmm/box.h>
#include <gtkmm/grid.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/notebook.h>
#include <gtkmm/sizegroup.h>
#include <gtkmm/frame.h>
#include <gtkmm/grid.h>
#include <gtkmm/messagedialog.h>

// additional scaling for the icons of the notebook
#define ICONSCALING 0.7

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent   the parent object
 **/
Rules::Rules(Base* const parent) :
  Base(parent),
  StickyDialog("FreeDoko – " + _("Window::Rules"), false)
{
  this->ui->add_window(*this);

  this->signal_realize().connect(sigc::mem_fun(*this, &Rules::init));
  this->create_backup();
  this->signal_show().connect(sigc::mem_fun(*this, &Rules::create_backup));

#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->signal_key_press_event().connect(sigc::mem_fun(*this, &Rules::on_key_press_event));
#endif
} // Rules::Rules(Base* parent)

/** destructor
 **/
Rules::~Rules()
{
  for (auto widget : this->type_bool)
    delete static_cast<Rule::Type::Bool*>(widget->steal_data("type"));

  for (auto widget : this->type_unsigned)
    delete static_cast<Rule::Type::Unsigned*>(widget->steal_data("type"));
} // Rules::~Rules()

/** creates all subelements
 **
 ** @todo   make the horizontal size always great enough
 ** @todo   visual difference between the groups (two background colours)
 **/
void
Rules::init()
{
  this->set_icon(this->ui->icon);

  this->reset_button = Gtk::manage(new Gtk::Button(_("Button::reset")));
  this->reset_button->set_image_from_icon_name("edit-clear-all");
  this->reset_button->set_always_show_image();

  this->load_button = Gtk::manage(new Gtk::Button(_("Button::load")));
  this->load_button->set_image_from_icon_name("document-open");
  this->load_button->set_always_show_image();

  this->save_button = Gtk::manage(new Gtk::Button(_("Button::save")));
  this->save_button->set_image_from_icon_name("document-save");
  this->save_button->set_always_show_image();

  this->save_file_chooser = make_unique<Gtk::FileChooserDialogWrapper>(_("Save rules"),
                                                                       Gtk::FILE_CHOOSER_ACTION_SAVE,
                                                                       sigc::mem_fun(*this, &Rules::save));

  {
    string dir = (::setting(Setting::Type::private_data_directory)
                  + "/"
                  + ::setting(Setting::Type::rules_directory)
                  + "/"
                 );
#ifdef WINDOWS
    String::replace_all(dir, "/", "\\");
#endif
    this->save_file_chooser->set_current_folder(dir);
  }

  { // create the buttons
    for (int t = Rule::Type::bool_first; t <= Rule::Type::bool_last; t++) {
      this->type_bool.push_back(Gtk::manage(new Gtk::CheckButton(_(Rule::Type::Bool(t)))));
      this->type_bool.back()->set_data("type", new Rule::Type::Bool(Rule::Type::Bool(t)));
      this->type_bool.back()->signal_toggled().connect(sigc::bind<int const>(sigc::mem_fun(*this, &Rules::changed),
                                                                             t)
                                                      );
    } // for (t \in Rule::Type::BOOL)
    for (int t = Rule::Type::unsigned_first; t <= Rule::Type::unsigned_last; t++) {
      switch (t) {
      case Rule::Type::marriage_determination:
      case Rule::Type::announcement_no_120:
      case Rule::Type::announcement_no_90:
      case Rule::Type::announcement_no_60:
      case Rule::Type::announcement_no_30:
      case Rule::Type::announcement_no_0:
      case Rule::Type::announcement_last:
        this->type_unsigned.push_back(Gtk::manage(new Gtk::LabelScaleSpinButton(_(Rule::Type::Unsigned(t)))));
        break;
      default:
        this->type_unsigned.push_back(Gtk::manage(new Gtk::LabelSpinButton(_(Rule::Type::Unsigned(t)))));
        break;
      } // switch(t)
      this->type_unsigned.back()->set_data("type", new Rule::Type::Unsigned(Rule::Type::Unsigned(t)));
      this->type_unsigned.back()->signal_value_changed().connect(sigc::bind<int const>(sigc::mem_fun(*this, &Rules::changed),
                                                                                       t)
                                                                );
    } // for (t \in Rule::Type::UNSIGNED)
    { // counting
      Gtk::RadioButton::Group counting_group;
      for (auto c : {Counting::plus, Counting::minus, Counting::plusminus}) {
        this->counting[c] = Gtk::manage(new Gtk::RadioButton(counting_group, _(c)));
        this->counting[c]->signal_toggled().connect(sigc::bind<int>(sigc::mem_fun(*this, &Rules::changed), static_cast<int>(Rule::Type::counting)));
      }
    } // counting
    { // same size of some widgets
      static auto announcement_scale_size_group
        = Gtk::SizeGroup::create(Gtk::SIZE_GROUP_BOTH);
      announcement_scale_size_group->add_widget(*dynamic_cast<Gtk::LabelScaleSpinButton*>(this->type_unsigned[Rule::Type::announcement_no_120 - Rule::Type::unsigned_first])->get_scale());
      announcement_scale_size_group->add_widget(*dynamic_cast<Gtk::LabelScaleSpinButton*>(this->type_unsigned[Rule::Type::announcement_no_90  - Rule::Type::unsigned_first])->get_scale());
      announcement_scale_size_group->add_widget(*dynamic_cast<Gtk::LabelScaleSpinButton*>(this->type_unsigned[Rule::Type::announcement_no_60  - Rule::Type::unsigned_first])->get_scale());
      announcement_scale_size_group->add_widget(*dynamic_cast<Gtk::LabelScaleSpinButton*>(this->type_unsigned[Rule::Type::announcement_no_30  - Rule::Type::unsigned_first])->get_scale());
      announcement_scale_size_group->add_widget(*dynamic_cast<Gtk::LabelScaleSpinButton*>(this->type_unsigned[Rule::Type::announcement_no_0   - Rule::Type::unsigned_first])->get_scale());
      announcement_scale_size_group->add_widget(*dynamic_cast<Gtk::LabelScaleSpinButton*>(this->type_unsigned[Rule::Type::announcement_last   - Rule::Type::unsigned_first])->get_scale());

    } // same size of some widgets
  } // create the buttons


  { // layout
    add_close_button(*this);

    { // layout for the rules
      this->group_notebook = Gtk::manage(new Gtk::Notebook());

      this->group_notebook->set_tab_pos(Gtk::POS_LEFT);
      //this->group_notebook->set_scrollable(true);
      this->group_notebook->popup_enable();
#ifdef TODO
      this->group_notebook->set_property("homogeneous", true);
#endif
      { // Tournament
        auto tournament_box = this->add_group("tournament");
        (void)_("Rule::Group::tournament");
        tournament_box->set_spacing(3 EX);

        { // with nines
          auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          vbox2->set_halign(Gtk::ALIGN_CENTER);
          tournament_box->add(*vbox2);

          vbox2->add(*this->type_bool[Rule::Type::with_nines - Rule::Type::bool_first]);
        } // with nines
        { // lower box
          auto hbox = Gtk::manage(new Gtk::Box());
          hbox->set_spacing(2 EM);
          { // limits
            auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 2 EX));
            {
              auto vbox3 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
              vbox3->add(*this->type_bool[Rule::Type::number_of_rounds_limited - Rule::Type::bool_first]);
              vbox3->add(*this->type_unsigned[Rule::Type::number_of_rounds - Rule::Type::unsigned_first]);
              vbox2->add(*vbox3);
            }
            {
              auto vbox3 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
              vbox3->add(*this->type_bool[Rule::Type::points_limited - Rule::Type::bool_first]);
              vbox3->add(*this->type_unsigned[Rule::Type::points - Rule::Type::unsigned_first]);
              vbox2->add(*vbox3);
            }
            hbox->add(*vbox2);
          } // limits
          { // duty soli
            auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 2 EX));
            {
              auto vbox3 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
              vbox3->add(*this->type_unsigned[Rule::Type::number_of_duty_soli - Rule::Type::unsigned_first]);
              vbox3->add(*this->type_unsigned[Rule::Type::number_of_duty_color_soli - Rule::Type::unsigned_first]);
              vbox3->add(*this->type_unsigned[Rule::Type::number_of_duty_picture_soli - Rule::Type::unsigned_first]);
              vbox2->add(*vbox3);
            }
            {
              auto vbox3 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
              vbox3->add(*this->type_bool[Rule::Type::offer_duty_solo - Rule::Type::bool_first]);
              vbox2->add(*vbox3);
            }
            hbox->pack_end(*vbox2, Gtk::PACK_SHRINK, 2 EX);
          } // duty soli
          tournament_box->add(*hbox);
        } // lower box
      } // Tournament
      { // counting
        auto vbox = this->add_group("counting");
        (void)_("Rule::Group::counting");
        auto counting_box = Gtk::manage(new Gtk::Box());
        counting_box->set_spacing(4 EM);
        { // counting type
          auto vbox_count = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 2 EX));
          {
            auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL));

            auto label = Gtk::manage(new Gtk::Label(_("Label::counting")));
            label->set_halign(Gtk::ALIGN_START);
            vbox->add(*label);

            for (auto widget : this->counting)
              vbox->add(*widget.second);

            vbox_count->add(*vbox);
          }
          {
            this->type_bool[Rule::Type::solo_always_counts_triple - Rule::Type::bool_first]->set_margin_top(1 EX / 2);
            vbox_count->add(*this->type_bool[Rule::Type::solo_always_counts_triple - Rule::Type::bool_first]);
          }
          counting_box->add(*vbox_count);
        } // counting type
        { // Bock
          auto bock_box = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL));
          bock_box->set_spacing(2 EX);
          counting_box->add(*bock_box);

          bock_box->add(*this->type_bool[Rule::Type::bock - Rule::Type::bool_first]);
          bock_box->add(*this->type_bool[Rule::Type::bock_append - Rule::Type::bool_first]);
          { // conditions
            auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
            vbox->set_homogeneous();
            bock_box->add(*vbox);

            vbox->add(*this->type_bool[Rule::Type::bock_120 - Rule::Type::bool_first]);
            vbox->add(*this->type_bool[Rule::Type::bock_solo_lost - Rule::Type::bool_first]);
            vbox->add(*this->type_bool[Rule::Type::bock_re_lost - Rule::Type::bool_first]);
            vbox->add(*this->type_bool[Rule::Type::bock_heart_trick - Rule::Type::bool_first]);
            vbox->add(*this->type_bool[Rule::Type::bock_black - Rule::Type::bool_first]);
          } // conditions
        } // Bock
        counting_box->set_halign(Gtk::ALIGN_CENTER);
        counting_box->set_valign(Gtk::ALIGN_CENTER);
        vbox->add(*counting_box);
      } // counting
      { // Solo
        auto solo_box = this->add_group("solo",
                                        this->ui->icons->new_managed_image(Icons::Type::solo_koehler, ICONSCALING));
        (void)_("Rule::Group::solo");
        solo_box->set_spacing(2 EX);

        solo_box->add(*this->type_bool[Rule::Type::solo - Rule::Type::bool_first]);
        { // solo types
          auto grid = Gtk::manage(new Gtk::Grid());
          grid->set_column_spacing(1 EM);
          grid->attach(*this->type_bool[Rule::Type::solo_color - Rule::Type::bool_first],
                       0, 0, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_koehler - Rule::Type::bool_first],
                       0, 1, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_meatless - Rule::Type::bool_first],
                       0, 2, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_jack - Rule::Type::bool_first],
                       1, 0, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_queen - Rule::Type::bool_first],
                       1, 1, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_king - Rule::Type::bool_first],
                       1, 2, 1, 1);

          grid->attach(*this->type_bool[Rule::Type::solo_queen_jack - Rule::Type::bool_first],
                       2, 0, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_king_jack - Rule::Type::bool_first],
                       2, 1, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::solo_king_queen - Rule::Type::bool_first],
                       2, 2, 1, 1);
          solo_box->add(*grid);
        } // solo types
        solo_box->add(*this->type_bool[Rule::Type::throwing_before_solo - Rule::Type::bool_first]);
        solo_box->add(*this->type_bool[Rule::Type::lustsolo_player_leads - Rule::Type::bool_first]);
      } // Solo
      { // Poverty
        auto poverty_box
          = this->add_group("poverty",
                            this->ui->icons->new_managed_image(Icons::Type::poverty, ICONSCALING));
        (void)_("Rule::Group::poverty");

        {
          auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          vbox->add(*this->type_bool[Rule::Type::poverty - Rule::Type::bool_first]);
          vbox->add(*this->type_bool[Rule::Type::poverty_shift - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::poverty_shift_only_trump - Rule::Type::bool_first]->set_margin_left(1 EM);
          vbox->add(*this->type_bool[Rule::Type::poverty_shift_only_trump - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::poverty_fox_do_not_count - Rule::Type::bool_first]->set_margin_left(1 EM);
          vbox->add(*this->type_bool[Rule::Type::poverty_fox_do_not_count - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::poverty_fox_shift_extra - Rule::Type::bool_first]->set_margin_left(2 EM);
          vbox->add(*this->type_bool[Rule::Type::poverty_fox_shift_extra - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::poverty_fox_shift_open - Rule::Type::bool_first]->set_margin_left(1 EM);
          vbox->add(*this->type_bool[Rule::Type::poverty_fox_shift_open - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::throw_with_one_trump - Rule::Type::bool_first]->set_margin_left(1 EM);
          vbox->add(*this->type_bool[Rule::Type::throw_with_one_trump - Rule::Type::bool_first]);
          poverty_box->add(*vbox);
        }
        {
          auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          vbox->add(*this->type_bool[Rule::Type::throw_when_fox_highest_trump - Rule::Type::bool_first]);
          vbox->add(*this->type_bool[Rule::Type::throw_with_nines - Rule::Type::bool_first]);
          vbox->add(*this->type_unsigned[Rule::Type::min_number_of_throwing_nines - Rule::Type::unsigned_first]);
          vbox->add(*this->type_bool[Rule::Type::throw_with_kings - Rule::Type::bool_first]);
          vbox->add(*this->type_unsigned[Rule::Type::min_number_of_throwing_kings - Rule::Type::unsigned_first]);
          vbox->add(*this->type_bool[Rule::Type::throw_with_nines_and_kings - Rule::Type::bool_first]);
          vbox->add(*this->type_unsigned[Rule::Type::min_number_of_throwing_nines_and_kings - Rule::Type::unsigned_first]);
          vbox->add(*this->type_bool[Rule::Type::throw_with_richness - Rule::Type::bool_first]);
          vbox->add(*this->type_unsigned[Rule::Type::min_richness_for_throwing - Rule::Type::unsigned_first]);
          poverty_box->add(*vbox);
        }
        {
          auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          vbox->add(*this->type_bool[Rule::Type::force_redistribute - Rule::Type::bool_first]);
          poverty_box->add(*vbox);
        }
      } // Poverty
      { // Marriage
        auto marriage_box = this->add_group("marriage",
                                            this->ui->icons->new_managed_image(Icons::Type::marriage, ICONSCALING));
        (void)_("Rule::Group::marriage");

        { // marriage
          marriage_box->add(*this->type_bool[Rule::Type::marriage_before_poverty - Rule::Type::bool_first]);
          {
            auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
            vbox2->add(*this->type_bool[Rule::Type::marriage_first_foreign - Rule::Type::bool_first]);
            vbox2->add(*this->type_bool[Rule::Type::marriage_first_color - Rule::Type::bool_first]);
            vbox2->add(*this->type_bool[Rule::Type::marriage_first_single_color - Rule::Type::bool_first]);
            vbox2->add(*this->type_bool[Rule::Type::marriage_first_trump - Rule::Type::bool_first]);
            marriage_box->add(*vbox2);
          }
          marriage_box->add(*this->type_bool[Rule::Type::marriage_first_one_decides - Rule::Type::bool_first]);
          marriage_box->add(*this->type_unsigned[Rule::Type::marriage_determination - Rule::Type::unsigned_first]);
        } // marriage
        { // genscher
          marriage_box->add(*this->type_bool[Rule::Type::genscher - Rule::Type::bool_first]);
        } // genscher
      } // Marriage
      { // Announcements
        auto announcements_box
          = this->add_group("announcements",
                            this->ui->icons->new_managed_image(Icons::Type::no_120, ICONSCALING));
        (void)_("Rule::Group::announcements");
        {
          auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          vbox->pack_start(*this->type_bool[Rule::Type::knocking - Rule::Type::bool_first]);
          announcements_box->add(*vbox);
        }

        {
          auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          auto label = Gtk::manage(new Gtk::Label(_("Label::announcement with maximal cards")));
          vbox->add(*label);

          vbox->pack_start(*this->type_unsigned[Rule::Type::announcement_no_120 - Rule::Type::unsigned_first]);
          vbox->pack_start(*this->type_unsigned[Rule::Type::announcement_no_90 - Rule::Type::unsigned_first]);
          vbox->pack_start(*this->type_unsigned[Rule::Type::announcement_no_60 - Rule::Type::unsigned_first]);
          vbox->pack_start(*this->type_unsigned[Rule::Type::announcement_no_30 - Rule::Type::unsigned_first]);
          vbox->pack_start(*this->type_unsigned[Rule::Type::announcement_no_0 - Rule::Type::unsigned_first]);
          vbox->pack_start(*this->type_unsigned[Rule::Type::announcement_last - Rule::Type::unsigned_first]);
          announcements_box->add(*vbox);
        }
        {
          auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));

          vbox->pack_start(*this->type_bool[Rule::Type::announcement_limit_only_for_current - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_till_full_trick - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_reply - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_strict_limit - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_re_doubles - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_contra_doubles_against_re - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_first_trick_thirty_points - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_first_trick_thirty_points_only_first - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_first_trick_thirty_points_only_re_contra - Rule::Type::bool_first]);
          vbox->pack_start(*this->type_bool[Rule::Type::announcement_first_trick_thirty_points_in_marriage - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::announcement_contra_doubles_against_re - Rule::Type::bool_first]->set_margin_left(1 EM);
          this->type_bool[Rule::Type::announcement_first_trick_thirty_points - Rule::Type::bool_first];
          this->type_bool[Rule::Type::announcement_first_trick_thirty_points_only_first - Rule::Type::bool_first]->set_margin_left(1 EM);
          this->type_bool[Rule::Type::announcement_first_trick_thirty_points_only_re_contra - Rule::Type::bool_first]->set_margin_left(1 EM);
          this->type_bool[Rule::Type::announcement_first_trick_thirty_points_in_marriage - Rule::Type::bool_first]->set_margin_left(1 EM);
          announcements_box->add(*vbox);
        }
      } // Announcements
      { // Dollen
        auto vbox = this->add_group("dollen",
                                    this->ui->icons->new_managed_image(Icons::Type::dollen, ICONSCALING));
        (void)_("Rule::Group::dollen");
        vbox->set_spacing(1 EX / 2);

        {
          vbox->add(*this->type_bool[Rule::Type::dollen - Rule::Type::bool_first]);
          vbox->add(*this->type_bool[Rule::Type::dollen_second_over_first - Rule::Type::bool_first]);
          vbox->add(*this->type_bool[Rule::Type::dollen_contrary_in_last_trick - Rule::Type::bool_first]);
          vbox->add(*this->type_bool[Rule::Type::dollen_first_over_second_with_swines - Rule::Type::bool_first]);
          this->type_bool[Rule::Type::dollen_second_over_first - Rule::Type::bool_first]->set_margin_left(1 EM);
          this->type_bool[Rule::Type::dollen_contrary_in_last_trick - Rule::Type::bool_first]->set_margin_left(2 EM);
          this->type_bool[Rule::Type::dollen_first_over_second_with_swines - Rule::Type::bool_first]->set_margin_left(2 EM);
        }
      } // Dollen
      { // Swines
        auto grid = Gtk::manage(new Gtk::Grid());
        this->add_group("swines",
                        this->ui->icons->new_managed_image(Icons::Type::swines,
                                                           ICONSCALING)
                       )->add(*grid);
        (void)_("Rule::Group::swines");

        grid->set_row_homogeneous(true);
        grid->set_column_homogeneous(true);
        grid->set_column_spacing(2 EM);
        grid->set_row_spacing(1 EX / 2);

        {
          grid->attach(*this->type_bool[Rule::Type::swines - Rule::Type::bool_first],
                       0, 0, 1, 1);
          this->type_bool[Rule::Type::swines_in_solo - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::swines_in_solo - Rule::Type::bool_first],
                       0, 1, 1, 1);
          this->type_bool[Rule::Type::swines_in_poverty - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::swines_in_poverty - Rule::Type::bool_first],
                       0, 2, 1, 1);
          this->type_bool[Rule::Type::swines_announcement_begin - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::swines_announcement_begin - Rule::Type::bool_first],
                       0, 3, 1, 1);
          this->type_bool[Rule::Type::swine_only_second - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::swine_only_second - Rule::Type::bool_first],
                       0, 4, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::hyperswines - Rule::Type::bool_first],
                       1, 0, 1, 1);
          this->type_bool[Rule::Type::hyperswines_in_solo - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::hyperswines_in_solo - Rule::Type::bool_first],
                       1, 1, 1, 1);
          this->type_bool[Rule::Type::hyperswines_in_poverty - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::hyperswines_in_poverty - Rule::Type::bool_first],
                       1, 2, 1, 1);
          this->type_bool[Rule::Type::hyperswines_announcement_begin - Rule::Type::bool_first]->set_margin_left(1 EM);
          grid->attach(*this->type_bool[Rule::Type::hyperswines_announcement_begin - Rule::Type::bool_first],
                       1, 3, 1, 1);
          grid->attach(*this->type_bool[Rule::Type::swines_and_hyperswines_joint - Rule::Type::bool_first],
                       0, 5, 2, 1);
        }
      } // Swines
      { // Extrapoints
        auto vbox = this->add_group("extrapoints",
                                    this->ui->icons->new_managed_image(Icons::Type::doppelkopf, ICONSCALING));
        (void)_("Rule::Group::extrapoints");

        {
          auto hbox = Gtk::manage(new Gtk::Box());
          hbox->set_spacing(1 EM);
          hbox->set_spacing(2 EM);
          vbox->add(*hbox);
          {
            auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_catch_fox - Rule::Type::bool_first]);
            this->type_bool[Rule::Type::extrapoint_catch_fox_last_trick - Rule::Type::bool_first]->set_margin_left(1 EM);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_catch_fox_last_trick - Rule::Type::bool_first]);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_fox_last_trick - Rule::Type::bool_first]);
            this->type_bool[Rule::Type::extrapoint_double_fox_last_trick - Rule::Type::bool_first]->set_margin_left(1 EM);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_double_fox_last_trick - Rule::Type::bool_first]);
            hbox->add(*vbox2);
          }
          {
            auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_charlie - Rule::Type::bool_first]);
            this->type_bool[Rule::Type::extrapoint_catch_charlie - Rule::Type::bool_first]->set_margin_left(1 EM);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_catch_charlie - Rule::Type::bool_first]);
            this->type_bool[Rule::Type::extrapoint_double_charlie - Rule::Type::bool_first]->set_margin_left(1 EM);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_double_charlie - Rule::Type::bool_first]);
            this->type_bool[Rule::Type::extrapoint_catch_double_charlie - Rule::Type::bool_first]->set_margin_left(1 EM);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_catch_double_charlie - Rule::Type::bool_first]);
            this->type_bool[Rule::Type::extrapoint_catch_charlie_only_with_diamond_queen - Rule::Type::bool_first]->set_margin_left(1 EM);
            vbox2->add(*this->type_bool[Rule::Type::extrapoint_catch_charlie_only_with_diamond_queen - Rule::Type::bool_first]);
            hbox->add(*vbox2);
          }
        }
        {
          auto vbox2 = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX / 2));
          vbox2->add(*this->type_bool[Rule::Type::extrapoint_dolle_jabs_dolle - Rule::Type::bool_first]);
          vbox2->add(*this->type_bool[Rule::Type::extrapoint_heart_trick - Rule::Type::bool_first]);
          vbox->add(*vbox2);
        }
      } // Extrapoints
    } // layout for the rules
    this->get_content_area()->pack_start(*this->group_notebook);


    { // put the buttons in the container
      auto file_box = Gtk::manage(new Gtk::ButtonBox());
      file_box->set_border_width(1 EX / 2);
      file_box->set_spacing(2 EM);
      file_box->set_halign(Gtk::ALIGN_CENTER);
      file_box->set_homogeneous();

      file_box->add(*this->reset_button);
      file_box->add(*this->load_button);
      file_box->add(*this->save_button);

      this->get_content_area()->pack_end(*file_box, Gtk::PACK_SHRINK, 1 EX);
    } // put the buttons in the container

  } // layout

#ifndef RELEASE
#ifdef DKNOF
  { // test, whether all rules-buttons are packed in a container
    for (int t = Rule::Type::bool_first; t <= Rule::Type::bool_last; t++) {
      if (t == Rule::Type::mutate)
        continue;
      if (!this->type_bool[t - Rule::Type::bool_first]->get_parent())
        cerr << "GTKMM_DOKO::Rules::Rules():\n"
          << "  rule '" << Rule::Type::Bool(t) << "' not packed\n";
    }
    for (int t = Rule::Type::unsigned_first; t <= Rule::Type::unsigned_last; t++) {
      if (   (t == Rule::Type::number_of_players)
          || (t == Rule::Type::number_of_players_in_game)
          || (t == Rule::Type::bock_multiplier) )
        continue;
      if (!this->type_unsigned[t - Rule::Type::unsigned_first]->get_parent())
        cerr << "GTKMM_DOKO::Rules::Rules():\n"
          << "  rule '" << Rule::Type::Unsigned(t) << "' not packed\n";
    }
  } // test, whether all rule-buttons are packed in a container
#endif
#endif

  { // signals
    this->reset_button->signal_clicked().connect(sigc::mem_fun(*this, &Rules::reset));

    this->load_button->signal_clicked().connect(sigc::mem_fun(*this, &Rules::load));
    this->save_button->signal_clicked().connect(sigc::mem_fun0(*this->save_file_chooser, &Gtk::FileChooserDialogWrapper::present));
  } // signals

  this->update_all();

  this->show_all_children();

  this->ui->party().signal_start().connect_back(*this, &Rules::sensitivity_update, this->disconnector_);
} // void Rules::init()

/** creates a backup
 **/
void
Rules::create_backup()
{
#ifdef WORKAROUND
  this->backup_ = std::make_unique<Rule>(::party->rule());
#else
  this->backup_ = std::make_unique<Rule>(this->ui->party().rule());
#endif

  this->sensitivity_update();
} // void Rules::create_backup()

/** @return   the backup
 **/
Rule const&
Rules::backup() const
{
  DEBUG_ASSERTION(this->backup_,
                  "Rule::Type::backup():\n"
                  "  'this->backup_' == nullptr");

  return *this->backup_;
} // Rule const& Rules::backup() const

/** resets the rules to the backup
 **/
void
Rules::reset()
{
  ::party->rule() = this->backup();
  this->update_all();
} // void Rules::reset()

/** creates a menu with the files in the rules directories
 **/
void
Rules::load()
{
  auto load_menu
    = Gtk::manage(new Gtk::FileMenu(sigc::mem_fun(*this, &Rules::load_file)));
  load_menu->add_directory(::setting(Setting::Type::private_data_directory)
                           + "/"
                           + ::setting(Setting::Type::rules_directory));
  load_menu->add_directory(::setting(Setting::Type::public_data_directory)
                           + "/"
                           + ::setting(Setting::Type::rules_directory));

  load_menu->show();
} // void Rules::load()

/** loads the rules
 **
 ** @param    filename   the rules file
 **/
void
Rules::load_file(string const filename)
{
  ::party->rule().load(filename);
  this->update_all();
} // void Rules::load_file(string filename)

/** save the rules.
 **
 ** @param    filename  path to save the rules
 **/
bool
Rules::save(string const filename)
{
  return this->ui->party().rule().save(filename);
} // bool Rules::save(string filename)

/** update the sensitivity of all widgets
 **/
void
Rules::sensitivity_update()
{
  if (!this->get_realized())
    return ;

#ifdef RELEASE
  bool sensitive = (::game_status == GameStatus::party_new);
#else
  bool sensitive = true;
#endif

  for (int i = 0; i < this->group_notebook->get_n_pages(); ++i)
    this->group_notebook->get_nth_page(i)->set_sensitive(sensitive);

  if (sensitive) {
    for (auto widget : this->type_bool)
      widget->set_sensitive(this->ui->party().rule().dependencies(*static_cast<Rule::Type::Bool*>(widget->get_data("type"))));
    for (auto widget : this->type_unsigned) {
      auto const type
        = *static_cast<Rule::Type::Unsigned*>(widget->get_data("type"));

      widget->set_sensitive(this->ui->party().rule().dependencies(type));
      widget->set_range(this->ui->party().rule().min(type),
                        this->ui->party().rule().max(type));
    } // for (auto widget : this->type_unsigned)

    for (auto widget : this->counting)
      widget.second->set_sensitive(true);

    this->load_button->set_sensitive(true);
    this->reset_button->set_sensitive(this->backup()
                                      != this->ui->party().rule());

  } else { // if (sensitive)
    for (auto widget : this->type_bool)
      widget->set_sensitive(false);
    for (auto widget : this->type_unsigned)
      widget->set_sensitive(false);
    for (auto widget : this->counting)
      widget.second->set_sensitive(false);

    this->load_button->set_sensitive(false);
    this->reset_button->set_sensitive(false);
  } // if (sensitive)
} // void Rules::sensitivity_update()

/** update the rule 'type'
 **
 ** @param    type                 the type of the rule
 ** @param    update_sensitivity   whether the sensitivity shall be updated
 **                (default: true)
 **/
void
Rules::update(int const type, bool const update_sensitivity)
{
  if (!this->get_realized())
    return ;

  if (update_sensitivity)
    this->sensitivity_update();

  Rule const& rule = this->ui->party().rule();

  if ((type >= Rule::Type::bool_first)
      && (type <= Rule::Type::bool_last)) {
    this->type_bool[type - Rule::Type::bool_first]->set_active(rule.value(Rule::Type::Bool(type)));
  } else if ((type >= Rule::Type::unsigned_first)
             && (type <= Rule::Type::unsigned_last)) {
    this->type_unsigned[type - Rule::Type::unsigned_first]->set_range(rule.min(Rule::Type::Unsigned(type)),
                                                                rule.max(Rule::Type::Unsigned(type)));
    this->type_unsigned[type - Rule::Type::unsigned_first]->set_value(rule.value(Rule::Type::Unsigned(type)));
  } else if (type == Rule::Type::counting) {
    this->counting[rule(Rule::Type::counting)]->set_active(true);
  } else  {
    DEBUG_ASSERTION(false,
                    "Rules::update(type, update_sensitivity):\n"
                    "  type '" << type << "' unknown.");
  }

  if (type == Rule::Type::knocking)
    this->update_announcement_text();

  if (update_sensitivity)
    this->reset_button->set_sensitive(this->backup() != rule);
} // void Rules::update(int type, bool update_sensitivity = true)

/** a rule has been changed by the user
 **
 ** @param    type   the type of the rule
 **/
void
Rules::changed(int const type)
{
  auto& rule = ::party->rule();

  if ( (type == Rule::Type::with_nines)
      && (this->type_bool[type - Rule::Type::bool_first]->get_active()
          != rule(Rule::Type::with_nines))
      && (this->without_nines_info_shown == false)
     ) {
    // show an information window for rule problems
    // @bug    memory bug: dialog is not destroyed (but is only created at most once)
    auto information
      = new Gtk::MessageDialog(!rule(Rule::Type::with_nines)
                               ? _("Rule::Information::with nines")
                               : _("Rule::Information::without nines"),
                               false,
                               Gtk::MESSAGE_INFO,
                               Gtk::BUTTONS_NONE,
                               false);
    information->set_icon(this->ui->icon);
    information->set_title(!rule(Rule::Type::with_nines)
                           ? _("Rule::Information::Title::with nines")
                           : _("Rule::Information::Title::without nines"));

    { // adjust rules button
      auto adjust_rules_button
        = Gtk::manage(new Gtk::Button(_("Button::adjust rules")));
      adjust_rules_button->show_all();

      adjust_rules_button->signal_clicked().connect(sigc::mem_fun(*this,
                                                                  !rule(Rule::Type::with_nines)
                                                                  ? &Rules::adjust_to_with_nines
                                                                  : &Rules::adjust_to_without_nines));
      adjust_rules_button->signal_clicked().connect(sigc::mem_fun(*information,
                                                                  &Gtk::Widget::hide)
                                                   );
    } // adjust rules button

    information->set_transient_for(*this);

    information->present();

    this->without_nines_info_shown = true;
  } // if ('with nines' changed)

  if ((type >= Rule::Type::bool_first)
      && (type <= Rule::Type::bool_last)) {
    rule.set(Rule::Type::Bool(type),
             this->type_bool[type - Rule::Type::bool_first]->get_active());
  } else if ((type >= Rule::Type::unsigned_first)
             && (type <= Rule::Type::unsigned_last)) {
    if (this->type_unsigned[type - Rule::Type::unsigned_first]->get_realized())
      rule.set(Rule::Type::Unsigned(type),
               this->type_unsigned[type - Rule::Type::unsigned_first]->get_value_as_int());
  } else if (type == Rule::Type::counting) {
    for (auto& c : this->counting)
      if (c.second->get_active())
        rule.set(Rule::Type::counting, c.first);
  } else
    DEBUG_ASSERTION(false,
                    "Rules::changed(type):\n"
                    "  type '" << type << "' unknown.");
} // void Rules::changed(int type)

/** updates all widgets
 **/
void
Rules::update_all()
{
  if (!this->get_realized())
    return ;

  this->sensitivity_update();
  for (int t = Rule::Type::bool_first; t <= Rule::Type::bool_last; t++)
    this->update(t, false);
  for (int t = Rule::Type::unsigned_first; t <= Rule::Type::unsigned_last; t++)
    this->update(t, false);
  this->update(Rule::Type::counting, false);

  this->reset_button->set_sensitive(this->backup() != this->ui->party().rule());
} // void Rules::update_all()

/** adjust the rules to the game with nines
 **/
void
Rules::adjust_to_with_nines()
{
  if (!this->get_realized())
    return ;

  auto& rule = ::party->rule();

  rule.set(Rule::Type::announcement_no_120, 11);
  rule.set(Rule::Type::announcement_no_90, 9);
  rule.set(Rule::Type::announcement_no_60, 6);
  rule.set(Rule::Type::announcement_no_30, 3);
  rule.set(Rule::Type::announcement_no_0, 3);
} // void Rules::adjust_to_with_nines()

/** adjust the rules to the game without nines
 **/
void
Rules::adjust_to_without_nines()
{
  if (!this->get_realized())
    return ;

  auto& rule = ::party->rule();

  rule.set(Rule::Type::announcement_no_120, 9);
  rule.set(Rule::Type::announcement_no_90, 7);
  rule.set(Rule::Type::announcement_no_60, 5);
  rule.set(Rule::Type::announcement_no_30, 3);
  rule.set(Rule::Type::announcement_no_0, 3);
} // void Rules::adjust_to_without_nines()

/** adds to the group notebook a page for group 'name'.
 **
 ** @param    name   name of the group
 ** @param    image   image before the label
 **         (default: nullptr)
 **
 ** @return   created vbox for the rules of the group
 **/
Gtk::Box*
Rules::add_group(string const& name, Gtk::Image* image)
{
  auto box = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 2 EX));
  box->set_border_width(1 EM);
  box->set_halign(Gtk::ALIGN_CENTER);
  box->set_valign(Gtk::ALIGN_CENTER);

  auto bookmark = Gtk::manage(new Gtk::Box());
  bookmark->set_spacing(1 EM / 2);
  if (image)
    bookmark->add(*image);
  if (!name.empty()) {
    auto label = Gtk::manage(new Gtk::Label(_("Rule::Group::" + name)));
    bookmark->add(*label);
  } // if (name)
  bookmark->show_all_children();

  auto menu_label = Gtk::manage(new Gtk::Label(_("Rule::Group::" + name)));
  this->group_notebook->append_page(*box, *bookmark, *menu_label);

  return box;
} // Gtk::Box* Rules::add_group(string name, Gtk::Image* image = nullptr)

/** a key has been pressed
 ** C-o: output of the rules on 'stdout'
 **
 ** @param    key   the key
 **
 ** @return   whether the key was managed
 **/
bool
Rules::on_key_press_event(GdkEventKey* const key)
{
  bool managed = false;

  if ((key->state & ~GDK_SHIFT_MASK) == GDK_CONTROL_MASK) {
    switch (key->keyval) {
    case GDK_KEY_o: // ouput of the rules
      cout << this->ui->party().rule() << endl;
      managed = true;
      break;
    case GDK_KEY_u: // update the rules
      this->update_all();
      managed = true;
      break;
    } // switch (key->keyval)
  } // if (key->state == GDK_CONTROL_MASK)

  return (managed
          || this->StickyDialog::on_key_press_event(key)
          || this->ui->key_press(key));
} // bool Rules::on_key_press_event(GdkEventKey* key)


/** Update the announcement texts (with or without knocking)
 **/
void
Rules::update_announcement_text()
{
  if (!this->get_realized())
    return ;

#ifdef OUTDATED
  // old translation system
  this->ui->translations->change(*this->type_unsigned[Rule::Type::announcement_no_120 - Rule::Type::unsigned_first]->get_label(),
                                 ::translation(Rule::Type::announcement_no_120,
                                               this->ui->party().rule()(Rule::Type::knocking) ? "knocking" : ""));
  this->ui->translations->change(*this->type_bool[Rule::Type::announcement_re_doubles - Rule::Type::bool_first],
                                 ::translation(Rule::Type::announcement_re_doubles,
                                               this->ui->party().rule()(Rule::Type::knocking) ? "knocking" : ""));

  this->ui->translations->change(*this->type_bool[Rule::Type::announcement_contra_doubles_against_re - Rule::Type::bool_first],
                                 ::translation(Rule::Type::announcement_contra_doubles_against_re,
                                               this->ui->party().rule()(Rule::Type::knocking) ? "knocking" : ""));

  this->ui->translations->change(*this->type_bool[Rule::Type::announcement_first_trick_thirty_points_only_re_contra - Rule::Type::bool_first],
                                 ::translation(Rule::Type::announcement_first_trick_thirty_points_only_re_contra,
                                               this->ui->party().rule()(Rule::Type::knocking) ? "knocking" : ""));
#endif
} // void Rules::update_announcement_text()

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
