/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "table.h"
#include "reservation.h"
#include "poverty.h"
#include "game_finished.h"
#include "game_review.h"
#include "party_points.h"
#include "party_finished.h"
#include "full_trick.h"
#include "last_trick.h"

#include "trick.h"
#include "hand.h"
#include "trickpile.h"
#include "icongroup.h"
#include "name.h"

#include "ui.h"
#include "main_window.h"
#include "card_suggestion.h"
#include "genscher_selection.h"
#include "gameinfo_dialog.h"
#include "gameinfos.h"

#include "../../party/party.h"
#include "../../game/game.h"
#include "../../player/player.h"
#include "../../player/ai/ai.h"
#include "../../card/hand.h"
#include "../../card/trick.h"
#include "../../os/bug_report_replay.h"
#include "../../utils/file.h"

#include <gtkmm/drawingarea.h>
#include <gtkmm/main.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent   parent object
 **/
Table::Table(Base* const parent) :
  Base(parent),
  party_points_(make_unique<PartyPoints>(this)),
  gameinfo_(make_unique<GameInfoDialog>(this)),
  surface_(Cairo::ImageSurface::create(Cairo::FORMAT_ARGB32, 1, 1))
{
  this->ui->main_window->container->add(*this);

  this->show();

  for (auto p : vector<Position>({Position::south,
                                 Position::west,
                                 Position::north,
                                 Position::east})) {
    this->hand_[p] = std::make_unique<Hand>(this, p);
    this->trickpile_[p] = std::make_unique<TrickPile>(this, p);
    this->icongroup_[p] = std::make_unique<Icongroup>(this, p);
    this->name_[p] = std::make_unique<Name>(this, p);
    this->reservation_[p] = std::make_unique<Reservation>(this, p);
  }
  this->trick_ = std::make_unique<Trick>(this);
  this->poverty_ = std::make_unique<Poverty>(this);

  for (auto& t : this->trickpile_)
    this->part_.push_back(t.second.get());
  for (auto& n : this->name_)
    this->part_.push_back(n.second.get());
  for (auto& i : this->icongroup_)
    this->part_.push_back(i.second.get());
  for (auto& h : this->hand_)
    this->part_.push_back(h.second.get());
  this->part_.push_back(this->trick_.get());
  this->part_.push_back(this->poverty_.get());

  this->add_events(Gdk::BUTTON_PRESS_MASK);
  this->add_events(Gdk::SCROLL_MASK);
  this->add_events(Gdk::POINTER_MOTION_MASK);

  this->load_background();

#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
  this->signal_expose_event().connect(sigc::mem_fun(*this, &Table::on_expose_event));
  this->signal_configure_event().connect(sigc::mem_fun(*this, &Table::on_configure_event));
  this->signal_button_press_event().connect(sigc::mem_fun(*this, &Table::on_button_press_event));
  this->signal_event().connect(sigc::mem_fun(*this, &Table::on_event));
  this->signal_motion_notify_event().connect(sigc::mem_fun(*this, &Table::on_motion_notify_event));
#endif

  Party::signal_open().connect_back([this](Party& party){ this->draw_all(); }, this->disconnector_);
  ::party->signal_finish().connect_back(*this, &Table::party_finish, this->disconnector_);
} // Table::Table(Base* parent)

/** destruktor
 **/
Table::~Table() = default;

/** show the party finished window
 **/
void
Table::party_finish()
{
  if (!this->party_finished_) {
    this->party_finished_ = std::make_unique<PartyFinished>(this);
  }

  this->draw_all();

  if (!(::fast_play & FastPlay::party_finished))
    this->party_finished_->party_finished();
} // void Table::party_finish()

/** a gameplay action has happened
 **
 ** @param    action   gameplay action
 **/
void
Table::gameplay_action(GameplayAction const& action)
{
  if (this->gameinfo_)
    this->gameinfo_->update_announcement_reply();
} // void Table::gameplay_action(GameplayAction action)

/** create the htin-objects
 **/
void
Table::game_open()
{
} // void Table::game_open()

/** the cards are distributed in the game
 **
 ** @todo     distribute in 3er packs (animation)
 **/
void
Table::game_cards_distributed()
{
  this->force_redraw_all();
} // void Table::game_cards_distributed()

/** the game is redistributed
 **/
void
Table::game_redistribute()
{
  this->draw_all();
} // void Table::game_redistribute()

/** ask 'player' for a reservation
 **
 ** @param    player   player who is asked
 **/
void
Table::reservation_ask(Player const& player)
{
  // redraw the names so that the current player is highlighted
  this->draw_all();

  // if this is the first reservation to ask, show the reservation window for the next human player
  if (!(   player.game().party().is_duty_soli_round()
        && !player.game().rule()(Rule::Type::offer_duty_solo))) {
    for (Player const* p = &player.game().players().following(player);
         *p != player;
         p = &player.game().players().following(*p)) {
      if (p->type() == Player::Type::human) {
        if (player == player.game().startplayer())
          this->reservation(*p).show_for_selection();
        else
          this->reservation(*p).update_for_reservation();
        break;
      } // if (p->type() == Player::Type::human)
    } // for (p)

  } // if (player is startplayer and not human
} // void Table::reservation_ask(Player player)

/** gets the reservation of the given player:
 **
 ** @param    player   player, whose reservation schould be get
 **
 ** @return   reservation of the player
 **/
::Reservation
Table::reservation_get(Player const& player)
{
  this->draw_all();
  return this->reservation_[this->position(player)]->get();
} // ::Reservation Table::reservation_get(player)

/** the game starts
 **/
void
Table::game_start()
{
  this->draw_all();
  this->party_points_->game_start();

  for (auto& r : this->reservation_)
    r.second->hide();

  // if we have a solo player show the info window
  if (this->ui->game().type() != GameType::normal) {

    if (   !(::fast_play & FastPlay::gameplay_information)
        && !(   ::bug_report_replay
             && ::bug_report_replay->auto_action())
        && ::setting(Setting::Type::show_gametype_window))
      this->gameinfo_->show(new GameInfoDialog::Information::GameType(this, this->ui->game()),
                            ::setting(Setting::Type::close_gametype_window_automatically),
                            ::setting(Setting::Type::gametype_window_close_delay));
  } // if (this->ui->game().type() != GameType::normal)

  this->draw_all();
} // void Table::game_start()

/** the player has to play a card
 **
 ** @param    player   the active player
 **/
void
Table::card_get(Player const& player)
{
  if (   this->card_suggestion_
      && this->card_suggestion_->is_visible())
    this->show_card_suggestion();

  this->get_card_ = true;
  this->mouse_cursor_update();
  this->hand(player).force_redraw();
  this->draw_all();
} // void Table::card_get(Player player)

/** the player has played a card
 **/
void
Table::card_got()
{
  this->get_card_ = false;
} // void Table::card_got()

/** the player has played a card
 **
 ** @param    player   the active player
 **/
void
Table::card_played(Player const& player)
{
  if (this->card_suggestion_)
    this->card_suggestion_->card_played();

  this->draw_all();

  if (::debug("screenshots")) {
    if (!this->ui->game().tricks().current().isfull()) {
      this->save_screenshot();
    }
  }
} // void Table::card_played(Player player)

/** the trick is full
 **/
void
Table::trick_open()
{
  this->draw_all();
  if (::debug("screenshots")) {
    this->save_screenshot();
  }
}

/** the trick is full
 **/
void
Table::trick_full()
{
  this->draw_all();
  if (   !(::fast_play & FastPlay::gameplay_information)
      && !(   ::bug_report_replay
           && ::bug_report_replay->auto_action())) {
    if (!this->full_trick_)
      this->full_trick_ = std::make_unique<FullTrick>(this);
    this->full_trick_->show_information(this->ui->game().tricks().current());
  }
  this->draw_all();
  if (::debug("screenshots")) {
    this->save_screenshot();
  }
} // void Table::trick_full()

/** move the trick in the trickpile
 **
 ** @todo   animation
 **/
void
Table::trick_move_in_pile()
{
  this->draw_all();
} // void Table::trick_move_in_pile()

/** show the last trick
 **/
void
Table::show_last_trick()
{
  if (!(   (::game_status >= GameStatus::game_play)
        && (::game_status < GameStatus::game_finished)))
    return ;

  Game const& game = this->ui->game();

  if (game.tricks().real_current_no() == 0)
    return ;

  this->show_trick(game.tricks().trick(game.tricks().real_current_no() - 1));
} // void Table::show_last_trick()

/** show the given trick
 **
 ** @param    trick   trick to show
 **/
void
Table::show_trick(::Trick const& trick)
{
  if (!this->last_trick_)
    this->last_trick_ = std::make_unique<LastTrick>(this);

  if (this->last_trick_->is_visible()) {
    this->last_trick_->set_trickno(trick.no());
    this->last_trick_->present();
    return ;
  } // if (this->last_trick_->is_visible())

  this->last_trick_->show_last_trick(trick);

  this->mouse_cursor_update();
} // void Table::show_trick(::Trick trick)

/** show the card suggestion
 **
 ** @param    show_window   whether to show the window
 **/
void
Table::show_card_suggestion(bool const show_window)
{
  DEBUG_ASSERTION((this->ui->game().players().current_player().type()
                   == Player::Type::human),
                  "Table::show_card_suggestion():\n"
                  "  current player is not human.");

  if (!this->card_suggestion_)
    this->card_suggestion_ = std::make_unique<CardSuggestion>(this);

  this->card_suggestion_->show_information(show_window);
} // void Table::show_card_suggestion(bool show_window)

/** show the game finished window
 **/
void
Table::game_finished()
{
  if (!this->game_finished_)
    this->game_finished_ = std::make_unique<GameFinished>(this);

  this->force_redraw_all();

  this->party_points_->game_finished();

  if (!(::fast_play & FastPlay::game_finished))
    this->game_finished_->game_finished();
} // void Table::game_finished()

/** destroy the htin-objects
 **/
void
Table::game_close()
{
  if (this->last_trick_)
    this->last_trick_->close();

  if (this->card_suggestion_)
    this->card_suggestion_->card_played();

  if (this->gameinfo_)
    this->gameinfo_->close();
} // void Table::game_close()

/** @return   whether the game is reviewed
 **/
bool
Table::in_game_review() const
{
  return (   (::game_status == GameStatus::game_finished)
          && (this->game_finished_)
          && (this->game_finished_->game_review));
} // bool Table::in_game_review() const

/** @return   current tricknumber in the game review
 **/
unsigned
Table::game_review_trickno() const
{
  DEBUG_ASSERTION(this->in_game_review(),
                  "Table::game_review_trickno():\n"
                  "  not in game review");
  return this->game_finished_->game_review->trickno();
} // unsigned Table::game_review_trickno() const

/** @return   current trick in the game review
 **/
::Trick const&
Table::game_review_trick() const
{
  DEBUG_ASSERTION(this->in_game_review(),
                  "Table::game_review_trickno():\n"
                  "  not in game review");
  return this->game_finished_->game_review->trick();
} // ::Trick const& Table::game_review_trick() const

/** @return   whether a trick is visible in the game review
 **/
bool
Table::game_review_trick_visible() const
{
  DEBUG_ASSERTION(this->in_game_review(),
                  "Table::game_review_trickno():\n"
                  "  not in game review");
  return this->game_finished_->game_review->trick_visible();
} // bool Table::game_review_trick_visible() const

/** 'player' has announced 'announcement'
 **
 ** @param    announcement   the announcement that has been made
 ** @param    player      the player that has announced swines
 **/
void
Table::announcement_made(Announcement const announcement,
                         Player const& player)
{
  this->force_redraw_all();

  if (   !(::fast_play & FastPlay::gameplay_information)
      && !(   ::bug_report_replay
           && ::bug_report_replay->auto_action())
      && ::setting(Setting::Type::show_announcement_window)) {
    this->gameinfo_->show(new GameInfoDialog::Information::Announcement(this, player, announcement),
                          ::setting(Setting::Type::close_announcement_window_automatically),
                          ::setting(Setting::Type::announcement_window_close_delay));
  }
} // void Table::announcement_made(Announcement const announcement, Player const& player)

/** swines have been announced
 **
 ** @param    player   the player that has announced swines
 **/
void
Table::swines_announced(Player const& player)
{
  this->force_redraw_all();

  if (player.type() != Player::Type::human) {

    if (   !(::fast_play & FastPlay::gameplay_information)
        && !(   ::bug_report_replay
             && ::bug_report_replay->auto_action())
        && ::setting(Setting::Type::show_swines_window))
      this->gameinfo_->show(new GameInfoDialog::Information::Swines(this, player),
                            ::setting(Setting::Type::close_swines_window_automatically),
                            ::setting(Setting::Type::swines_window_close_delay));

  } // if (player.type() != Player::Type::human)
} // void Table::swines_announced(Player const& player)

/** hyperswines have been announced
 **
 ** @param    player   the player that has announced hyperswines
 **/
void
Table::hyperswines_announced(Player const& player)
{
  this->force_redraw_all();

  if (player.type() != Player::Type::human) {

    if (   !(::fast_play & FastPlay::gameplay_information)
        && !(   ::bug_report_replay
             && ::bug_report_replay->auto_action())
        && ::setting(Setting::Type::show_swines_window)) {
      this->gameinfo_->show(new GameInfoDialog::Information::Hyperswines(this, player),
                            ::setting(Setting::Type::close_swines_window_automatically),
                            ::setting(Setting::Type::swines_window_close_delay));
    } // if (!(::fast_play & FastPlay::gameplay_information))
  } // if (player.type() != Player::Type::human)
} // void Table::hyperswines_announced(Player const& player)

/** marriage: Information of the new team
 ** if the bride is the bridegroom, the bridegroom must play a solo
 **
 ** @param    bridegroom   the player with the marriage
 ** @param    bride      the bride
 **/
void
Table::marriage(Player const& bridegroom,
                Player const& bride)
{
  this->force_redraw_all();

  if (   !(::fast_play & FastPlay::gameplay_information)
      && !(   ::bug_report_replay
           && ::bug_report_replay->auto_action())
      && ::setting(Setting::Type::show_marriage_window)) {
    this->gameinfo_->show(new GameInfoDialog::Information::Marriage(this, bridegroom, bride),
                          ::setting(Setting::Type::close_marriage_window_automatically),
                          ::setting(Setting::Type::marriage_window_close_delay));
  }
} // void Table::marriage(Player const& bridegroom, Player const& bride)

/** select a partner
 **
 ** @param    genscher   player with the genscher
 **
 ** @return   pointer to the new teammate,
 **           nullptr if the genscher is not to be announced
 **/
Player const*
Table::select_genscher_partner(Player const& genscher)
{
  if (!this->genscher_selection_)
    this->genscher_selection_ = std::make_unique<GenscherSelection>(this);

  return this->genscher_selection_->select_partner(genscher);
} // Player const* Table::select_genscher_partner(Player genscher)

/** genscher information
 **
 ** @param    genscher   player with the genscher
 ** @param    partner    new partner
 **/
void
Table::genscher(Player const& genscher,
                Player const& partner)
{
  this->force_redraw_all();

  if (genscher.type() != Player::Type::human) {

    if (   !(::fast_play & FastPlay::gameplay_information)
        && !(   ::bug_report_replay
             && ::bug_report_replay->auto_action())
        && ::setting(Setting::Type::show_genscher_window)) {
      this->gameinfo_->show(new GameInfoDialog::Information::Genscher(this, genscher, partner),
                            ::setting(Setting::Type::close_genscher_window_automatically),
                            ::setting(Setting::Type::genscher_window_close_delay));
    }
  } // if (genscher.type() != Player::Type::human)
} // void Table::genscher(Player const& genscher, Player const& partner)

/** save the table as screenshot
 **/
void
Table::save_screenshot() const
{
  auto pixbuf = Gdk::Pixbuf::create(this->surface_, 0, 0, this->width(), this->height());
  string filename;
  if (::game_status & GameStatus::game) {
    auto const& game = this->ui->game();
    if (   ::game_status >= GameStatus::game_play
        && ::game_status < GameStatus::game_finished) {
      auto const& trick = game.tricks().current();
      auto const trickno = game.tricks().real_current_no() + 1;
      filename = (File::desktop_directory()
                  + "/FreeDoko_Screenshot_"
                  + std::to_string(game.seed()) + "_"
                  + (trickno < 10 ? "0" : "")
                  + std::to_string(trickno)
                  + "-"
                  + std::to_string(trick.actcardno())
                  + ".png");
    } else if (::game_status == GameStatus::game_finished) {
      if (this->in_game_review()) {
      auto const trickno = this->game_review_trickno() + 1;
      filename = (File::desktop_directory()
                  + "/FreeDoko_Screenshot_"
                  + std::to_string(game.seed()) + "_"
                  + (trickno < 10 ? "0" : "")
                  + std::to_string(trickno)
                  + ".png");
      } else {
        filename = (File::desktop_directory()
                    + "/FreeDoko_Screenshot_"
                    + std::to_string(game.seed()) + "_"
                    + "finished"
                    + ".png");
      }

    } else {
      filename = (File::desktop_directory()
                  + "/FreeDoko_Screenshot_"
                  + std::to_string(game.seed())
                  + ".png");
    }
  } else {
    filename = (File::desktop_directory()
                + "/FreeDoko_Screenshot.png");
  }
  pixbuf->save(filename, "png");
}

/** event: mouse click
 **
 ** @param    button_event   the event of the mouse clock
 **
 ** @return   whether there was a reaction
 **/
bool
Table::on_button_press_event(GdkEventButton* button_event)
{
  int const x = static_cast<int>(button_event->x);
  int const y = static_cast<int>(button_event->y);

  if (this->in_game_review()) {
    switch (button_event->button) {
    case 3:
    case 4:
      this->game_finished_->game_review->previous_trick();
      return true;
    case 2:
      this->game_finished_->game_review->set_trickno(0);
      return true;
    case 1:
    case 5:
      this->game_finished_->game_review->following_trick();
      return true;
    default:
      break;
    }
  } // if (this->in_game_review)

  // shifting cards in a poverty
  if (::game_status == GameStatus::game_poverty_shift)
    if (this->poverty().button_press_event(button_event)) {
      this->mouse_cursor_update();
      return true;
    }

  { // test, whether the last trick shall be shown
    if (::in_running_game()) {
      Game const& game = this->ui->game();
      for (auto const& p : game.players()) {
        auto const trick = this->trickpile(p).over_trick(x, y);
        if (trick) {
          switch (button_event->button) {
          case 1: // left mouse button
            this->show_trick(*trick);
            this->mouse_cursor_update();
            return true;
          default:
            break;
          } // switch (button_event->button)
        } // if (trick)
      } // for (p \in game.player)
    } // if (::game_status in game)
  } // test, whether the last trick shall be shown

  { // test whether an announcement shall be made
    if (button_event->button == 1) {
      if (::in_running_game()) {
        for (auto& p : this->ui->game().players()) {
          if (   (p.type() == Player::Type::human)
              && this->icongroup(p).mouse_over_next_announcement()) {
            this->ui->game().announcements().make_announcement(next(p.game().announcements().last(p.team())),
                                                               p);
            this->mouse_cursor_update();
            return true;
          }
        } // for (p)
      } // if (::in_running_game())
    } // if (button_event->button == 3)
  } // test whether an announcement shall be made

  { // test whether the full trick shall be closed
    if (button_event->button == 1) {
      if (   ::in_running_game()
          && (::game_status == GameStatus::game_full_trick)
          && this->ui->game().tricks().current().isfull()
          && this->full_trick_
          && this->trick().mouse_over_cards()) {
        this->full_trick_->close();
        return true;
      } // if (::in_running_game() && trick full && ...)
    } // if (button_event->button == 1)
  } // test whether the full trick shall be closed

  { // close the gameinfo dialogs
    if (button_event->button == 3) {
      this->gameinfo_->close();
      if (   this->full_trick_
          && (::game_status == GameStatus::game_full_trick)
          && this->ui->game().tricks().current().isfull()) {
        this->full_trick_->close();
      } else if (this->last_trick_) {
        this->last_trick_->hide();
      } else if (this->card_suggestion_) {
        this->card_suggestion_->hide();
      }
    } // if (button_event->button == 3)
  } // close the gameinfo dialogs

  { // test, whether a card shall be played
    if (this->get_card()) {
      Player& player = this->ui->game().players().current_player();
      if (player.type() == Player::Type::human) {
        auto const cardno = this->hand(player).cardno_at_position(x, y);

        if (cardno) {
          switch (button_event->button) {
          case 1: // left mouse button
            player.sorted_hand().request_position_all(*cardno);
            if (   !this->card_suggestion_
                || (player.sorted_hand().card_all(*cardno)
                    != this->card_suggestion_->suggested_card) )
              dynamic_cast<Ai&>(player).set_last_heuristic_to_manual();
            // ToDo: only do this, if the card is in the hand
            //       (for a bug report replay)
            this->ui->thrower(player.sorted_hand().requested_card(),
                              __FILE__, __LINE__);

            this->mouse_cursor_update();
            return true;
          case 3: // right mouse button
            if (this->card_suggestion_)
              this->card_suggestion_->play_card();
            this->mouse_cursor_update();
            return true;
          default:
            break;
          } // switch (button_event->button)
        } // if (cardno)
      } // if (player is human)
    } // if (::game_status == GameStatus::game_play)
  } // test, whether a card shall be played


  switch (button_event->button) {
  case 1: // left mouse button
    DEBUG("mouse") << "mouse: " << x << '+' << y << '\n';
  } // switch (button_event->button)

#ifdef WINDOWS
#ifdef WORKAROUND
  // Under MS-Windows the information windows can get in the background of the main window, although they should stay in the front.
  // So, when a mouse click is made on the table which is not otherwise interpreted, the information windows are put in the foreground.
  { // put a gameinfo dialog, ... in the foreground
    if (this->game_finished_
        && this->game_finished_->is_visible())
      this->game_finished_->raise();
    if (this->party_finished_
        && this->party_finished_->is_visible())
      this->party_finished_->raise();
    if (this->card_suggestion_
        && this->card_suggestion_->is_visible())
      this->card_suggestion_->raise();
    if (this->gameinfo_
        && this->gameinfo_->is_visible())
      this->gameinfo_->raise();
    if (this->full_trick_
        && this->full_trick_->is_visible())
      this->full_trick_->raise();
    if (this->last_trick_
        && this->last_trick_->is_visible())
      this->last_trick_->raise();
    if (this->genscher_selection_
        && this->genscher_selection_->is_visible())
      this->genscher_selection_->raise();
  } // put gameinfo dialog, ... in the foreground
#endif
#endif

  this->mouse_cursor_update();
  return false;
} // bool Table::on_button_press_event(GdkEventButton* event)

/** event: mouse scroll
 **
 ** @param    event   the event of the scrolling
 **
 ** @return   whether there was a reaction
 **/
bool
Table::on_scroll_event(GdkEventScroll* const event)
{
  if (this->in_game_review()) {
    switch (event->direction) {
    case GDK_SCROLL_UP:
      this->game_finished_->game_review->previous_trick();
      break;
    case GDK_SCROLL_DOWN:
      this->game_finished_->game_review->following_trick();
      break;
    default:
      break;
    } // switch (event->direction)
  } // if (this->in_game_review)

  return true;
} // bool Table::on_scroll_event(GdkEventScroll* event)

/** conversion into a string
 **
 ** @param    layout   layout
 **
 ** @return   string of the layout
 **/
string
to_string(Table::Layout const layout)
{
  switch (layout) {
  case Table::Layout::standard:
    return "standard"s;
  case Table::Layout::widescreen:
    return "widescreen"s;
  } // switch (layout)
  return {};
} // string to_string(Table::Layout layout)

/** write the layout name into the output stream
 **
 ** @param    ostr     output stream
 ** @param    layout   layout
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, Table::Layout const layout)
{
  ostr << to_string(layout);
  return ostr;
}

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
