/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "trick.h"
#include "table.h"
#include "hand.h"
#include "name.h"
#include "icongroup.h"
#include "trickpile.h"
#include "ui.h"
#include "cards.h"

#include "../../party/party.h"
#include "../../game/game.h"
#include "../../player/player.h"

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    table   corresponding table
 **/
Trick::Trick(Table* const table) :
  Table::Part(table),
  TrickDrawing(table)
{ }

/** destructor
 **/
Trick::~Trick() = default;

/**
 **/
bool
Trick::changed() const
{
  if (!this->surface_)
    return true;
  if (::game_status >= GameStatus::game_play)
    return (this->trick_drawed_ != this->game().tricks().current());
  return true;
} // bool Trick::changed() const

/** @return   the geometry of the trick
 **/
Trick::Geometry
Trick::geometry() const
{
  Geometry geometry;
  geometry.width = [=](){ return static_cast<int>(this->TrickDrawing::width()); };
  geometry.height = [=](){ return static_cast<int>(this->TrickDrawing::height()); };
  switch (this->table().layout()) {
  case Table::Layout::standard:
    geometry.pos_x = [=](){
      return static_cast<int>(this->table().width() / 2);
    };
    geometry.pos_y = [=](){
      return static_cast<int>(this->table().height() / 2);
    };
    break;
  case Table::Layout::widescreen:
    auto hand_north = this->table().hand(Position::north).geometry();
    auto name_north = this->table().name(Position::north).geometry();
    auto hand_south = this->table().hand(Position::south).geometry();
    auto hand_east = this->table().hand(Position::east).geometry();
    auto icongroup_south = this->table().icongroup(Position::south).geometry();
    auto trickpile_east = this->table().trickpile(Position::east).geometry();
    geometry.pos_x = [=](){
      return (hand_north.pos_x() + hand_north.margin_x()
                 + (trickpile_east.pos_x() - trickpile_east.height())) / 2;
      return max(hand_north.pos_x() + hand_north.margin_x()
                 + geometry.width() / 2,
                 max(trickpile_east.pos_x() - trickpile_east.height()
                     - hand_east.margin_y()
                     - geometry.width() / 2,
                     hand_south.pos_x() + hand_south.width() / 2));
    };
    geometry.pos_y = [=](){
      auto const y_min = (name_north.pos_y()
                          + hand_north.margin_y());
      auto const y_max = (icongroup_south.pos_y()
                          - icongroup_south.height()
                          - hand_south.margin_y());
      return (y_max - hand_south.margin_y() - geometry.height() / 2);
      return (y_max + y_min) / 2;
    };
    break;
  } // switch (this->table().layout())
  return geometry;
} // Geometry Trick::geometry() const

/** @return   the outline of the trick
 **/
Trick::Outline
Trick::outline() const
{
  return this->geometry().outline();
}

/** @return   the outline of the trick
 **/
Trick::Outline
Trick::Geometry::outline() const
{
  return this->transform({-this->width() / 2,
                         -this->height() / 2,
                         this->width(),
                         this->height()});
} // Outline Trick::Geometry::outline() const

/** draws the trick
 **
 ** @param    cr       drawing context
 **/
void
Trick::draw(Cairo::RefPtr<::Cairo::Context> cr)
{
  if (!(   (::game_status >= GameStatus::game_play)
        && (   (::game_status < GameStatus::game_finished)
            || (this->table().in_game_review()) ) ) )
    return ;

  if (this->table().in_game_review()) {
    if (this->table().game_review_trick_visible())
      this->set_trick(this->table().game_review_trick());
    else
      this->remove_trick();
  } else {
    this->set_trick(this->TrickDrawing::ui->game().tricks().current());
  }

  if (!this->trick_set()) {
    this->trick_drawed_ = {};
    return;
  }

  auto const& trick = this->trick();

  // only show the trick if it is in the middle
  // (and not in the trickpile)
  if (    !trick.isfull()
       || trick.winnerplayer().trickpile().empty()
       || (trick.winnerplayer().trickpile().last_trick().no()
           < trick.no())
       || this->table().in_game_review()) {
    cr->push_group();
    this->geometry().transform(cr);
    this->set_center({0, 0});
    this->TrickDrawing::draw(cr);
    cr->pop_group_to_source();
    cr->paint();
  }

  this->trick_drawed_ = trick;
} // void Trick::draw(Cairo::RefPtr<::Cairo::Context> cr)

/** @return   whether the mouse is over the trick
 **/
bool
Trick::mouse_over_cards() const
{
  if (!this->trick_set())
    return false;

  int x, y;
  this->table().get_pointer(x, y);
  auto const geometry = this->geometry();
  return this->TrickDrawing::mouse_over_cards(x - geometry.pos_x(),
                                              y - geometry.pos_y());
} // bool Trick::mouse_over_cards() const

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
