/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Trick
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "trick_summary.h"

#include "trick_drawing.h"
#include "ui.h"
#include "table.h"

#include "../../game/game.h"
#include "../../player/player.h"
#include "../../card/trick.h"

#include <gtkmm/drawingarea.h>
#include <gtkmm/label.h>
#include <gtkmm/treeview.h>

namespace UI_GTKMM_NS {

/** constructor
 **
 ** @param    parent               parent widget
 ** @param    show_trick_picture   whether to show the trick picture
 **/
TrickSummary::TrickSummary(Base* const parent,
                           bool const show_trick_picture) :
  Base{parent},
  Box(),
  show_trick_picture{show_trick_picture},
  trick_drawing{this}
{
  this->signal_realize().connect(sigc::mem_fun(*this,
                                               &TrickSummary::init));
} // TrickSummary::TrickSummary(Base* parent, bool show_trick_picture = true)

/** destruktor
 **/
TrickSummary::~TrickSummary() = default;

/** create all subelements
 **/
void
TrickSummary::init()
{
  this->set_spacing(10);

  if (show_trick_picture) {
    if (this->trick)
      this->trick_drawing.set_trick(*this->trick);

    this->trick_picture = Gtk::manage(new Gtk::DrawingArea());
    this->trick_picture->set_size_request(this->trick_drawing.width()
                                          * 6 / 5,
                                          this->trick_drawing.height()
                                          * 6 / 5);
    this->trick_picture->signal_draw().connect(sigc::mem_fun(*this,
                                                             &TrickSummary::on_draw_trick));

    this->pack_start(*this->trick_picture, false, true);
  } // if (show_trick_picture)

  { // info box
    auto info_box = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL));
    info_box->set_homogeneous();
    info_box->set_border_width(2 EX);
    this->pack_end(*info_box);

    { // the labels
      auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX));
      vbox->set_border_width(2 EX);

      // gettext: %s = winner
      this->winner = Gtk::manage(new Gtk::Label(_("The trick goes to %s.",
                                                  this->trick->winnerplayer().name())));
      vbox->pack_start(*this->winner, false, true);

      this->points = Gtk::manage(new Gtk::Label(_("Points in the trick: %u",
                                                  this->trick->points())));

      vbox->pack_start(*this->points, false, true);

      info_box->pack_start(*vbox, false, true);
    } // the labels

    { // specialpoints
      auto vbox = Gtk::manage(new Gtk::Box(Gtk::ORIENTATION_VERTICAL, 1 EX));
      vbox->set_border_width(2 EX);

      this->specialpoints_list
        = Gtk::ListStore::create(this->specialpoints_model);
      this->specialpoints_treeview
        = Gtk::manage(new Gtk::TreeView(this->specialpoints_list));
      this->specialpoints_treeview->get_selection()->set_mode(Gtk::SELECTION_NONE);

      this->specialpoints_treeview->append_column(_("specialpoint"),
                                                  this->specialpoints_model.special_point);
      this->specialpoints_treeview->append_column(_("player"),
                                                  this->specialpoints_model.player);

      vbox->pack_start(*this->specialpoints_treeview, false, true);

      info_box->pack_end(*vbox, false, true);
    } // specialpoints
  } // info box

  this->show_all();
  this->update();
} // void TrickSummary::init()

/** set the trick
 **
 ** @param    trick   the trick to show the info of
 **/
void
TrickSummary::set_trick(::Trick const& trick)
{
  this->trick = &trick;
  this->trick_drawing.set_trick(trick);
  this->update();
} // void TrickSummary::set_trick(::Trick trick)

/** removes the trick
 **/
void
TrickSummary::remove_trick()
{
  this->trick = nullptr;
  this->trick_drawing.remove_trick();
} // void TrickSummary::remove_trick()

/** update (rewrite) all information
 **
 ** @todo   if there is no 'player_of' don't show the column
 **/
void
TrickSummary::update()
{
  if (!this->get_realized())
    return ;

  if (this->trick == nullptr)
    return ;

  auto const& trick = *this->trick;
  auto const& game = trick.game();

  // gettext: %s = winner
  this->winner->set_label(_("The trick goes to %s.",
                            trick.winnerplayer().name()));
  this->points->set_label(_("Points in the trick: %u",
                            trick.points()));

  this->update_trick_picture();

  { // special points
    this->specialpoints_list->clear();

    bool valid_point = false;
    auto const specialpoints = trick.specialpoints();
    for (auto sp : specialpoints) {
      // test whether to show the point
      if (sp.player_get_no == UINT_MAX)
#ifndef RELEASE
        // no 'player of' (can this happen?)
        DEBUG_ASSERTION(false,
                        "Gtkmm::TrickSummary(trick):\n"
                        "  no 'player_get_no' in specialpoint");
#else
      continue;
#endif
      else if (sp.player_of_no != UINT_MAX) {
        auto const& teaminfo = game.teaminfo();
        auto const team_get = teaminfo.get_human_teaminfo(game.player(sp.player_get_no));
        auto const team_of = teaminfo.get_human_teaminfo(game.player(sp.player_of_no));

        if (   is_real(team_get)
            && (team_get == team_of) )
          continue;
      } // if !(player_get_no == UINT_MAX) && (player_of_no != UINT_MAX)

      valid_point |= true;

      auto row = *(this->specialpoints_list->append());
      row[this->specialpoints_model.special_point]
        = _(sp.type);
      if (sp.player_of_no != UINT_MAX)
        row[this->specialpoints_model.player]
          = game.player(sp.player_of_no).name();
      else
        row[this->specialpoints_model.player] = "";
    } // for (sp : specialpoints)

    this->specialpoints_treeview->show_all();

    if (valid_point)
      this->specialpoints_treeview->show();
    else
      this->specialpoints_treeview->hide();
  } // special points
} // void TrickSummary::update()

/** update the trick in the drawing area
 **/
void
TrickSummary::update_trick_picture()
{
  if (   (this->trick == nullptr)
      || (this->trick_picture == nullptr))
    return ;

  this->trick_drawing.set_center({this->trick_picture->get_width() / 2,
                                 this->trick_picture->get_height() / 2});
  this->trick_drawing.draw(this->trick_picture->get_window()->create_cairo_context());
} // void TrickSummary::update_trick_picture()

/** update the trick in the drawing area
 **
 ** @param    cr    cairo context
 **
 ** @return   true
 **/
bool
TrickSummary::on_draw_trick(Cairo::RefPtr<Cairo::Context> const& cr)
{
  this->trick_drawing.set_center({this->trick_picture->get_width() / 2,
                                 this->trick_picture->get_height() / 2});
  this->trick_drawing.draw(cr);
  return true;
} // bool TrickSummary::on_draw_trick(Cairo::RefPtr<Cairo::Context> const& cr)

/** updates the cards
 **/
void
TrickSummary::cards_update()
{
  if (!this->get_realized())
    return ;

  if (this->trick_picture != nullptr)
    this->trick_picture->set_size_request(this->trick_drawing.width()
                                          * 6 / 5,
                                          this->trick_drawing.height()
                                          * 6 / 5);
} // void TrickSummary::cards_update()

/** the setting has changed
 **
 ** @param    type   the type that has changed
 **/
void
TrickSummary::setting_update(int const type)
{
  switch(type) {
  case Setting::Type::rotate_trick_cards:
    if (this->trick_picture != nullptr) {
      this->trick_picture->set_size_request(this->trick_drawing.width()
                                            * 6 / 5,
                                            this->trick_drawing.height()
                                            * 6 / 5);
    } // if (this->trick_picture != nullptr)
    this->update_trick_picture();
    break;
  case Setting::Type::language:
    this->update();
    break;
  default:
    break;
  } // switch(type)
} // void TrickSummary::setting_update(int const type)

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
