/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "ui.h"

#include "main_window.h"
#include "menu.h"
#include "table.h"
#include "name.h"
#include "hand.h"
#include "trick.h"
#include "poverty.h"
#include "reservation.h"
#include "players_db.h"

#include "../../party/party.h"
#include "../../party/rule.h"
#include "../../game/game.h"
#include "../../game/gameplay_actions.h"
#include "../../card/trick.h"
#include "../../player/player.h"
#include "../../os/bug_report_replay.h"
#include "../../basetypes.h"

namespace UI_GTKMM_NS {

/** the game is opened
 **
 ** @param    game   the game that is opened
 **/
void
  UI_GTKMM::game_open(Game& game)
  {
    this->UI::game_open(game);
    this->table->game_open();
  } // void UI_GTKMM::game_open(Game& game)

/** the cards are distributed
 **/
void
  UI_GTKMM::game_cards_distributed()
  {
    this->table->game_cards_distributed();
  } // void UI_GTKMM::game_cards_distributed()

/** ask 'player' for a reservation
 **
 ** @param    player   player who is asked
 **/
void
  UI_GTKMM::reservation_ask(Player const& player)
  {
    this->table->reservation_ask(player);
  } // void UI_GTKMM::reservation_ask(Player player)

/** Gets the reservation of the given player:
 ** pops up a window
 ** This is the first time, the hands can (should) be shown.
 **
 ** @param    player   player, whose reservation schould be get
 **
 ** @return   reservation of the player
 **/
::Reservation
  UI_GTKMM::get_reservation(Player const& player)
  {
    auto const result = this->table->reservation_get(player);
    return result;
  } // ::Reservation UI_GTKMM::get_reservation(player)

/** 'player' has a poverty and shifts cards
 **
 ** @param    player   the player who has the poverty
 **/
void
  UI_GTKMM::poverty_shifting(Player const& player)
  {
    this->table->draw_all();
  } // void UI_GTKMM::poverty_shifting(Player player)

/** 'player' shifts 'cardno' cards
 **
 ** @param    player   the player who shifts the cards
 ** @param    cardno   the number of cards that are shifted
 **/
void
  UI_GTKMM::poverty_shift(Player const& player, unsigned const cardno)
  { 
    this->table->poverty().shift(player, cardno);
  } // void UI_GTKMM::poverty_shift(Player player, unsigned cardno)

/** ask 'player' whether to accept the poverty
 **
 ** @param    player   the player who is asked
 ** @param    cardno   the number of cards that are shifted
 **/
void
  UI_GTKMM::poverty_ask(Player const& player, unsigned const cardno)
  { 
    this->table->poverty().ask(player, cardno);
  } // void UI_GTKMM::poverty_ask(Player player, unsigned cardno)

/** 'player' denied to take the shifted cards
 **
 ** @param    player   the player who has denied to take the cards
 **/
void
  UI_GTKMM::poverty_take_denied(Player const& player)
  { 
    this->table->poverty().take_denied(player);
  } // void UI_GTKMM::poverty_take_denied(Player player)

/** all players have denied to take the cards
 **
 ** @todo   show a window
 **/
void
  UI_GTKMM::poverty_take_denied_by_all()
  { 
    this->table->poverty().take_denied_by_all();
  } // void UI_GTKMM::poverty_take_denied_by_all()

/** 'player' accepts to take the shifted cards
 ** and returns 'cardno' cards with 'trumpno' trumps
 **
 ** @param    player   the player who has denied to take the cards
 ** @param    cardno   number of cards that are given back
 ** @param    trumpno   number of trumps of the cards
 **/
void
  UI_GTKMM::poverty_take_accepted(Player const& player,
                                  unsigned const cardno,
                                  unsigned const trumpno)
  { 
    this->table->poverty().take_accepted(player, cardno, trumpno);
  } // void UI_GTKMM::poverty_take_accepted(Player player, unsigned cardno, unsigned trumpno)

/** returns which cards the player shifts
 **
 ** @param    player   the player who shifts the cards
 **
 ** @return   the cards that are to be shifted
 **/
HandCards
  UI_GTKMM::get_poverty_cards_to_shift(Player& player)
  { 
    auto const result = this->table->poverty().shift(player);
    return result;
  } // HandCards UI_GTKMM::get_poverty_cards_to_shift(Player& player)

/** returns whether 'player' accepts the shifted cards
 **
 ** @param    player   the player who shifts the cards
 ** @param    cardno   the number of shifted cards
 **
 ** @return   whether to accept the cards
 **/
bool
  UI_GTKMM::get_poverty_take_accept(Player const& player, unsigned cardno)
  { 
    auto const result = this->table->poverty().take_accept(player, cardno);
    return result;
  } // void UI_GTKMM::get_poverty_take_accept(Player const& player, unsigned cardno)

/** changes the cards from the poverty-player
 **
 ** @param    player   the player who has accepted the cards
 ** @param    cards   the cards that are given to the player
 **
 ** @return   the cards that are returned to the poverty-player
 **/
HandCards
  UI_GTKMM::get_poverty_exchanged_cards(Player& player,
                                 HandCards const& cards)
  { 
    auto const result = this->table->poverty().cards_change(player, cards);
    return result;
  } // HandCards UI_GTKMM::get_poverty_exchanged_cards(Player& player, HandCards cards)

/** the poverty player 'player' gets 'cards'
 **
 ** @param    player   the player gets the cards
 ** @param    cards   the cards that are given to the player
 **/
void
  UI_GTKMM::poverty_cards_get_back(Player const& player,
                                   HandCards const& cards)
  { 
    this->table->poverty().cards_get_back(player, cards);
  } // void UI_GTKMM::poverty_cards_get_back(Player& player, HandCards cards)

/** the game is redistributed
 **/
void
  UI_GTKMM::game_redistribute()
  {
    this->table->game_redistribute();
  } // void UI_GTKMM::game_redistribute()

/** the game is started:
 ** the gametype and announcements of the other players are shown
 **/
void
  UI_GTKMM::game_start()
  {
    this->main_window->menu->game_start();
    this->table->game_start();
  } // void UI_GTKMM::game_start()

/** a new trick is opened
 **
 ** @param    trick   trick that is opened
 **/
void
  UI_GTKMM::trick_open(::Trick const& trick)
  {
    this->UI::trick_open(trick);
    this->main_window->menu->trick_open();
    this->table->trick_open();
  } // void UI_GTKMM::trick_open(::Trick trick)

/** the trick is full.
 ** show the 'full trick'-window or wait
 **/
void
  UI_GTKMM::trick_full()
  {
    this->main_window->menu->announcements_update();
    this->table->trick_full();
  } // void UI_GTKMM::trick_full()

/** move the trick in the pile of the winnerplayer
 **/
void
  UI_GTKMM::trick_move_in_pile()
  {
    this->table->trick_move_in_pile();
  } // void UI_GTKMM::trick_move_in_pile()

/** the trick is closed
 **/
void
  UI_GTKMM::trick_close()
  {
    this->UI::trick_close();
  } // void UI_GTKMM::trick_close()

/** a gameplay action
 **
 ** @param    action   the action
 **/
void
  UI_GTKMM::gameplay_action(GameplayAction const& action)
  {
    this->thrower.inc_depth();

    this->table->gameplay_action(action);

    switch (action.type()) {
    case GameplayActions::Type::announcement: {
      auto const& announcement
        = dynamic_cast<GameplayActions::Announcement const&>(action).announcement;
      auto const& player
        = this->game().player(dynamic_cast<GameplayActions::Announcement const&>(action).player);
      this->main_window->menu->announcements_update();
      this->table->announcement_made(announcement, player);
      break;
    }

    case GameplayActions::Type::swines: {
      auto const& player
        = this->game().player(dynamic_cast<GameplayActions::Swines const&>(action).player);
      this->main_window->menu->announcements_update();
      this->table->swines_announced(player);
    }
      break;

    case GameplayActions::Type::hyperswines: {
      auto const& player
        = this->game().player(dynamic_cast<GameplayActions::Hyperswines const&>(action).player);
      this->main_window->menu->announcements_update();
      this->table->hyperswines_announced(player);
      break;
    }

    case GameplayActions::Type::marriage: {
      auto const& bridegroom
        = this->game().player(dynamic_cast<GameplayActions::Marriage const&>(action).player);
      auto const& bride
        = this->game().player(dynamic_cast<GameplayActions::Marriage const&>(action).bride);
      this->main_window->menu->announcements_update();
      this->table->marriage(bridegroom, bride);
      break;
    }

    case GameplayActions::Type::genscher: {
      auto const& genscher
        = this->game().player(dynamic_cast<GameplayActions::Genscher const&>(action).player);
      auto const& partner
        = this->game().player(dynamic_cast<GameplayActions::Genscher const&>(action).partner);
      this->main_window->menu->announcements_update();
      this->table->genscher(genscher, partner);
      break;
    }

    case GameplayActions::Type::reservation:
      // nothing to do
      break;
    case GameplayActions::Type::card_played:
      // much to do, see this->card_played(HandCard const& card)
      break;

    case GameplayActions::Type::poverty_shift:
    case GameplayActions::Type::poverty_accepted:
    case GameplayActions::Type::poverty_returned:
    case GameplayActions::Type::poverty_denied:
    case GameplayActions::Type::poverty_denied_by_all:
    case GameplayActions::Type::trick_open:
    case GameplayActions::Type::trick_full:
    case GameplayActions::Type::trick_taken:
    case GameplayActions::Type::check:
    case GameplayActions::Type::print:
      // ToDo
      break;
    } // switch (action.type())

    try {
      this->thrower.dec_depth();
    } catch (GameStatus gs) {
      if (gs == GameStatus::game_trick_taken) {
        ::game_status = GameStatus::game_trick_taken;
      } else {
        throw;
      }
    }
  } // void UI_GTKMM::gameplay_action(GameplayAction action)

/** @return   selected partner of the player
 **   (nullptr if the genscher is not announced)
 **/
Player const*
  UI_GTKMM::get_genscher_partner()
  {
    auto const result = this->table->select_genscher_partner(this->game().players().current_player());
    return result;
  } // Player const* UI_GTKMM::get_genscher_partner()

/** let the user select a card for the given player
 **
 ** @param    player   the player who is to play a card
 **
 ** @return   the selected card
 **/
HandCard
  UI_GTKMM::get_card(Player const& player)
  {
    this->table->card_get(player);

    this->main_window->menu->card_get();
    const_cast<Player&>(player).hand().forget_request();

    try {
      if (   ::setting(Setting::Type::automatic_card_suggestion)
          || (   ::bug_report_replay
              && (player.type() == Player::Type::human) )
         )
        this->table->show_card_suggestion(false);

      do {
        try {
          while (   !this->ui->thrower
                 && (::game_status == GameStatus::game_play))
            ::ui->wait();
        } catch (Card const& card) {
          DEBUG_ASSERTION((card == player.hand().requested_card()),
                          "UI_GTKMM::card_get(player):\n"
                          "  did not catch the requested card:\n"
                          "  " << card << " != " << player.hand().requested_card()
                         );
        }

        if (player.hand().requested_card().is_empty()) {
          const_cast<Player&>(player).hand().forget_request();
          continue;
        }
        // the player cannot play an invalid card
        if (!player.game().tricks().current().isvalid(player.hand().requested_card(),
                                                      player.hand())) {
          const_cast<Player&>(player).hand().forget_request();
          continue;
        }

      } while (player.hand().requested_position() == UINT_MAX);
    } catch (...) {
      this->table->card_got();
      this->main_window->menu->card_got();
      throw;
    } // catch(...)

    this->table->card_got();
    this->main_window->menu->card_got();
    return player.hand().requested_card();
  } // HandCard UI_GTKMM::get_card(Player player)

/** the given player has played 'card.
 ** remove the card from the hand
 **
 ** @param    card   the played card
 **/
void
  UI_GTKMM::card_played(HandCard const& card)
  { 
    auto const& player = card.player();

    // for the sensitivity of the swines button
    this->main_window->menu->announcements_update();

    // for updating the heuristics
    this->players_db->update_db();

    this->table->card_played(player);

    this->update();
  } // void UI_GTKMM::card_player(HandCard card)

/** the game is finished:
 ** the winner and the points of the game are shown.
 **/
void
  UI_GTKMM::game_finish()
  { 
    this->players_db->update_db();
    this->main_window->menu->game_finish();
    this->table->game_finished();
  } // void UI_GTKMM::game_finish()

/** the game is closed
 **/
void
  UI_GTKMM::game_close()
  {
    this->UI::game_close();

    this->main_window->menu->game_close();
    this->table->game_close();
  } // void UI_GTKMM::game_close()

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
