/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#ifdef USE_UI_GTKMM

#include "ui.h"
#include "splash_screen.h"
#include "cards.h"
#include "icons.h"
#include "party_summary.h"
#include "party_settings.h"
#include "players_db.h"
#include "game_debug.h"
#include "rules.h"
#include "table.h"
#include "game_finished.h"
#include "game_summary.h"
#include "party_points.h"
#include "party_finished.h"
#include "bug_report_replay.h"
#include "hand.h"
#include "icongroup.h"
#include "license.h"
#include "main_window.h"

#include "../../basetypes.h"
#include "../../party/party.h"
#include "../../misc/setting.h"

#include <gtkmm/label.h>
#include <gtkmm/button.h>
#include <gtkmm/textview.h>

namespace UI_GTKMM_NS {

/** updates all cards
 **/
void
  UI_GTKMM::cards_changed()
  {
    if (this->main_window)
      this->main_window->set_minimal_size();

    if (this->cards)
      this->cards->update_managed();

    if (this->icons)
      this->icons->update_managed();

    if (this->license)
      this->license->license_update();

    for (auto p : this->parts)
      p->cards_update();
  } // void UI_GTKMM::cards_changed()

/** updates all cards back
 **/
void
  UI_GTKMM::cards_back_changed()
  {
    for (auto p : this->parts)
      p->cards_back_update();
  } // void UI_GTKMM::cards_back_changed()

/** updates all icons
 **/
void
  UI_GTKMM::icons_changed()
  {
    this->icons->update_managed();

    if (this->table)
      this->table->update_icons();

    if (this->license)
      this->license->license_update();
  } // void UI_GTKMM::icons_changed()

/** redraw all (of the playfield)
 **/
void
  UI_GTKMM::redraw_all()
  { 
    if (this->table)
      this->table->force_redraw_all();
  } // void UI_GTKMM::redraw_all()

/** update the gametype icon
 **/
void
  UI_GTKMM::gametype_changed()
  { 
    if (this->table)
      this->table->draw_all();
  } // void UI_GTKMM::gametype_changed()

/** the players 'player_a' and 'player_b' have been switched
 **
 ** @param    player_a   first player
 ** @param    player_b   second player
 **/
void
  UI_GTKMM::players_switched(Player const& player_a, Player const& player_b)
  {
    if (this->party_settings)
      this->party_settings->players_swapped(player_a, player_b);
  } // void UI_GTKMM::players_switched(Player player_a, Player player_b)

/** update the 'player'
 **
 ** @param    player   the player
 **/
void
  UI_GTKMM::player_changed(Player const& player)
  {
    if (this->party_settings)
      this->party_settings->player_update(player);
    this->name_changed(player);
    this->voice_changed(player);
  } // void UI_GTKMM::player_changed(Player player)

/** add the 'player'
 **
 ** @param    player   the player
 **/
void
  UI_GTKMM::player_added(Player const& player)
  {
    if (this->table)
      this->table->party_points_->player_added(player);
    this->name_changed(player);
    this->voice_changed(player);
  } // void UI_GTKMM::player_added(Player player)

/** update the name of 'player'
 **
 ** @param    player   the player
 **/
void
  UI_GTKMM::name_changed(Player const& player)
  {
#ifdef WORKAROUND
    { // name: change to utf8
      auto const name = UI_GTKMM::to_utf8(player.name());
      if (name != player.name()) {
        const_cast<Player&>(player).set_name(name);
        return ;
      }
    } // name: change to utf8
#endif

    if (this->players_db)
      this->players_db->name_changed(player);
    if (this->party_settings)
      this->party_settings->name_update(player);
    if (this->party_summary)
      this->party_summary->name_update(player);
    if (this->table) {
      this->table->name_changed(player);
      this->table->mouse_cursor_update();
    }
    if (this->game_debug)
      this->game_debug->name_changed(player);
    if (this->bug_report_replay)
      this->bug_report_replay->name_changed(player);
  } // void UI_GTKMM::name_changed(Player player)

/** update the voice of 'player'
 **
 ** @param    player   the player
 **/
void
  UI_GTKMM::voice_changed(Player const& player)
  {
    if (this->party_settings)
      this->party_settings->voice_update(player);
  } // void UI_GTKMM::voice_changed(Player player)

/** update the hand of 'player'
 **
 ** @param    player   the player
 **/
void
  UI_GTKMM::hand_changed(Player const& player)
  { 
    if (this->table)
      this->table->force_redraw_all();
  } // void UI_GTKMM::hand_changed(Player player)

/** update the team icon of 'player'
 **
 ** @param    player   the player
 **/
void
  UI_GTKMM::teaminfo_changed(Player const& player)
  { 
    if (this->table)
      this->table->draw_all();
  } // void UI_GTKMM::teaminfo_changed(Player player)

/** update the 'aiconfig'
 **
 ** @param    aiconfig   the aiconfig
 **/
void
  UI_GTKMM::aiconfig_changed(Aiconfig const& aiconfig)
  {
    if (this->party_settings)
      this->party_settings->aiconfig_update(aiconfig);
  } // void UI_GTKMM::aiconfig_changed(Aiconfig aiconfig)

/** update the rule
 **
 ** @param    type      the rule, that has changed
 ** @param    old_value   old value
 **/
void
  UI_GTKMM::rule_changed(int const type, void const* const old_value)
  {
    switch (type) {
    case Rule::Type::counting: {
      if (this->table
          && this->table->game_finished_
          && this->table->game_finished_->game_summary)
        this->table->game_finished_->game_summary->update();

      if (this->table
          && this->table->party_points_)
        this->table->party_points_->recalc_all();

      if (this->table
          && this->table->party_finished_)
        this->table->party_finished_->redraw_points_graph();

      if (this->table
          && (::game_status == GameStatus::party_finished))
        this->table->draw_all();
    }
      break;

    case Rule::Type::with_nines: {
      if (this->party_settings)
        this->party_settings->update();
    }
      break;

    case Rule::Type::number_of_duty_soli:
    case Rule::Type::number_of_duty_color_soli:
    case Rule::Type::number_of_duty_picture_soli:
      if (this->table
          && this->table->party_points_)
        this->table->party_points_->recreate_all();

    default:
      break;
    } // switch (type)

#ifndef RELEASE
    this->table->force_redraw_all();
#endif
  } // void UI_GTKMM::rule_changed(int type, void const* old_value)

/** update the setting
 **
 ** @param    type      the setting, that has changed
 ** @param    old_value   old value
 **/
void
  UI_GTKMM::setting_changed(int const type, void const* const old_value)
  {
    switch(type) {
    case Setting::Type::show_ai_information_hands:
    case Setting::Type::show_ai_information_teams:
      this->redraw_all();
      break;

    case Setting::Type::announce_in_table:
      this->redraw_all();
      break;

    case Setting::Type::name:
#ifdef WORKAROUND
      { // convert latin to utf8
        auto const name = UI_GTKMM::to_utf8(::setting(Setting::Type::name));
        if (name != ::setting(Setting::Type::name)) {
          ::setting.set(Setting::Type::name, name);
          return ;
        }
      } // convert latin to utf8
#endif
      for (auto const& p : ::party->players())
        this->name_changed(p);
      break;

    case Setting::Type::cardset:
      if (this->cards)
        this->cards->load_all();
      if (this->icons)
        this->icons->load();

      if (this->cards)
        this->cards_changed();
      break;

    case Setting::Type::cards_back:
      if (this->cards) {
        this->cards->load_back();
        this->cards_changed();
      }
      break;

    case Setting::Type::iconset:
      if (this->icons)
        this->icons->load();

      if (this->cards)
        this->cards_changed();
      break;

    case Setting::Type::original_cards_size:
      if (::setting(Setting::Type::original_cards_size) == true)
        ::setting.set(Setting::Type::cards_height, this->cards->height_original());
    case Setting::Type::cards_height:
      if (this->cards)
        this->cards->update_scaling();
      if (this->icons)
        this->icons->update_scaling();

      this->cards_changed();
      break;

    case Setting::Type::show_known_teams_in_game:
    case Setting::Type::show_soloplayer_in_game:
      this->redraw_all();
      break;

    case Setting::Type::additional_party_settings:
      if (this->party_settings)
        this->party_settings->update();
      break;

    default:
      break;
    } // switch(type)

    for (auto p : this->parts)
      p->setting_update(type);
  } // void UI_GTKMM::setting_changed(int type, void const* old_value)

/** the status message has changed
 **
 ** @param    status_message   the status message
 **/
void
  UI_GTKMM::status_message_changed(string const& status_message)
  {
    if (this->splash_screen)
      this->splash_screen->status_message_changed(status_message);
  } // void UI_GTKMM::status_message_changed(string status_message)

/** the program is busy
 **/
void
  UI_GTKMM::set_busy()
  {
    this->UI::set_busy();
    this->table->mouse_cursor_update();
  } // void UI_GTKMM::set_busy()

/** the program is not busy any more
 **/
void
  UI_GTKMM::set_not_busy()
  {
    this->UI::set_not_busy();
    this->table->mouse_cursor_update();
  } // void UI_GTKMM::set_not_busy()

/** the progress has changed
 **
 ** @param    progress   the progress
 **/
void
  UI_GTKMM::progress_changed(double const progress)
  {
    if (this->splash_screen)
      this->splash_screen->progress_changed(progress);

    if (this->table)
      this->table->progress_changed(progress);
  } // void UI_GTKMM::progress_changed(double progress)

/** the progress is finished
 **/
void
  UI_GTKMM::progress_finished()
  {
    if (this->table)
      this->table->progress_finished();
  } // void UI_GTKMM::progress_finished()

} // namespace UI_GTKMM_NS

#endif // #ifdef USE_UI_GTKMM
