/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/

#include "constants.h"

#include "ui.h"

#include <unistd.h>

#include "ui.dummy.h"
#ifdef USE_UI_TEXT
#include "text/ui.h"
#endif // #ifdef USE_UI_TEXT
#ifdef USE_UI_GTKMM
#include "gtkmm/ui.h"
#endif // #ifdef USE_UI_GTKMM

#include "../card/hand.h"
#include "../misc/setting.h"
#include "../misc/bug_report.h"
#include "../os/bug_report_replay.h"

/** Constructor
 ** (to be called from a child class)
 **
 ** @param    type   the UIType
 **/
UI::UI(UIType const type) :
  type_(type)
{ }

/** create a new ui
 ** (it should just call the new-operator of the child class)
 **
 ** @param    type   type of the UI to create
 **
 ** @return   pointer of UI
 **/
unique_ptr<UI>
UI::create(UIType const type)
{
  switch (type) {
  case UIType::DUMMY:
    return make_unique<UI_Dummy>();
#ifdef USE_UI_TEXT
  case UIType::TEXT:
    return make_unique<UI_Text>();
#endif // #ifdef USE_UI_TEXT
#ifdef USE_UI_GTKMM
  case UIType::GTKMM_DOKO:
    return make_unique<UI_GTKMM_NS::UI_GTKMM>();
#endif // #ifdef USE_UI_GTKMM
  default:
    return {};
  } // switch(type)

  return {};
} // static unique_ptr<UI> UI::create(UIType const type)

/** @return   type of the UI
 **/
UIType
UI::type() const
{ return this->type_; }

/** @return   the party
 **/
Party&
UI::party()
{
  DEBUG_ASSERTION(this->party_,
                  "UI::party()\n"
                  "  party is not set");
  return *this->party_;
}

/** @return   the party
 **/
Party const&
UI::party() const
{
  DEBUG_ASSERTION(this->party_,
                  "UI::party()\n"
                  "  party is not set");
  return *this->party_;
} // Party UI::party() const

/** @return   the game
 **/
Game&
UI::game()
{
  DEBUG_ASSERTION(this->game_,
                  "UI::game()\n"
                  "  game is not set");
  return *this->game_;
} // Game& UI::game()

/** @return   the game
 **/
Game const&
UI::game() const
{
  DEBUG_ASSERTION(this->game_,
                  "UI::game()\n"
                  "  game is not set");
  return *this->game_;
} // Game UI::game() const

/** sleeps for 'sleep_usec' micro seconds
 **
 ** @param    sleep_usec  micro seconds to sleep
 **/
void
UI::sleep(unsigned const sleep_msec)
{
#ifdef LINUX
  usleep(1000 * sleep_msec);
#endif
  this->update();
} // virtual void sleep(unsigned sleep_msec)

/** wait a bit (sleep for 20 msec)
 **/
void
UI::wait()
{
  if (!(::fast_play & FastPlay::pause))
    this->sleep(20);
  else
    this->sleep(1);
} // void UI::wait()

/** the party is opened
 **
 ** @param    party   party that is opened
 **/
void
UI::party_open(Party& party)
{
  DEBUG_ASSERTION(!this->party_,
                  "UI::party_open(party):\n"
                  "  party already set");
  this->party_ = &party;
} // void UI::party_open(Party& party)

/** the party will close
 **/
void
UI::party_close()
{
  this->party_ = {};
} // void UI::party_close()

/** a game is opened
 **
 ** @param    game   game that is opened
 **/
void
UI::game_open(Game& game)
{
  DEBUG_ASSERTION(!this->game_,
                  "UI::game_open(game):\n"
                  "  game already set");
  this->game_ = &game;
} // void UI::game_open(Game& game)

/** the game will close
 **/
void
UI::game_close()
{
  this->game_ = {};
} // void UI::game_close()

/** @return   the latest status message
 **/
string const&
UI::status_message() const
{
  if (this->status_message_.empty()) {
    static string const translation_empty = ""s;
    return translation_empty;
  }

  return this->status_message_.back();
} // string UI::status_message() const

/** adds the status message
 **
 ** @param    status_message   the new status message
 **/
void
UI::add_status_message(string const& status_message)
{
  this->status_message_.push_back(status_message);
  this->status_message_changed(status_message);
} // void UI::add_status_message(string status_message)

/** remove the status message
 **
 ** @param    status_message   the status message that is to be removed
 **/
void
UI::remove_status_message(string const& status_message)
{
  auto pos
    = std::find(this->status_message_, status_message);

  DEBUG_ASSERTION((pos != this->status_message_.end()),
                  "UI::remove_status_message(status_message):\n"
                  "  message '" << status_message << "' "
                  "not found in the list");

  if (pos != this->status_message_.end())
    pos = this->status_message_.erase(pos);

  if (pos == this->status_message_.end())
    this->status_message_changed(this->status_message());
} // void UI::remove_status_message(string status_message)

/** @return   whether the program is busy
 **/
bool
UI::busy() const
{ return this->busy_; }

/** the program is busy
 **/
void
UI::set_busy()
{
  this->busy_ = true;
} // void UI::set_busy()

/** the program is not busy any more
 **/
void
UI::set_not_busy()
{
  this->busy_ = false;
} // void UI::set_not_busy()

/** @return   the progress
 **/
double
UI::progress() const
{
  if (this != ::ui.get())
    return ::ui->progress();

  if (this->progress_max_ == 0)
    return 1;

  return (this->progress_ / this->progress_max_);
} // double UI::progress() const

/** @return   whether the ui is in a progress (t.i. progress_max is set)
 **/
bool
UI::in_progress() const
{
  if (this != ::ui.get())
    return ::ui->in_progress();

  return (this->progress_max_ > 0);
} // bool UI::in_progress() const

/** sets the maximal progress
 **
 ** @param    progress_max   the maximal progress
 **/
void
UI::set_max_progress(double const progress_max)
{
  if (this != ::ui.get()) {
    ::ui->set_max_progress(progress_max);
    return ;
  }

  DEBUG_ASSERTION((this->progress_max_ == 0),
                  "UI::set_max_progress(progress_max):\n"
                  "  'this->progress_max_' is already set: "
                  << this->progress_max_ << ".");

  this->progress_max_ = progress_max;
  this->progress_ = 0;
  this->progress_changed(this->progress());
} // void UI::set_max_progress(double progress_max)

/** set 'progress' to the progress
 **
 ** @param    progress   the progress
 **/
void
UI::set_progress(double const progress)
{
  if (this != ::ui.get()) {
    ::ui->add_progress(progress);
    return ;
  }

  DEBUG_ASSERTION(progress <= this->progress_max_,
                  "UI::set_progress(" << progress << ")\n"
                  "The progress is greater then the maximum one (" << this->progress_max_ << ")");
  DEBUG_ASSERTION(progress >= this->progress_,
                  "UI::set_progress(" << progress << ")\n"
                  "The progress is less then the current one (" << this->progress_ << ")");
  this->progress_ = progress;

  this->progress_changed(this->progress());
} // void UI::add_progress(double progress)

/** adds 'progress' to the progress
 **
 ** @param    progress   the progress
 **/
void
UI::add_progress(double const progress)
{
  if (this != ::ui.get()) {
    ::ui->add_progress(progress);
    return ;
  }

  this->progress_ += progress;

#ifdef DKNOF
  DEBUG_ASSERTION((this->progress() <= 1.000001),
                  "UI::add_progress(progress):\n"
                  "  'progress' is greater than 1: "
                  << this->progress());
#else
  if (this->progress() > 1.000001)
    cerr << "UI::add_progres():\n"
      << "  progress is " << this->progress() << ", greater than 1" << endl;
  else
#endif
    if (this->progress() > 1 - 0.000001)
      this->progress_ = this->progress_max_;

  this->progress_changed(this->progress());
} // void UI::add_progress(double progress)

/** the progress is finished
 **
 ** @param    set_full   whether to set to full before finishing the progress
 **/
void
UI::finish_progress(bool const set_full)
{
  if (this != ::ui.get()) {
    ::ui->finish_progress();
    return ;
  }

#ifndef RELEASE
#ifdef POSTPONED
  DEBUG_ASSERTION((this->progress() == 1),
                  "UI::finish_progress():\n"
                  "  'progress' is not equal to 1: "
                  << this->progress()
                  << ", diff: " << this->progress() - 1);
#else
  if (this->progress() != 1)
    cerr << "UI::finish_progres()\n"
      << "  progress is " << this->progress() << ", not 1" << endl;
#endif
#endif
  if (set_full) {
    this->progress_ = this->progress_max_;
    this->progress_changed(this->progress());
  } // if (set_full)

  this->progress_finished();
  this->progress_max_ = 0;
  this->progress_ = 0;
} // void UI::finish_progress(bool set_full = true)

/** adds the message to the shown messages set
 **
 ** @param    message   message to add
 **
 ** @return   whether the message is added to the set (if false: the message was already in the set)
 **/
bool
UI::add_to_messages_shown(string const& message,
                          InformationType const type)
{
  if (contains(this->messages_shown_, Message(message, type)))
    return false;

  this->messages_shown_.push_back(Message(message, type));
  return true;
} // bool UI::add_to_messages_shown(string message, InformationType type)

/** -> result
 **
 ** @param    type   ui type
 **
 ** @return   name of the ui type
 **/
string
to_string(UIType const type)
{
  switch (type) {
  case UIType::DUMMY:
    return "dummy";
  case UIType::WRAP:
    return "wrap";
#ifdef USE_UI_TEXT
  case UIType::TEXT:
    return "text";
#endif
#ifdef USE_UI_AATEXT
  case UIType::AATEXT:
    return "aatext";
#endif
#ifdef USE_UI_GTKMM
  case UIType::GTKMM_DOKO:
    return "gtkmm";
#endif
  } // switch (type)
  return {};
} // string to_string(UIType type)

/** write the uitype
 **
 ** @param    ostr   output stream
 ** @param    type   uitype
 **
 ** @return   output stream
 **/
ostream&
operator<<(ostream& ostr, UIType const type)
{ return (ostr << to_string(type)); }


#if 0
// pure virtual functions
// should be = 0 in the declaration
Reservation
UI::get_reservation(Player const& player) // = 0
{
  DEBUG_ASSERTION(false, "Call of Ui<" + to_string(this->type_) + ">::get_reservation(player)");
  return ::Reservation();
}

HandCard
UI::get_card(Player const& player) // = 0
{
  DEBUG_ASSERTION(false, "Call of Ui<" + to_string(this->type_) + ">::get_card(player)");
  return HandCard();
}

HandCards
UI::get_poverty_cards_to_shift(Player& player) // = 0
{
  DEBUG_ASSERTION(false, "Call of Ui<" + to_string(this->type_) + ">::get_poverty_cards_to_shift(player)");
  return HandCards();
}

bool
UI::get_poverty_take_accept(Player const& player, unsigned const cardno) // = 0
{
  DEBUG_ASSERTION(false, "Call of Ui<" + to_string(this->type_) + ">::get_poverty_take_accept(player, cardno)");
  return false;
}

HandCards
UI::get_poverty_exchanged_cards(Player& player, HandCards const& cards) // = 0
{
  DEBUG_ASSERTION(false, "Call of Ui<" + to_string(this->type_) + ">::get_poverty_exchanged_cards(player, cards)");
  return HandCards();
}

Player const*
UI::get_genscher_partner() // = 0
{
  DEBUG_ASSERTION(false, "Call of Ui<" + to_string(this->type_) + ">::get_genscher_partner()");
  return nullptr;
}
#endif
