/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001 – 2018 by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@posteo.de
 *
 *********************************************************************/


#include "constants.h"

#include <sys/stat.h>

// all things written into this stream are ignored
NullOStr null_ostr;

/** writes a mark with a counter
 **
 ** @param    string      text to write (default: NULL)
 ** @param    increase   whether to increase the counter (default: true)
 **
 ** @return   new counter
 **/
unsigned MARK(char const* const string, bool increase)
{
  static unsigned int mark_nr = 0;
  if (increase)
    mark_nr += 1;

  if (string != nullptr)
    cerr << mark_nr << ": " << string << endl;
  else
    cerr << mark_nr << endl;

  return mark_nr;
} // unsigned MARK(char const* const string = NULL, bool increase = true);

/** reads a line from the input stream
 ** empty lines are skipped and the DOS EOL is recognized
 **
 ** @param    istr   input stream
 **
 ** @return   read line
 **/
string
read_line(std::istream& istr)
{
  string line;
  do {
    std::getline(istr, line);
    if ( !line.empty() && (*(line.end() - 1) == '\r') )
      line.erase(line.end() - 1);
  } while (line.empty() && istr.good()) ;
  return line;
} // string read_line(istream& istr)

/** generates an error
 **
 ** @param    error_message   error message
 **/
void
generate_error(string const error_message)
{
  DEBUG_ASSERTION(false, error_message);
}

/** constructor
 **
 ** @param    s   string with the unsigned
 **/
Unsigned::Unsigned(string const& s)
{
  istringstream istr(s);
  istr >> *this;
} // Unsigned::Unsigned(string s)

/** writes an 'Unsigned' into the stream
 ** 'UINT_MAX' is written as '-'
 **
 ** @param    ostr   stream to write into
 ** @param    u      unsigned to write
 **
 ** @return   output stream
 **/
std::ostream&
operator<<(std::ostream& ostr, Unsigned const& u)
{
  if (u == UINT_MAX)
    ostr << '-';
  else
    ostr << static_cast<unsigned>(u);
  return ostr;
} // std::ostream& operator<<(std::ostream& ostr, Unsigned u)

/** reads an 'Unsigned' from the stream
 ** a '-' is interpreted as 'UINT_MAX'
 **
 ** @param    istr   stream to read from
 ** @param    u      unsigned to read
 **
 ** @return   input stream
 **/
std::istream&
operator>>(std::istream& istr, Unsigned& u)
{
  while (istr.good()
         && std::isspace(istr.peek()))
    istr.get();
  if (istr.peek() == '-') {
    u = UINT_MAX;
    istr.get();
  } else {
    istr >> static_cast<unsigned&>(u);
  }
  return istr;
} // std::istream& operator>>(std::istream& istr, Unsigned& u)

/** writes an optional unsigned into the stream
 ** not set is written as '-'
 **
 ** @param    ostr   stream to write into
 ** @param    u      unsigned to write
 **
 ** @return   output stream
 **/
std::ostream&
operator<<(std::ostream& ostr, optional<unsigned> const& u)
{
  if (!u)
    ostr << '-';
  else
    ostr << *u;
  return ostr;
}

/** reads an optional unsigned from the stream
 ** a '-' is interpreted as not set
 **
 ** @param    istr   stream to read from
 ** @param    u      unsigned to read
 **
 ** @return   input stream
 **/
std::istream&
operator>>(std::istream& istr, optional<unsigned>& u)
{
  while (istr.good()
         && std::isspace(istr.peek()))
    istr.get();
  if (istr.peek() == '-') {
    u.reset();
    istr.get();
  } else {
    unsigned u2 = 0;
    istr >> u2;
    if (istr)
      *u = u2;
  }
  return istr;
} // std::istream& operator>>(std::istream& istr, Unsigned& u)

